/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  arithlow.c
 *
 *         One image grayscale arithmetic (8 bpp)
 *                  void       addConstantGrayLow()
 *
 *         Two image grayscale arithmetic (8 bpp)
 *                  void       addGrayLow()
 *                  void       subtractGrayLow()
 *
 *         Image accumulator arithmetic operations (onto 16 bpp)
 *                  void       init8To16Low()
 *                  void       accumulate8To16Low()
 *                  void       divide8Into16Low()
 *                  void       multConstant16Low()
 *                  void       final16To8Low()
 *
 *         Display in full dynamic range in 8 bpp image:
 *                  void       maxDynamicRangeLow()
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "allheaders.h"


    /* mid-range uint16 value */
#define   INIT_16_VALUE       0x8000


/*------------------------------------------------------------------*
 *               One image grayscale arithmetic (8 bpp)             *
 *------------------------------------------------------------------*/
/*!
 *  addConstantGrayLow()
 *
 *     Note: val is limited to the range [-255, +255]
 */
void
addConstantGrayLow(UINT32   *data,
                   INT32     w,
                   INT32     h,
	           INT32     wpl,
	           INT32     val)
{
INT32     i, j, pval;
UINT32   *line;

    if (SIGN(val) == 1) {  /* add */
	for (i = 0; i < h; i++) {
	    line = data + i * wpl;
	    for (j = 0; j < w; j++) {
		pval = GET_DATA_BYTE(line, j);
		pval = MIN(pval + val, 255);
		SET_DATA_BYTE(line, j, (UINT8)pval);
	    }
	}
    } else {  /* subtract */
	for (i = 0; i < h; i++) {
	    line = data + i * wpl;
	    for (j = 0; j < w; j++) {
		pval = GET_DATA_BYTE(line, j);
		pval = MAX(pval + val, 0);
		SET_DATA_BYTE(line, j, (UINT8)pval);
	    }
	}
    }

    return;
}


/*------------------------------------------------------------------*
 *               Two image grayscale arithmetic (8 bpp)             *
 *------------------------------------------------------------------*/
/*!
 *  addGrayLow()
 *
 */
void
addGrayLow(UINT32   *datad,
           INT32     w,
           INT32     h,
	   INT32     wpld,
           UINT32   *datas,
           INT32     wpls)
{
INT32         i, j;
UINT8         val;
INT32         sum;
UINT32       *lines, *lined;

    for (i = 0; i < h; i++) {
	lined = datad + i * wpld;
	lines = datas + i * wpls;
	for (j = 0; j < w; j++) {
	    sum = GET_DATA_BYTE(lines, j) + GET_DATA_BYTE(lined, j);
	    val = MIN(sum, 255);
	    SET_DATA_BYTE(lined, j, val);
	}
    }

    return;
}


/*!
 *  subtractGrayLow()
 *
 */
void
subtractGrayLow(UINT32   *datad,
                INT32     w,
                INT32     h,
	        INT32     wpld,
                UINT32   *datas,
                INT32     wpls,
                INT32     swap)
{
INT32         i, j;
UINT8         val;
INT32         diff;
UINT32       *lines, *lined;

    for (i = 0; i < h; i++) {
	lined = datad + i * wpld;
	lines = datas + i * wpls;
	if (swap == 0) {
	    for (j = 0; j < w; j++) {
		diff = GET_DATA_BYTE(lined, j) - GET_DATA_BYTE(lines, j);
		val = MAX(diff, 0);
		SET_DATA_BYTE(lined, j, val);
	    }
	}
	else {   /* swap == 1 */
	    for (j = 0; j < w; j++) {
		diff = GET_DATA_BYTE(lines, j) - GET_DATA_BYTE(lined, j);
		val = MAX(diff, 0);
		SET_DATA_BYTE(lined, j, val);
	    }
	}
    }

    return;
}


/*-------------------------------------------------------------*
 *     Image accumulator arithmetic operations (onto 16 bpp)   *
 *-------------------------------------------------------------*/
/*!
 *  init8To16Low()
 *
 */
void
init8To16Low(UINT32   *datad,
             INT32     w,
             INT32     h,
	     INT32     wpld,
             UINT32   *datas,
             INT32     wpls)
{
INT32         i, j, val;
UINT32       *lines, *lined;

    for (i = 0; i < h; i++) {
	lines = datas + i * wpls;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    val = INIT_16_VALUE + GET_DATA_BYTE(lines, j);
	    SET_DATA_TWO_BYTES(lined, j, (UINT16)val);
	}
    }

    return;
}
   

/*!
 *  final16To8Low()
 *
 */
void
final16To8Low(UINT32   *datad,
              INT32     w,
              INT32     h,
	      INT32     wpld,
              UINT32   *datas,
              INT32     wpls)
{
INT32         i, j, val;
UINT32       *lines, *lined;

    for (i = 0; i < h; i++) {
	lines = datas + i * wpls;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    val = GET_DATA_TWO_BYTES(lines, j) - INIT_16_VALUE;
	    val = MAX(0, val);
	    val = MIN(255, val);
	    SET_DATA_BYTE(lined, j, (UINT8)val);
	}
    }

    return;
}
   

/*!
 *  accumulate8To16Low()
 *
 */
void
accumulate8To16Low(UINT32   *datad,
                   INT32     w,
                   INT32     h,
	           INT32     wpld,
                   UINT32   *datas,
                   INT32     wpls,
                   INT32     op)
{
INT32         i, j, val;
UINT32       *lines, *lined;

    for (i = 0; i < h; i++) {
	lines = datas + i * wpls;
	lined = datad + i * wpld;
	if (op == ADD) {
	    for (j = 0; j < w; j++) {
		val = GET_DATA_TWO_BYTES(lined, j) + GET_DATA_BYTE(lines, j);
		SET_DATA_TWO_BYTES(lined, j, (UINT16)val);
	    }
	}
	else {  /* op == SUBTRACT */
	    for (j = 0; j < w; j++) {
		val = GET_DATA_TWO_BYTES(lined, j) - GET_DATA_BYTE(lines, j);
		SET_DATA_TWO_BYTES(lined, j, (UINT16)val);
	    }
	}
    }

    return;
}
   


/*!
 *  divide8Into16Low()
 *
 */
void
divide8Into16Low(UINT32   *datad,
                 INT32     w,
                 INT32     h,
	         INT32     wpld,
                 UINT32   *datas,
                 INT32     wpls)
{
INT32         i, j, val;
UINT8         val8;
UINT32       *lines, *lined;

    for (i = 0; i < h; i++) {
	lines = datas + i * wpls;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    val = 64 * (GET_DATA_TWO_BYTES(lined, j) - INIT_16_VALUE);
	    if (val < 0)
		val = 0;
	    if ((val8 = GET_DATA_BYTE(lines, j)) > 0)
		val = val / val8;
	    if (val > 0xffff)
		val = 0xffff;
	    SET_DATA_TWO_BYTES(lined, j, (UINT16)val);
	}
    }

    return;
}
   


/*!
 *  multConstant16Low()
 *
 */
void
multConstant16Low(UINT32   *data,
                  INT32     w,
                  INT32     h,
	          INT32     wpl,
                  FLOAT32   factor)
{
INT32         i, j, val;
UINT32       *line;

    for (i = 0; i < h; i++) {
	line = data + i * wpl;
	for (j = 0; j < w; j++) {
	    val = GET_DATA_TWO_BYTES(line, j) - INIT_16_VALUE;
	    val = (INT32)(val * factor);
	    val = val + INIT_16_VALUE;
	    SET_DATA_TWO_BYTES(line, j, (UINT16)val);
	}
    }

    return;
}
   


/*-----------------------------------------------------------------------*
 *              Display in full dynamic range in 8 bpp image             *
 *-----------------------------------------------------------------------*/
/*!
 *  maxDynamicRangeLow()
 *
 *       type:  LINEAR_SCALE or LOG_SCALE
 */
void
maxDynamicRangeLow(UINT32  *datad,
		   INT32    w,
		   INT32    h,
		   INT32    wpld,
		   UINT32  *datas,
		   INT32    d,
		   INT32    wpls,
		   INT32    type)
{
UINT8     max, sval, dval;
UINT16    max16, sval16;
INT32     i, j;
UINT32    word;
UINT32   *lines, *lined;
FLOAT32   factor;

    switch (d)
    {
    case 8:
	max = 0;
	for (i = 0; i < h; i++) {
	    lines = datas + i * wpls;
	    for (j = 0; j < wpls; j++) {
		word = *(lines + j);
		max = MAX(max, word >> 24);
		max = MAX(max, (word >> 16) & 0xff);
		max = MAX(max, (word >> 8) & 0xff);
		max = MAX(max, word & 0xff);
	    }
	}
/*	fprintf(stderr, "max = %d\n", max); */

	if (type == LINEAR_SCALE) {
	    factor = 255 / (FLOAT32)max;
	    for (i = 0; i < h; i++) {
		lines = datas + i * wpls;
		lined = datad + i * wpld;
		for (j = 0; j < w; j++) {
		    sval = GET_DATA_BYTE(lines, j);
		    dval = (UINT8)(factor * (FLOAT32)sval);
		    SET_DATA_BYTE(lined, j, dval);
		}
	    }
	}
	else  /* type == LOG_SCALE) */ {
	    factor = 255 / log((FLOAT32)max);
	    for (i = 0; i < h; i++) {
		lines = datas + i * wpls;
		lined = datad + i * wpld;
		for (j = 0; j < w; j++) {
		    sval = GET_DATA_BYTE(lines, j);
		    dval = (UINT8)(factor * log((FLOAT32)sval));
		    SET_DATA_BYTE(lined, j, dval);
		}
	    }
	}
	break;

    case 16:
	max16 = 0;
	for (i = 0; i < h; i++) {
	    lines = datas + i * wpls;
	    for (j = 0; j < wpls; j++) {
		word = *(lines + j);
		max16 = MAX(max16, word >> 16);
		max16 = MAX(max16, word & 0xffff);
	    }
	}

	if (type == LINEAR_SCALE) {
	    factor = 65535 / (FLOAT32)max16;
	    for (i = 0; i < h; i++) {
		lines = datas + i * wpls;
		lined = datad + i * wpld;
		for (j = 0; j < w; j++) {
		    sval16 = GET_DATA_TWO_BYTES(lines, j);
		    dval = (UINT8)(factor * (FLOAT32)sval16);
		    SET_DATA_BYTE(lined, j, dval);
		}
	    }
	}
	else  /* type == LOG_SCALE) */ {
	    factor = 65535 / log((FLOAT32)max16);
	    for (i = 0; i < h; i++) {
		lines = datas + i * wpls;
		lined = datad + i * wpld;
		for (j = 0; j < w; j++) {
		    sval16 = GET_DATA_TWO_BYTES(lines, j);
		    dval = (UINT8)(factor * log((FLOAT32)sval16));
		    SET_DATA_BYTE(lined, j, dval);
		}
	    }
	}
	break;

    default:
	fprintf(stderr, "Error: source depth must be 8 or 16 bpp\n");
	break;
    }

    return;
}

