/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/

/*
 *  binarize.c
 *                     
 *      Floyd-Steinberg dithering
 *          PIX        *pixDitherFS()
 *
 *      Simple (pixelwise) binarization with fixed threshold
 *          PIX        *pixThresholdToBinary()
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"


/*------------------------------------------------------------------*
 *             Binarization by Floyd-Steinberg Dithering            *
 *------------------------------------------------------------------*/
/*!
 *  pixDitherFS()
 *
 *      Input:  pixs
 *      Return: pixd (dithered binary), or null on error
 *
 *      The Floyd-Steinberg error diffusion dithering algorithm
 *      binarizes an 8 bpp grayscale image to a threshold of 128.
 *      If a pixel has a value above 127, it is binarized to white
 *      and the excess (below 255) is subtracted from three 
 *      neighboring pixels in the fractions 3/8 to (i, j+1),
 *      3/8 to (i+1, j) and 1/4 to (i+1,j+1), truncating to 0
 *      if necessary.  Likewise, if it the pixel has a value 
 *      below 128, it is binarized to black and the excess above 0
 *      is added to the neighboring pixels, truncating to 255 if
 *      necessary.
 *
 *      We differ from straight dithering in that we allow
 *      clipping of grayscale to 0 or 255 if the values are
 *      sufficiently close (CLIP_LOWER_TO_BLACK, etc.).  This
 *      reduces the snake patterns in parts of the image that
 *      are nearly black or white.  But it also prevents the attempt
 *      to reproduce gray for those values.
 *
 *      The implementation is slow but straightforward.
 *      It uses a pair of line buffers to avoid changing pixs.
 */
PIX *
pixDitherFS(PIX    *pixs)
{
INT32      w, h, wpls, wpld;
UINT32    *datas, *datad;
UINT32    *bufs1, *bufs2;
PIX       *pixd;

    PROCNAME("pixDitherFS");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
	return (PIX *)ERROR_PTR("must be 8 bpp for dithering", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

	/* two line buffers, 1 for current line and 2 for next line */
    if ((bufs1 = (UINT32 *)CALLOC(wpls, sizeof(UINT32))) == NULL)
	return (PIX *)ERROR_PTR("bufs1 not made", procName, NULL);
    if ((bufs2 = (UINT32 *)CALLOC(wpls, sizeof(UINT32))) == NULL)
	return (PIX *)ERROR_PTR("bufs2 not made", procName, NULL);

    if ((pixd = pixCreate(w, h, 1)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    ditherFSLow(datad, w, h, wpld, datas, wpls, bufs1, bufs2);

    FREE((void *)bufs1);
    FREE((void *)bufs2);

    return pixd;
}


/*------------------------------------------------------------------*
 *       Simple (pixelwise) binarization with fixed threshold       *
 *------------------------------------------------------------------*/
/*!
 *  pixThresholdToBinary()
 *
 *      Input:  pixs (4 or 8 bpp)
 *              threshold value
 *      Return: pixd (1 bpp), or null on error
 *
 *      Note: if the source pixel is less than the
 *            threshold value, the dest will be 1;
 *            otherwise, it will be 0
 */
PIX *
pixThresholdToBinary(PIX    *pixs,
                     INT32   thresh)
{
INT32      d, w, h, wpls, wpld;
UINT32    *datas, *datad;
PIX       *pixd;

    PROCNAME("pixThresholdToBinary");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    d = pixGetDepth(pixs);
    if (d != 4 && d != 8)
	return (PIX *)ERROR_PTR("must be 4 or 8 bpp", procName, NULL);
    if (thresh < 0)
	return (PIX *)ERROR_PTR("thresh must be non-negative", procName, NULL);
    if (d == 4 && thresh > 15)
	return (PIX *)ERROR_PTR("4 bpp thresh not in {0-15}", procName, NULL);
    if (d == 8 && thresh > 255)
	return (PIX *)ERROR_PTR("8 bpp thresh not in {0-255}", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

    if ((pixd = pixCreate(w, h, 1)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    thresholdToBinaryLow(datad, w, h, wpld, datas, d, wpls, thresh);
    return pixd;
}


