/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *   box.c
 *
 *      Box creation, copy, clone, destruction
 *           BOX    *boxCreate()
 *           BOX    *boxCopy()
 *           BOX    *boxClone()
 *           void    boxDestroy()
 *
 *      Box accessors
 *           INT32   boxGetRefcount()
 *           INT32   boxChangeRefcount()
 *
 *      Boxa creation, copy, destruction
 *           BOXA   *boxaCreate()
 *           BOXA   *boxaCopy()
 *           void    boxaDestroy()
 *
 *      Boxa array extension
 *           INT32   boxaAddBox()
 *           INT32   boxaExtendArray()
 *
 *      Boxa accessors
 *           INT32   boxaGetCount()
 *           INT32   boxaGetBox()
 *
 *      Boxa properties
 *           INT32   boxaGetExtent()
 *
 *      Boxa serialized I/O
 *           INT32   boxaWriteStream()
 *           BOXA   *boxaReadStream()
 *
 *      Box print (for debug)
 *           INT32   boxPrintStream()
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"


#define  INITIAL_PTR_ARRAYSIZE      20    /* n'import quoi */


/*---------------------------------------------------------------------*
 *                  Box creation, destruction and copy                 *
 *---------------------------------------------------------------------*/
/*!
 *  boxCreate()
 *
 *       Input:  x, y, width, height
 *       Return: box, or null on error
 *
 *       Note:  if either width < 0 or height < 0, this is an
 *              error and null is returned
 */
BOX *
boxCreate(INT32  x,
          INT32  y,
          INT32  w,
          INT32  h)
{
BOX   *box;

    PROCNAME("boxCreate");

    if ((w < 0) || (h < 0))
	return (BOX *)ERROR_PTR("w and h must be >= 0", procName, NULL);

    if ((box = (BOX *) CALLOC(1, sizeof(BOX))) == NULL)
	return (BOX *)ERROR_PTR("box not made", procName, NULL);
    box->x = x;
    box->y = y;
    box->w = w;
    box->h = h;
    box->refcount = 1;

    return box;
}


/*!
 *  boxCopy()
 *
 *       Input:  box
 *       Return: copy of box, or null on error
 */
BOX *
boxCopy(BOX  *box)
{
BOX   *boxc;

    PROCNAME("boxCopy");

    if (!box)
	return (BOX *)ERROR_PTR("box not defined", procName, NULL);

    boxc = boxCreate(box->x, box->y, box->w, box->h);

    return boxc;
}


/*! 
 *  boxClone()
 *
 *       Input:  box
 *       Return: ptr to same box, or null on error
 */
BOX *
boxClone(BOX  *box)
{

    PROCNAME("boxClone");

    if (!box)
	return (BOX *)ERROR_PTR("box not defined", procName, NULL);

    boxChangeRefcount(box, 1);

    return box;
}


/*!
 *  boxDestroy()
 *
 *       Input:   &box (<can be nulled>)
 *       Return:  void
 *
 *       Action:  decrements the ref count and, if 0, destroys the box
 *                and nulls the input ptr.
 */
void
boxDestroy(BOX  **pbox)
{
BOX   *box;

    PROCNAME("boxDestroy");

    if (pbox == NULL) {
        WARNING("ptr address is null!", procName);
	return;
    }

    if ((box = *pbox) == NULL)
	return;

	/* decrement the ref count.  If it is still > 0, just return */
    boxChangeRefcount(box, -1);
    if (boxGetRefcount(box) > 0)
	return;

    FREE((void *)box);
    *pbox = NULL;

    return;
}


/*---------------------------------------------------------------------*
 *                        Box refcount accessors                       *
 *---------------------------------------------------------------------*/
INT32
boxGetRefcount(BOX  *box)
{
    PROCNAME("boxGetRefcount");

    if (!box)
	return ERROR_INT("box not defined", procName, UNDEF);

    return box->refcount;
}


INT32
boxChangeRefcount(BOX   *box,
                  INT32  delta)
{
    PROCNAME("boxChangeRefcount");

    if (!box)
	return ERROR_INT("box not defined", procName, 1);

    box->refcount += delta;
    return 0;
}



/*---------------------------------------------------------------------*
 *             Boxa creation, destruction, copy, extension             *
 *---------------------------------------------------------------------*/
/*!
 *  boxaCreate()
 *
 *       Input:  n  (initial number of ptrs)
 *       Return: boxa, or null on error
 */
BOXA *
boxaCreate(INT32  n)
{
BOXA   *boxa;

    PROCNAME("boxaCreate");

    if (n <= 0)
	n = INITIAL_PTR_ARRAYSIZE;

    if ((boxa = (BOXA *)CALLOC(1, sizeof(BOXA))) == NULL)
	return (BOXA *)ERROR_PTR("boxa not made", procName, NULL);
    boxa->n = 0;
    boxa->nalloc = n;
    
    if ((boxa->box = (BOX **)CALLOC(n, sizeof(BOX *))) == NULL)
	return (BOXA *)ERROR_PTR("boxa ptrs not made", procName, NULL);

    return boxa;
}


/*!
 *  boxaCopy()
 *
 *     Input:  boxa
 *             copytype (COPY, CLONE)
 *     Return: new boxa
 *
 *     Note: the copytype describes what happens to the individual
 *           boxes that are added to the array.
 */
BOXA *
boxaCopy(BOXA  *boxa,
         INT32  copytype)
{
INT32   i;
BOX    *boxc;
BOXA   *boxac;

    PROCNAME("boxaCopy");

    if (!boxa)
	return (BOXA *)ERROR_PTR("boxa not defined", procName, NULL);
    if (copytype != COPY && copytype != CLONE)
	return (BOXA *)ERROR_PTR("copytype must be in {1,2}", procName, NULL);

    if ((boxac = boxaCreate(boxa->nalloc)) == NULL)
	return (BOXA *)ERROR_PTR("boxac not made", procName, NULL);

    for (i = 0; i < boxa->n; i++) {
        if ((boxc = boxaGetBox(boxa, i, copytype)) == NULL)
	    return (BOXA *)ERROR_PTR("boxc not made", procName, NULL);
	boxaAddBox(boxac, boxc, INSERT);
    }

    return boxac;
}


/*!
 *  boxaDestroy()
 *
 *       Input:   &boxa (<set to null>)
 *       Return:  void
 */
void
boxaDestroy(BOXA  **pboxa)
{
INT32   i;
BOXA   *boxa;

    PROCNAME("boxaDestroy");

    if (pboxa == NULL) {
        WARNING("ptr address is null!", procName);
	return;
    }

    if ((boxa = *pboxa) == NULL)
	return;

    for (i = 0; i < boxa->n; i++)
	boxDestroy(&boxa->box[i]);
    FREE((void *)boxa->box);

    FREE((void *)boxa);
    *pboxa = NULL;

    return;
}


/*!
 *  boxaAddBox()
 *
 *       Input:  boxa
 *               box  (to be added)
 *               copyflag (INSERT, COPY, CLONE)
 *       Return: 0 if OK, 1 on error
 */
INT32
boxaAddBox(BOXA   *boxa,
           BOX    *box,
           INT32   copyflag)
{
INT32   n;
BOX    *boxc;

    PROCNAME("boxaAddBox");

    if (!boxa)
	return ERROR_INT("boxa not defined", procName, 1);
    if (!box)
	return ERROR_INT("box not defined", procName, 1);
    if (copyflag != INSERT && copyflag != COPY && copyflag != CLONE)
	return ERROR_INT("copyflag must be in {0,1,2}", procName, 1);

    if (copyflag == INSERT)
	boxc = box;
    else if (copyflag == COPY) {
	if ((boxc = boxCopy(box)) == NULL)
	    return ERROR_INT("boxc not made", procName, 1);
    }
    else  /* CLONE */
	boxc = boxClone(box);

    n = boxaGetCount(boxa);
    if (n >= boxa->nalloc)
	boxaExtendArray(boxa);
    boxa->box[n] = boxc;
    boxa->n++;

    return 0;
}


/*!
 *  boxaExtendArray()
 *
 *       Input:  boxa
 *       Return: 0 if OK; 1 on error
 */
INT32
boxaExtendArray(BOXA   *boxa)
{

    PROCNAME("boxaExtendArray");

    if (!boxa)
	return ERROR_INT("boxa not defined", procName, 1);

    if ((boxa->box = (BOX **)reallocNew((void **)&boxa->box,
                                4 * boxa->nalloc,
				8 * boxa->nalloc)) == NULL)
	return ERROR_INT("new ptr array not returned", procName, 1);

    boxa->nalloc = 2 * boxa->nalloc;
    return 0;
}



/*---------------------------------------------------------------------*
 *                             Boxa accessors                          *
 *---------------------------------------------------------------------*/
/*!
 *  boxaGetCount()
 *
 *       Input:  boxa
 *       Return: count, or 0 on error
 */
INT32
boxaGetCount(BOXA   *boxa)
{

    PROCNAME("boxaGetCount");

    if (!boxa)
	return ERROR_INT("boxa not defined", procName, 0);

    return boxa->n;
}


/*!
 *  boxaGetBox()
 *
 *       Input:  boxa
 *               index  (to the index-th box)
 *               accessflag  (COPY or CLONE)
 *       Return: box, or null on error
 */
BOX *
boxaGetBox(BOXA   *boxa,
           INT32   index,
	   INT32   accessflag)
{
    PROCNAME("boxaGetBox");

    if (!boxa)
	return (BOX *)ERROR_PTR("boxa not defined", procName, NULL);
    if (index < 0 || index >= boxa->n)
	return (BOX *)ERROR_PTR("index not valid", procName, NULL);
    if (accessflag != COPY && accessflag != CLONE)
	return (BOX *)ERROR_PTR("accessflag must be in {1,2}", procName, NULL);

    if (accessflag == COPY)
	return boxCopy(boxa->box[index]);
    else  /* CLONE */
	return boxClone(boxa->box[index]);
}



/*---------------------------------------------------------------------*
 *                             Boxa properties                         *
 *---------------------------------------------------------------------*/
/*!
 *  boxaGetExtent()
 *
 *       Input:  boxa
 *               &w  (<return> width)
 *               &h  (<return> height)
 *       Return: 0 if OK, 1 on error
 *
 *       Note: the returned w and h are the minimum size image
 *             that would contain all boxes untranslated.
 */
INT32
boxaGetExtent(BOXA   *boxa,
              INT32  *pw,
              INT32  *ph)
{
INT32   i, n;
BOX    *box;

    PROCNAME("boxaGetExtent");

    if (!boxa)
	return ERROR_INT("boxa not defined", procName, 1);
    if (!pw || !ph)
	return ERROR_INT("ptrs not defined", procName, 1);

    *pw = *ph = 0;
    n = boxaGetCount(boxa);
    for (i = 0; i < n; i++) {
	box = boxaGetBox(boxa, i, CLONE);
	*pw = MAX(*pw, box->x + box->w);
	*ph = MAX(*ph, box->y + box->h);
	boxDestroy(&box);
    }

    return 0;
}



/*---------------------------------------------------------------------*
 *                        Boxa serialized I/O                          *
 *---------------------------------------------------------------------*/
/*!
 *  boxaWriteStream()
 *
 *       Input: stream
 *              boxa
 *       Return: 0 if OK, 1 on error
 */
INT32
boxaWriteStream(FILE   *fp,
	        BOXA   *boxa)
{
INT32    n, i;
BOX     *box;

    PROCNAME("boxaWriteStream");

    if (!fp)
	return ERROR_INT("stream not defined", procName, 1);
    if (!boxa)
	return ERROR_INT("boxa not defined", procName, 1);

    n = boxaGetCount(boxa);
    fprintf(fp, "\nboxa: number of boxes = %d\n", n);
    for (i = 0; i < n; i++) {
	if ((box = boxaGetBox(boxa, i, CLONE)) == NULL)
	    return ERROR_INT("box not found", procName, 1);
	fprintf(fp, "  Box[%d]: x = %d, y = %d, w = %d, h = %d\n",
	        i, box->x, box->y, box->w, box->h);
	boxDestroy(&box);
    }
    return 0;
}


/*!
 *  boxaReadStream()
 *
 *       Input:  stream
 *       Return: boxa, or null on error
 */
BOXA *
boxaReadStream(FILE   *fp)
{
INT32    n, i, x, y, w, h;
INT32    ignore;
BOX     *box;
BOXA    *boxa;

    PROCNAME("boxReadStream");

    if (!fp)
	return (BOXA *)ERROR_PTR("stream not defined", procName, NULL);

    if (fscanf(fp, "\nboxa: number of boxes = %d\n", &n) != 1)
	return (BOXA *)ERROR_PTR("not a boxa file", procName, NULL);
    if ((boxa = boxaCreate(n)) == NULL)
	return (BOXA *)ERROR_PTR("boxa not made", procName, NULL);
	
    for (i = 0; i < n; i++) {
	if (fscanf(fp, "  Box[%d]: x = %d, y = %d, w = %d, h = %d\n",
	        &ignore, &x, &y, &w, &h) != 5)
	    return (BOXA *)ERROR_PTR("box descr not valid", procName, NULL);
	if ((box = boxCreate(x, y, w, h)) == NULL)
	    return (BOXA *)ERROR_PTR("box not made", procName, NULL);
	boxaAddBox(boxa, box, INSERT);
    }

    return boxa;
}



/*---------------------------------------------------------------------*
 *                            Debug printing                           *
 *---------------------------------------------------------------------*/
/*!
 *  boxPrintStream()
 */
INT32
boxPrintStream(FILE   *fp,
	       BOX    *box)
{
    PROCNAME("boxPrintStream");

    if (!fp)
	return ERROR_INT("stream not defined", procName, 1);
    if (!box)
	return ERROR_INT("box not defined", procName, 1);

    fprintf(fp, " Box x (pixels) =           %d\n", box->x);
    fprintf(fp, " Box y (pixels) =           %d\n", box->y);
    fprintf(fp, " Box width (pixels) =       %d\n", box->w);
    fprintf(fp, " Box height (pixels) =      %d\n", box->h);

    return 0;
}


