/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/

/*
 *  colorquant.c
 *                     
 *      One-pass color quantization from 24 bit full color,
 *      with fixed partitioning and 256 colors only
 *          PIX        *pixColorQuant1Pass()
 *
 *      Two-pass octree color quantization from 24 bit full color,
 *      with adaptive tree and variable number of colors
 *          PIX        *pixOctreeColorQuant()
 *
 *        which calls
 *          CQCELL   ***octreeGenerateAndPrune()
 *          PIX        *pixOctreeQuantizePixels()
 *
 *        which calls
 *          INT32       octreeFindColorCell()
 *          
 *      Helper cqcell functions
 *          CQCELL   ***cqcellTreeCreate()
 *          void        cqcellTreeDestroy()
 *
 *      Helper index functions
 *          INT32       makeRGBToIndexTables()
 *          void        getRGBFromOctcube()
 *          INT32       getOctcubeIndices()
 *          void        getOctcubeIndexFromRGB()
 *
 *      Simple octcube quantization to arbitrary depth, with fixed partitions
 *          PIX        *pixSimpleOctcubeQuant()
 *          NUMA       *pixOctcubeHistogram()
 */

#include <stdio.h>
#include <stdlib.h>

#include "allheaders.h"


    /* in pixOctreeColorQuant() */
#define  SUBSAMPLE_FACTOR    0.25


#ifndef   NO_CONSOLE_IO
#define   DEBUG_OCTINDEX     0
#define   DEBUG_CMAP         0
#define   PRINT_STATISTICS   1
#endif   /* ~NO_CONSOLE_IO */


/*------------------------------------------------------------------*
 *                 Simple octree color quantization                 *
 *------------------------------------------------------------------*/
/*!
 *  pixColorQuant1Pass()
 *
 *      Input:  pixs  (32 bpp; 24-bit color)
 *              ditherflag  (1 for dithering; 0 for no dithering)
 *      Return: pixd (8 bit with colormap), or null on error
 *
 *      This simple 1-pass color quantization works by breaking the
 *      color space into 256 pieces, with 3 bits quantized for each of
 *      red and green, and 2 bits quantized for blue.  We shortchange
 *      blue because the eye is least sensitive to blue.  This
 *      division of the color space is into two levels of octrees,
 *      followed by a further division by 4 (not 8), where both
 *      blue octrees have been combined in the third level. 
 *
 *      The color map is generated from the 256 color centers by
 *      taking the representative color to be the center of the
 *      cell volume.  This gives a maximum error in the red and
 *      green values of 16 levels, and a maximum error in the
 *      blue sample of 32 levels. 
 *
 *      Each pixel in the 24-bit color image is placed in its containing
 *      cell, given by the relevant MSbits of the red, green and blue
 *      samples.  An error-diffusion dithering is performed on each
 *      color sample to give the appearance of good average local color.
 *      Dithering is required; without it, the contouring and visible
 *      color errors are very bad.
 *
 *      I originally implemented this algorithm in two passes,
 *      where the first pass was used to compute the weighted average
 *      of each sample in each pre-allocated region of color space.
 *      The idea was to use these centroids in the dithering algorithm
 *      of the second pass, to reduce the average error that was
 *      being dithered.  However, with dithering, there is
 *      virtually no difference, so there is no reason to make the
 *      first pass.  Consequently, this 1-pass version just assigns
 *      the pixels to the centers of the pre-allocated cells.
 *      We use dithering to spread the difference between the sample
 *      value and the location of the center of the cell.  For speed
 *      and simplicity, we use integer dithering and propagate only
 *      to the right, down, and diagonally down-right, with ratios
 *      3/8, 3/8 and 1/4, respectively.  The results should be nearly
 *      as good, and a bit faster, with propagation only to the right
 *      and down.
 * 
 *      The algorithm is very fast, because there is no search,
 *      only fast generation of the cell index for each pixel.
 *      We use a simple mapping from the three 8 bit rgb samples
 *      to the 8 bit cell index; namely, (r7 r6 r5 g7 g6 g5 b7 b6).
 *      This is not in an octcube format, but it doesn't matter.
 *      There are no storage requirements.  We could keep a
 *      running average of the center of each sample in each
 *      cluster, rather than using the center of the cell, but
 *      this is just extra work, esp. with dithering.
 *
 *      The method implemented here is very simple and fast, and
 *      gives surprisingly good results with dithering.
 *      However, without dithering, the loss of color accuracy is
 *      evident in regions that are very light or that have subtle
 *      blending of colors.
 *
 */
PIX *
pixColorQuant1Pass(PIX    *pixs,
                   INT32   ditherflag)
{
UINT8                rval, gval, bval;
UINT8               *bufu8r, *bufu8g, *bufu8b;
INT32               *buf1r, *buf1g, *buf1b, *buf2r, *buf2g, *buf2b;
INT32                val1, val2, val3, dif;
UINT8                index;
INT32                w, h, wpls, wpld, i, j, cindex;
UINT32              *ppixel;
UINT32              *datas, *datad, *lines, *lined;
GIFColorTableEntry  *gifcmap;
PIX                 *pixd;
PIXCMAP             *colormap;

    PROCNAME("pixColorQuant1Pass");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
	return (PIX *)ERROR_PTR("must be 24 bit color", procName, NULL);

	/* integerize centers to UINT8 and generate a gif colormap */
    if ((gifcmap =
	(GIFColorTableEntry *)CALLOC(sizeof(GIFColorTableEntry), 256)) == NULL)
	    return (PIX *)ERROR_PTR("gifcmap not made", procName, NULL);
    for (cindex = 0; cindex < 256; cindex++) {
	gifcmap[cindex].cmRed = (cindex & 0xe0) | 0x10;
	gifcmap[cindex].cmGreen = ((cindex << 3) & 0xe0) | 0x10;
	gifcmap[cindex].cmBlue = ((cindex << 6) & 0xc0) | 0x20;
    }
    if ((colormap = (PIXCMAP *)CALLOC(1, sizeof(PIXCMAP))) == NULL)
	return (PIX *)ERROR_PTR("colormap not made", procName, NULL);
    colormap->type = CM_GIF;
    colormap->array = gifcmap;
    colormap->size = 256 * sizeof(GIFColorTableEntry);

	/* make output 8 bpp palette image */
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    if ((pixd = pixCreate(w, h, 8)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    pixSetColormap(pixd, colormap);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

	/* set dest pix values to colortable indices */
    if (ditherflag == 0) {   /* no dithering */
	for (i = 0; i < h; i++) {
	    lines = datas + i * wpls;
	    lined = datad + i * wpld;
	    for (j = 0; j < w; j++) {
		ppixel = lines + j;
		rval = GET_DATA_BYTE(ppixel, 0);
		gval = GET_DATA_BYTE(ppixel, 1);
		bval = GET_DATA_BYTE(ppixel, 2);
		index = (rval & 0xe0) | ((gval >> 3) & 0x1c) | (bval >> 6);
		SET_DATA_BYTE(lined, j, index);
	    }
	}
    }
    else {  /* ditherflag == 1 */
	bufu8r = (UINT8 *)CALLOC(w, sizeof(UINT8));
	bufu8g = (UINT8 *)CALLOC(w, sizeof(UINT8));
	bufu8b = (UINT8 *)CALLOC(w, sizeof(UINT8));
	buf1r = (INT32 *)CALLOC(w, sizeof(INT32));
	buf1g = (INT32 *)CALLOC(w, sizeof(INT32));
	buf1b = (INT32 *)CALLOC(w, sizeof(INT32));
	buf2r = (INT32 *)CALLOC(w, sizeof(INT32));
	buf2g = (INT32 *)CALLOC(w, sizeof(INT32));
	buf2b = (INT32 *)CALLOC(w, sizeof(INT32));
	if (!bufu8r || !bufu8g || !bufu8b)
	    return (PIX *)ERROR_PTR("uint8 mono line buf not made",
	        procName, NULL);
	if (!buf1r || !buf1g || !buf1b || !buf2r || !buf2g || !buf2b)
	    return (PIX *)ERROR_PTR("mono line buf not made", procName, NULL);

	    /* start by priming buf2; line 1 is above line 2 */
	pixGetRGBLine(pixs, 0, bufu8r, bufu8g, bufu8b);
	for (j = 0; j < w; j++) {
	    buf2r[j] = 64 * bufu8r[j];
	    buf2g[j] = 64 * bufu8g[j];
	    buf2b[j] = 64 * bufu8b[j];
	}

	for (i = 0; i < h - 1; i++) {
		/* swap data 2 --> 1, and read in new line 2 */
	    memcpy(buf1r, buf2r, 4 * w);
	    memcpy(buf1g, buf2g, 4 * w);
	    memcpy(buf1b, buf2b, 4 * w);
	    pixGetRGBLine(pixs, i + 1, bufu8r, bufu8g, bufu8b);
	    for (j = 0; j < w; j++) {
		buf2r[j] = 64 * bufu8r[j];
		buf2g[j] = 64 * bufu8g[j];
		buf2b[j] = 64 * bufu8b[j];
	    }

		/* dither */
	    lined = datad + i * wpld;
	    for (j = 0; j < w - 1; j++) {
		rval = buf1r[j] / 64;
		gval = buf1g[j] / 64;
		bval = buf1b[j] / 64;
		index = (rval & 0xe0) | ((gval >> 3) & 0x1c) | (bval >> 6);
		SET_DATA_BYTE(lined, j, index);

		dif = buf1r[j] / 8 - 8 * ((rval | 0x10) & 0xf0);
		if (dif != 0) {
		    val1 = buf1r[j + 1] + 3 * dif;
		    val2 = buf2r[j] + 3 * dif;
		    val3 = buf2r[j + 1] + 2 * dif;
		    if (dif > 0) {
			buf1r[j + 1] = MIN(16383, val1);
			buf2r[j] = MIN(16383, val2);
			buf2r[j + 1] = MIN(16383, val3);
		    }
		    else if (dif < 0) {
			buf1r[j + 1] = MAX(0, val1);
			buf2r[j] = MAX(0, val2);
			buf2r[j + 1] = MAX(0, val3);
		    }
		}

		dif = buf1g[j] / 8 - 8 * ((gval | 0x10) & 0xf0);
		if (dif != 0) {
		    val1 = buf1g[j + 1] + 3 * dif;
		    val2 = buf2g[j] + 3 * dif;
		    val3 = buf2g[j + 1] + 2 * dif;
		    if (dif > 0) {
			buf1g[j + 1] = MIN(16383, val1);
			buf2g[j] = MIN(16383, val2);
			buf2g[j + 1] = MIN(16383, val3);
		    }
		    else if (dif < 0) {
			buf1g[j + 1] = MAX(0, val1);
			buf2g[j] = MAX(0, val2);
			buf2g[j + 1] = MAX(0, val3);
		    }
		}

		dif = buf1b[j] / 8 - 8 * ((bval | 0x20) & 0xe0);
		if (dif != 0) {
		    val1 = buf1b[j + 1] + 3 * dif;
		    val2 = buf2b[j] + 3 * dif;
		    val3 = buf2b[j + 1] + 2 * dif;
		    if (dif > 0) {
			buf1b[j + 1] = MIN(16383, val1);
			buf2b[j] = MIN(16383, val2);
			buf2b[j + 1] = MIN(16383, val3);
		    }
		    else if (dif < 0) {
			buf1b[j + 1] = MAX(0, val1);
			buf2b[j] = MAX(0, val2);
			buf2b[j + 1] = MAX(0, val3);
		    }
		}
	    }

		/* get last pixel in row; no downward propagation */
	    rval = buf1r[w - 1] / 64;
	    gval = buf1g[w - 1] / 64;
	    bval = buf1b[w - 1] / 64;
	    index = (rval & 0xe0) | ((gval >> 3) & 0x1c) | (bval >> 6);
	    SET_DATA_BYTE(lined, w - 1, index);
	}

	    /* get last row of pixels; no leftward propagation */
	lined = datad + (h - 1) * wpld;
	for (j = 0; j < w; j++) {
	    rval = buf2r[j] / 64;
	    gval = buf2g[j] / 64;
	    bval = buf2b[j] / 64;
	    index = (rval & 0xe0) | ((gval >> 3) & 0x1c) | (bval >> 6);
	    SET_DATA_BYTE(lined, j, index);
	}

	FREE((void *)bufu8r);
	FREE((void *)bufu8g);
	FREE((void *)bufu8b);
	FREE((void *)buf1r);
	FREE((void *)buf1g);
	FREE((void *)buf1b);
	FREE((void *)buf2r);
	FREE((void *)buf2g);
	FREE((void *)buf2b);
    }

    return pixd;
}



/*------------------------------------------------------------------*
 *                 Better octree color quantization                 *
 *------------------------------------------------------------------*/
/*!
 *  pixOctreeColorQuant()
 *
 *      Input:  pixs  (32 bpp; 24-bit color)
 *              colors  (in colormap; some number in range [32 ... 256];
 *                      the actual number of colors used will be smaller)
 *              ditherflag  (1 to dither, 0 otherwise)
 *              &actualcolors  (returns the number actually used)
 *      Return: pixd (8 bit with colormap), or null on error
 *
 *      I found one description in the literature of octree color
 *      quantization, using progressive truncation of the octree,
 *      by M. Gervautz and W. Purgathofer in Graphics Gems, pp.
 *      287-293, ed. A. Glassner, Academic Press, 1990.
 *      Rather than setting up a fixed partitioning of the color
 *      space ab initio, as we do here, they allow the octree to be
 *      progressively truncated as new pixels are added.  They
 *      need to set up some data structures that are traversed
 *      with the addition of each 24 bit pixel, in order to decide
 *      either (1) in which cluster (sub-branch of the octree) to put
 *      the pixel, or (2) whether to truncate the octree further
 *      to place the pixel in an existing cluster, or (3) which
 *      two existing clusters should be merged so that the pixel
 *      can be left to start a truncated leaf of the octree.  Such dynamic
 *      truncation is considerably more complicated, and Gervautz et
 *      al. did not explain how they did it in anywhere near the
 *      detail required to check their implementation.
 *
 *      The simple method in pixColorQuant1Pass() is very
 *      fast, and with dithering the results are good, but you
 *      can do better if the color clusters are selected adaptively
 *      from the image.  We want a method that makes much better
 *      use of color samples in regions of color space with high
 *      pixel density, while also fairly representing small numbers
 *      of color pixels in low density regions.  This will require
 *      two passes through the image: the first for generating the pruned
 *      tree of color cubes and the second for computing the index
 *      into the color table for each pixel.  We perform the first
 *      pass on a subsampled image, because we do not need to use
 *      all the pixels in the image to generate the tree.  Subsampling
 *      down to 0.25 (1/16 of the pixels) makes the program run
 *      about 1.3 times faster.
 *
 *      Instead of dividing the color space into 256 equal-sized
 *      regions, we initially divide it into 2^12 or 2^15 or 2^18
 *      equal-sized octcubes.  Suppose we choose to use 2^18 octcubes.  
 *      This gives us 6 octree levels.  We then prune back,
 *      starting from level 6.  For every cube at level 6, there
 *      are 8 cubes at level 5.  Call the operation of putting a
 *      cube aside as a color table entry (CTE) a "saving." 
 *      We use a (in general) level-dependent threshold, and save
 *      those level 6 cubes that are above threshold.
 *      The rest are combined into the containing level 5 cube.
 *      If between 1 and 7 level 6 cubes within a level 5
 *      cube have been saved by thresholding, then the remaining
 *      level 6 cubes in that level 5 cube are automatically
 *      saved as well, without applying a threshold.  This greatly
 *      simplifies both the description of the CTEs and the later
 *      classification of each pixel as belonging to a CTE.
 *      This procedure is iterated through every cube, starting at
 *      level 5, and then 4, 3, and 2, successively.  The result is that
 *      each CTE contains the entirety of a set of from 1 to 7 cubes
 *      from a given level that all belong to a single cube at the
 *      level above.   We classify the CTEs in terms of the
 *      condition in which they are made as either being "threshold"
 *      or "residual."  They are "threshold" CTEs if no subcubes
 *      are CTEs (that is, they contain every pixel within the cube)
 *      and the number of pixels exceeds the threshold for making 
 *      a CTE.  They are "residual" CTEs if at least one but not more
 *      than 7 of the subcubes have already been determined to be CTEs;
 *      this happens automatically -- no threshold is applied.
 *      If all 8 subcubes are determined to be CTEs, the cube is
 *      marked as having all pixels accounted for ('bleaf' = 1) but
 *      is not saved as a CTE.
 *
 *      We stop the pruning at level 2, at which there are 64
 *      sub-cubes.  Any pixels not already claimed in a CTE are
 *      put in these cubes.  
 *
 *      As the cubes are saved as color samples in the color table,
 *      the number of remaining pixels P and the number of
 *      remaining colors in the color table N are recomputed,
 *      along with the average number of pixels P/N (ppc) to go in
 *      each of the remaining colors.  This running average number is
 *      used to set the threshold at the current level.
 *
 *      Because we are going to very small cubes at levels 6 or 5,
 *      and will dither the colors for errors, it is not necessary 
 *      to compute the color center of each cluster; we can simply
 *      use the center of the cube.  This gives us a minimax error
 *      condition: the maximum error is half the width of the 
 *      level 2 cubes -- 32 color values out of 256 -- for each color
 *      sample.  In practice, most of the pixels will be very much
 *      closer to the center of their cells.  And with dithering,
 *      the average pixel color in a small region will be closer still.
 *      Thus with the octree quantizer, we are able to capture
 *      regions of high color pdf (probability density function) in small
 *      but accurate CTEs, and to have only a small number of pixels
 *      that end up a significant distance (with a guaranteed maximum)
 *      from their true color.
 *
 *      How should the threshold factor vary?  Threshold factors
 *      are required for levels 2, 3, 4 and 5 in the pruning stage.
 *      The threshold for level 5 is actually applied to cubes at
 *      level 6, etc.  From various experiments, it appears that
 *      the results do not vary appreciably for threshold values near 1.0.
 *      If you want more colors in smaller cubes, the threshold
 *      factors can be set lower than 1.0 for cubes at levels 4 and 5. 
 *      However, if the factor is set much lower than 1.0 for
 *      levels 2 and 3, we can easily run out of colors.
 *      We put aside 64 colors in the calculation of the threshold
 *      values, because we must have 64 color centers at level 2,
 *      that will have very few pixels in most of them.
 *      If we reduce the factor for level 5 to 0.4, this will
 *      generate many level 6 CTEs, and consequently
 *      many residual cells will be formed up from those leaves,
 *      resulting in the possibility of running out of colors.
 *      Remember, the residual CTEs are mandatory, and are formed
 *      without using the threshold, regardless of the number of
 *      pixels that are absorbed.
 *      
 *      The implementation logically has four parts:
 *
 *           (1) accumulation into small, fixed cells
 *           (2) pruning back into selected CTE cubes
 *           (3) organizing the CTEs for fast search to find
 *               the CTE to which any image pixel belongs
 *           (4) doing a second scan to code the image pixels
 *               by CTE
 *
 *      Step (1) is straightforward; we use 2^15 cells.
 *
 *      We've already discussed how the pruning step (2) will
 *      be performed. 
 *
 *      Steps (3) and (4) are related, in that the organization
 *      used by step (3) determines how the search actually
 *      takes place for each pixel in step (4).
 *
 *      There are many ways to do step (3).  Let's explore a few.
 *
 *      (a) The simplest is to order the cubes from highest occupancy
 *          to lowest, and traverse the list looking for the deepest
 *          match.  To make this more efficient, so that we know when
 *          to stop looking, any cube that has separate CTE subcubes
 *          would be marked as such, so that we know when we hit a 
 *          true leaf.
 *
 *      (b) Alternatively, we can order the cubes by highest
 *          occupancy separately each level, and work upward,
 *          starting at level 5, so that when we find a match we
 *          know that it will be correct. 
 *
 *      (c) Another approach would be to order the cubes by
 *          "address" and use a hash table to find the cube
 *          corresponding to a pixel color.  I don't know how to
 *          do this with a variable length address, as each CTE
 *          will have 3*n bits, where n is the level.
 *
 *      (d) Another approach entirely is to put the CTE cubes into
 *          a tree, in such a way that starting from the root, and
 *          using 3 bits of address at a time, the correct branch of
 *          each octree can be taken until a leaf is found.  Because
 *          a given cube can be both a leaf and also have branches
 *          going to sub-cubes, the search stops only when no
 *          marked subcubes have addresses that match the given pixel.
 *
 *          In the tree method, we can start with a dense infrastructure,
 *          and place the leaves corresponding to the N colors
 *          in the tree, or we can grow from the root only those
 *          branches that end directly on leaves.
 *
 *      What we do here is to take approach (d), and implement the tree
 *      "virtually", as a set of arrays, one array for each level
 *      of the tree.   Initially we start at level 5, an array with
 *      2^15 cubes, each with 8 subcubes.  We then build nodes at
 *      levels closer to the root; at level 4 there are 2^12 nodes
 *      each with 8 subcubes; etc.  Using these arrays has
 *      several advantages:
 *
 *         -  We don't need to keep track of links between cubes
 *            and subcubes, because we can use the canonical
 *            addressing on the cell arrays directly to determine
 *            which nodes are parent cubes and which are sub-cubes.
 *
 *         -  We can prune directly on this tree
 *
 *         -  We can navigate the pruned tree quickly to classify
 *            each pixel in the image.
 *      
 *      Canonical addressing guarantees that the i-th node at level k
 *      has 8 subnodes given by the 8*i ... 8*i+7 nodes at level k+1.
 *
 *      The pruning step works as follows.  We go from the lowest
 *      level up.  At each level, the threshold is found from the
 *      product of a factor near 1.0 and the ratio of unmarked pixels
 *      to remaining colors (minus the 64).  We march through
 *      the space, sequentially considering a cube and its 8 subcubes.
 *      We first check those subcubes that are not already
 *      marked as CTE to see if any are above threshold, and if so,
 *      generate a CTE and mark them as such.
 *      We then determine if any of the subcubes have been marked.
 *      If so, and there are subcubes that are not marked,
 *      we generate a CTE for the cube from the remaining unmarked
 *      subcubes; this is mandatory and does not depend on how many
 *      pixels are in the set of subcubes.  If none of the subcubes
 *      are marked, we aggregate their pixels into the cube
 *      containing them, but do not mark it as a CTE; that
 *      will be determined when iterating through the next level up.
 *
 *      When all the pixels in a cube are accounted for in one or more
 *      colors, we set the boolean 'bleaf' to true.  This is the
 *      flag used to mark the cubes in the pruning step.  If a cube
 *      is marked, and all 8 subcubes are marked, then it is not
 *      itself given a CTE because all pixels have already been
 *      accounted for.
 *
 *      Note that the pruning of the tree and labelling of the CTEs
 *      (step 2) accomplishes step 3 implicitly, because the marked
 *      and pruned tree is ready for use in labelling each pixel
 *      in step 4.  We now, for every pixel in the image, traverse
 *      the tree from the root, looking for the lowest cube that is a leaf.
 *      At each level we have a cube and subcube.  If we reach a subcube
 *      leaf that is marked 0, we know that the color is stored in the
 *      cube above, and we've found the CTE.  Otherwise, the subcube
 *      leaf is marked 1.  If we're at the last level, we've reached
 *      the final leaf and must use it.  Otherwise, continue the
 *      process at the next level down.
 */
PIX *
pixOctreeColorQuant(PIX    *pixs,
                    INT32   colors,
		    INT32   ditherflag,
		    INT32  *pactualcolors)
{
INT32                actualcolors, extrareserved;
CQCELL            ***cqcaa;
GIFColorTableEntry  *gifcmap;
PIX                 *pixd, *pixsub;
PIXCMAP             *colormap;

    PROCNAME("pixOctreeColorQuant");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
	return (PIX *)ERROR_PTR("must be 24 bit color", procName, NULL);
    if (colors < 128 || colors > 256)
	return (PIX *)ERROR_PTR("colors must be in [128, 256]", procName, NULL);

	/* subsample to speed up the first pass */
    if ((pixsub = pixScaleBySampling(pixs, SUBSAMPLE_FACTOR, SUBSAMPLE_FACTOR))
	    == NULL)
	return (PIX *)ERROR_PTR("pixsub not made", procName, NULL);

	/* make the pruned octree */
    cqcaa = octreeGenerateAndPrune(pixsub, colors, CQ_RESERVED_COLORS,
                                   &gifcmap, &actualcolors);
    if (!cqcaa)
	return (PIX *)ERROR_PTR("tree not made", procName, NULL);
    fprintf(stderr, " Colors requested = %d, actual colors = %d\n",
           colors, actualcolors);

	/* traverse tree from root, looking for lowest cube
	 * that is a leaf, and set dest pix value to its 
	 * colortable index */
    if ((pixd = pixOctreeQuantizePixels(pixs, cqcaa, ditherflag)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

	/* make and attach the colormap */
    if ((colormap = (PIXCMAP *)CALLOC(1, sizeof(PIXCMAP))) == NULL)
	return (PIX *)ERROR_PTR("colormap not made", procName, NULL);
    colormap->type = CM_GIF;
    colormap->array = gifcmap;
    colormap->size = actualcolors * sizeof(GIFColorTableEntry);
    pixSetColormap(pixd, colormap);

    cqcellTreeDestroy(&cqcaa);
    pixDestroy(&pixsub);
    return pixd;
}


/*!
 *  octreeGenerateAndPrune()
 *
 *      Input:  pixs
 *              number of colors to use (between 128 and 256)
 *              number of reserved colors
 *              &gifcmap  (made and returned)
 *              &actualcolors (may differ from the number of colors
 *                             requested, but will not be larger than 256)
 *
 *      Return: octree, colormap and number of colors used, or null
 *              on error
 */
CQCELL ***
octreeGenerateAndPrune(PIX                 *pixs,
                       INT32                colors,
		       INT32                reservedcolors,
		       GIFColorTableEntry **pgifcmap,
		       INT32               *pactualcolors)
{
UINT8                rval, gval, bval;
UINT32              *rtab, *gtab, *btab;
INT32                level, ncells, rgbindex;
INT32                numleaves;  /* number assigned; color table index */
INT32                w, h, wpls, wpld;
INT32                i, j, isub;
INT32                npix;  /* number of remaining pixels to be assigned */
INT32                ncolor; /* number of remaining color cells to be used */
INT32                ppc;  /* ave number of pixels left for each color cell */
INT32                tc[] = {0, 0, 0, 0, 0, 0, 0};
INT32                rc[] = {0, 0, 0, 0, 0, 0, 0};
INT32                nt, nr, ival;
INT32                rv, gv, bv;
FLOAT32              thresholdFactor[] = {0.01, 0.01, 1.0, 1.0, 1.0, 1.0};
FLOAT32              thresh;  /* factor of ppc for this level */
UINT32              *ppixel;
UINT32              *datas, *lines;
CQCELL            ***cqcaa;   /* one array for each octree level */
CQCELL             **cqca, **cqcasub;
CQCELL              *cqc, *cqcsub;
GIFColorTableEntry  *gifcmap;
NUMA                *nat;  /* accumulates levels for threshold cells */
NUMA                *nar;  /* accumulates levels for residual cells */

    PROCNAME("octreeGenerateAndPrune");

    if (!pixs)
	return (CQCELL ***)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
	return (CQCELL ***)ERROR_PTR("must be 24 bit color", procName, NULL);
    if (colors < 128 || colors > 256)
	return (CQCELL ***)ERROR_PTR("colors not in [128, 256]", procName, NULL);
    if (!pgifcmap)
	return (CQCELL ***)ERROR_PTR("&gifcmap not defined", procName, NULL);
    if (!pactualcolors)
	return (CQCELL ***)ERROR_PTR("&actualcolors not defined", procName, NULL);

	/* make the canonical index tables */
    if (makeRGBToIndexTables(&rtab, &gtab, &btab, CQ_NLEVELS))
	return (CQCELL ***)ERROR_PTR("tables not made", procName, NULL);

    if ((cqcaa = cqcellTreeCreate()) == NULL)
	return (CQCELL ***)ERROR_PTR("cqcaa not made", procName, NULL);

	/* generate a gif CTE array of size 256 */
    if ((gifcmap =
	(GIFColorTableEntry *)CALLOC(sizeof(GIFColorTableEntry), 256))
		== NULL)
	    return (CQCELL ***)ERROR_PTR("gifcmap not made", procName, NULL);
    *pgifcmap = gifcmap;

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    npix = w * h;  /* initialize to all pixels */
    ncolor = colors - reservedcolors;  /* init to (almost) all colors */
    ppc = npix / ncolor;
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

	/* accumulate the centers of each cluster at level CQ_NLEVELS */
    ncells = 1 << (3 * CQ_NLEVELS);
    cqca = cqcaa[CQ_NLEVELS];
    for (i = 0; i < h; i++) {
	lines = datas + i * wpls;
	for (j = 0; j < w; j++) {
	    ppixel = lines + j;
	    rval = GET_DATA_BYTE(ppixel, 0);
	    gval = GET_DATA_BYTE(ppixel, 1);
	    bval = GET_DATA_BYTE(ppixel, 2);
	    rgbindex = rtab[rval] | gtab[gval] | btab[bval];
	    cqc = cqca[rgbindex];
	    cqc->n++;
	}
    }

	/* arrays for storing statistics */
    if ((nat = numaCreate(0)) == NULL)
	return (CQCELL ***)ERROR_PTR("nat not made", procName, NULL);
    if ((nar = numaCreate(0)) == NULL)
	return (CQCELL ***)ERROR_PTR("nar not made", procName, NULL);

	/* prune back from the lowest level and generate the colormap */
    numleaves = 0;
    for (level = CQ_NLEVELS - 1; level >= 2; level--) {
	thresh = thresholdFactor[level];
	cqca = cqcaa[level];
	cqcasub = cqcaa[level + 1];
	ncells = 1 << (3 * level);
	for (i = 0; i < ncells; i++) {  /* i is rgbindex at level */
	    cqc = cqca[i];
	    for (j = 0; j < 8; j++) {  /* check all subnodes */
		isub = 8 * i + j;   /* isub is rgbindex at level+1 */
		cqcsub = cqcasub[isub];
		if (cqcsub->bleaf == 1) {  /* already a leaf? */
		    cqc->nleaves++;   /* count the subcube leaves */
		    continue;
		}
		if (cqcsub->n >= thresh * ppc) {  /* make it a true leaf? */
		    cqcsub->bleaf = 1;
		    if (numleaves < 256) {
			cqcsub->index = numleaves;  /* assign the color index */
			getRGBFromOctcube(isub, level + 1, &rv, &gv, &bv); 
			gifcmap[numleaves].cmRed = (UINT8)rv;
			gifcmap[numleaves].cmGreen = (UINT8)gv;
			gifcmap[numleaves].cmBlue = (UINT8)bv;
#if 1   /* save values */
			cqcsub->rc = rv;
			cqcsub->gc = gv;
			cqcsub->bc = bv;
#endif
		    }
		    else {
			WARNING("possibly assigned pixels to wrong color",
			        procName);
			cqcsub->index = 255;  /* assign to the last; bad */
			cqcsub->rc = gifcmap[255].cmRed;
			cqcsub->gc = gifcmap[255].cmGreen;
			cqcsub->bc = gifcmap[255].cmBlue;
		    }
		    numleaves++;
		    cqc->nleaves++;
		    npix -= cqcsub->n;
		    ncolor--;
		    if (ncolor > 0)
			ppc = npix / ncolor;
		    else if (ncolor + reservedcolors > 0)
		        ppc = npix / (ncolor + reservedcolors);
		    else
		        ppc = 1000000;  /* make it big */
		    numaAddNumber(nat, level + 1);

#if  DEBUG_CMAP
    fprintf(stderr, "Exceeds threshold: colors used = %d, colors remaining = %d\n",
                     numleaves, ncolor + reservedcolors);
    fprintf(stderr, "  cell with %d pixels, npix = %d, ppc = %d\n",
                     cqcsub->n, npix, ppc);
    fprintf(stderr, "  index = %d, level = %d, subindex = %d\n",
                     i, level, j);
    fprintf(stderr, "  rv = %d, gv = %d, bv = %d\n", rv, gv, bv);
#endif  /* DEBUG_CMAP */

		}
	    }
	    if (cqc->nleaves > 0 || level == 2) { /* make the cube a leaf now */
		cqc->bleaf = 1;
		if (cqc->nleaves < 8) {  /* residual CTE cube: acquire the
		                          * remaining pixels */
		    for (j = 0; j < 8; j++) {  /* check all subnodes */
			isub = 8 * i + j;
			cqcsub = cqcasub[isub];
			if (cqcsub->bleaf == 0)  /* absorb */
			    cqc->n += cqcsub->n;
		    }
		    if (numleaves < 256) {
			getRGBFromOctcube(i, level, &rv, &gv, &bv); 
			cqc->index = numleaves;  /* assign the color index */
			gifcmap[numleaves].cmRed = (UINT8)rv;
			gifcmap[numleaves].cmGreen = (UINT8)gv;
			gifcmap[numleaves].cmBlue = (UINT8)bv;
#if 1   /* save values */
			cqc->rc = rv;
			cqc->gc = gv;
			cqc->bc = bv;
#endif
		    }
		    else {
			WARNING("possibly assigned pixels to wrong color",
			        procName);
			cqc->index = 255;  /* assign to the last */
			cqc->rc = gifcmap[255].cmRed;
			cqc->gc = gifcmap[255].cmGreen;
			cqc->bc = gifcmap[255].cmBlue;
		    }
		    numleaves++;
		    npix -= cqc->n;
		    ncolor--;
		    if (ncolor > 0)
			ppc = npix / ncolor;
		    else if (ncolor + reservedcolors > 0)
		        ppc = npix / (ncolor + reservedcolors);
		    else
		        ppc = 1000000;  /* make it big */
		    numaAddNumber(nar, level);

#if  DEBUG_CMAP
    fprintf(stderr, "By remainder: colors used = %d, colors remaining = %d\n",
                     numleaves, ncolor + reservedcolors);
    fprintf(stderr, "  cell with %d pixels, npix = %d, ppc = %d\n",
                     cqc->n, npix, ppc);
    fprintf(stderr, "  index = %d, level = %d\n", i, level);
    fprintf(stderr, "  rv = %d, gv = %d, bv = %d\n", rv, gv, bv);
#endif  /* DEBUG_CMAP */

		}
	    }
	    else {  /* absorb all the subpixels but don't make it a leaf */
		for (j = 0; j < 8; j++) {  /* absorb from all subnodes */
		    isub = 8 * i + j;
		    cqcsub = cqcasub[isub];
		    cqc->n += cqcsub->n;
		}
	    }
	}
    }

    *pactualcolors = MIN(256, numleaves);

#if  PRINT_STATISTICS
    nt = numaGetCount(nat);
    nr = numaGetCount(nar);
    for (i = 0; i < nt; i++) {
	numaGetIValue(nat, i, &ival);
	tc[ival]++;
    }
    for (i = 0; i < nr; i++) {
	numaGetIValue(nar, i, &ival);
	rc[ival]++;
    }
    fprintf(stderr, " Threshold cells formed: %d\n", nt);
    for (i = 1; i < CQ_NLEVELS + 1; i++)
	fprintf(stderr, "   level %d:  %d\n", i, tc[i]);
    fprintf(stderr, "\n Residual cells formed: %d\n", nr);
    for (i = 0; i < CQ_NLEVELS ; i++)
	fprintf(stderr, "   level %d:  %d\n", i, rc[i]);
#endif  /* PRINT_STATISTICS */

    numaDestroy(&nat);
    numaDestroy(&nar);
    FREE((void *)rtab);
    FREE((void *)gtab);
    FREE((void *)btab);

    return cqcaa;
}


/*!
 *  pixOctreeQuantizePixels()
 *
 *     Input:  pixs
 *             octree in array format
 *             ditherflag (1 for dithering, 0 for no dithering)
 *     Return: pixd or null on error
 *
 *     Note: This routine doesn't need to use the CTEs (colormap
 *           table entries) because the color indices are embedded
 *           in the octree.  Thus, the calling program must make
 *           and attach the colormap to pixd after it is returned.
 *
 *     Dithering is performed in integers, effectively rounding
 *     to 1/8 sample increment.  The data in the integer buffers is
 *     64 times the sample values.  The 'dif' is 8 times the
 *     sample values, and this spread, multiplied by 8, to the
 *     integer buffers.  Because the dif is truncated to an
 *     integer, the dither is accurate to 1/8 of a sample increment,
 *     or 1/2048 of the color range.
 */
PIX *
pixOctreeQuantizePixels(PIX       *pixs,
                        CQCELL  ***cqcaa,
			INT32      ditherflag)
{
UINT8             rval, gval, bval;
UINT8            *bufu8r, *bufu8g, *bufu8b;
INT32            *buf1r, *buf1g, *buf1b, *buf2r, *buf2g, *buf2b;
INT32             rgbindex, index;
INT32             val1, val2, val3, dif;
INT32             w, h, wpls, wpld, i, j;
INT32             rc, gc, bc;
UINT32           *rtab, *gtab, *btab;
UINT32           *ppixel;
UINT32           *datas, *datad, *lines, *lined;
PIX              *pixd;

    PROCNAME("pixOctreeQuantizePixels");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
	return (PIX *)ERROR_PTR("must be 24 bit color", procName, NULL);
    if (!cqcaa)
	return (PIX *)ERROR_PTR("cqcaa not defined", procName, NULL);

	/* make the canonical index tables */
    if (makeRGBToIndexTables(&rtab, &gtab, &btab, CQ_NLEVELS))
	return (PIX *)ERROR_PTR("tables not made", procName, NULL);

	/* make output 8 bpp palette image */
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    if ((pixd = pixCreate(w, h, 8)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

	/* Traverse tree from root, looking for lowest cube
	 * that is a leaf, and set dest pix to its 
	 * colortable index value.  The results are far
	 * better when dithering to get a more accurate
	 * average color.  */
    if (ditherflag == 0) {    /* no dithering */
	for (i = 0; i < h; i++) {
	    lines = datas + i * wpls;
	    lined = datad + i * wpld;
	    for (j = 0; j < w; j++) {
		ppixel = lines + j;
		rval = GET_DATA_BYTE(ppixel, 0);
		gval = GET_DATA_BYTE(ppixel, 1);
		bval = GET_DATA_BYTE(ppixel, 2);
		rgbindex = rtab[rval] | gtab[gval] | btab[bval];
		octreeFindColorCell(rgbindex, cqcaa, &index, &rc, &gc, &bc);
		SET_DATA_BYTE(lined, j, index);
	    }
	}
    }
    else {  /* dither */
	bufu8r = (UINT8 *)CALLOC(w, sizeof(UINT8));
	bufu8g = (UINT8 *)CALLOC(w, sizeof(UINT8));
	bufu8b = (UINT8 *)CALLOC(w, sizeof(UINT8));
	buf1r = (INT32 *)CALLOC(w, sizeof(INT32));
	buf1g = (INT32 *)CALLOC(w, sizeof(INT32));
	buf1b = (INT32 *)CALLOC(w, sizeof(INT32));
	buf2r = (INT32 *)CALLOC(w, sizeof(INT32));
	buf2g = (INT32 *)CALLOC(w, sizeof(INT32));
	buf2b = (INT32 *)CALLOC(w, sizeof(INT32));
	if (!bufu8r || !bufu8g || !bufu8b)
	    return (PIX *)ERROR_PTR("uint8 mono line buf not made",
	        procName, NULL);
	if (!buf1r || !buf1g || !buf1b || !buf2r || !buf2g || !buf2b)
	    return (PIX *)ERROR_PTR("mono line buf not made", procName, NULL);

	    /* start by priming buf2; line 1 is above line 2 */
	pixGetRGBLine(pixs, 0, bufu8r, bufu8g, bufu8b);
	for (j = 0; j < w; j++) {
	    buf2r[j] = 64 * bufu8r[j];
	    buf2g[j] = 64 * bufu8g[j];
	    buf2b[j] = 64 * bufu8b[j];
	}

	for (i = 0; i < h - 1; i++) {
		/* swap data 2 --> 1, and read in new line 2 */
	    memcpy(buf1r, buf2r, 4 * w);
	    memcpy(buf1g, buf2g, 4 * w);
	    memcpy(buf1b, buf2b, 4 * w);
	    pixGetRGBLine(pixs, i + 1, bufu8r, bufu8g, bufu8b);
	    for (j = 0; j < w; j++) {
		buf2r[j] = 64 * bufu8r[j];
		buf2g[j] = 64 * bufu8g[j];
		buf2b[j] = 64 * bufu8b[j];
	    }

		/* dither */
	    lined = datad + i * wpld;
	    for (j = 0; j < w - 1; j++) {
		rval = buf1r[j] / 64;
		gval = buf1g[j] / 64;
		bval = buf1b[j] / 64;
		rgbindex = rtab[rval] | gtab[gval] | btab[bval];
		octreeFindColorCell(rgbindex, cqcaa, &index, &rc, &gc, &bc);
		SET_DATA_BYTE(lined, j, index);

		dif = buf1r[j] / 8 - 8 * rc;
		if (dif != 0) {
		    val1 = buf1r[j + 1] + 3 * dif;
		    val2 = buf2r[j] + 3 * dif;
		    val3 = buf2r[j + 1] + 2 * dif;
		    if (dif > 0) {
			buf1r[j + 1] = MIN(16383, val1);
			buf2r[j] = MIN(16383, val2);
			buf2r[j + 1] = MIN(16383, val3);
		    }
		    else if (dif < 0) {
			buf1r[j + 1] = MAX(0, val1);
			buf2r[j] = MAX(0, val2);
			buf2r[j + 1] = MAX(0, val3);
		    }
		}

		dif = buf1g[j] / 8 - 8 * gc;
		if (dif != 0) {
		    val1 = buf1g[j + 1] + 3 * dif;
		    val2 = buf2g[j] + 3 * dif;
		    val3 = buf2g[j + 1] + 2 * dif;
		    if (dif > 0) {
			buf1g[j + 1] = MIN(16383, val1);
			buf2g[j] = MIN(16383, val2);
			buf2g[j + 1] = MIN(16383, val3);
		    }
		    else if (dif < 0) {
			buf1g[j + 1] = MAX(0, val1);
			buf2g[j] = MAX(0, val2);
			buf2g[j + 1] = MAX(0, val3);
		    }
		}

		dif = buf1b[j] / 8 - 8 * bc;
		if (dif != 0) {
		    val1 = buf1b[j + 1] + 3 * dif;
		    val2 = buf2b[j] + 3 * dif;
		    val3 = buf2b[j + 1] + 2 * dif;
		    if (dif > 0) {
			buf1b[j + 1] = MIN(16383, val1);
			buf2b[j] = MIN(16383, val2);
			buf2b[j + 1] = MIN(16383, val3);
		    }
		    else if (dif < 0) {
			buf1b[j + 1] = MAX(0, val1);
			buf2b[j] = MAX(0, val2);
			buf2b[j + 1] = MAX(0, val3);
		    }
		}
	    }

		/* get last pixel in row; no downward propagation */
	    rval = buf1r[w - 1] / 64;
	    gval = buf1g[w - 1] / 64;
	    bval = buf1b[w - 1] / 64;
	    rgbindex = rtab[rval] | gtab[gval] | btab[bval];
	    octreeFindColorCell(rgbindex, cqcaa, &index, &rc, &gc, &bc);
	    SET_DATA_BYTE(lined, w - 1, index);
	}

	    /* get last row of pixels; no leftward propagation */
	lined = datad + (h - 1) * wpld;
	for (j = 0; j < w; j++) {
	    rval = buf2r[j] / 64;
	    gval = buf2g[j] / 64;
	    bval = buf2b[j] / 64;
	    rgbindex = rtab[rval] | gtab[gval] | btab[bval];
	    octreeFindColorCell(rgbindex, cqcaa, &index, &rc, &gc, &bc);
	    SET_DATA_BYTE(lined, j, index);
	}

	FREE((void *)bufu8r);
	FREE((void *)bufu8g);
	FREE((void *)bufu8b);
	FREE((void *)buf1r);
	FREE((void *)buf1g);
	FREE((void *)buf1b);
	FREE((void *)buf2r);
	FREE((void *)buf2g);
	FREE((void *)buf2b);
    }

	/* cleanup */
    FREE((void *)rtab);
    FREE((void *)gtab);
    FREE((void *)btab);

    return pixd;
}


/*!
 *  octreeFindColorCell()
 *
 *    Input:  rgbindex 
 *            cqcaa
 *            &index   (<return> index of CTE; returned to set pixel value)
 *            &rval    (<return> of CTE)
 *            &gval    (<return> of CTE)
 *            &bval    (<return> of CTE)
 *    Return: 0 if OK; 1 on error
 *
 *    Note: as this is in inner loop, we don't check input pointers!
 *
 *    Action: traverses from root (well, actually from level 2, because
 *            the level 2 cubes are the largest CTE cubes), and finds
 *            the index number of the cell and the color values,
 *            which can be used either directly or in a (Floyd-Steinberg)
 *            error-diffusion dithering algorithm.
 */
INT32
octreeFindColorCell(INT32      rgbindex,
                    CQCELL  ***cqcaa,
		    INT32     *pindex,
		    INT32     *prval,
		    INT32     *pgval,
		    INT32     *pbval)
{
INT32             level;
INT32             baseindex, subindex;
INT32             rv, gv, bv;  /* use only for generating on the fly */
CQCELL           *cqc, *cqcsub;

        /* use rgb values stored in the cubes; a little faster */
    for (level = 2; level < CQ_NLEVELS; level++) {
	getOctcubeIndices(rgbindex, level, &baseindex, &subindex);
	cqc = cqcaa[level][baseindex];
	cqcsub = cqcaa[level + 1][subindex];
	if (cqcsub->bleaf == 0) {  /* use cell at level above */
	    *pindex = cqc->index;
	    *prval = cqc->rc;
	    *pgval = cqc->gc;
	    *pbval = cqc->bc;
	    break;
	}
	else if (level == CQ_NLEVELS - 1) {  /* reached the bottom */
	   *pindex = cqcsub->index;
	   *prval = cqcsub->rc;
	   *pgval = cqcsub->gc;
	   *pbval = cqcsub->bc;
	    break;
	}
    }

#if 0
        /* generate rgb values for each cube on the fly; slower */
    for (level = 2; level < CQ_NLEVELS; level++) {
	getOctcubeIndices(rgbindex, level, &baseindex, &subindex);
	cqc = cqcaa[level][baseindex];
	cqcsub = cqcaa[level + 1][subindex];
	if (cqcsub->bleaf == 0) {  /* use cell at level above */
	    getRGBFromOctcube(baseindex, level, &rv, &gv, &bv);
	    *pindex = cqc->index;
	    *prval = rv;
	    *pgval = gv;
	    *pbval = bv;
	    break;
	}
	else if (level == CQ_NLEVELS - 1) {  /* reached the bottom */
	    getRGBFromOctcube(subindex, level + 1, &rv, &gv, &bv);
	   *pindex = cqcsub->index;
	    *prval = rv;
	    *pgval = gv;
	    *pbval = bv;
	    break;
	}
    }
#endif

    return 0;
}



/*------------------------------------------------------------------*
 *                      Helper cqcell functions                     *
 *------------------------------------------------------------------*/
/*!
 *  cqcellTreeCreate()
 *
 *       Input: none
 *       Return: cqcell array tree
 */
CQCELL ***
cqcellTreeCreate(void)
{
INT32          level, ncells, i;
CQCELL      ***cqcaa;  
CQCELL       **cqca;   /* one array for each octree level */

    PROCNAME("cqcellTreeCreate");

	/* make array of accumulation cell arrays from levels 1 to 5 */
    if ((cqcaa = (CQCELL ***)CALLOC(CQ_NLEVELS + 1, sizeof(CQCELL **))) == NULL)
	return (CQCELL ***)ERROR_PTR("cqcaa not made", procName, NULL);
    for (level = 0; level <= CQ_NLEVELS; level++) {
	ncells = 1 << (3 * level);
	if ((cqca = (CQCELL **)CALLOC(ncells, sizeof(CQCELL *))) == NULL)
	    return (CQCELL ***)ERROR_PTR("cqca not made", procName, NULL);
	cqcaa[level] = cqca;
	for (i = 0; i < ncells; i++) {
	    if ((cqca[i] = (CQCELL *)CALLOC(1, sizeof(CQCELL))) == NULL)
		return (CQCELL ***)ERROR_PTR("cqc not made", procName, NULL);
	}
    }

    return cqcaa;
}


/*!
 *  cqcellTreeDestroy()
 *
 *     Input:  &cqcaa (<to be nulled>
 *     Return: void
 */
void
cqcellTreeDestroy(CQCELL ****pcqcaa)
{
INT32        level, ncells, i;
CQCELL    ***cqcaa;
CQCELL     **cqca;

    PROCNAME("cqcellTreeDestroy");

    if (pcqcaa == NULL) {
	WARNING("ptr address is NULL", procName);
	return;
    }

    if ((cqcaa = *pcqcaa) == NULL)
	return;

    for (level = 0; level <= CQ_NLEVELS; level++) {
	cqca = cqcaa[level];
	ncells = 1 << (3 * level);
	for (i = 0; i < ncells; i++)
	    FREE((void *)cqca[i]);
	FREE((void *)cqca);
    }
    FREE((void *)cqcaa);
    *pcqcaa = NULL;

    return;
}



/*------------------------------------------------------------------*
 *                       Helper index functions                     *
 *------------------------------------------------------------------*/
/*!
 *  makeRGBToIndexTables()
 *
 *     Input:  &rtab, &gtab, &btab  (<return> tables)
 *             cqlevels (can be 2, 3, 4, 5 or 6)
 *     Return: 0 if OK; 1 on error
 *
 *     Set up tables.  e.g., for cqlevels = 5, we need an integer
 *             0 < i < 2^15:
 *        rtab = (0  i7  0   0  i6  0   0  i5  0   0   i4  0   0   i3  0   0)
 *        gtab = (0  0   i7  0   0  i6  0   0  i5  0   0   i4  0   0   i3  0)
 *        btab = (0  0   0   i7  0  0   i6  0  0   i5  0   0   i4  0   0   i3)
 *
 *     The tables are then used to map from rbg --> index as follows:
 *       index = (0  r7  g7  b7  r6  g6  b6  r5  g5  b5  r4  g4  b4  r3  g3  b3)
 *
 *     e.g., for cqlevels = 4, we map to
 *       index = (0  0   0   0   r7  g7  b7  r6  g6  b6  r5  g5  b5  r4  g4  b4)
 *
 */
INT32
makeRGBToIndexTables(UINT32  **prtab,
                     UINT32  **pgtab,
                     UINT32  **pbtab,
		     INT32     cqlevels)
{
INT32      i;
UINT32    *rtab, *gtab, *btab;

    PROCNAME("makeRGBToIndexTables");

    if (cqlevels < 2 || cqlevels > 6)
	return ERROR_INT("cqlevels must be in {2,...6}", procName, 1);

    if (!prtab || !pgtab || !pbtab)
	return ERROR_INT("&*tab not defined", procName, 1);
    if ((rtab = (UINT32 *)CALLOC(256, sizeof(UINT32))) == NULL)
	return ERROR_INT("rtab not made", procName, 1);
    if ((gtab = (UINT32 *)CALLOC(256, sizeof(UINT32))) == NULL)
	return ERROR_INT("gtab not made", procName, 1);
    if ((btab = (UINT32 *)CALLOC(256, sizeof(UINT32))) == NULL)
	return ERROR_INT("btab not made", procName, 1);
    *prtab = rtab;
    *pgtab = gtab;
    *pbtab = btab;
	
    switch (cqlevels)
    {
    case 2:
	for (i = 0; i < 256; i++) {
	    rtab[i] = ((i >> 2) & 0x0020) | ((i >> 4) & 0x0004);
	    gtab[i] = ((i >> 3) & 0x0010) | ((i >> 5) & 0x0002);
	    btab[i] = ((i >> 4) & 0x0008) | ((i >> 6) & 0x0001);
	}
	break;
    case 3:
	for (i = 0; i < 256; i++) {
	    rtab[i] = ((i << 1) & 0x0100) | ((i >> 1) & 0x0020) |
		      ((i >> 3) & 0x0004);
	    gtab[i] = (i & 0x0080) | ((i >> 2) & 0x0010) |
		      ((i >> 4) & 0x0002);
	    btab[i] = ((i >> 1) & 0x0040) | ((i >> 3) & 0x0008) |
		      ((i >> 5) & 0x0001);
	}
	break;
    case 4:
	for (i = 0; i < 256; i++) {
	    rtab[i] = ((i << 4) & 0x0800) | ((i << 2) & 0x0100) |
		      (i & 0x0020) | ((i >> 2) & 0x0004);
	    gtab[i] = ((i << 3) & 0x0400) | ((i << 1) & 0x0080) |
		      ((i >> 1) & 0x0010) | ((i >> 3) & 0x0002);
	    btab[i] = ((i << 2) & 0x0200) | (i & 0x0040) |
		      ((i >> 2) & 0x0008) | ((i >> 4) & 0x0001);
	}
	break;
    case 5:
	for (i = 0; i < 256; i++) {
	    rtab[i] = ((i << 7) & 0x4000) | ((i << 5) & 0x0800) |
		      ((i << 3) & 0x0100) | ((i << 1) & 0x0020) |
		      ((i >> 1) & 0x0004);
	    gtab[i] = ((i << 6) & 0x2000) | ((i << 4) & 0x0400) |
		      ((i << 2) & 0x0080) | (i & 0x0010) |
		      ((i >> 2) & 0x0002);
	    btab[i] = ((i << 5) & 0x1000) | ((i << 3) & 0x0200) |
		      ((i << 1) & 0x0040) | ((i >> 1) & 0x0008) |
		      ((i >> 3) & 0x0001);
	}
	break;
    case 6:
	for (i = 0; i < 256; i++) {
	    rtab[i] = ((i << 10) & 0x20000) | ((i << 8) & 0x4000) |
		      ((i << 6) & 0x0800) | ((i << 4) & 0x0100) |
		      ((i << 2) & 0x0020) | (i & 0x0004);
	    gtab[i] = ((i << 9) & 0x10000) | ((i << 7) & 0x2000) |
		      ((i << 5) & 0x0400) | ((i << 3) & 0x0080) |
		      ((i << 1) & 0x0010) | ((i >> 1) & 0x0002);
	    btab[i] = ((i << 8) & 0x8000) | ((i << 6) & 0x1000) |
		      ((i << 4) & 0x0200) | ((i << 2) & 0x0040) |
		      (i & 0x0008) | ((i >> 2) & 0x0001);
	}
	break;
    default:
	fprintf(stderr, "not supposed to get here!\n");
	return 1;
    }

    return 0;
}


/*!
 *  getRGBFromOctcube()
 *
 *    Input:  cube index
 *            level (at which index is expressed)
 *            &rval  (<return> r val of this cube)
 *            &gval  (<return> g val of this cube)
 *            &bval  (<return> b val of this cube)
 *    Return: void
 *
 *    Note: for example, if index = 110101 (binary), level = 2,
 *          then the rgbindex is 000110101111000000000000
 *          and rval = 11100000, gval = 10100000, bval = 01100000
 */
void
getRGBFromOctcube(INT32  cubeindex,
                  INT32  level,
		  INT32  *prval,
		  INT32  *pgval,
		  INT32  *pbval)
{
INT32    rgbindex;

	/* Bring to format in 21 bits: (r7 g7 b7 r6 g6 b6 ...) */
	/* This is valid for levels from 0 to 6 */
    rgbindex = cubeindex << (3 * (7 - level));  /* upper corner of cube */
    rgbindex |= (0x7 << (3 * (6 - level)));   /* index to center of cube */

	/* then extract separate pieces */
    *prval = ((rgbindex >> 13) & 0x80) |
             ((rgbindex >> 11) & 0x40) |
             ((rgbindex >> 9) & 0x20) |
             ((rgbindex >> 7) & 0x10) |
             ((rgbindex >> 5) & 0x08) |
             ((rgbindex >> 3) & 0x04) |
             ((rgbindex >> 1) & 0x02);
    *pgval = ((rgbindex >> 12) & 0x80) |
             ((rgbindex >> 10) & 0x40) |
             ((rgbindex >> 8) & 0x20) |
             ((rgbindex >> 6) & 0x10) |
             ((rgbindex >> 4) & 0x08) |
             ((rgbindex >> 2) & 0x04) |
	     (rgbindex & 0x02);
    *pbval = ((rgbindex >> 11) & 0x80) |
             ((rgbindex >> 9) & 0x40) |
             ((rgbindex >> 7) & 0x20) |
             ((rgbindex >> 5) & 0x10) |
             ((rgbindex >> 3) & 0x08) |
             ((rgbindex >> 1) & 0x04) |
             ((rgbindex << 1) & 0x02);

    return;
}


    
/*!
 *  getOctcubeIndices()
 *
 *    Input:  RGBIndex
 *            octree level (0, 1, 2, 3, 4, 5)
 *            &octcube base index (<return> index at the octree level)
 *            &octcube sub index (<return> index at the next lower level)
 *    Return: 0 if OK, 1 on error
 *
 *    for CQ_NLEVELS = 6, the full RGB index is in the form:
 *       index = (0[13] 0 r7 g7 b7 r6 g6 b6 r5 g5 b5 r4 g4 b4 r3 g3 b3 r2 g2 b2)
 *    for CQ_NLEVELS = 5, the full RGB index is in the form:
 *       index = (0[16] 0 r7 g7 b7 r6 g6 b6 r5 g5 b5 r4 g4 b4 r3 g3 b3)
 *    for CQ_NLEVELS = 4, the full RGB index is in the form:
 *       index = (0[19] 0 r7 g7 b7 r6 g6 b6 r5 g5 b5 r4 g4 b4)
 *
 *    The base index is the index of the octcube at the level given,
 *    whereas the sub index is the index at the next level down.
 *
 *    For level 0: base index = 0
 *                 sub index is the 3 bit number (r7 g7 b7)
 *    For level 1: base index = (r7 g7 b7)
 *                 sub index = (r7 g7 b7 r6 g6 b6)
 *    For level 2: base index = (r7 g7 b7 r6 g6 b6)
 *                 sub index = (r7 g7 b7 r6 g6 b6 r5 g5 b5)
 *    For level 3: base index = (r7 g7 b7 r6 g6 b6 r5 g5 b5)
 *                 sub index = (r7 g7 b7 r6 g6 b6 r5 g5 b5 r4 g4 b4)
 *    For level 4: base index = (r7 g7 b7 r6 g6 b6 r5 g5 b5 r4 g4 b4)
 *                 sub index = (r7 g7 b7 r6 g6 b6 r5 g5 b5 r4 g4 b4 r3 g3 b3)
 *    For level 5: base index = (r7 g7 b7 r6 g6 b6 r5 g5 b5 r4 g4 b4 r3 g3 b3)
 *                 sub index = (r7 g7 b7 r6 g6 b6 r5 g5 b5 r4 g4 b4 r3 g3 b3
 *                              r2 g2 b2)
 *      
 */
INT32
getOctcubeIndices(INT32   rgbindex,
                  INT32   level,
	          INT32  *pbindex,
		  INT32  *psindex)
{
    PROCNAME("getOctcubeIndex");

    if (level < 0 || level > CQ_NLEVELS - 1)
	return ERROR_INT("level must be in e.g., [0 ... 5]", procName, 1);
    if (!pbindex)
	return ERROR_INT("&bindex not defined", procName, 1);
    if (!psindex)
	return ERROR_INT("&sindex not defined", procName, 1);

    *pbindex = rgbindex >> (3 * (CQ_NLEVELS - level));
    *psindex = rgbindex >> (3 * (CQ_NLEVELS - 1 - level));
    return 0;
}


/*!
 *   getOctcubeIndexFromRGB()
 *
 *      Input:  rval, gval, bval
 *              rtab, gtab, btab
 *              &index (<return>)
 *      Return: void
 *      Note: no error checking!
 */
void
getOctcubeIndexFromRGB(INT32    rval,
	               INT32    gval,
	               INT32    bval,
	               UINT32  *rtab,
	               UINT32  *gtab,
	               UINT32  *btab,
		       UINT32  *pindex)
{
    *pindex = rtab[rval] | gtab[gval] | btab[bval];
    return;
}



/*---------------------------------------------------------------------------*
 *           Fixed partition octcube quantization and histogram              *
 *---------------------------------------------------------------------------*/
/*!
 *   pixSimpleOctcubeQuant()
 *
 *      Input:  pixs (32 bpp rgb)
 *              level (significant bits for each of RGB)
 *      Return: pixd (quantized to octcube) or null on error
 *
 */     
PIX *
pixSimpleOctcubeQuant(PIX   *pixs,
	              INT32  level)
{
INT32    w, h, wpls, wpld, i, j;
INT32    rval, gval, bval;
UINT32   octindex;
UINT32  *rtab, *gtab, *btab;
UINT32  *lines, *lined, *datas, *datad, *pspixel, *pdpixel;
PIX     *pixd;

    PROCNAME("pixSimpleOctcubeQuant");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (level < 2 || level > 6)
	return (PIX *)ERROR_PTR("level not in {2,...6}", procName, NULL);

    if (makeRGBToIndexTables(&rtab, &gtab, &btab, level))
	return (PIX *)ERROR_PTR("tables not made", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    pixd = pixCreate(w, h, 32);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

    for (i = 0; i < h; i++) {
	lines = datas + i * wpls;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    pspixel = lines + j;
	    pdpixel = lined + j;
	    rval = GET_DATA_BYTE(pspixel, 0);
	    gval = GET_DATA_BYTE(pspixel, 1);
	    bval = GET_DATA_BYTE(pspixel, 2);
	    octindex = rtab[rval] | gtab[gval] | btab[bval];
	    getRGBFromOctcube(octindex, level, &rval, &gval, &bval);
	    composeRGBPixel(rval, gval, bval, pdpixel);
	}
    }

    FREE((void *)rtab);
    FREE((void *)gtab);
    FREE((void *)btab);
    return pixd;
}


/*!
 *   pixOctcubeHistogram()
 *
 *      Input:  pixs (32 bpp rgb)
 *              level (significant bits for each of RGB)
 *      Return: numa (histogram of color pixels, or null on error)
 *
 */     
NUMA *
pixOctcubeHistogram(PIX   *pixs,
	            INT32  level)
{
INT32     size, i, j, w, h, wpl;
UINT32    rval, gval, bval, rgbindex;
UINT32   *rtab, *gtab, *btab;
UINT32   *ppixel, *data, *line;
FLOAT32  *array;
NUMA     *na;

    PROCNAME("pixOctcubeHistogram");

    if (!pixs)
	return (NUMA *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
	return (NUMA *)ERROR_PTR("depth not 32 bpp", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    wpl = pixGetWpl(pixs);
    data = pixGetData(pixs);

	/* size = 2 ** (3 * level)  */
    if (level == 2)
	size = 64;   /* 2 ** 6 */
    else if (level == 3)
	size = 512;  /* 2 ** 9 */
    else if (level == 4)
	size = 4098;   /* 2 ** 12 */
    else if (level == 5)
	size = 32784;   /* 2 ** 15 */
    else if (level == 6)
	size = 262272;   /* 2 ** 18 */
    else
	return (NUMA *)ERROR_PTR("level not in {2, ... 6}", procName, NULL);

    if (makeRGBToIndexTables(&rtab, &gtab, &btab, level))
	return (NUMA *)ERROR_PTR("tables not made", procName, NULL);

    if ((na = numaCreate(size)) == NULL)
	return (NUMA *)ERROR_PTR("na not made", procName, NULL);
    na->n = size;  /* fake storage of n zeroes */
    array = na->array;  /* don't do this at home */

    for (i = 0; i < h; i++) {
	line = data + i * wpl;
	for (j = 0; j < w; j++) {
	    ppixel = line + j;
	    rval = GET_DATA_BYTE(ppixel, 0);
	    gval = GET_DATA_BYTE(ppixel, 1);
	    bval = GET_DATA_BYTE(ppixel, 2);
	    rgbindex = rtab[rval] | gtab[gval] | btab[bval];
#if DEBUG_OCTINDEX
	    if ((level == 2 && rgbindex > 63) ||
	        (level == 3 && rgbindex > 511) ||
	        (level == 4 && rgbindex > 4097) ||
	        (level == 5 && rgbindex > 32783) ||
	        (level == 6 && rgbindex > 262271)) {
		fprintf(stderr, "level = %d, rgbindex = %d, index error!\n",
		        level, rgbindex);
	    } else
#endif  /* DEBUG_OCTINDEX */
		array[rgbindex] += 1.0;
	}
    }

    return na;
}

