/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  enhance.c
 *
 *      Gamma-correction
 *           PIX        *pixGammaCorrect()
 *           PIX        *pixGammaCorrectColor()
 *           PIX        *pixGammaCorrectGray()
 *
 *      Contrast enhancement
 *           PIX        *pixEnhanceContrast()
 *           PIX        *pixEnhanceContrastColor()
 *           PIX        *pixEnhanceContrastGray()
 *
 *      Unsharp-masking
 *           PIX        *pixUnsharpMask()
 *           PIX        *pixUnsharpMaskColor()
 *           PIX        *pixUnsharpMaskGray()
 *
 *
 *      Gamma correction and contrast enhancement apply a
 *      simple mapping function to each pixel (or, for color
 *      images, to each sample (i.e., r,g,b) of the pixel).
 *      Gamma correction either lightens or darkens the
 *      entire image, depending on whether the gamma factor
 *      is greater or less than 0.0, respectively.
 *      Contrast enhancement darkens the pixels that are already
 *      darker than the middle of the dynamic range (128)
 *      and lightens pixels that are lighter than 128.
 *
 *      Unsharp masking is a more complicated enhancement.
 *      A "high frequency" image, generated by subtracting
 *      the smoothed ("low frequency") part of the image from
 *      itself, has all the energy at the edges.  This "edge image"
 *      has 0 average value.  A fraction of the edge image is
 *      then added to the original, enhancing the differences
 *      between pixel values at edges.  Because we represent
 *      images as UINT8 arrays, we preserve dynamic range and
 *      handle negative values by doing all the arithmetic on
 *      shifted UINT16 arrays; the UINT8 values are recovered
 *      at the end.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "allheaders.h"

    /* scales contrast enhancement factor to have a useful range
     * between 0.0 and 1.0 */
#define  ENHANCE_SCALE_FACTOR      5.


/*-------------------------------------------------------------*
 *                         Gamma correction                    *
 *-------------------------------------------------------------*/
/*!
 *  pixGammaCorrect()
 *
 *          Input:  pixs (8 or 32 bpp)
 *                  gam  (gamma correction)
 *          Return: pixd, or null on error
 *
 *          Note: If gam < 1.0, the image will appear darker;
 *                if gam > 1.0, the image will appear lighter;
 *                if gam == 1.0, return a clone
 */
PIX  *
pixGammaCorrect(PIX      *pixs,
                FLOAT32   gam)
{
INT32     d;

    PROCNAME("pixGammaCorrect");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (gam <= 0.0)
	return (PIX *)ERROR_PTR("gamma must be > 0.0", procName, NULL);

    d = pixGetDepth(pixs);
    if (d != 8 && d != 32)
	return (PIX *)ERROR_PTR("d not 8 or 32 bpp", procName, NULL);

    if (gam == 1.0)
	return pixClone(pixs);

    if (d == 8)
	return pixGammaCorrectGray(pixs, gam);
    else    /* d == 32 */
	return pixGammaCorrectColor(pixs, gam);
}


/*!
 *  pixGammaCorrectColor()
 *
 *          Input:  pixs (32 bpp; 24 bpp RGB color)
 *                  gam  (gamma correction)
 *          Return: pixd, or null on error
 *
 *          Note: If gam < 1.0, the image will appear darker;
 *                if gam > 1.0, the image will appear lighter;
 *                if gam == 1.0, return a clone
 */
PIX  *
pixGammaCorrectColor(PIX      *pixs,
                     FLOAT32   gam)
{
PIX      *pixRed, *pixGreen, *pixBlue;
PIX      *pixRedGam, *pixGreenGam, *pixBlueGam;
PIX      *pixd;

    PROCNAME("pixGammaCorrectColor");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (gam <= 0.0)
	return (PIX *)ERROR_PTR("gamma must be > 0.0", procName, NULL);

    if (pixGetDepth(pixs) != 32)
	return (PIX *)ERROR_PTR("pix not 32 bpp", procName, NULL);

    if (gam == 1.0)
        return pixClone(pixs);

    pixRed = pixGetRGBComponent(pixs, COLOR_RED);
    pixRedGam = pixGammaCorrectGray(pixRed, gam);
    pixDestroy(&pixRed);
    pixGreen = pixGetRGBComponent(pixs, COLOR_GREEN);
    pixGreenGam = pixGammaCorrectGray(pixGreen, gam);
    pixDestroy(&pixGreen);
    pixBlue = pixGetRGBComponent(pixs, COLOR_BLUE);
    pixBlueGam = pixGammaCorrectGray(pixBlue, gam);
    pixDestroy(&pixBlue);

    if ((pixd = pixCreateRGBImage(pixRedGam, pixGreenGam, pixBlueGam)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixDestroy(&pixRedGam);
    pixDestroy(&pixGreenGam);
    pixDestroy(&pixBlueGam);

    return pixd;
}


/*!
 *  pixGammaCorrectGray()
 *
 *          Input:  pixs (8 bpp)
 *                  gam  (gamma correction)
 *          Return: pixd, or null on error
 *
 *          Note: If gam < 1.0, the image will appear darker;
 *                if gam > 1.0, the image will appear lighter;
 *                if gam == 1.0, return a clone
 */
PIX  *
pixGammaCorrectGray(PIX      *pixs,
                    FLOAT32   gam)
{
UINT8     val;
INT32     i, j, wpls, wpld, w, h;
INT32    *tab;
FLOAT64   invgam, x;
UINT32   *datas, *datad, *lines, *lined;
PIX      *pixd;

    PROCNAME("pixGammaCorrectGray");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (gam <= 0.0)
	return (PIX *)ERROR_PTR("gamma must be > 0.0", procName, NULL);

    if (pixGetDepth(pixs) != 8)
	return (PIX *)ERROR_PTR("pix not 8 bpp", procName, NULL);

    if (gam == 1.0)
        return pixClone(pixs);

	/* build mapping table */
    if ((tab = (INT32 *)CALLOC(256, sizeof(INT32))) == NULL)
	return (PIX *)ERROR_PTR("tab not made", procName, NULL);
    invgam = 1. / gam;
    for (i = 0; i < 256; i++) {
	x = (FLOAT64)i / 255.;
        tab[i] = (INT32)(255. * pow(x, invgam) + 0.5);
/*	fprintf(stderr, "tab[%d] = %d\n", i, tab[i]); */
    }

    if ((pixd = pixCreateTemplate(pixs)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datas = pixGetData(pixs);
    datad = pixGetData(pixd);
    wpls = pixGetWpl(pixs);
    wpld = pixGetWpl(pixd);
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    for (i = 0; i < h; i++) {
	lines = datas + wpls * i;
	lined = datad + wpld * i;
	for (j = 0; j < w; j++) {
	    val = GET_DATA_BYTE(lines, j);
	    SET_DATA_BYTE(lined, j, tab[val]);
	}
    }

    FREE((void *)tab);
    return pixd;
}

	    

/*-------------------------------------------------------------*
 *                      Contrast enhancement                   *
 *-------------------------------------------------------------*/
/*!
 *  pixEnhanceContrast()
 *
 *          Input:  pixs (8 or 32 bpp)
 *                  factor  (0.0 is no enhancement)
 *          Return: pixd, or null on error
 *
 *          Note: Useful range is scaled to be in (0.0 to 1.0),
 *                but larger values can also be used.
 */
PIX  *
pixEnhanceContrast(PIX      *pixs,
                   FLOAT32   factor)
{
INT32     d;

    PROCNAME("pixEnhanceContrast");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (factor < 0.0) {
        WARNING("factor must be >= 0.0; using 0.0", procName);
	return pixClone(pixs);
    }

    d = pixGetDepth(pixs);
    if (d == 8)
	return pixEnhanceContrastGray(pixs, factor);
    else if (d == 32)
	return pixEnhanceContrastColor(pixs, factor);
    else
	return (PIX *)ERROR_PTR("d not 8 or 32 bpp", procName, NULL);
}


/*!
 *  pixEnhanceContrastColor()
 *
 *          Input:  pixs (32 bpp; 24 bpp RGB color)
 *                  factor (0.0 is no enhancement; )
 *          Return: pixd, or null on error
 *
 *          Note: Useful range is scaled to be in (0.0 to 1.0),
 *                but larger values can also be used.
 */
PIX  *
pixEnhanceContrastColor(PIX      *pixs,
                        FLOAT32   factor)
{
PIX      *pixRed, *pixGreen, *pixBlue;
PIX      *pixRedGam, *pixGreenGam, *pixBlueGam;
PIX      *pixd;

    PROCNAME("pixEnhanceContrastColor");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (factor < 0.0) {
        WARNING("factor must be >= 0.0; using 0.0", procName);
	return pixClone(pixs);
    }

    if (pixGetDepth(pixs) != 32)
	return (PIX *)ERROR_PTR("pix not 32 bpp", procName, NULL);

    pixRed = pixGetRGBComponent(pixs, COLOR_RED);
    pixRedGam = pixEnhanceContrastGray(pixRed, factor);
    pixDestroy(&pixRed);
    pixGreen = pixGetRGBComponent(pixs, COLOR_GREEN);
    pixGreenGam = pixEnhanceContrastGray(pixGreen, factor);
    pixDestroy(&pixGreen);
    pixBlue = pixGetRGBComponent(pixs, COLOR_BLUE);
    pixBlueGam = pixEnhanceContrastGray(pixBlue, factor);
    pixDestroy(&pixBlue);

    if ((pixd = pixCreateRGBImage(pixRedGam, pixGreenGam, pixBlueGam)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixDestroy(&pixRedGam);
    pixDestroy(&pixGreenGam);
    pixDestroy(&pixBlueGam);

    return pixd;
}



/*!
 *  pixEnhanceContrastGray()
 *
 *          Input:  pixs (8 bpp)
 *                  factor (0.0 is no enhancement; )
 *          Return: pixd, or null on error
 *
 *          Note: Useful range is scaled to be in (0.0 to 1.0),
 *                but larger values can also be used.
 */
PIX  *
pixEnhanceContrastGray(PIX      *pixs,
                       FLOAT32   factor)
{
UINT8     val;
INT32     i, j, wpls, wpld, w, h;
INT32    *tab;
FLOAT64   x, ymax, ymin, dely, scale;
UINT32   *datas, *datad, *lines, *lined;
PIX      *pixd;

    PROCNAME("pixEnhanceContrastGray");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (factor < 0.0) {
        WARNING("factor must be >= 0.0; using 0.0", procName);
	return pixClone(pixs);
    }

    if (pixGetDepth(pixs) != 8)
	return (PIX *)ERROR_PTR("pix not 8 bpp", procName, NULL);

	/* build atan mapping table */
    scale = ENHANCE_SCALE_FACTOR;
    if ((tab = (INT32 *)CALLOC(256, sizeof(INT32))) == NULL)
	return (PIX *)ERROR_PTR("tab not made", procName, NULL);
    ymax = atan((FLOAT64)(1.0 * factor * scale));
    ymin = atan((FLOAT64)(-127. * factor * scale / 128.));
    dely = ymax - ymin;
    for (i = 0; i < 256; i++) {
	x = (FLOAT64)i;
        tab[i] = (INT32)((255. / dely) *
	         (-ymin + atan((FLOAT64)(factor * scale * (x - 127.) / 128.))) +
		 0.5);
/*	fprintf(stderr, "tab[%d] = %d\n", i, tab[i]); */
    }

    if ((pixd = pixCreateTemplate(pixs)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datas = pixGetData(pixs);
    datad = pixGetData(pixd);
    wpls = pixGetWpl(pixs);
    wpld = pixGetWpl(pixd);
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    for (i = 0; i < h; i++) {
	lines = datas + wpls * i;
	lined = datad + wpld * i;
	for (j = 0; j < w; j++) {
	    val = GET_DATA_BYTE(lines, j);
	    SET_DATA_BYTE(lined, j, tab[val]);
	}
    }

    FREE((void *)tab);
    return pixd;
}

	    

/*-------------------------------------------------------------*
 *                        Unsharp masking                      *
 *-------------------------------------------------------------*/
/*!
 *  pixUnsharpMask()
 *
 *          Input:  pixs (8 or 32 bpp)
 *                  smooth  ("half-width" of smoothing filter)
 *                  fract  (fraction of edge added back into image)
 *          Return: pixd, or null on error
 *
 *          Note: (1) We use symmetric smoothing filters of odd dimension,
 *                    typically use 3, 5, 7, etc.  The smooth parameter
 *                    for these is (size - 1)/2; i.e., 1, 2, 3, etc.
 *                (2) The fract parameter is typically taken in the
 *                    range:  0.2 < fract < 0.7
 */
PIX  *
pixUnsharpMask(PIX      *pixs,
	       INT32     smooth,
               FLOAT32   fract)
{
INT32     d;

    PROCNAME("pixUnsharpMask");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (fract <= 0.0) {
	WARNING("no fraction added back in", procName);
	return pixClone(pixs);
    }

    d = pixGetDepth(pixs);
    if (d == 8)
	return pixUnsharpMaskGray(pixs, smooth, fract);
    else if (d == 32)
	return pixUnsharpMaskColor(pixs, smooth, fract);
    else
	return (PIX *)ERROR_PTR("d not 8 or 32 bpp", procName, NULL);
}


/*!
 *  pixUnsharpMaskColor()
 *
 *          Input:  pixs (32 bpp; 24 bpp RGB color)
 *                  smooth  ("half-width" of smoothing filter)
 *                  fract  (fraction of edge added back into image)
 *          Return: pixd, or null on error
 *
 *          Note: (1) We use symmetric smoothing filters of odd dimension,
 *                    typically use 3, 5, 7, etc.  The smooth parameter
 *                    for these is (size - 1)/2; i.e., 1, 2, 3, etc.
 *                (2) The fract parameter is typically taken in the
 *                    range:  0.2 < fract < 0.7
 */
PIX  *
pixUnsharpMaskColor(PIX      *pixs,
		    INT32     smooth,
                    FLOAT32   fract)
{
PIX      *pixRed, *pixGreen, *pixBlue;
PIX      *pixRedSharp, *pixGreenSharp, *pixBlueSharp;
PIX      *pixd;

    PROCNAME("pixUnsharpMaskColor");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
	return (PIX *)ERROR_PTR("pixs not 32 bpp", procName, NULL);

    if (fract <= 0.0) {
	WARNING("no fraction added back in", procName);
	return pixClone(pixs);
    }

    pixRed = pixGetRGBComponent(pixs, COLOR_RED);
    pixRedSharp = pixUnsharpMaskGray(pixRed, smooth, fract);
    pixDestroy(&pixRed);
    pixGreen = pixGetRGBComponent(pixs, COLOR_GREEN);
    pixGreenSharp = pixUnsharpMaskGray(pixGreen, smooth, fract);
    pixDestroy(&pixGreen);
    pixBlue = pixGetRGBComponent(pixs, COLOR_BLUE);
    pixBlueSharp = pixUnsharpMaskGray(pixBlue, smooth, fract);
    pixDestroy(&pixBlue);

    if ((pixd = pixCreateRGBImage(pixRedSharp, pixGreenSharp, pixBlueSharp))
            == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixDestroy(&pixRedSharp);
    pixDestroy(&pixGreenSharp);
    pixDestroy(&pixBlueSharp);

    return pixd;
}


/*!
 *  pixUnsharpMaskGray()
 *
 *          Input:  pixs (8 bpp)
 *                  smooth  ("half-width" of smoothing filter)
 *                  fract  (fraction of edge added back into image)
 *          Return: pixd, or null on error
 *
 *          Note: (1) We use symmetric smoothing filters of odd dimension,
 *                    typically use 3, 5, 7, etc.  The smooth parameter
 *                    for these is (size - 1)/2; i.e., 1, 2, 3, etc.
 *                (2) The fract parameter is typically taken in the
 *                    range:  0.2 < fract < 0.7
 *                
 */
PIX  *
pixUnsharpMaskGray(PIX      *pixs,
		   INT32     smooth,
                   FLOAT32   fract)
{
PIX      *pixc, *pixt, *pixd;

    PROCNAME("pixUnsharpMaskGray");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
	return (PIX *)ERROR_PTR("pixs not 8 bpp", procName, NULL);

    if (fract <= 0.0) {
	WARNING("no fraction added back in", procName);
	return pixClone(pixs);
    }

    if ((pixc = pixBlockconvGray(pixs, NULL, smooth, smooth)) == NULL)
	return (PIX *)ERROR_PTR("pixc not made", procName, NULL);

	/* steps:
	 *    (1) edge image is pixs - pixc  (this is highpass part)
	 *    (2) multiply edge image by fract
	 *    (3) add fraction of edge to pixs
	 */
    if ((pixt = pixInit8To16(pixs)) == NULL)
	return (PIX *)ERROR_PTR("pixt not made", procName, NULL);
    pixAccumulate8To16(pixt, pixc, SUBTRACT);
    pixMultConstant16(pixt, fract);
    pixAccumulate8To16(pixt, pixs, ADD);
    if ((pixd = pixFinal16To8(pixt)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixDestroy(&pixc);
    pixDestroy(&pixt);

    return pixd;
}


