/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/

/*
 *  graphics.c
 *                     
 *      Line rendering
 *          PTA        *ptaFindLine()
 *          INT32       pixRenderPta()
 *          INT32       pixRenderLine()
 *          INT32       pixRenderBox()
 *
 */

#include <stdio.h>
#include <string.h>

#include "allheaders.h"



/*------------------------------------------------------------------*
 *                            Line rendering                        *
 *------------------------------------------------------------------*/
/*!
 *  ptaFindLine()
 *
 *      Input:  x1, y1  (end point 1)
 *              x2, y2  (end point 2)
 *      Return: pta, or null on error
 *
 */
PTA  *
ptaFindLine(INT32  x1,
            INT32  y1,
	    INT32  x2,
	    INT32  y2)
{
INT32     npts, diff, getyofx, sign, i, x, y;
FLOAT32   slope;
PTA      *pta;

    PROCNAME("ptaFindLine");

	/* generate line parameters */
    if (ABS(x2 - x1) >= ABS(y2 - y1)) {
	getyofx = TRUE;
	npts = ABS(x2 - x1) + 1;
	diff = x2 - x1;
	sign = SIGN(x2 - x1);
	slope = (FLOAT32)(sign * (y2 - y1)) / (FLOAT32)diff;
    }
    else {
	getyofx = FALSE;
	npts = ABS(y2 - y1) + 1;
	diff = y2 - y1;
	sign = SIGN(y2 - y1);
	slope = (FLOAT32)(sign * (x2 - x1)) / (FLOAT32)diff;
    }

    if ((pta = ptaCreate(npts)) == NULL)
	return (PTA *)ERROR_PTR("pta not made", procName, NULL);

    if (npts == 1) {  /* degenerate case */
	ptaAddPt(pta, x1, y1);
	return pta;
    }

	/* generate the set of points */
    if (getyofx) {  /* y = y(x) */
	for (i = 0; i < npts; i++) {
	    x = x1 + sign * i;
	    y = (INT32)(y1 + (FLOAT32)i * slope + 0.5);
	    ptaAddPt(pta, x, y);
	}
    }
    else {   /* x = x(y) */
	for (i = 0; i < npts; i++) {
	    x = (INT32)(x1 + (FLOAT32)i * slope + 0.5);
	    y = y1 + sign * i;
	    ptaAddPt(pta, x, y);
	}
    }

    return pta;
}


/*!
 *  pixRenderPta()
 *
 *      Input:  pix
 *              pta  (for line)
 *              op   (one of {SET, CLEAR, FLIP})
 *      Return: 0 if OK, 1 on error
 */
INT32
pixRenderPta(PIX    *pix,
             PTA    *pta,
	     INT32   op)
{
INT32     i, n, x, y, w, h;

    PROCNAME("pixRenderPta");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (!pta)
	return ERROR_INT("pta not defined", procName, 1);
    if (op != SET && op != CLEAR && op != FLIP)
	return ERROR_INT("op must be one of {SET, CLEAR, FLIP}",
	     procName, 1);

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);

    n = ptaGetCount(pta);
    for (i = 0; i < n; i++) {
	ptaGetIPt(pta, i, &x, &y);
	if (x < 0 || x >= w)
	    continue;
	if (y < 0 || y >= h)
	    continue;
	if (op == SET)
	    pixSetPixel(pix, x, y, 1);
	else if (op == CLEAR)
	    pixClearPixel(pix, x, y);
	else
	    pixFlipPixel(pix, x, y);
    }

    return 0;
}


/*!
 *  pixRenderLine()
 *
 *      Input:  pix
 *              x1, y1
 *              x2, y2
 *              width  (thickness of line)
 *              op  (one of {SET, CLEAR, FLIP})
 *      Return: 0 if OK, 1 on error
 */
INT32
pixRenderLine(PIX    *pix,
              INT32   x1,
              INT32   y1,
              INT32   x2,
              INT32   y2,
	      INT32   width,
	      INT32   op)
{
INT32     i, x1a, x2a, y1a, y2a;
PTA      *pta;

    PROCNAME("pixRenderLine");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (width <= 0) {
	WARNING("width must be > 0; setting to 1", procName);
	width = 1;
    }

    if ((pta = ptaFindLine(x1, y1, x2, y2)) == NULL)
	return ERROR_INT("pta not made", procName, 1);
    pixRenderPta(pix, pta, op);
    ptaDestroy(&pta);
    if (width == 1)
	return 0;

	/* width > 1; estimate line direction & add new ptas */
    if (ABS(x1 - x2) > ABS(y1 - y2)) {  /* "horizontal" line  */
	for (i = 1; i < width; i++) {
	    if ((i & 1) == 1) {   /* place above */
		y1a = y1 - (i + 1) / 2;
		y2a = y2 - (i + 1) / 2;
	    }
	    else {  /* place below */
		y1a = y1 + (i + 1) / 2;
		y2a = y2 + (i + 1) / 2;
	    }
	    if ((pta = ptaFindLine(x1, y1a, x2, y2a)) == NULL)
		return ERROR_INT("pta not made", procName, 1);
	    pixRenderPta(pix, pta, op);
	    ptaDestroy(&pta);
	}
    }
    else  {  /* "vertical" line  */
	for (i = 1; i < width; i++) {
	    if ((i & 1) == 1) {   /* place to left */
		x1a = x1 - (i + 1) / 2;
		x2a = x2 - (i + 1) / 2;
	    }
	    else {  /* place to right */
		x1a = x1 + (i + 1) / 2;
		x2a = x2 + (i + 1) / 2;
	    }
	    if ((pta = ptaFindLine(x1a, y1, x2a, y2)) == NULL)
		return ERROR_INT("pta not made", procName, 1);
	    pixRenderPta(pix, pta, op);
	    ptaDestroy(&pta);
	}
    }

    return 0;
}


/*!
 *  pixRenderBox()
 *
 *      Input:  pix
 *              box
 *              width  (thickness of line)
 *              op  (one of {SET, CLEAR, FLIP})
 *      Return: 0 if OK, 1 on error
 *
 */
INT32
pixRenderBox(PIX    *pix,
             BOX    *box,
	     INT32   width,
	     INT32   op)
{
INT32    x, y, w, h;

    PROCNAME("pixRenderBox");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (!box)
	return ERROR_INT("box not defined", procName, 1);

    x = box->x;
    y = box->y;
    w = box->w;
    h = box->h;
    if ((width & 1) == 1) {   /* odd width */
	pixRenderLine(pix, x - width / 2, y,
		      x + w - 1 + width / 2, y, width, op);
	pixRenderLine(pix, x - width / 2, y + h - 1,
		      x + w - 1 + width / 2, y + h - 1, width, op);
	pixRenderLine(pix, x, y + 1 + width / 2,
		      x, y + h - 2 - width / 2, width, op);
	pixRenderLine(pix, x + w - 1, y + 1 + width / 2,
		      x + w - 1, y + h - 2 - width / 2, width, op);
    }
    else {   /* even width */
	pixRenderLine(pix, x - width / 2, y,
		      x + w - 2 + width / 2, y, width, op);
	pixRenderLine(pix, x - width / 2, y + h - 1,
		      x + w - 2 + width / 2, y + h - 1, width, op);
	pixRenderLine(pix, x, y + 0 + width / 2,
		      x, y + h - 2 - width / 2, width, op);
	pixRenderLine(pix, x + w - 1, y + 0 + width / 2,
		      x + w - 1, y + h - 2 - width / 2, width, op);
    }
    return 0;
}


