/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/

/*
 *  jpegio.c
 *                     
 *          PIX        *pixReadStreamJpeg()
 *          INT32       pixWriteStreamJpeg()
 *
 *
 *    Documentation: libjpeg.doc can be found, along with all
 *    source code, at ftp://ftp.uu.net/graphics/jpeg
 *    Download and untar the file:  jpegsrc.v6b.tar.gz
 *    A good paper on jpeg can also be found there: wallace.ps.gz
 *
 *    The functions in libjpeg make it very simple to compress
 *    and decompress images.  On input (decompression from file),
 *    3 component color images can be read into either an 8 bpp Pix
 *    with a colormap or a 32 bpp Pix with RGB components.  For output
 *    (compression to file), all color Pix, whether 8 bpp with a
 *    colormap or 32 bpp, are written compressed as a set of three
 *    8 bpp (rgb) images.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"


/*!
 *  pixReadStreamJpeg()
 *
 *      Input:  stream
 *              colormap flag (0 means full color;
 *                             1 means create and use 8 bpp palette)
 *      Return: pix, or null on error
 */
PIX *
pixReadStreamJpeg(FILE  *fp,
	          INT32  cmflag)
{
INT32		               i, j, k;
INT32                          w, h, wpl, spp, ncolors, cindex;
UINT32                        *data;
UINT32                        *line, *ppixel;
GIFColorTableEntry            *gifcmap;
JSAMPROW                       rowbuffer;
PIX 		              *pix;
PIXCMAP                       *colormap;
struct jpeg_decompress_struct  decomp;
struct jpeg_error_mgr          decomperror;


    PROCNAME("pixReadStreamJpeg");

    if (!fp)
        return (PIX *)ERROR_PTR("fp not defined", procName, NULL);

    if (BITS_IN_JSAMPLE != 8)  /* set in jmorecfg.h */
	return (PIX *)ERROR_PTR("BITS_IN_JSAMPLE != 8", procName, NULL);

    rewind(fp);

    decomp.err = jpeg_std_error(&decomperror);
    jpeg_create_decompress(&decomp);
    jpeg_stdio_src(&decomp, fp);
    jpeg_read_header(&decomp, TRUE);
    jpeg_calc_output_dimensions(&decomp);

	/* allocate the image and a row buffer */
    spp = decomp.out_color_components;
    if (spp != 1 && spp != 3)
	return (PIX *)ERROR_PTR("spp must be 1 or 3", procName, NULL);
    w = decomp.output_width;
    h = decomp.output_height;
    if ((spp == 3) && (cmflag == 0))  /* get full color pix */
    {
        if ((rowbuffer = (JSAMPROW)CALLOC(sizeof(JSAMPLE), spp * w)) == NULL)
            return (PIX *)ERROR_PTR("rowbuffer not made", procName, NULL);
        if ((pix = pixCreate(w, h, 32)) == NULL)
	    return (PIX *)ERROR_PTR("pix not made", procName, NULL);
    }
    else {  /* 8 bpp gray or colormapped */
        if ((rowbuffer = (JSAMPROW)CALLOC(sizeof(JSAMPLE), w)) == NULL)
            return (PIX *)ERROR_PTR("rowbuffer not made", procName, NULL);
        if ((pix = pixCreate(w, h, 8)) == NULL)
	    return (PIX *)ERROR_PTR("pix not made", procName, NULL);
    }

    if (spp == 1)  /* grayscale or colormapped */
        jpeg_start_decompress(&decomp);
    else  {        /* color; spp == 3 */
	if (cmflag == 0) {   /* -- 24 bit color in 32 bit pix -- */
            decomp.quantize_colors = FALSE;
            jpeg_start_decompress(&decomp);
	}
	else {      /* -- color quantize to 8 bits with GIF colormap -- */
            decomp.quantize_colors = TRUE;
            decomp.desired_number_of_colors = 256;
            jpeg_start_decompress(&decomp);

	        /* construct a colormap array in GIF cm format */
	    ncolors = decomp.actual_number_of_colors;
            if ((gifcmap =
	      (GIFColorTableEntry *)CALLOC(sizeof(GIFColorTableEntry), ncolors))
		    == NULL)
	        return (PIX *)ERROR_PTR("gifcmap not made", procName, NULL);
            for (cindex = 0; cindex < ncolors; cindex++)
	    {
	        gifcmap[cindex].cmRed = decomp.colormap[0][cindex];
	        gifcmap[cindex].cmGreen = decomp.colormap[1][cindex];
	        gifcmap[cindex].cmBlue = decomp.colormap[2][cindex];
            }

	        /* make a pix colormap */
	    colormap = (PIXCMAP *)CALLOC(1, sizeof(PIXCMAP));
	    if (!colormap)
	        return (PIX *)ERROR_PTR("colormap not made", procName, NULL);
	    pixSetColormap(pix, colormap);
            colormap->type = CM_GIF;
            colormap->array  = gifcmap;
            colormap->size  = ncolors * sizeof(GIFColorTableEntry);
	}
    }

    wpl  = pixGetWpl(pix);
    data = pixGetData(pix);

	/* decompress */
    if ((spp == 3) && (cmflag == 0))
    {	/* -- 24 bit color -- */
        for (i = 0; i < h; i++) {
            if (jpeg_read_scanlines(&decomp, &rowbuffer, (JDIMENSION)1) != 1)
	        return (PIX *)ERROR_PTR("bad read scanline", procName, NULL);
	    ppixel = data + i * wpl;
            for (j = k = 0; j < w; j++)
	    {
	        SET_DATA_BYTE(ppixel, 0, rowbuffer[k++]);        /* R */
	        SET_DATA_BYTE(ppixel, 1, rowbuffer[k++]);      	 /* G */
	        SET_DATA_BYTE(ppixel, 2, rowbuffer[k++]);        /* B */
		ppixel++;
            }
        }
    }
    else {    /* -- 8 bpp grayscale or colormapped pix -- */
        for (i = 0; i < h; i++) {
            if (jpeg_read_scanlines(&decomp, &rowbuffer, (JDIMENSION)1) != 1)
	        return (PIX *)ERROR_PTR("bad read scanline", procName, NULL);
	    line = data + i * wpl;
            for (j = 0; j < w; j++)
	        SET_DATA_BYTE(line, j, rowbuffer[j]);
        }
    }

    jpeg_finish_decompress(&decomp);
    jpeg_destroy_decompress(&decomp);
    FREE((void *)rowbuffer);

    return pix;
}



/*!
 *  pixWriteStreamJpeg()
 *
 *      Input:  stream
 *              pix  (8 or 32 bpp)
 *              quality  (1 - 100; 75 is default value; 0 is also default)
 *      Return: 0 if OK, 1 on error
 *
 *      Action: there are three possibilities
 *         (1) grayscale image, no colormap: compress as 8 bpp image.
 *	   (2) 24 bpp full color image: copy each line into the color
 *             line buffer, and compress as three 8 bpp images.
 *         (3) 8 bpp colormapped image: convert each line to three
 *             8 bpp line images in the color line buffer, and
 *             compress as three 8 bpp images.
 *
 *      Note: under the covers, the library transforms rgb to a
 *      luminence-chromaticity triple, each component of which is
 *      also 8 bits, and compresses that.  It uses 2 Huffman tables,
 *      a higher resolution one (with more quantization levels)
 *      for luminosity and a lower resolution one for the chromas.
 *
 */
INT32
pixWriteStreamJpeg(FILE   *fp,
	           PIX    *pix,
		   INT32   quality)
{
UINT8                       byteval;
UINT8                      *rmap, *gmap, *bmap;
INT32                       i, j, k;
INT32                       w, h, d, wpl, spp, colorflg;
INT32                       numcolors, rowsamples;
UINT32                     *ppixel, *line, *data;
JSAMPROW                    rowbuffer;
PIXCMAP                    *colormap;
struct jpeg_compress_struct comp;
struct jpeg_error_mgr       comperror;

    PROCNAME("pixWriteStreamJfif");

    if (!fp)
	return ERROR_INT("stream not open", procName, 1);
    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    rewind(fp);

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);
    d = pixGetDepth(pix);
    if (d != 8 && d != 32)
	return ERROR_INT("bpp must be 8 or 32", procName, 1);

    if (quality <= 0)
        quality = 75;  /* default */

    if (d == 32)
	colorflg = 2;    /* 24 bpp rgb; no colormap */
    else if ((colormap = pixGetColormap(pix)) == NULL)
	colorflg = 0;    /* 8 bpp grayscale; no colormap */
    else {   
	colorflg = 1;    /* 8 bpp; colormap */
	if (colormapToArrays(colormap, &rmap, &gmap, &bmap, &numcolors))
	    return ERROR_INT("unidentified colormap type", procName, 1);
    }

    comp.err = jpeg_std_error(&comperror);
    jpeg_create_compress(&comp);
    jpeg_stdio_dest(&comp, fp);

    comp.image_width  = w;
    comp.image_height = h;

    if (colorflg == 0) {
        comp.input_components = 1;
        comp.in_color_space = JCS_GRAYSCALE;
    }
    else {  /* colorflg == 1 or 2 */
        comp.input_components = 3;
        comp.in_color_space = JCS_RGB;
    }

    jpeg_set_defaults(&comp);
    jpeg_set_quality(&comp, quality, TRUE);
    jpeg_start_compress(&comp, TRUE);

	/* alloc row buffer */
    spp = comp.input_components;
    rowsamples = spp * w;
    if ((rowbuffer = (JSAMPROW)CALLOC(sizeof(JSAMPLE), rowsamples)) == NULL)
	return ERROR_INT("calloc fail for rowbuffer", procName, 1);

    data = pixGetData(pix);
    wpl  = pixGetWpl(pix);
    for (i = 0; i < h; i++) {
        line = data + i * wpl;
	if (colorflg == 0) {	/* -- 8 bpp gray -- */
	    for (j = 0; j < w; j++)
		rowbuffer[j] = GET_DATA_BYTE(line, j);
	}
	else if (colorflg == 1) {  /* -- 8 bpp colormapped -- */
	    for (j = 0; j < w; j++) {
		byteval = GET_DATA_BYTE(line, j);
		rowbuffer[3 * j] = rmap[byteval];
		rowbuffer[3 * j + 1] = gmap[byteval];
		rowbuffer[3 * j + 2] = bmap[byteval];
	    }
	}
	else { /* -- 24 bpp color -- */
	    ppixel = line;
	    for (j = k = 0; j < w; j++) {
		rowbuffer[k++] = GET_DATA_BYTE(ppixel, 0);
		rowbuffer[k++] = GET_DATA_BYTE(ppixel, 1);
		rowbuffer[k++] = GET_DATA_BYTE(ppixel, 2);
		ppixel++;
	    }
	}
        jpeg_write_scanlines(&comp, &rowbuffer, 1);
    }

    jpeg_finish_compress(&comp);

    FREE((void *)rowbuffer);
    if (colorflg == 1) {
	FREE((void *)rmap);
	FREE((void *)gmap);
	FREE((void *)bmap);
    }

    jpeg_destroy_compress(&comp);
    
    return 0;
}

