/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  morphapp.c
 *
 *      Selective connected component closing
 *            PIX     *pixSelectiveConnCompClose()
 *
 *      Selective connected component filling
 *            PIX     *pixSelectiveConnCompFill()
 *
 *      Iterative morphological seed filling (don't use for real work)
 *            PIX     *pixSeedfillMorph()
 *      
 *      Granulometry on binary images
 *            NUMA    *pixRunHistogram()
 *
 *      Composite operations on grayscale images
 *            PIX     *pixTophat()
 *            PIX     *pixHDome()
 *
 *      Centroids of PIXA
 *            PTA     *pixaCentroids()
 *
 *      These are some useful and/or interesting composite
 *      image processing operations, of the type that are often
 *      useful in applications.  Most are morphological in
 *      nature.
 */

#include <stdio.h>

#include "allheaders.h"


#define   SWAP(x, y)   {temp = (x); (x) = (y); (y) = temp;}



/*-----------------------------------------------------------------*
 *             Selective connected component closing               *
 *-----------------------------------------------------------------*/
/*!
 *    pixSelectiveConnCompClose()
 *
 *        Input:  pixs
 *                sel1
 *                sel2  (<optional>)
 *                connectivity (4 or 8)
 *                minw  (minimum width to consider; use 0 or 1 for any width)
 *                minh  (minimum height to consider; use 0 or 1 for any height)
 *        Return: pixd, or null on error
 *
 *        This closes each c.c in the pix that is larger
 *        than a certain minimum size with one or two sels.
 */
PIX *
pixSelectiveConnCompClose(PIX   *pixs,
                          SEL   *sel1,
                          SEL   *sel2,
         		  INT32  connectivity,
                          INT32  minw,
                          INT32  minh)
{
INT32   n, i;
BOX    *box;
BOXA   *boxa;
PIX    *pixt1, *pixt2, *pixd;
PIXA   *pixa;

    PROCNAME("pixSelectiveConnCompClose");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (!sel1)
	return (PIX *)ERROR_PTR("sel1 not defined", procName, NULL);

    if (minw <= 0) minw = 1;
    if (minh <= 0) minh = 1;

    if ((pixd = pixCopy(NULL, pixs)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    if ((boxa = pixConnComp(pixs, &pixa, connectivity)) == NULL)
	return (PIX *)ERROR_PTR("boxa not made", procName, NULL);
    n = boxaGetCount(boxa);
    for (i = 0; i < n; i++) {
	if ((box = boxaGetBox(boxa, i, CLONE)) == NULL)
	    return (PIX *)ERROR_PTR("box not found", procName, NULL);
	if (box->w >= minw && box->h >= minh) {
	    if ((pixt1 = pixaGetPix(pixa, i, CLONE)) == NULL)
		return (PIX *)ERROR_PTR("pixt1 not found", procName, NULL);
	    if ((pixt2 = pixCloseSafe(NULL, pixt1, sel1)) == NULL)
		return (PIX *)ERROR_PTR("pixt2 not made", procName, NULL);
	    if (sel2)
		pixCloseSafe(pixt2, pixt2, sel2);
	    pixRasterop(pixd, box->x, box->y, box->w, box->h, PIX_PAINT,
	                pixt2, 0, 0);
	    pixDestroy(&pixt1);
	    pixDestroy(&pixt2);
	}
	boxDestroy(&box);
    }
    pixaDestroy(&pixa);
    boxaDestroy(&boxa);

    return pixd;
}


/*-----------------------------------------------------------------*
 *             Selective connected component filling               *
 *-----------------------------------------------------------------*/
/*!
 *  pixSelectiveConnCompFill()
 *
 *      Input:  pixs (binary)
 *              connectivity (4 or 8)
 *              minw  (minimum width to consider; use 0 or 1 for any width)
 *              minh  (minimum height to consider; use 0 or 1 for any height)
 *      Return: pix with holes filled in selected c.c., or null on error
 * 
 */
PIX *
pixSelectiveConnCompFill(PIX   *pixs,
		         INT32  connectivity,
                         INT32  minw,
                         INT32  minh)
{
INT32   n, i;
BOX    *box;
BOXA   *boxa;
PIX    *pixt1, *pixt2, *pixd;
PIXA   *pixa;

    PROCNAME("pixSelectiveConnCompFill");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 1)
	return (PIX *)ERROR_PTR("pixs not 1 bpp", procName, NULL);

    if (minw <= 0) minw = 1;
    if (minh <= 0) minh = 1;

    if ((pixd = pixCopy(NULL, pixs)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    if ((boxa = pixConnComp(pixs, &pixa, connectivity)) == NULL)
	return (PIX *)ERROR_PTR("boxa not made", procName, NULL);
    n = boxaGetCount(boxa);
    for (i = 0; i < n; i++) {
	if ((box = boxaGetBox(boxa, i, CLONE)) == NULL)
	    return (PIX *)ERROR_PTR("box not found", procName, NULL);
	if (box->w >= minw && box->h >= minh) {
	    if ((pixt1 = pixaGetPix(pixa, i, CLONE)) == NULL)
		return (PIX *)ERROR_PTR("pixt1 not found", procName, NULL);
	    if ((pixt2 = pixHolesByFilling(pixt1, 12 - connectivity)) == NULL)
		return (PIX *)ERROR_PTR("pixt2 not made", procName, NULL);
	    pixRasterop(pixd, box->x, box->y, box->w, box->h, PIX_PAINT,
	                pixt2, 0, 0);
	    pixDestroy(&pixt1);
	    pixDestroy(&pixt2);
	}
	boxDestroy(&box);
    }
    pixaDestroy(&pixa);
    boxaDestroy(&boxa);

    return pixd;
}



/*-----------------------------------------------------------------*
 *             Iterative morphological seed filling                *
 *-----------------------------------------------------------------*/
/*!
 *  pixSeedfillMorph()
 *
 *       Input: pixs (seed)
 *              pixm (mask)
 *              connectivity (4 or 8)
 *       Return: pix where seed has been grown to completion
 *               into the mask, or null on error
 *
 *       Notes:  (1) This is in general a very inefficient method
 *                   for filling from a seed into a mask.  I've
 *                   included it here for pedagogical reasons, but
 *                   it should NEVER be used if efficiency is any
 *                   consideration -- use pixSeedfillBinary()!
 *               (2) We use a 3x3 brick SEL for 8-cc filling
 *                   and a 3x3 plus SEL for 4-cc.
 */
PIX  *
pixSeedfillMorph(PIX   *pixs,
		 PIX   *pixm,
		 INT32  connectivity)
{
INT32    same, iter;
PIX     *pixt1, *pixd, *temp;
SEL     *sel_3;

    PROCNAME("pixSeedfillMorph");

    if (!pixs)
	return (PIX *)ERROR_PTR("seed pix not defined", procName, NULL);
    if (!pixm)
	return (PIX *)ERROR_PTR("mask pix not defined", procName, NULL);
    if (connectivity != 4 && connectivity != 8)
	return (PIX *)ERROR_PTR("connectivity not in {4,8}", procName, NULL);

    if (pixSizesEqual(pixs, pixm) == 0)
	return (PIX *)ERROR_PTR("pix sizes unequal", procName, NULL);
    if (pixGetDepth(pixs) != 1)
	return (PIX *)ERROR_PTR("pix not binary", procName, NULL);

    if ((sel_3 = selCreateBrick(3, 3, 1, 1, 1)) == NULL)
	return (PIX *)ERROR_PTR("sel_3 not made", procName, NULL);
    if (connectivity == 4) {  /* remove corner hits to make a '+' */
	selSetElement(sel_3, 0, 0, DONT_CARE);
	selSetElement(sel_3, 2, 2, DONT_CARE);
	selSetElement(sel_3, 2, 0, DONT_CARE);
	selSetElement(sel_3, 0, 2, DONT_CARE);
    }

    if ((pixt1 = pixCopy(NULL, pixs)) == NULL)
	return (PIX *)ERROR_PTR("pixt1 not made", procName, NULL);
    if ((pixd = pixCreateTemplate(pixs)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    iter = 0;
    while (1) {
	iter++;
	pixDilate(pixd, pixt1, sel_3);
	pixAnd(pixd, pixd, pixm);
	pixEqual(pixd, pixt1, &same);
	if (same)
	    break;
	else
	    SWAP(pixt1, pixd);
    }
    fprintf(stderr, " Num iters in binary reconstruction = %d\n", iter);

    pixDestroy(&pixt1);
    selDestroy(&sel_3);

    return pixd;
}



/*-----------------------------------------------------------------*
 *                   Granulometry on binary images                 *
 *-----------------------------------------------------------------*/
/*!
 *  pixRunHistogram()
 *
 *       Input: pixs
 *              runtype (OFF, ON)
 *              direction (HORIZ, VERT)
 *              maxsize  (size of largest runlength counted)
 *       Return: numa of run-lengths
 *
 */
NUMA  *
pixRunHistogram(PIX    *pixs,
		INT32   runtype,
		INT32   direction,
		INT32   maxsize)
{
INT32     count, i;
FLOAT32   val;
NUMA     *na, *nah;
PIX      *pixt1, *pixt2, *pixt3;
SEL      *sel_2h;

    PROCNAME("pixRunHistogram");

    if (!pixs)
	return (NUMA *)ERROR_PTR("seed pix not defined", procName, NULL);
    if (runtype != OFF && runtype != ON)
	return (NUMA *)ERROR_PTR("runtype must be {OFF,ON}", procName, NULL);
    if (direction != HORIZ && direction != VERT)
	return (NUMA *)ERROR_PTR("direction must be {HORIZ,VERT}",
	                         procName, NULL);

    if (pixGetDepth(pixs) != 1)
	return (NUMA *)ERROR_PTR("pixs must be binary", procName, NULL);

    if ((na = numaCreate(0)) == NULL)
	return (NUMA *)ERROR_PTR("na not made", procName, NULL);

    if ((sel_2h = selCreateBrick(2, 1, 0, 0, 1)) == NULL)
	return (NUMA *)ERROR_PTR("sel_2h not made", procName, NULL);

    if (runtype == OFF) {
	if ((pixt1 = pixCopy(NULL, pixs)) == NULL)
	    return (NUMA *)ERROR_PTR("pix1 not made", procName, NULL);
	pixInvert(pixt1, pixt1);
    }
    else  /* runtype == ON */
	pixt1 = pixClone(pixs);

    if ((pixt2 = pixCreateTemplate(pixs)) == NULL)
	return (NUMA *)ERROR_PTR("pix2 not made", procName, NULL);
    if ((pixt3 = pixCreateTemplate(pixs)) == NULL)
	return (NUMA *)ERROR_PTR("pix3 not made", procName, NULL);

	/* get pixel counts at different stages of erosion */
    pixCountPixels(pixt1, &count, NULL);
    numaAddNumber(na, count);
    pixErode(pixt2, pixt1, sel_2h);
    pixCountPixels(pixt2, &count, NULL);
    numaAddNumber(na, count);
    for (i = 0; i < maxsize / 2; i++) {
	pixErode(pixt3, pixt2, sel_2h);
	pixCountPixels(pixt3, &count, NULL);
	numaAddNumber(na, count);
	pixErode(pixt2, pixt3, sel_2h);
	pixCountPixels(pixt2, &count, NULL);
	numaAddNumber(na, count);
/*	if (i == 4) pixWrite("junkoutt", pixt2, IFF_PNG); */
    }

	/* compute length histogram */
    if ((nah = numaCreate(na->n)) == NULL)
	return (NUMA *)ERROR_PTR("nah not made", procName, NULL);
    numaAddNumber(nah, 0); /* number at length 0 */
    for (i = 1; i < na->n - 1; i++) {
	val = na->array[i+1] - 2 * na->array[i] + na->array[i-1];
	numaAddNumber(nah, val);
/*        fprintf(stderr, "i = %d, val = %f\n", i, nah->array[i]); */
    }

    pixDestroy(&pixt1);
    pixDestroy(&pixt2);
    pixDestroy(&pixt3);
    selDestroy(&sel_2h);
    numaDestroy(&na);

    return nah;
}


/*-----------------------------------------------------------------*
 *            Composite operations on grayscale images             *
 *-----------------------------------------------------------------*/
/*!
 *   pixTophat()
 *
 *     Input:  pixs
 *             size (of SEL; must be odd; center implicitly in center)
 *             shape  (HORIZ, VERT, SQUARE)
 *             type   (WHITE: image - opening
 *                     BLACK: closing - image)
 *     Return: pixd, or null on error
 *
 *     Note: The WHITE tophat emphasizes small bright regions,
 *           whereas the BLACK tophat emphasizes small dark regions.
 *           The WHITE tophat can be accomplished by doing a BLACK
 *           tophat on the inverse, or v.v.
 */
PIX  *
pixTophat(PIX   *pixs,
	  INT32  size,
	  INT32  shape,
	  INT32  type)
{
PIX     *pixd;

    PROCNAME("pixTophat");

    if (!pixs)
	return (PIX *)ERROR_PTR("seed pix not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
	return (PIX *)ERROR_PTR("pixs not 8 bpp", procName, NULL);
    if (shape != HORIZ && shape != VERT && shape != SQUARE)
	return (PIX *)ERROR_PTR("shape not {HORIZ, VERT, SQUARE}",
		procName, NULL);
    if ((size & 1) == 0) {
	WARNING("SEL size must be odd; increasing by 1", procName);
	size++;
    }
    if (type != WHITE && type != BLACK)
	return (PIX *)ERROR_PTR("type must be BLACK or WHITE", procName, NULL);

    switch (type)
    {
    case WHITE:
	if ((pixd = pixOpenGray(pixs, size, shape)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
	pixSubtractGray(pixd, pixs, pixd);
	break;
    case BLACK:
	if ((pixd = pixCloseGray(pixs, size, shape)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
	pixSubtractGray(pixd, pixd, pixs);
	break;
    default:
	WARNING("invalid type", procName);
    }

    return pixd;
}
	

/*!
 *   pixHDome()
 *
 *     Input:  pixs
 *             height (of hdome; should be > 0)
 *     Return: pixd, or null on error
 *
 *     Note:    we use connectivity of 4 for the fill
 *     Action:  pixd starts as a seed, with every pixel value set
 *              to be h less than the corresponding pixel in pixs.
 *              We then do a grayscale seed fill, clipping pixd to
 *              the mask pixs.  Finally, we subtract the result from
 *              the original.
 *
 */
PIX  *
pixHDome(PIX   *pixs,
	 INT32  height)
{
PIX     *pixd;

    PROCNAME("pixHDome");

    if (!pixs)
	return (PIX *)ERROR_PTR("src pix not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
	return (PIX *)ERROR_PTR("pixs not 8 bpp", procName, NULL);

    if ((pixd = pixCopy(NULL, pixs)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    pixAddConstantGray(pixd, -ABS(height));
    pixSeedfillGray(pixd, pixs, 4);
    pixSubtractGray(pixd, pixs, pixd);

    return pixd;
}



/*-----------------------------------------------------------------*
 *                            Center of mass                       *
 *-----------------------------------------------------------------*/
/*!
 *  pixaCentroids()
 *
 *     Input:  pixa of components
 *     Return: pta of centroids relative to the UL corner of
 *             each pix, or null on error
 */
PTA  *
pixaCentroids(PIXA   *pixa)
{
INT32    i, j, k, n, w, h, wpl, npix;
FLOAT32  xsum, ysum, xave, yave;
UINT32  *data, *line;
PIX     *pix;
PTA     *pta;

    PROCNAME("pixaCentroids");

    if (!pixa)
	return (PTA *)ERROR_PTR("pixa not defined", procName, NULL);

    n = pixaGetCount(pixa);
    if ((pta = ptaCreate(n)) == NULL)
	return (PTA *)ERROR_PTR("pta not defined", procName, NULL);

    for (k = 0; k < n; k++) {
	pix = pixaGetPix(pixa, k, CLONE);
	w = pixGetWidth(pix);
	h = pixGetHeight(pix);
	if (pixGetDepth(pix) != 1)
	    return (PTA *)ERROR_PTR("pix not 1 bpp", procName, NULL);
	data = pixGetData(pix);
	wpl = pixGetWpl(pix);
	xsum = ysum = 0;
	npix = 0;
	for (i = 0; i < h; i++) {
	    line = data + wpl * i;
	    for (j = 0; j < w; j++) {
		if (GET_DATA_BIT(line, j)) {
		    xsum += j;
		    ysum += i;
		    npix++;
		}
	    }
	}
	if (npix == 0) {
	    WARNING("no ON pixels in pix", procName);
	    ptaAddPt(pta, 0.0, 0.0);   /* this shouldn't happen */
	}
	else {
	    xave = xsum / npix;
	    yave = ysum / npix;
	    ptaAddPt(pta, xave, yave);
	}
	pixDestroy(&pix);
    }

    return pta;
}


