/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *   numarray.c
 *
 *      Create/Destroy/Copy/Clone/Empty Numa
 *          NUMA      *numaCreate()
 *          void      *numaDestroy()
 *          NUMA      *numaCopy()
 *          NUMA      *numaClone()
 *          INT32      numaEmpty()
 *
 *      Add number (float or integer) to Numa
 *          INT32      numaAddNumber()
 *          INT32      numaExtendArray()
 *
 *      Numa Accessors
 *          INT32      numaGetCount()
 *          INT32      numaGetIValue()
 *          INT32      numaGetFValue()
 *          INT32      numaSetValue()
 *          INT32     *numaGetIArray()
 *          FLOAT32   *numaGetFArray()
 *          INT32      numaGetRefcount()
 *          INT32      numaChangeRefcount()
 *
 *      Misc computation
 *          INT32      numaGetMax()
 *          INT32      numaGetMin()
 *          INT32      numaFitMax()
 *          INT32      numaGetSum()
 *          NUMA      *numaSort()
 *          INT32      numaGetMedian()
 *          INT32      numaGetMode()
 *          NUMA      *numaConvertToInt()
 *          NUMA      *numaGetHistogram()
 *          NUMA      *numaGetHistogramClipped()
 *          NUMA      *numaConvolve()
 *          NUMA      *numaFindPeaks()
 *          NUMA      *numaMakeSequence()
 *
 *      Serialize Numa for I/O
 *          INT32      numaRead()
 *          INT32      numaReadStream()
 *          INT32      numaWrite()
 *          INT32      numaWriteStream()
 *
 *      Create/Destroy Numaa
 *          NUMAA     *numaaCreate()
 *          void      *numaaDestroy()
 *
 *      Add Numa to Numaa
 *          INT32      numaaAddNuma()
 *          INT32      numaaExtendArray()
 *
 *      Numaa Accessors
 *          INT32      numaaGetCount()
 *          NUMA      *numaaGetNuma()
 *       
 *
 *    The number array holds FLOAT32 values.  It can also
 *    be used to store INT32 values.  This versatility is the
 *    reason it's called a "number" array.
 *
 *    To handle integer values:
 *
 *       - to add a new one to the array, use numaAddNumber() with
 *         the integer input.  The type will automatically be converted
 *         to FLOAT32 and stored.
 *
 *       - to set a value in the array, use numaSetValue() with
 *         the integer input.  It will be converted to FLOAT32 and stored.
 *
 *       - to get a value in the array, use either numaGetIValue()
 *         or numaGetFValue(), depending on whether you are retrieving
 *         an integer or a float.  This avoids forcing the user to do
 *         an explicit cast, which would happen in one of two ways:
 *           (a) return a FLOAT32 and cast it to an INT32
 *           (b) cast the return directly to (FLOAT32 *) to
 *               satisfy the function prototype, as in
 *                 numaGetFValue(na, index, (FLOAT32 *)&ival);
 *              (OK, that is quite ugly!)
 *
 *    Tradition dictates that type conversions go automatically from
 *    INT32 --> FLOAT32, even though it is possible to lose
 *    precision for large integers, whereas you must cast (INT32)
 *    to go from FLOAT32 --> INT32 because you're truncating
 *    to the integer value.
 *                   
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "allheaders.h"

    /* bin sizes in numaGetHistogram() */
static const INT32 BinSizeArray[] = {2, 5, 10, 20, 50, 100, 200, 500, 1000,\
                      2000, 5000, 10000, 20000, 50000, 100000, 200000,\
		      500000, 1000000, 2000000, 5000000, 10000000,\
		      200000000, 50000000, 100000000};
static const INT32 NBinSizes = 24;


#define   INITIAL_PTR_ARRAYSIZE     50      /* n'importe quoi */


/*--------------------------------------------------------------------------*
 *                     Create/destroy/copy/empty Numa                       *
 *--------------------------------------------------------------------------*/
/*!
 *  numaCreate()
 *
 *       Input:  size of number array to be alloc'd (0 for default)
 *       Return: na, or null on error
 *
 */
NUMA *
numaCreate(INT32   n)
{
NUMA   *na;

    PROCNAME("numaCreate");

    if (n <= 0)
	n = INITIAL_PTR_ARRAYSIZE;

    if ((na = (NUMA *)CALLOC(1, sizeof(NUMA))) == NULL)
        return (NUMA *)ERROR_PTR("na not made", procName, NULL);
    if ((na->array = (FLOAT32 *)CALLOC(n, sizeof(FLOAT32))) == NULL)
        return (NUMA *)ERROR_PTR("number array not made", procName, NULL);

    na->nalloc = n;
    na->n = 0;
    na->refcount = 1;

    return na;
}


/*!
 *  numaDestroy()
 *
 *       Input:  &numa <may be nulled>
 *       Return: void
 *       Action: destroys the numa and nulls the contents
 *               of the input ptr.  If more than one ref,
 *               it just decrements the refcount and
 *               nulls the content of the input ptr.
 */
void
numaDestroy(NUMA  **pna)
{
NUMA   *na;

    PROCNAME("numaDestroy");

    if (pna == NULL) {
	WARNING("ptr address is NULL", procName);
	return;
    }

    if ((na = *pna) == NULL)
	return;

	/* decrement the ref count.  If it is still > 0, just return */
    numaChangeRefcount(na, -1);
    if (numaGetRefcount(na) > 0) {
	*pna = NULL;
	return;
    }

    if (na->array)
	FREE((void *)na->array);
    FREE((void *)na);
    *pna = NULL;

    return;
}

        
/*!
 *  numaCopy()
 *
 *     Input:  numa
 *     Return: copy of numa, or null on error
 */
NUMA *
numaCopy(NUMA  *na)
{
INT32      i;
NUMA      *cna;

    PROCNAME("numaCopy");

    if (!na)
        return (NUMA *)ERROR_PTR("na not defined", procName, NULL);

    if ((cna = numaCreate(na->nalloc)) == NULL)
        return (NUMA *)ERROR_PTR("cna not made", procName, NULL);

    for (i = 0; i < na->n; i++)
	numaAddNumber(cna, na->array[i]);

    return cna;
}


/*!
 *  numaClone()
 *
 *     Input:  numa
 *     Return: ptr to same numa, or null on error
 */
NUMA *
numaClone(NUMA  *na)
{

    PROCNAME("numaClone");

    if (!na)
        return (NUMA *)ERROR_PTR("na not defined", procName, NULL);

    numaChangeRefcount(na, 1);
    return na;
}


/*!
 *  numaEmpty()
 *
 *     Input:  numa
 *     Return: 0 if OK; 1 on error
 *     Note: this does not change the allocation of the array;
 *           it just clears the number of stored numbers.
 */
INT32
numaEmpty(NUMA  *na)
{

    PROCNAME("numaEmpty");

    if (!na)
        return ERROR_INT("na not defined", procName, 1);

    na->n = 0;
    return 0;
}



/*--------------------------------------------------------------------------*
 *                 Number array: add number and extend array                *
 *--------------------------------------------------------------------------*/
/*!
 *     numaAddNumber()
 *
 *         Input:  numa
 *                 val  (float32 or integer to be added)
 *
 *         Return: 0 if OK, 1 on error
 *
 */
INT32
numaAddNumber(NUMA    *na,
              FLOAT32  val)
{
INT32     n;

    PROCNAME("numaAddNumber");

    if (!na)
        return ERROR_INT("na not defined", procName, 1);
    
    n = numaGetCount(na);
    if (n >= na->nalloc)
	numaExtendArray(na);
    na->array[n] = val;
    na->n++;
    return 0;
}


/*!
 *     numaExtendArray()
 *
 *         Input:  numa
 *         Return: 0 if OK, 1 on error
 */
INT32
numaExtendArray(NUMA  *na)
{

    PROCNAME("numaExtendArray");

    if (!na)
        return ERROR_INT("na not defined", procName, 1);

    if ((na->array = (FLOAT32 *)reallocNew((void **)&na->array,
				4 * na->nalloc,
				8 * na->nalloc)) == NULL)
	    return ERROR_INT("new ptr array not returned", procName, 1);

    na->nalloc *= 2;
    return 0;
}



/*----------------------------------------------------------------------*
 *                            Numa Accessors                            *
 *----------------------------------------------------------------------*/
/*!
 *  numaGetCount()
 *
 *       Input: numa
 *       Return: count, or 0 if no numbers or on error
 */
INT32
numaGetCount(NUMA  *na)
{

    PROCNAME("numaGetCount");

    if (!na)
        return ERROR_INT("na not defined", procName, 0);
    return na->n;
}
	

/*!
 *     numaGetFValue()
 *
 *         Input:  numa
 *                 index (into numa)
 *                 &val  (<return> float value)
 *         Return: 0 if OK; 1 on error
 *
 */
INT32
numaGetFValue(NUMA      *na,
              INT32      index,
	      FLOAT32   *pval)
{
INT32    *array;

    PROCNAME("numaGetFValue");

    if (!na)
        return ERROR_INT("na not defined", procName, 1);
    if (!pval)
        return ERROR_INT("&val not defined", procName, 1);
    *pval = 0.0;

    if (index < 0 || index >= na->n)
        return ERROR_INT("index not valid", procName, 1);

    *pval = na->array[index];
    return 0;
}


/*!
 *     numaGetIValue()
 *
 *         Input:  numa
 *                 index (into numa)
 *                 &ival  (<return> integer value)
 *         Return: 0 if OK; 1 on error
 *
 */
INT32
numaGetIValue(NUMA    *na,
              INT32    index,
	      INT32   *pival)
{
INT32    *array;

    PROCNAME("numaGetIValue");

    if (!na)
        return ERROR_INT("na not defined", procName, 1);
    if (!pival)
        return ERROR_INT("&ival not defined", procName, 1);
    *pival = 0;

    if (index < 0 || index >= na->n)
        return ERROR_INT("index not valid", procName, 1);

    *pival = (INT32)(na->array[index] + 0.5);
    return 0;
}


/*!
 *     numaSetValue()
 *
 *         Input:  numa
 *                 index   (to element to be set)
 *                 val  (to set element)
 *         Return: 0 if OK; 1 on error
 */
INT32
numaSetValue(NUMA    *na,
             INT32    index,
             FLOAT32  val)
{

    PROCNAME("numaSetValue");

    if (!na)
        return ERROR_INT("na not defined", procName, 1);
    if (index < 0 || index >= na->n)
        return ERROR_INT("index not valid", procName, 1);

    na->array[index] = val;
    return 0;
}


/*!
 *     numaGetIArray()
 *
 *         Input:  numa
 *         Return: a copy of the bare internal array, integerized
 *                 by rounding, or null on error
 *
 */
INT32 *
numaGetIArray(NUMA    *na)
{
INT32     i, ival;
INT32    *array;

    PROCNAME("numaGetIArray");

    if (!na)
        return (INT32 *)ERROR_PTR("na not defined", procName, NULL);

    if ((array = (INT32 *)CALLOC(na->n, sizeof(INT32))) == NULL)
        return (INT32 *)ERROR_PTR("array not made", procName, NULL);
    for (i = 0; i < na->n; i++) {
	numaGetIValue(na, i, &ival);
	array[i] = ival;
    }

    return array;
}


/*!
 *     numaGetFArray()
 *
 *         Input:  numa
 *         Return: a copy of the bare internal array, or null on error
 *
 */
FLOAT32 *
numaGetFArray(NUMA    *na)
{
INT32     i;
FLOAT32   fval;
FLOAT32  *array;

    PROCNAME("numaGetFArray");

    if (!na)
        return (FLOAT32 *)ERROR_PTR("na not defined", procName, NULL);

    if ((array = (FLOAT32 *)CALLOC(na->n, sizeof(FLOAT32))) == NULL)
        return (FLOAT32 *)ERROR_PTR("array not made", procName, NULL);
    for (i = 0; i < na->n; i++) {
	numaGetFValue(na, i, &fval);
	array[i] = fval;
    }

    return array;
}


/*!
 *   numaGetRefCount()
 *
 *       Input: na
 *       Return: refcount, or UNDEF on error
 */
INT32
numaGetRefcount(NUMA   *na)
{
    PROCNAME("numaGetRefcount");

    if (!na)
        return ERROR_INT("na not defined", procName, UNDEF);
    return na->refcount;
}


/*!
 *   numaChangeRefCount()
 *
 *       Input: na
 *              delta (change to be applied)
 *       Return: 0 if OK, 1 on error
 */
INT32
numaChangeRefcount(NUMA   *na,
                   INT32   delta)
{
    PROCNAME("numaChangeRefcount");

    if (!na)
        return ERROR_INT("na not defined", procName, 1);
    na->refcount += delta;
    return 0;
}



/*----------------------------------------------------------------------*
 *                          Misc computation                            *
 *----------------------------------------------------------------------*/
/*!
 *  numaGetMax()
 *
 *       Input: na
 *              &maxval (<return> max value)
 *              &imaxloc (<return> index of max location)
 *       Return: 0 if OK; 1 on error
 */
INT32
numaGetMax(NUMA      *na,
           FLOAT32   *pmaxval,
           INT32     *pimaxloc)
{
INT32     i, n, imaxloc;
FLOAT32   val, maxval;

    PROCNAME("numaGetMax");

    *pmaxval = 0.0;  /* init */
    *pimaxloc = 0;

    if (!na)
        return ERROR_INT("na not defined", procName, 1);
    if (!pmaxval)
        return ERROR_INT("&maxval not defined", procName, 1);
    if (!pimaxloc)
        return ERROR_INT("&imaxloc not defined", procName, 1);

    maxval = -1000000000.;
    imaxloc = 0;
    n = numaGetCount(na);
    for (i = 0; i < n; i++) {
	numaGetFValue(na, i, &val);
	if (val > maxval) {
	    maxval = val;
	    imaxloc = i;
	}
    }
    *pmaxval = maxval;
    *pimaxloc = imaxloc;
    return 0;
}


/*!
 *  numaGetMin()
 *
 *       Input: na
 *              &minval (<return> min value)
 *              &iminloc (<return> index of min location)
 *       Return: 0 if OK; 1 on error
 */
INT32
numaGetMin(NUMA      *na,
           FLOAT32   *pminval,
           INT32     *piminloc)
{
INT32     i, n, iminloc;
FLOAT32   val, minval;

    PROCNAME("numaGetMin");

    *pminval = 0.0;  /* init */
    *piminloc = 0;

    if (!na)
        return ERROR_INT("na not defined", procName, 1);
    if (!pminval)
        return ERROR_INT("&minval not defined", procName, 1);
    if (!piminloc)
        return ERROR_INT("&iminloc not defined", procName, 1);

    minval = +1000000000.;
    iminloc = 0;
    n = numaGetCount(na);
    for (i = 0; i < n; i++) {
	numaGetFValue(na, i, &val);
	if (val < minval) {
	    minval = val;
	    iminloc = i;
	}
    }
    *pminval = minval;
    *piminloc = iminloc;
    return 0;
}


/*!
 *  numaGetSum()
 *
 *     Input: na
 *           &sum (<return> sum of values)
 *     Return: 0 if OK, 1 on error
 */
INT32
numaGetSum(NUMA     *na,
           FLOAT32  *psum)
{
INT32    i, n;
FLOAT32  val, sum;

    PROCNAME("numaGetSum");

    if (!na)
        return ERROR_INT("na not defined", procName, 1);

    sum = 0.0;
    n = numaGetCount(na);
    for (i = 0; i < n; i++) {
        numaGetFValue(na, i, &val);	
	sum += val;
    }
    *psum = sum;
    return 0;
}


/*!
 *  numaFitMax()
 *
 *       Input: na  (numa of ordinate values, to fit a max to)
 *              &maxval (<return> max value)
 *              naloc (<optional> associated numa of abscissa values)
 *              &maxloc (<return> abscissa value that gives max value in na;
 *                    if naloc == null, this is given as an interpolated
 *                    index value)
 *       Return: 0 if OK; 1 on error
 *
 *       Note: if naloc is given, there is no requirement that the
 *             data points are evenly spaced.  Lagrangian interpolation
 *             handles that.  The only requirement is that the
 *             data points are ordered so that the values in naloc
 *             are either increasing or decreasing.  We test to make
 *             sure that the sizes of na and naloc are equal, and it
 *             is assumed that the correspondences na[i] as a function
 *             of naloc[i] are properly arranged for all i.
 *
 *       The formula for Lagrangian interpolation through 3 data pts is:
 *            y(x) = y1(x-x2)(x-x3)/((x1-x2)(x1-x3)) +
 *                   y2(x-x1)(x-x3)/((x2-x1)(x2-x3)) +
 *                   y3(x-x1)(x-x2)/((x3-x1)(x3-x2))
 *
 *       Then the derivative, using the constants (c1,c2,c3) defined below,
 *       is set to 0:
 *            y'(x) = 2x(c1+c2+c3) - c1(x2+x3) - c2(x1+x3) - c3(x1+x2) = 0
 */
INT32
numaFitMax(NUMA     *na,
           FLOAT32  *pmaxval,
           NUMA     *naloc,
           FLOAT32  *pmaxloc)
{
FLOAT32   maxval, maxloc, val;
FLOAT32   smaxval;  /* start value of maximum sample, before interpolating */
INT32     n, imaxloc;
FLOAT32   x1, x2, x3, y1, y2, y3, c1, c2, c3, a, b, xmax, ymax;

    PROCNAME("numaFitMax");

    *pmaxval = *pmaxloc = 0.0;  /* init */

    if (!na)
        return ERROR_INT("na not defined", procName, 1);
    if (!pmaxval)
        return ERROR_INT("&maxval not defined", procName, 1);
    if (!pmaxloc)
        return ERROR_INT("&maxloc not defined", procName, 1);

    n = numaGetCount(na);
    if (naloc) {
	if (n != numaGetCount(naloc))
	    return ERROR_INT("na and naloc of unequal size", procName, 1);
    }

    numaGetMax(na, &smaxval, &imaxloc);

	/* simple case: max is at end point */
    if (imaxloc == 0 || imaxloc == n - 1) {
	*pmaxval = smaxval;
	if (naloc) {
	    numaGetFValue(naloc, imaxloc, &val);
	    *pmaxloc = val;
	}
	else 
	    *pmaxloc = imaxloc;
	return 0;
    }

	/* interior point; use quadratic interpolation */
    y2 = smaxval;
    numaGetFValue(na, imaxloc - 1, &val);
    y1 = val;
    numaGetFValue(na, imaxloc + 1, &val);
    y3 = val;
    if (naloc) {
	numaGetFValue(naloc, imaxloc - 1, &val);
	x1 = val;
	numaGetFValue(naloc, imaxloc, &val);
	x2 = val;
	numaGetFValue(naloc, imaxloc + 1, &val);
	x3 = val;
    }
    else {
	x1 = imaxloc - 1;
	x2 = imaxloc;
	x3 = imaxloc + 1;
    }

	/* can't interpolate; just use the max val in na
	 * and the corresponding one in naloc */
    if (x1 == x2 || x1 == x3 || x2 == x3) {
	*pmaxval = y2;
	*pmaxloc = x2;
	return 0;
    }

	/* use lagrangian interpolation; set dy/dx = 0 */
    c1 = y1 / ((x1 - x2) * (x1 - x3));
    c2 = y2 / ((x2 - x1) * (x2 - x3));
    c3 = y3 / ((x3 - x1) * (x3 - x2));
    a = c1 + c2 + c3;
    b = c1 * (x2 + x3) + c2 * (x1 + x3) + c3 * (x1 + x2);
    xmax = b / (2 * a);
    ymax = c1 * (xmax - x2) * (xmax - x3) +
           c2 * (xmax - x1) * (xmax - x3) +
           c3 * (xmax - x1) * (xmax - x2);
    *pmaxval = ymax;
    *pmaxloc = xmax;

    return 0;
}


/*!
 *  numaSort()
 *
 *       Input: na
 *       Return: numa sorted in decreasing order, or null on error
 *       
 *       Source: K&R, 2nd edition, p.62.  They produce a sort in
 *               increasing order, using the ">" comparison in the
 *               innermost "for" loop.
 */
NUMA *
numaSort(NUMA   *na)
{
INT32     i, n, gap, j;
FLOAT32   tmp;
FLOAT32  *array;
NUMA     *nas;

    PROCNAME("numaSort");

    if (!na)
        return (NUMA *)ERROR_PTR("na not defined", procName, NULL);

    if ((array = numaGetFArray(na)) == NULL)
        return (NUMA *)ERROR_PTR("array not made", procName, NULL);
    n = numaGetCount(na);

	/* shell sort */
    for (gap = n/2; gap > 0; gap = gap / 2) {
	for (i = gap; i < n; i++) {
	    for (j = i - gap; j >= 0 && array[j] < array[j + gap]; j -= gap) {
		tmp = array[j];
		array[j] = array[j + gap];
		array[j + gap] = tmp;
	    }
	}
    }

    nas = numaCreate(n);
    for (i = 0; i < n; i++) 
	numaAddNumber(nas, array[i]);

    FREE((void *)array);
    return nas;
}


/*!
 *  numaGetMedian()
 *
 *       Input: na
 *              &val  (<return> median val)
 *       Return: 0 if OK; 1 on error
 */
INT32
numaGetMedian(NUMA     *na,
              FLOAT32  *pval)
{
INT32     n;
NUMA     *nasort;

    PROCNAME("numaGetMedian");

    if (!na)
        return ERROR_INT("na not defined", procName, 1);
    if (!pval)
        return ERROR_INT("&val not defined", procName, 1);
    *pval = 0.0;  /* init */

	/* get array sorted in decreasing order */
    n = numaGetCount(na);
    if (n == 0)
	return 1;
    if ((nasort = numaSort(na)) == NULL)
	return ERROR_INT("nasort not made", procName, 1);
    numaGetFValue(nasort, n / 2, pval);

    numaDestroy(&nasort);
    return 0;
}




/*!
 *  numaGetMode()
 *
 *       Input: na
 *              &val  (<return> mode val)
 *              &count  (<return> mode count)
 *       Return: 0 if OK; 1 on error
 *       
 */
INT32
numaGetMode(NUMA     *na,
            FLOAT32  *pval,
	    INT32    *pcount)
{
INT32     i, n, maxcount, prevcount;
FLOAT32   val, maxval, prevval;
FLOAT32  *array;
NUMA     *nasort;

    PROCNAME("numaGetMode");

    if (!na)
        return ERROR_INT("na not defined", procName, 1);
    if (!pval)
        return ERROR_INT("&val not defined", procName, 1);
    if (!pcount)
        return ERROR_INT("&count not defined", procName, 1);

    *pval = 0.0;
    *pcount = 0;
    if ((n = numaGetCount(na)) == 0)
	return 1;

	/* get array sorted in decreasing order */
    if ((nasort = numaSort(na)) == NULL)
        return ERROR_INT("nas not made", procName, 1);
    if ((array = numaGetFArray(nasort)) == NULL)
        return ERROR_INT("array not made", procName, 1);

        /* initialize with array[0] */
    prevval = array[0];
    prevcount = 1;
    maxval = prevval;
    maxcount = prevcount;

       /* scan the sorted array */
    for (i = 1; i < n; i++) {
	val = array[i];
	if (val == prevval)
	    prevcount++;
	else {  /* new value */
	    if (prevcount > maxcount) {  /* new max */
	        maxcount = prevcount;
		maxval = prevval;
	    }
	    prevval = val;
	    prevcount = 1;
	}
    }

	/* was the mode the last run of elements? */
    if (prevcount > maxcount) {
        maxcount = prevcount;
	maxval = prevval;
    }

    *pval = maxval;
    *pcount = maxcount;

    numaDestroy(&nasort);
    FREE((void *)array);
    return 0;
}


/*!
 *  numaConvertToInt()
 *
 *       Input: na
 *       Return: na with all values rounded to nearest integer, or
 *               null on error
 */
NUMA *
numaConvertToInt(NUMA  *nas)
{
INT32     i, n, ival;
NUMA     *nad;

    PROCNAME("numaConvertToInt");

    if (!nas)
        return (NUMA *)ERROR_PTR("nas not defined", procName, NULL);

    n = numaGetCount(nas);
    if ((nad = numaCreate(n)) == NULL)
        return (NUMA *)ERROR_PTR("nad not made", procName, NULL);
    for (i = 0; i < n; i++) {
	numaGetIValue(nas, i, &ival);
	numaAddNumber(nad, ival);
    }
    return nad;
}


/*!
 *  numaMakeHistogram()
 *
 *       Input: na
 *              maxnbins (max number of histogram bins)
 *             &binsize  (<return> size of histogram bins)
 *             &binstart (<optional return> start val of minimum bin;
 *                        input NULL to force start at 0)
 *       Return: na consisiting of histogram of integerized values,
 *               or null on error.
 *
 *       Note: We specify the max number of input bins, and
 *             are returned the size of bins necessary to accommodate
 *             the input data.  The size is one of the sequence:
 *             {1, 2, 5, 10, 20, 50, ...}.
 *             If &binstart is given, all values are accommodated,
 *             and the min value of the starting bin is returned;
 *             otherwise, all negative values are discarded and
 *             the histogram bins start at 0.
 *
 */
NUMA *
numaMakeHistogram(NUMA   *na,
                  INT32   maxbins,
		  INT32  *pbinsize,
		  INT32  *pbinstart)
{
INT32     i, n, ival, iloc, nhist, hval;
INT32     iminval, imaxval, range, binsize, nbins, ibin;
FLOAT32   val, ratio;
NUMA     *nai, *nahist;

    PROCNAME("numaMakeHistogram");

    if (!na)
        return (NUMA *)ERROR_PTR("na not defined", procName, NULL);
    if (!pbinsize)
        return (NUMA *)ERROR_PTR("&binsize not defined", procName, NULL);

        /* determine input range */
    numaGetMin(na, &val, &iloc);
    iminval = (INT32)(val + 0.5);
    numaGetMax(na, &val, &iloc);
    imaxval = (INT32)(val + 0.5);
    if (pbinstart == NULL) {  /* clip negative vals; start from 0 */
        iminval = 0;
	if (imaxval < 0)
	    return (NUMA *)ERROR_PTR("all values < 0", procName, NULL);
    }

        /* determine binsize */
    range = imaxval - iminval + 1;
    if (range > maxbins - 1) {
        ratio = (FLOAT64)range / (FLOAT64)maxbins;
	binsize = 0;
	for (i = 0; i < NBinSizes; i++) {
	    if (ratio < BinSizeArray[i]) {
	        binsize = BinSizeArray[i];
		break;
	    }
	}
	if (binsize == 0)
	    return (NUMA *)ERROR_PTR("numbers too large", procName, NULL);
    }
    else
        binsize = 1;
    *pbinsize = binsize;
    nbins = 1 + range / binsize;  /* +1 seems to be sufficient */

        /* redetermine iminval */
    if (pbinstart && binsize > 1) {
        if (iminval >= 0)
	    iminval = binsize * (iminval / binsize);
	else
	    iminval = binsize * ((iminval - binsize + 1) / binsize);
    }
    if (pbinstart)
	*pbinstart = iminval;
    fprintf(stderr, " imaxval = %d, range = %d, nbins = %d\n",
            imaxval, range, nbins);

        /* use integerized data for input */
    if ((nai = numaConvertToInt(na)) == NULL)
        return (NUMA *)ERROR_PTR("nai not made", procName, NULL);
    n = numaGetCount(nai);

        /* make histogram, converting value in input array 
	 * into a bin number for this histogram array. */
    if ((nahist = numaCreate(nbins)) == NULL)
        return (NUMA *)ERROR_PTR("nahist not made", procName, NULL);
    nahist->n = nbins;  /* fake the storage of nbins zeroes */
    for (i = 0; i < n; i++) {
	numaGetIValue(nai, i, &ival);
        ibin = (ival - iminval) / binsize;
	if (ibin >= 0 && ibin < nbins) {
	    numaGetIValue(nahist, ibin, &hval);
	    numaSetValue(nahist, ibin, hval + 1.0);
	}
    }

    numaDestroy(&nai);
    return nahist;
}


/*!
 *  numaMakeHistogramClipped()
 *
 *       Input: na
 *              maxsize (of histogram ordinate)
 *       Return: na consisiting of histogram of unit bins, non-negative
 *               starting at 0 and going up to a maximum of maxsize,
 *               or null on error.
 *
 *       Note: This is a very simple function.
 *             We integerize the array, ignore all negative
 *             values and all values larger than the maximum
 *             value (maxsize).
 */
NUMA *
numaMakeHistogramClipped(NUMA   *na,
                         INT32   maxsize)
{
INT32     i, n, nhist, imaxval, imaxloc, ival, hval;
FLOAT32   maxval;
NUMA     *nai, *nahist;

    PROCNAME("numaMakeHistogram");

    if (!na)
        return (NUMA *)ERROR_PTR("na not defined", procName, NULL);

    numaGetMax(na, &maxval, &imaxloc);
    imaxval = (INT32)(maxval + 0.5);
    nhist = MIN(imaxval, maxsize + 1);  /* size of histogram array */

    if ((nai = numaConvertToInt(na)) == NULL)
        return (NUMA *)ERROR_PTR("nai not made", procName, NULL);
    n = numaGetCount(nai);

    if ((nahist = numaCreate(nhist)) == NULL)
        return (NUMA *)ERROR_PTR("nahist not made", procName, NULL);
    nahist->n = nhist;  /* fake the storage of nhist zeroes */
    for (i = 0; i < n; i++) {
	numaGetIValue(nai, i, &ival);
	if (ival >= 0 && ival < nhist) {
	    numaGetIValue(nahist, ival, &hval);
	    numaSetValue(nahist, ival, hval + 1.0);
	}
    }

    numaDestroy(&nai);
    return nahist;
}


/*!
 *  numaConvolve()
 *
 *       Input: na
 *              halfwidth (of rectangular filter, minus the center)
 *       Return: na (after low-pass filtering), or null on error
 *
 *       Note: full convolution takes place only from i = halfwidth to
 *             i = n - halfwidth - 1.  We do the end parts using only
 *             the partial array available.  We do not pad the ends
 *             with 0.
 *       Note: this implementation assumes specific fields in the Numa!
 */
NUMA *
numaConvolve(NUMA   *na,
             INT32   halfwidth)
{
INT32     i, n, rval;
FLOAT32   sum, norm;
FLOAT32  *array, *carray, *sumarray;
NUMA     *nai, *nac;

    PROCNAME("numaConvolve");

    if (!na)
        return (NUMA *)ERROR_PTR("na not defined", procName, NULL);
    n = numaGetCount(na);
    if (2 * halfwidth + 1 > n)
        WARNING("filter wider than input array!", procName);
    array = na->array;

    if ((nac = numaCreate(n)) == NULL)
        return (NUMA *)ERROR_PTR("nac not made", procName, NULL);
    carray = nac->array;
    nac->n = n;  /* fill with zeroes */

        /* make sum array; note the indexing */
    if ((sumarray = (FLOAT32 *)CALLOC(n + 1, sizeof(FLOAT32))) == NULL)
        return (NUMA *)ERROR_PTR("sumarray not made", procName, NULL);
    sum = 0.0;
    sumarray[0] = 0.0;
    for (i = 0; i < n; i++) {
        sum += array[i];
	sumarray[i + 1] = sum;
    }

        /* central part */
    norm = 1. / (2 * halfwidth + 1);
    rval = n - halfwidth;
    for (i = halfwidth; i < rval; ++i)
        carray[i] = norm *
	              (sumarray[i + halfwidth + 1] - sumarray[i - halfwidth]);

	/* left side */
    for (i = 0; i < halfwidth; i++)
        carray[i] = sumarray[i + halfwidth + 1] / (halfwidth + i + 1);

	/* right side */
    for (i = rval; i < n; i++)
        carray[i] = (1. / (n - i + halfwidth)) *
	               (sumarray[n] - sumarray[i - halfwidth]);
        
    FREE((void *)sumarray);
    return nac;
}


/*!
 *  numaFindPeaks()
 *
 *       Input: source na
 *              max number of peaks to be found
 *              fract1  (min fraction of peak value)
 *              fract2  (min slope)
 *       Return: peak na (consisting of quads:
 *                          - left edge
 *                          - peak center
 *                          - right edge
 *                          - normalized peak area
 *                        or null on error)
 */
NUMA *
numaFindPeaks(NUMA   *nas,
	      INT32   nmax,
	      FLOAT32 fract1,
	      FLOAT32 fract2)
{
INT32     i, k, n, maxloc, lloc, rloc;
FLOAT32   fmaxval, sum, total, newtotal, val, lastval;
FLOAT32   peakfract;
NUMA     *na, *napeak;

    PROCNAME("numaFindPeaks");

    if (!nas)
        return (NUMA *)ERROR_PTR("nas not defined", procName, NULL);
    n = numaGetCount(nas);
    numaGetSum(nas, &total);

	/* we munge this copy */
    if ((na = numaCopy(nas)) == NULL)
        return (NUMA *)ERROR_PTR("na not made", procName, NULL);
    if ((napeak = numaCreate(4 * nmax)) == NULL)
        return (NUMA *)ERROR_PTR("napeak not made", procName, NULL);

    for (k = 0; k < nmax; k++) {
	numaGetSum(na, &newtotal);
	if (newtotal == 0.0)   /* sanity check */
	    break;
	numaGetMax(na, &fmaxval, &maxloc);
	sum = fmaxval;
	lastval = fmaxval;
	lloc = 0;
	for (i = maxloc - 1; i >= 0; --i) {
	    numaGetFValue(na, i, &val);
	    if (val == 0.0) {
		lloc = i + 1;
		break;
	    }
	    if (val > fract1 * fmaxval) {
		sum += val;
		lastval = val;
		continue;
	    }
	    if (lastval - val > fract2 * lastval) {
		sum += val;
		lastval = val;
		continue;
	    }
	    lloc = i;
	    break;
	}
	lastval = fmaxval;
	rloc = n - 1;
	for (i = maxloc + 1; i < n; ++i) {
	    numaGetFValue(na, i, &val);
	    if (val == 0.0) {
		rloc = i - 1;
		break;
	    }
	    if (val > fract1 * fmaxval) {
		sum += val;
		lastval = val;
		continue;
	    }
	    if (lastval - val > fract2 * lastval) {
		sum += val;
		lastval = val;
		continue;
	    }
	    rloc = i;
	    break;
	}
	peakfract = sum / total;
	numaAddNumber(napeak, lloc);
	numaAddNumber(napeak, maxloc);
	numaAddNumber(napeak, rloc);
	numaAddNumber(napeak, peakfract);

	for (i = lloc; i <= rloc; i++)
	    numaSetValue(na, i, 0.0);
    }

    numaDestroy(&na);
    return napeak;
}


/*!
 *  numaMakeSequence()
 *
 *           Input:  startval
 *                   increment
 *                   size (of sequence)
 *           Return: numa of sequence of evenly spaced values,
 *                   or null on error
 */
NUMA *
numaMakeSequence(FLOAT32   startval,
		 FLOAT32   increment,
		 INT32     size)
{
INT32     i;
FLOAT32   val;
NUMA     *na;

    PROCNAME("numaMakeSequence");

    if ((na = numaCreate(size)) == NULL)
        return (NUMA *)ERROR_PTR("na not made", procName, NULL);

    for (i = 0; i < size; i++) {
        val = startval + i * increment;
	numaAddNumber(na, val);
    }

    return na;
}


/*----------------------------------------------------------------------*
 *                        Serialize for I/O                             *
 *----------------------------------------------------------------------*/
/*!
 *  numaRead()
 *
 *       Input: filename
 *       Return: na, or null on error
 */
NUMA *
numaRead(char   *filename)
{
FILE    *fp;
NUMA    *na;

    PROCNAME("numaRead");

    if (!filename)
        return (NUMA *)ERROR_PTR("filename not defined", procName, NULL);

    if ((fp = fopenReadStream(filename)) == NULL)
        return (NUMA *)ERROR_PTR("stream not opened", procName, NULL);

    if ((na = numaReadStream(fp)) == NULL) {
	fclose(fp);
        return (NUMA *)ERROR_PTR("na not read", procName, NULL);
    }

    fclose(fp);
    return na;
}
	

/*!
 *  numaReadStream()
 *
 *       Input: stream
 *       Return: numa, or null on error
 */
NUMA *
numaReadStream(FILE    *fp)
{
INT32     i, n, index;
FLOAT32   val;
NUMA     *na;

    PROCNAME("numaReadStream");

    if (!fp)
        return (NUMA *)ERROR_PTR("stream not defined", procName, NULL);

    if (fscanf(fp, "\nnumarray: number of numbers = %d\n", &n) != 1)
	return (NUMA *)ERROR_PTR("not a numarray file", procName, NULL);

    if ((na = numaCreate(n)) == NULL)
	return (NUMA *)ERROR_PTR("na not made", procName, NULL);

    for (i = 0; i < n; i++) {
	if ((fscanf(fp, "  array[%d] = %f\n", &index, &val)) != 2)
	    return (NUMA *)ERROR_PTR("bad input data", procName, NULL);
	numaAddNumber(na, val);
    }

    return na;
}
	

/*!
 *  numaWrite()
 *
 *       Input: filename, na
 *       Return: 0 if OK, 1 on error
 */
INT32
numaWrite(char  *filename,
          NUMA  *na)
{
FILE    *fp;

    PROCNAME("numaWrite");

    if (!filename)
        return ERROR_INT("filename not defined", procName, 0);
    if (!na)
        return ERROR_INT("na not defined", procName, 0);

    if ((fp = fopen(filename, "w")) == NULL)
        return ERROR_INT("stream not opened", procName, 0);
    if (numaWriteStream(fp, na))
        return ERROR_INT("na not written to stream", procName, 0);
    fclose(fp);

    return 0;
}
	

/*!
 *  numaWriteStream()
 *
 *       Input: stream, na
 *       Return: 0 if OK, 1 on error
 */
INT32
numaWriteStream(FILE  *fp,
                NUMA  *na)
{
INT32   i, n;

    PROCNAME("numaWriteStream");

    if (!fp)
        return ERROR_INT("stream not defined", procName, 0);
    if (!na)
        return ERROR_INT("na not defined", procName, 0);

    n = numaGetCount(na);
    fprintf(fp, "\nnumarray: number of numbers = %d\n", n);
    for (i = 0; i < n; i++)
	fprintf(fp, "  array[%d] = %f\n", i, na->array[i]);
    fprintf(fp, "\n");

    return 0;
}
	


/*--------------------------------------------------------------------------*
 *                    Create/destroy/copy/empty Numaa                       *
 *--------------------------------------------------------------------------*/
/*!
 *  numaaCreate()
 *       Input:  size of numa ptr array to be alloc'd (0 for default)
 *       Return: naa, or null on error
 *
 */
NUMAA *
numaaCreate(INT32   n)
{
NUMAA   *naa;

    PROCNAME("numaaCreate");

    if (n <= 0)
	n = INITIAL_PTR_ARRAYSIZE;

    if ((naa = (NUMAA *)CALLOC(1, sizeof(NUMAA))) == NULL)
        return (NUMAA *)ERROR_PTR("naa not made", procName, NULL);
    if ((naa->numa = (NUMA **)CALLOC(n, sizeof(NUMA *))) == NULL)
        return (NUMAA *)ERROR_PTR("numa ptr array not made", procName, NULL);

    naa->nalloc = n;
    naa->n = 0;

    return naa;
}


/*!
 *  numaaDestroy()
 *       Input:  &numaa <to be nulled>
 *       Return: void
 */
void
numaaDestroy(NUMAA  **pnaa)
{
INT32    i;
NUMAA   *naa;

    PROCNAME("numaaDestroy");

    if (pnaa == NULL) {
	WARNING("ptr address is NULL!", procName);
	return;
    }

    if ((naa = *pnaa) == NULL)
	return;

    for (i = 0; i < naa->n; i++)
	numaDestroy(&naa->numa[i]);
    FREE((void *)naa->numa);
    FREE((void *)naa);
    *pnaa = NULL;

    return;
}

        

/*--------------------------------------------------------------------------*
 *                       Numaa: add Numa and extend array                   *
 *--------------------------------------------------------------------------*/
/*!
 *     numaaAddNuma()
 *
 *         Input:  numaa
 *                 numa     (to be added)
 *                 copyflag  (INSERT, COPY, CLONE)
 *
 *         Return: 0 if OK, 1 on error
 *
 */
INT32
numaaAddNuma(NUMAA  *naa,
             NUMA   *na,
	     INT32   copyflag)
{
INT32     n;
NUMA     *nac;

    PROCNAME("numaaAddNuma");

    if (!naa)
        return ERROR_INT("naa not defined", procName, 1);
    if (!na)
        return ERROR_INT("na not defined", procName, 1);
    if (copyflag != INSERT && copyflag != COPY && copyflag != CLONE)
        return ERROR_INT("copyflag must be in {0,1,2}", procName, 1);
    
    if (copyflag == INSERT)
	nac = na;
    else if (copyflag == COPY) {
	if ((nac = numaCopy(na)) == NULL)
	    return ERROR_INT("nac not made", procName, 1);
    }
    else  /* CLONE */
        nac = numaClone(na);

    n = numaaGetCount(naa);
    if (n >= naa->nalloc)
	numaaExtendArray(naa);
    naa->numa[n] = nac;
    naa->n++;
    return 0;
}


/*!
 *     numaaExtendArray()
 *
 *         Input:  numaa
 *         Return: 0 if OK, 1 on error
 */
INT32
numaaExtendArray(NUMAA  *naa)
{

    PROCNAME("numaaExtendArray");

    if (!naa)
        return ERROR_INT("naa not defined", procName, 1);

    if ((naa->numa = (NUMA **)reallocNew((void **)&naa->numa,
				4 * naa->nalloc,
				8 * naa->nalloc)) == NULL)
	    return ERROR_INT("new ptr array not returned", procName, 1);

    naa->nalloc *= 2;
    return 0;
}


/*----------------------------------------------------------------------*
 *                           Numaa Accessors                            *
 *----------------------------------------------------------------------*/
/*!
 *  numaaGetCount()
 *
 *       Input: numaa
 *       Return: count, or 0 if no numbers or on error
 */
INT32
numaaGetCount(NUMAA  *naa)
{
    PROCNAME("numaaGetCount");

    if (!naa)
        return ERROR_INT("naa not defined", procName, 0);
    return naa->n;
}
	

/*!
 *  numaaGetNuma()
 *
 *     Input:  numaa
 *             index  (to the index-th numa)
 *             accessflag   (COPY or CLONE)
 *     Return: numa, or null on error
 */
NUMA *
numaaGetNuma(NUMAA  *naa,
             INT32   index,
	     INT32   accessflag)
{
    PROCNAME("numaaGetNuma");

    if (!naa)
	return (NUMA *)ERROR_PTR("naa not defined", procName, NULL);
    if (index < 0 || index >= naa->n)
	return (NUMA *)ERROR_PTR("index not valid", procName, NULL);
    if (accessflag != COPY && accessflag != CLONE)
	return (NUMA *)ERROR_PTR("accessflag must be in {1,2}", procName, NULL);

    if (accessflag == COPY)
	return numaCopy(naa->numa[index]);
    else  /* CLONE */
	return numaClone(naa->numa[index]);
}

