/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/

#ifndef  PIX_H_INCLUDED
#define  PIX_H_INCLUDED

/*
 *   pix.h
 *
 *   This includes the following structures:
 *        struct Pix
 *        struct Pixa
 *        struct Pixaa
 *        struct Box
 *        struct Boxa
 *        struct Pta
 *        struct Ptaa
 *   plus macro definitions for
 *        colors
 *        colormap conversions
 *        rasterop bit flags
 *        structure access flags (for copy, clone, newref, etc.)
 */


/*------------------------------------------------------------------------*
 *                           Basic Pix                                    *
 *------------------------------------------------------------------------*/
struct Pix
{
    UINT32               w;           /* width in pixels                  */
    UINT32               h;           /* height in pixels                 */
    UINT32               d;           /* depth in bits                    */
    UINT32               wpl;         /* 32-bit words/line                */
    UINT32               refcount;    /* reference count (1 if no clones) */
    UINT32               xres;        /* image res (ppi) in x direction   */
				      /* (use 0 if unknown)               */
    UINT32               yres;        /* image res (ppi) in y direction   */
				      /* (use 0 if unknown)               */
    char                *text;        /* text string associated with pix  */
    struct PixColormap  *colormap;    /* colormap (may be null)           */
    UINT32              *data;        /* the image data                   */
};
typedef struct Pix PIX;




/*-----------------------------------------------------------------------*
 *                            Colors for 32 bpp                          *
 *-----------------------------------------------------------------------*/
/* Note: colors are used in 32 bpp images.  The 4th byte can be
 * used for blending, but is not used here.  */
#define  COLOR_RED         0
#define  COLOR_GREEN       1
#define  COLOR_BLUE        2


/*-----------------------------------------------------------------------*
 *                       Flags for colormap conversion                   *
 *-----------------------------------------------------------------------*/
#define  CONVERT_TO_GRAYSCALE         1
#define  CONVERT_TO_FULL_COLOR        2


/*-----------------------------------------------------------------------*
 *
 * The following operation bit flags have been modified from
 * Sun's pixrect.h.
 *
 * The 'op' in 'rasterop' is represented by an integer
 * composed with Boolean functions using the set of five integers
 * given below.  The integers, and the op codes resulting from
 * boolean expressions on them, need only be in the range from 0 to 15.
 * The function is applied on a per-pixel basis.
 *
 * Examples: the op code representing ORing the src and dest
 * is computed using the bit OR, as PIX_SRC | PIX_DST;  the op
 * code representing XORing src and dest is found from
 * PIX_SRC ^ PIX_DST;  the op code representing ANDing src and dest
 * is found from PIX_SRC & PIX_DST.  Note that
 * PIX_NOT(PIX_CLR) = PIX_SET, and v.v., as they must be.
 *
 * We would like to use the following set of definitions:
 *
 *      #define   PIX_SRC      0xc
 *      #define   PIX_DST      0xa
 *      #define   PIX_NOT(op)  ((op) ^ 0xf)
 *      #define   PIX_CLR      0x0
 *      #define   PIX_SET      0xf
 *
 * Now, these definitions differ from Sun's, in that Sun
 * left-shifted each value by 1 pixel, and used the least
 * significant bit as a flag for the "pseudo-operation" of
 * clipping.  We don't need this bit, because it is both
 * efficient and safe ALWAYS to clip the rectangles to the src
 * and dest images, which is what we do.  See the notes in rop.h
 * on the general choice of these bit flags.
 *
 * However, if you include Sun's xview package, you will get their
 * definitions, and because I like using these flags, we will
 * adopt the original Sun definitions to avoid redefinition conflicts.
 *
 * Then we have, for reference, the following 16 unique op flags:
 *
 *      PIX_CLR                           00000             0x0
 *      PIX_SET                           11110             0x1e
 *      PIX_SRC                           11000             0x18
 *      PIX_DST                           10100             0x14
 *      PIX_NOT(PIX_SRC)                  00110             0x06
 *      PIX_NOT(PIX_DST)                  01010             0x0a
 *      PIX_SRC | PIX_DST                 11100             0x1c
 *      PIX_SRC & PIX_DST                 10000             0x10
 *      PIX_SRC ^ PIX_DST                 01100             0x0c
 *      PIX_NOT(PIX_SRC) | PIX_DST        10110             0x16
 *      PIX_NOT(PIX_SRC) & PIX_DST        00100             0x04
 *      PIX_SRC | PIX_NOT(PIX_DST)        11010             0x1a
 *      PIX_SRC & PIX_NOT(PIX_DST)        01000             0x08
 *      PIX_NOT(PIX_SRC | PIX_DST)        00010             0x02
 *      PIX_NOT(PIX_SRC & PIX_DST)        01110             0x0e
 *      PIX_NOT(PIX_SRC ^ PIX_DST)        10010             0x12
 *
 *-----------------------------------------------------------------------*/
#define   PIX_SRC      (0xc << 1)
#define   PIX_DST      (0xa << 1)
#define   PIX_NOT(op)  ((op) ^ 0x1e)
#define   PIX_CLR      (0x0 << 1)
#define   PIX_SET      (0xf << 1)

#define   PIX_PAINT    (PIX_SRC | PIX_DST)
#define   PIX_MASK     (PIX_SRC & PIX_DST)
#define   PIX_XOR      (PIX_SRC ^ PIX_DST)

#define   SET          PIX_SET
#define   CLEAR        PIX_CLR
#define   FLIP         PIX_NOT(PIX_DST)


/*-----------------------------------------------------------------------*
 *
 *   Important Notes:
 *
 *       (1) The image data is stored in a single contiguous
 *           array of UINT32, into which the pixels are packed.
 *           By "packed" we mean that there are no unused bits
 *           between pixels, except for end-of-line padding to
 *           satisfy item (2) below.
 *
 *       (2) Every image raster line begins on a 32-bit word
 *           boundary within this array.
 *
 *       (3) Pix image data is stored in 32-bit units, with the
 *           pixels ordered from left to right in the image being
 *           stored in order from the MSB to LSB within the word,
 *           for both big-endian and little-endian machines.
 *           This is the natural ordering for big-endian machines,
 *           as successive bytes are stored and fetched progressively
 *           to the right.  However, for little-endians, when storing
 *           we re-order the bytes from this byte stream order, and
 *           reshuffle again for byte access on 32-bit entities.
 *           So if the bytes come in sequence from left to right, we
 *           store them on little-endians in byte order: 
 *                3 2 1 0 7 6 5 4 ...
 *           This MSB to LSB ordering allows left and right shift
 *           operations on 32 bit words to move the pixels properly.
 *
 *       (4) For 24-bit color images, use 32 bpp data, leaving
 *           the fourth byte unused.  Within each 4 byte pixel, the
 *           colors are ordered from MSB to LSB, as follows:
 *
 *                |  MSB  |  2nd MSB  |  3rd MSB  |  LSB  |
 *                   red      green       blue      unused
 *                    0         1           2         3   (big-endian)
 *                    3         2           1         0   (little-endian)
 *
 *           Because we use MSB to LSB ordering within the 32-bit word,
 *           the individual 8-bit samples can be accessed with 
 *           GET_DATA_BYTE and SET_DATA_BYTE macros, using the
 *           (implicitly big-ending) ordering
 *                 red:    byte 0  (MSB)
 *                 green:  byte 1  (2nd MSB)
 *                 blue:   byte 2  (3rd MSB)
 *
 *           For simplicity and consistency, we have defined (COLOR_RED,
 *           COLOR_GREEN, COLOR_BLUE) so that we can retrieve red, green
 *           and blue sample values using
 *              redval = GET_DATA_BYTE(&pixel, COLOR_RED);
 *              greenval = GET_DATA_BYTE(&pixel, COLOR_GREEN);
 *              blueval = GET_DATA_BYTE(&pixel, COLOR_BLUE);
 *           and we can set them in the pixel using
 *              SET_DATA_BYTE(&pixel, COLOR_RED, redval);
 *              SET_DATA_BYTE(&pixel, COLOR_GREEN, greenval);
 *              SET_DATA_BYTE(&pixel, COLOR_BLUE, blueval);
 *
 *       (5) A reference count is held within each pix, giving the
 *           number of ptrs to the pix.  When a pixClone() call
 *           is made, the ref count is increased by 1, and
 *           when a pixDestroy() call is made, the reference count
 *           of the pix is decremented.  The pix is only destroyed
 *           when the reference count goes to zero.
 *
 *-----------------------------------------------------------------------*/



/*-----------------------------------------------------------------------*
 *                            Array of pix                               *
 *-----------------------------------------------------------------------*/
struct Pixa
{
    INT32               n;              /* number of Pix in ptr array      */
    INT32               nalloc;         /* number of Pix ptrs allocated    */
    UINT32              refcount;       /* reference count (0 if no refs)  */
    struct Pix        **pix;            /* the array of ptrs to pix        */
    struct Boxa        *boxa;           /* array of boxes                  */
};
typedef struct Pixa PIXA;


struct Pixaa
{
    INT32               n;              /* number of Pixa in ptr array     */
    INT32               nalloc;         /* number of Pixa ptrs allocated   */
    struct Pixa       **pixa;           /* array of ptrs to pixa           */
};
typedef struct Pixaa PIXAA;



/*-----------------------------------------------------------------------*
 *               Basic rectangle and rectangle arrays                    *
 *-----------------------------------------------------------------------*/
struct Box
{
    INT32              x;
    INT32              y;
    INT32              w;
    INT32              h;
    UINT32             refcount;      /* reference count (1 if no clones)  */

};
typedef struct Box    BOX;

struct Boxa
{
    INT32              n;             /* number of box in ptr array        */
    INT32              nalloc;        /* number of box ptrs allocated      */
    struct Box       **box;           /* box ptr array                     */
};
typedef struct Boxa  BOXA;


/*-----------------------------------------------------------------------*
 *                           Array of points                             *
 *-----------------------------------------------------------------------*/
struct Pta
{
    INT32              n;              /* actual number of pts              */
    INT32              nalloc;         /* size of allocated arrays          */
    INT32              refcount;       /* reference count (1 if no clones)  */
    FLOAT32           *x, *y;          /* arrays of floats                  */
};
typedef struct Pta PTA;


/*-----------------------------------------------------------------------*
 *                            Array of Pta                               *
 *-----------------------------------------------------------------------*/
struct Ptaa
{
    INT32                n;              /* number of pta in ptr array    */
    INT32                nalloc;         /* number of pta ptrs allocated  */
    struct Pta         **pta;            /* pta ptr array                 */
};
typedef struct Ptaa PTAA;




/*-----------------------------------------------------------------------*
 *                              Access flags                             *
 *-----------------------------------------------------------------------*/
#define    INSERT           0       /* stuff it in; no copy, clone, refcount */
#define    COPY             1       /* use copy of the data                  */
#define    CLONE            2       /* use clone (ref count) of the data     */
#define    NEWREF           3       /* higher-level refcount; be sure to     */
                                    /*    call destroyPixa()                 */



#endif  /* PIX_H_INCLUDED */


