/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  pix2.c
 *
 *      Pixel poking
 *           INT32       pixGetPixel()
 *           INT32       pixSetPixel()
 *           INT32       pixClearPixel()
 *           INT32       pixFlipPixel()
 *
 *      Full image clear/set/set-to-arbitrary-value
 *           INT32       pixClearAll()
 *           INT32       pixSetAll()
 *           INT32       pixSetAllArbitrary()
 *           INT32       pixSetPadBits()
 *           INT32       pixSetPadBitsBand()
 *
 *      Set border pixels to arbitrary value
 *           INT32       pixSetOrClearBorder()
 *           INT32       pixSetBorderVal()
 *
 *      Masked set to arbitrary value
 *           INT32       pixSetMasked()
 *
 *      One and two-image boolean operations on arbitrary depth images
 *           PIX        *pixInvert()
 *           PIX        *pixOr()
 *           PIX        *pixAnd()
 *           PIX        *pixXor()
 *           PIX        *pixSubtract()
 *           INT32       pixEqual()
 *
 *      Pixel counting
 *           INT32       pixZero()
 *           INT32       pixCountPixels()
 *           NUMA       *pixaCountPixels()
 *           INT32       pixCountPixelsInRow()
 *           NUMA       *pixCountPixelsByRow()
 *           INT32       pixThresholdPixels()
 *           INT32      *makePixelSumTab8()
 *
 *      Pixel histogram
 *           NUMA       *pixGrayHistogram()
 *
 *      Conversion between big and little endians
 *           INT32       pixEndianByteSwap()
 *
 *      Color sample setting and extraction
 *           PIX        *pixCreateRGBImage()
 *           PIX        *pixGetRGBComponent()
 *           INT32       pixSetRGBComponent()
 *           INT32       composeRGBPixel()
 *           INT32       pixGetRGBLine()
 *
 *      Conversion to full color or grayscale from colormapped pix
 *           PIX        *pixRemoveColormap()
 *           INT32       colormapToArrays()
 *
 *      Conversion from RGB color to grayscale
 *           PIX        *pixConvertRGBToGray()
 *
 *      Add and remove border
 *           PIX        *pixAddBorder()
 *           PIX        *pixRemoveBorder()
 *           PIX        *pixAddBorderGeneral()
 *           PIX        *pixRemoveBorderGeneral()
 *
 *      Extract rectangle
 *           PIX        *pixClipRectangle()
 *
 *      Clip to foreground
 *           PIX        *pixClipToForeground()
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "allheaders.h"

static const UINT32 rmask32[] = {0x0,
    0x00000001, 0x00000003, 0x00000007, 0x0000000f,
    0x0000001f, 0x0000003f, 0x0000007f, 0x000000ff,
    0x000001ff, 0x000003ff, 0x000007ff, 0x00000fff,
    0x00001fff, 0x00003fff, 0x00007fff, 0x0000ffff,
    0x0001ffff, 0x0003ffff, 0x0007ffff, 0x000fffff,
    0x001fffff, 0x003fffff, 0x007fffff, 0x00ffffff,
    0x01ffffff, 0x03ffffff, 0x07ffffff, 0x0fffffff,
    0x1fffffff, 0x3fffffff, 0x7fffffff, 0xffffffff};


#define   EQUAL_SIZE_WARNING      0

    /* weighting factor for conversion of RGB to grayscale */
#define   RED_WEIGHT        0.25
#define   GREEN_WEIGHT      0.60
#define   BLUE_WEIGHT       0.15



/*-------------------------------------------------------------*
 *                         Pixel poking                        *
 *-------------------------------------------------------------*/
/*!
 *  pixGetPixel()
 *
 *       Input: pix, (x,y), and ptr to returned value.
 *       Return: 0 if OK; 1 on error.
 */
INT32
pixGetPixel(PIX    *pix,
	    INT32   x,
	    INT32   y,
	    UINT32 *pval)
{
INT32        w, h, d, wpl, val;
UINT32      *line, *data;

    PROCNAME("pixGetPixel");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (!pval)
	return ERROR_INT("pval not defined", procName, 1);
    *pval = 0;

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);
    if (x < 0 || x >= w)
	return ERROR_INT("x out of bounds", procName, 1);
    if (y < 0 || y >= h)
	return ERROR_INT("y out of bounds", procName, 1);

    wpl = pixGetWpl(pix);
    data = pixGetData(pix);
    line = data + y * wpl;

    d = pixGetDepth(pix);
    switch (d)
    {
    case 1:
	val = GET_DATA_BIT(line, x);
	break;
    case 2:
	val = GET_DATA_DIBIT(line, x);
	break;
    case 4:
	val = GET_DATA_QBIT(line, x);
	break;
    case 8:
	val = GET_DATA_BYTE(line, x);
	break;
    case 16:
	val = GET_DATA_TWO_BYTES(line, x);
	break;
    case 32:
	val = line[x];
	break;
    default:
	return ERROR_INT("depth must be in {1,2,4,8,16,32} bpp", procName, 1);
    }

    *pval = val;
    return 0;
}


/*!
 *  pixSetPixel()
 *
 *       Input: pix, (x,y), and val to be inserted
 *       Return: 0 if OK; 1 on error
 *
 *       Note: input value is not checked for overflow, and
 *             the sign bit (if any) is ignored.
 */
INT32
pixSetPixel(PIX    *pix,
	    INT32   x,
	    INT32   y,
	    UINT32  val)
{
INT32        w, h, d, wpl;
UINT32      *line, *data;

    PROCNAME("pixSetPixel");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);
    if (x < 0 || x >= w)
	return ERROR_INT("x out of bounds", procName, 1);
    if (y < 0 || y >= h)
	return ERROR_INT("y out of bounds", procName, 1);

    data = pixGetData(pix);
    wpl = pixGetWpl(pix);
    line = data + y * wpl;

    d = pixGetDepth(pix);
    switch (d)
    {
    case 1:
	if (val)
	    SET_DATA_BIT(line, x);
	else
	    CLEAR_DATA_BIT(line, x);
	break;
    case 2:
	SET_DATA_DIBIT(line, x, val);
	break;
    case 4:
	SET_DATA_QBIT(line, x, val);
	break;
    case 8:
	SET_DATA_BYTE(line, x, val);
	break;
    case 16:
	SET_DATA_TWO_BYTES(line, x, val);
	break;
    case 32:
	line[x] = val;
	break;
    default:
	return ERROR_INT("depth must be in {1,2,4,8,16,32} bpp", procName, 1);
    }

    return 0;
}



/*!
 *  pixClearPixel()
 *
 *       Input: pix, (x,y)
 *       Return: 0 if OK; 1 on error.
 */
INT32
pixClearPixel(PIX    *pix,
	      INT32   x,
	      INT32   y)
{
INT32        w, h, d, wpl;
UINT32      *line, *data;

    PROCNAME("pixClearPixel");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);
    if (x < 0 || x >= w)
	return ERROR_INT("x out of bounds", procName, 1);
    if (y < 0 || y >= h)
	return ERROR_INT("y out of bounds", procName, 1);

    wpl = pixGetWpl(pix);
    data = pixGetData(pix);
    line = data + y * wpl;

    d = pixGetDepth(pix);
    switch (d)
    {
    case 1:
	CLEAR_DATA_BIT(line, x);
	break;
    case 2:
	CLEAR_DATA_DIBIT(line, x);
	break;
    case 4:
	CLEAR_DATA_QBIT(line, x);
	break;
    case 8:
	SET_DATA_BYTE(line, x, 0);
	break;
    case 16:
	SET_DATA_TWO_BYTES(line, x, 0);
	break;
    case 32:
	line[x] = 0;
	break;
    default:
	return ERROR_INT("depth must be in {1,2,4,8,16,32} bpp", procName, 1);
    }

    return 0;
}


/*!
 *  pixFlipPixel()
 *
 *       Input: pix, (x,y)
 *       Return: 0 if OK; 1 on error
 */
INT32
pixFlipPixel(PIX    *pix,
	     INT32   x,
	     INT32   y)
{
INT32        w, h, d, wpl;
UINT32       val;
UINT32      *line, *data;

    PROCNAME("pixFlipPixel");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);
    if (x < 0 || x >= w)
	return ERROR_INT("x out of bounds", procName, 1);
    if (y < 0 || y >= h)
	return ERROR_INT("y out of bounds", procName, 1);

    data = pixGetData(pix);
    wpl = pixGetWpl(pix);
    line = data + y * wpl;

    d = pixGetDepth(pix);
    switch (d)
    {
    case 1:
	val = GET_DATA_BIT(line, x);
	if (val)
	    CLEAR_DATA_BIT(line, x);
	else
	    SET_DATA_BIT(line, x);
	break;
    case 2:
	val = GET_DATA_DIBIT(line, x);
	val ^= 0x3;
	SET_DATA_DIBIT(line, x, val);
	break;
    case 4:
	val = GET_DATA_QBIT(line, x);
	val ^= 0xf;
	SET_DATA_QBIT(line, x, val);
	break;
    case 8:
	val = GET_DATA_BYTE(line, x);
	val ^= 0xff;
	SET_DATA_BYTE(line, x, val);
	break;
    case 16:
	val = GET_DATA_TWO_BYTES(line, x);
	val ^= 0xffff;
	SET_DATA_TWO_BYTES(line, x, val);
	break;
    case 32:
	val = line[x] ^ 0xffffffff;
	line[x] = val;
	break;
    default:
	return ERROR_INT("depth must be in {1,2,4,8,16,32} bpp", procName, 1);
    }

    return 0;
}




/*-------------------------------------------------------------*
 *     Full image clear/set/set-to-arbitrary-value/invert      *
 *-------------------------------------------------------------*/
/*!
 *  pixClearAll()
 *
 *    Input: pix
 *    Return: 0 if OK, 1 on error.
 *    Action: clears all data to 0
 */
INT32
pixClearAll(PIX   *pix)
{
    PROCNAME("pixClearAll");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    pixRasterop(pix, 0, 0, pixGetWidth(pix), pixGetHeight(pix),
                PIX_CLR, NULL, 0, 0);
    return 0;
}


/*!
 *  pixSetAll()
 *
 *    Input: pix
 *    Return: 0 if OK, 1 on error.
 *    Action: sets all data to 1
 */
INT32
pixSetAll(PIX   *pix)
{
    PROCNAME("pixSetAll");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    pixRasterop(pix, 0, 0, pixGetWidth(pix), pixGetHeight(pix),
                PIX_SET, NULL, 0, 0);
    return 0;
}


/*!
 *  pixSetAllArbitrary()
 *
 *        Input:  pix
 *                val  (set all pixels to this)
 *        Return: 0 if OK; 1 on error
 */
INT32
pixSetAllArbitrary(PIX    *pix,
                   UINT32  val)
{
INT32    i, j, w, h, d, wpl, npix;
UINT32   maxval, wordval;
UINT32  *data, *line;

    PROCNAME("pixSetAllArbitrary");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    d = pixGetDepth(pix); 
    w = pixGetWidth(pix);
    h = pixGetHeight(pix);

    if (d == 32)
	maxval = 0xffffffff;
    else
	maxval = (1 << d) - 1;
    if (val < 0) {
        WARNING("invalid pixel value; set to 0", procName);
	val = 0;
    }
    if (val > maxval) {
        WARNING_INT("invalid pixel val; set to maxval = %d", procName, maxval);
	val = maxval;
    }

	/* set up word to tile with */
    wordval = 0;
    npix = 32 / d;    /* number of pixels per 32 bit word */
    for (j = 0; j < npix; j++)
	wordval |= (val << (j * d));

    wpl = pixGetWpl(pix);
    data = pixGetData(pix);
    for (i = 0; i < h; i++) {
	line = data + i * wpl;
	for (j = 0; j < wpl; j++) {
	    *(line + j) = wordval;
	}
    }

    return 0;
}


/*!
 *  pixSetPadBits()
 *
 *        Input:  binary pix
 *                val  (0 or 1)
 *        Return: 0 if OK; 1 on error
 *        Action: sets the val of the pad bits (if any)
 *                on the RHS of the last word in each scanline.
 */
INT32
pixSetPadBits(PIX    *pix,
              INT32   val)
{
INT32    i, w, h, wpl, index, nfullwords;
UINT32   mask;
UINT32  *data, *pword;

    PROCNAME("pixSetPadBits");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (pixGetDepth(pix) != 1)
	return ERROR_INT("pix not binary", procName, 1);

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);
    data = pixGetData(pix);
    wpl = pixGetWpl(pix);
    index = 32 - (w % 32);
    if (index == 32)  /* no partial word */
	return 0;
    nfullwords = w / 32;

    mask = rmask32[index];
    if (val == 0)
	mask = ~mask;

    for (i = 0; i < h; i++) {
	pword = data + i * wpl + nfullwords;
	if (val == 0) /* clear */
	    *pword = *pword & mask;
	else  /* set */
	    *pword = *pword | mask;
    }

    return 0;
}


/*!
 *  pixSetPadBitsBand()
 *
 *        Input:  binary pix
 *                by  (starting y value of band)
 *                bh  (height of band)
 *                val  (0 or 1)
 *        Return: 0 if OK; 1 on error
 *
 *        Action: sets the val of the pad bits (if any)
 *                on the RHS of the last word in each scanline,
 *                within the specified band.
 */
INT32
pixSetPadBitsBand(PIX    *pix,
                  INT32   by,
                  INT32   bh,
                  INT32   val)
{
INT32    i, w, h, wpl, index, nfullwords;
UINT32   mask;
UINT32  *data, *pword;

    PROCNAME("pixSetPadBits");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (pixGetDepth(pix) != 1)
	return ERROR_INT("pix not binary", procName, 1);

    h = pixGetHeight(pix);
    if (by < 0)
	by = 0;
    if (by >= h)
	return ERROR_INT("start y not in image", procName, 1);
    if (by + bh > h)
	bh = h - by;

    w = pixGetWidth(pix);
    data = pixGetData(pix);
    wpl = pixGetWpl(pix);
    index = 32 - (w % 32);
    if (index == 32)  /* no partial word */
	return 0;
    nfullwords = w / 32;

    mask = rmask32[index];
    if (val == 0)
	mask = ~mask;

    for (i = by; i < by + bh; i++) {
	pword = data + i * wpl + nfullwords;
	if (val == 0) /* clear */
	    *pword = *pword & mask;
	else  /* set */
	    *pword = *pword | mask;
    }

    return 0;
}


/*-------------------------------------------------------------*
 *                       Set border pixels                     *
 *-------------------------------------------------------------*/
/*!
 *  pixSetOrClearBorder()
 *
 *     Input:  pixs (all depths)
 *             leftpix, rightpix, toppix, bottompix
 *             operation (SET or CLEAR)
 *     Return: 0 if OK; 1 on error
 *
 *     Action: The border region is defined to be the region in the
 *     image within a specific distance of each edge.  We
 *     allow the max distances to each edge to be set
 *     independently.  This either sets or clears all pixels
 *     in the border region.
 *
 *     Note: for binary images, use SET for black and CLEAR for white;
 *     for grayscale or color images, use SET for white and CLEAR for black.
 *
 */
INT32
pixSetOrClearBorder(PIX   *pixs,
                    INT32  leftpix,
                    INT32  rightpix,
                    INT32  toppix,
                    INT32  bottompix,
	            INT32  op)
{
INT32     w, h;

    PROCNAME("pixSetOrClearBorder");

    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);
    if (op != SET && op != CLEAR)
	return ERROR_INT("op must be SET or CLEAR", procName, 1);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);

    pixRasterop(pixs, 0, 0, leftpix, h, op, NULL, 0, 0);
    pixRasterop(pixs, w - rightpix, 0, rightpix, h, op, NULL, 0, 0);
    pixRasterop(pixs, 0, 0, w, toppix, op, NULL, 0, 0);
    pixRasterop(pixs, 0, h - bottompix, w, bottompix, op, NULL, 0, 0);

    return 0;
}


/*!
 *  pixSetBorderVal()
 *
 *     Input:  pixs (8 or 32 bpp)
 *             leftpix, rightpix, toppix, bottompix
 *             val (value to set at each border pixel)
 *     Return: 0 if OK; 1 on error
 *
 *     Action: The border region is defined to be the region in the
 *     image within a specific distance of each edge.  We
 *     allow the max distances to each edge to be set
 *     independently.  This sets the pixels in the border region
 *     to the given input value.
 *
 *     Note 1: For efficiency, use pixSetOrClearBorder() if
 *     you're setting the border to either black or white.
 *
 *     Note 2: if d != 32, the input value should be masked off
 *             to the appropriate number of least significant bits.
 *
 *     Note 3: if you need to do this for 2, 4 or 16 bpp, the
 *             generalization should be obvious.
 */
INT32
pixSetBorderVal(PIX   *pixs,
                INT32  leftpix,
                INT32  rightpix,
                INT32  toppix,
                INT32  bottompix,
	        UINT32 val)
{
INT32     w, h, d, wpls, i, j, bstart, rstart;
UINT32   *datas, *lines;

    PROCNAME("pixSetBorderVal");

    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);
    d = pixGetDepth(pixs);
    if (d != 8 && d != 32)
	return ERROR_INT("depth must be 8 or 32 bpp", procName, 1);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

    if (d == 8) {
	val &= 0xff;
	for (i = 0; i < toppix; i++) {
	    lines = datas + i * wpls;
	    for (j = 0; j < w; j++)
		SET_DATA_BYTE(lines, j, val);
	}
	rstart = w - rightpix;
	bstart = h - bottompix;
	for (i = toppix; i < bstart; i++) {
	    lines = datas + i * wpls;
	    for (j = 0; j < leftpix; j++)
		SET_DATA_BYTE(lines, j, val);
	    for (j = rstart; j < w; j++)
		SET_DATA_BYTE(lines, j, val);
	}
	for (i = bstart; i < h; i++) {
	    lines = datas + i * wpls;
	    for (j = 0; j < w; j++)
		SET_DATA_BYTE(lines, j, val);
	}
    }
    else {   /* d == 32 */
	for (i = 0; i < toppix; i++) {
	    lines = datas + i * wpls;
	    for (j = 0; j < w; j++)
	        *(lines + j) = val;
	}
	rstart = w - rightpix;
	bstart = h - bottompix;
	for (i = toppix; i < bstart; i++) {
	    lines = datas + i * wpls;
	    for (j = 0; j < leftpix; j++)
	        *(lines + j) = val;
	    for (j = rstart; j < w; j++)
	        *(lines + j) = val;
	}
	for (i = bstart; i < h; i++) {
	    lines = datas + i * wpls;
	    for (j = 0; j < w; j++)
	        *(lines + j) = val;
	}
    }

    return 0;
}
	


/*-------------------------------------------------------------*
 *                 Masked set to arbitrary value               *
 *-------------------------------------------------------------*/
/*!
 *  pixSetMasked()
 *
 *     Input:  pixd (8 or 32 bpp)
 *             pixm (1 bpp mask)
 *             val (value to set at each masked pixel)
 *     Return: 0 if OK; 1 on error
 *
 *     Action: sets each pixel in pixd that co-locates with an ON
 *             pixel in pixm to the input value.  For 8 bpp,
 *             val is masked to the 8 lsb
 */
INT32
pixSetMasked(PIX    *pixd,
             PIX    *pixm,
	     UINT32  val)
{
UINT8     qbit;
INT32     wd, hd, wm, hm, w, h, d, wpld, wplm, i, j, bitval;
UINT32   *datad, *datam, *lined, *linem;
UINT32   *tab;

    PROCNAME("pixSetMasked");

    if (!pixd)
	return ERROR_INT("pixd not defined", procName, 1);
    if (!pixm)
	return ERROR_INT("pixm not defined", procName, 1);

    wd = pixGetWidth(pixd);
    hd = pixGetHeight(pixd);
    wm = pixGetWidth(pixm);
    hm = pixGetHeight(pixm);
    w = MIN(wd, wm);
    h = MIN(hd, hm);
    if (wd != wm || hd != hm)
        WARNING("pixd and pixm sizes differ", procName);
    d = pixGetDepth(pixd);
    if (d != 8 && d != 32)
	return ERROR_INT("pixd not 8 or 32 bpp", procName, 1);
    if (pixGetDepth(pixm) != 1)
	return ERROR_INT("pixm not 1 bpp", procName, 1);

    datad = pixGetData(pixd);
    datam = pixGetData(pixm);
    wpld = pixGetWpl(pixd);
    wplm = pixGetWpl(pixm);
    if (d == 8)
    {
	    /* make 16-entry table of 32-bit words to be
	     * written selectively into pixd, that maps
	     * 4 bits in pixm to 32 bits in pixd. */
	val &= 0xff;
	if ((tab = (UINT32 *)CALLOC(16, sizeof(UINT32))) == NULL)
	    return ERROR_INT("tab not made", procName, 1);
	for (i = 0; i < 16; i++) {
	    if (i & 1)
		tab[i] = val;
	    if ((i >> 1) & 1)
		tab[i] |= (val << 8);
	    if ((i >> 2) & 1)
		tab[i] |= (val << 16);
	    if (i >> 3)
		tab[i] |= (val << 24);
	}

	for (i = 0; i < h; i++) {
	    lined = datad + i * wpld;
	    linem = datam + i * wplm;
	    for (j = 0; j < wpld; j++) {
		qbit = GET_DATA_QBIT(linem, j);
		*(lined + j) = tab[qbit];
	    }
	}
	FREE((void *)tab);
    }
    else {  /*  d == 32 */
	for (i = 0; i < h; i++) {
	    lined = datad + i * wpld;
	    linem = datam + i * wplm;
	    for (j = 0; j < w; j++) {
		bitval = GET_DATA_BIT(linem, j);
		if (bitval)
		    *(lined + j) = val;
	    }
	}
    }

    return 0;
}


/*-------------------------------------------------------------*
 *    One and two-image boolean ops on arbitrary depth images  *
 *-------------------------------------------------------------*/
/*!
 *  pixInvert()
 *
 *    Input:  pixd  (<optional> destination: this can be null, 
 *                   equal to pixs, or different from pixs)
 *            pixs
 *    Return:  pixd always
 *
 *    Action:  Inversion of pixs
 *              o   pixd == null,   ~src --> new pixd
 *              o   pixd == pixs,   ~src --> src  (in-place)
 *              o   pixd != pixs,  ~src --> input pixd
 *
 *    Note: this works on any depth
 */
PIX *
pixInvert(PIX     *pixd,
          PIX     *pixs)
{

    PROCNAME("pixInvert");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, pixd);

    if (pixs != pixd) {
	if ((pixd = pixCopy(pixd, pixs)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    }

	/* invert pixd */
    pixRasterop(pixd, 0, 0, pixGetWidth(pixd), pixGetHeight(pixd), FLIP,
                NULL, 0, 0);

    return pixd;
}


/*!
 *  pixOr()
 *
 *    Input:  pixd  (<optional> destination: this can be null, 
 *                   equal to pixs1, or different from pixs1)
 *            pixs1 (can be == to pixd)
 *            pixs2 
 *    Return:  pixd always
 *
 *    Action:  Union of two images.  There are 3 cases:
 *              o   pixd == null,   (src1 | src2) --> new pixd
 *              o   pixd == pixs1,  (src1 | src2) --> src1  (in-place)
 *              o   pixd != pixs1,  (src1 | src2) --> input pixd
 *
 *    Note: this works on any depth
 */
PIX *
pixOr(PIX     *pixd,
      PIX     *pixs1,
      PIX     *pixs2)
{
    PROCNAME("pixOr");

    if (!pixs1)
	return (PIX *)ERROR_PTR("pixs1 not defined", procName, pixd);
    if (!pixs2)
	return (PIX *)ERROR_PTR("pixs2 not defined", procName, pixd);
    if (pixs1 == pixs2)
	return (PIX *)ERROR_PTR("pixs1 and pixs2 must differ", procName, pixd);

#if  EQUAL_SIZE_WARNING
    if (!pixSizesEqual(pixs1, pixs2))
	WARNING("pixs1 and pixs2 not equal sizes", procName);
#endif  /* EQUAL_SIZE_WARNING */

    if (pixs1 != pixd) {
	if ((pixd = pixCopy(pixd, pixs1)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    }

	/* src1 | src2 --> dest */
    pixRasterop(pixd, 0, 0, pixGetWidth(pixd), pixGetHeight(pixd),
                PIX_SRC | PIX_DST, pixs2, 0, 0);

    return pixd;
}


/*!
 *  pixAnd()
 *
 *    Input:  pixd  (<optional> destination: this can be null, 
 *                   equal to pixs1, or different from pixs1)
 *            pixs1 (can be == to pixd)
 *            pixs2 
 *    Return:  pixd always
 *
 *    Action:  Intersection of two images.  There are 3 cases:
 *              o   pixd == null,   (src1 & src2) --> new pixd
 *              o   pixd == pixs1,  (src1 & src2) --> src1  (in-place)
 *              o   pixd != pixs1,  (src1 & src2) --> input pixd
 *
 *    Note: this works on any depth
 */
PIX *
pixAnd(PIX     *pixd,
       PIX     *pixs1,
       PIX     *pixs2)
{
    PROCNAME("pixAnd");

    if (!pixs1)
	return (PIX *)ERROR_PTR("pixs1 not defined", procName, pixd);
    if (!pixs2)
	return (PIX *)ERROR_PTR("pixs2 not defined", procName, pixd);
    if (pixs1 == pixs2)
	return (PIX *)ERROR_PTR("pixs1 and pixs2 must differ", procName, pixd);

#if  EQUAL_SIZE_WARNING
    if (!pixSizesEqual(pixs1, pixs2))
	WARNING("pixs1 and pixs2 not equal sizes", procName);
#endif  /* EQUAL_SIZE_WARNING */

    if (pixs1 != pixd) {
	if ((pixd = pixCopy(pixd, pixs1)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    }

	/* src1 & src2 --> dest */
    pixRasterop(pixd, 0, 0, pixGetWidth(pixd), pixGetHeight(pixd),
                PIX_SRC & PIX_DST, pixs2, 0, 0);

    return pixd;
}


/*!
 *  pixXor()
 *
 *    Input:  pixd  (<optional> destination: this can be null, 
 *                   equal to pixs1, or different from pixs1)
 *            pixs1 (can be == to pixd)
 *            pixs2 
 *    Return:  pixd always
 *
 *    Action:  XOR of two images.  There are 3 cases:
 *              o   pixd == null,   (src1 ^ src2) --> new pixd
 *              o   pixd == pixs1,  (src1 ^ src2) --> src1  (in-place)
 *              o   pixd != pixs1,  (src1 ^ src2) --> input pixd
 *
 *    Note: this works on any depth
 */
PIX *
pixXor(PIX     *pixd,
       PIX     *pixs1,
       PIX     *pixs2)
{
    PROCNAME("pixXor");

    if (!pixs1)
	return (PIX *)ERROR_PTR("pixs1 not defined", procName, pixd);
    if (!pixs2)
	return (PIX *)ERROR_PTR("pixs2 not defined", procName, pixd);
    if (pixs1 == pixs2)
	return (PIX *)ERROR_PTR("pixs1 and pixs2 must differ", procName, pixd);

#if  EQUAL_SIZE_WARNING
    if (!pixSizesEqual(pixs1, pixs2))
	WARNING("pixs1 and pixs2 not equal sizes", procName);
#endif  /* EQUAL_SIZE_WARNING */

    if (pixs1 != pixd) {
	if ((pixd = pixCopy(pixd, pixs1)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    }

	/* src1 ^ src2 --> dest */
    pixRasterop(pixd, 0, 0, pixGetWidth(pixd), pixGetHeight(pixd),
                PIX_SRC ^ PIX_DST, pixs2, 0, 0);

    return pixd;
}


/*!
 *  pixSubtract()
 *
 *    Input:  pixd  (<optional> destination: this can be null, 
 *                   equal to pixs1, equal to pixs2, or different
 *                   from both pixs1 and pixs2)
 *            pixs1 (can be == to pixd)
 *            pixs2 (can be == to pixd)
 *    Return:  pixd always
 *
 *    Action:  Set subtraction of two binary images.  Source pixs2 is
 *             always subtracted from source pixs1.  The result is
 *                  pixs1 \ pixs2 = pixs1 & (~pixs2)
 *
 *    Action:  There are 4 cases.  The result can go to a new dest,
 *             in-place to either pixs1 or pixs2, or to an
 *             existing input dest:
 *              o  pixd == null   (src1 - src2) --> new pixd
 *              o  pixd == pixs1  (src1 - src2) --> src1  (in-place)
 *              o  pixd == pixs2  (src1 - src2) --> src2  (in-place)
 *              o  pixd != pixs1 && pixd != pixs2)
 *                                (src1 - src2) --> input pixd
 *
 *    Note: this operation can work on all depths
 */
PIX *
pixSubtract(PIX     *pixd,
            PIX     *pixs1,
            PIX     *pixs2)
{
INT32     w, h;

    PROCNAME("pixSubtract");

    if (!pixs1)
	return (PIX *)ERROR_PTR("pixs1 not defined", procName, pixd);
    if (!pixs2)
	return (PIX *)ERROR_PTR("pixs2 not defined", procName, pixd);
    if (pixs1 == pixs2)
	return (PIX *)ERROR_PTR("pixs1 and pixs2 must differ", procName, pixd);
    if (pixGetDepth(pixs1) != 1)
	return (PIX *)ERROR_PTR("pixs1 not binary", procName, pixd);
    if (pixGetDepth(pixs2) != 1)
	return (PIX *)ERROR_PTR("pixs2 not binary", procName, pixd);

#if  EQUAL_SIZE_WARNING
    if (!pixSizesEqual(pixs1, pixs2))
	WARNING("pixs1 and pixs2 not equal sizes", procName);
#endif  /* EQUAL_SIZE_WARNING */

    w = pixGetWidth(pixs1);
    h = pixGetHeight(pixs1);

    if (!pixd) {
	if ((pixd = pixCopy(NULL, pixs1)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
	pixRasterop(pixd, 0, 0, w, h, PIX_DST & PIX_NOT(PIX_SRC),
	    pixs2, 0, 0);   /* src1 & (~src2)  */
    }
    else if (pixd == pixs1) {
	pixRasterop(pixd, 0, 0, w, h, PIX_DST & PIX_NOT(PIX_SRC),
	    pixs2, 0, 0);   /* src1 & (~src2)  */
    }
    else if (pixd == pixs2) {
	pixRasterop(pixd, 0, 0, w, h, PIX_NOT(PIX_DST) & PIX_SRC,
	    pixs1, 0, 0);   /* src1 & (~src2)  */
    }
    else  { /* pixd != pixs1 && pixd != pixs2 */
	if (pixGetDepth(pixd) != 1)
	    return (PIX *)ERROR_PTR("pixd not binary", procName, pixd);
#if  EQUAL_SIZE_WARNING
	if (!pixSizesEqual(pixd, pixs1))
	    WARNING("pixd and pixs1 not equal sizes", procName);
#endif  /* EQUAL_SIZE_WARNING */
	pixRasterop(pixd, 0, 0, w, h, PIX_SRC, pixs1, 0, 0);   /* copy */
	pixRasterop(pixd, 0, 0, w, h, PIX_DST & PIX_NOT(PIX_SRC),
	    pixs2, 0, 0);   /* src1 & (~src2)  */
    }

    return pixd;
}


/*!
 *  pixEqual()
 *
 *       Input: pix1
 *              pix2
 *              &bool  (<return> 1 if same; 0 if different)
 *       Return: 0 if OK; 1 on error
 *
 *       This defines equality as having the same bits in their
 *       respective image data arrays.
 *
 *       Notes:  (1) requires 2 pix of equal size
 *               (2) works on pix of any depth
 */
INT32
pixEqual(PIX    *pix1,
         PIX    *pix2,
	 INT32  *pbool)
{
INT32        w, h, wpl1, wpl2, i, j, fullwords, endbits;
UINT32       endmask;
UINT32      *data1, *data2, *line1, *line2;

    PROCNAME("pixEqual");

    if (!pix1)
	return ERROR_INT("pix1 not defined", procName, 1);
    if (!pix2)
	return ERROR_INT("pix2 not defined", procName, 1);
    if (!pbool)
	return ERROR_INT("pbool not defined", procName, 1);

    *pbool = 0;  /* different unless we exit after checking all data */

    if (pixSizesEqual(pix1, pix2) == 0)
	return ERROR_INT("pix sizes unequal", procName, 1);
    w = pixGetWidth(pix1) * pixGetDepth(pix1);   /* little trick def'n */
    h = pixGetHeight(pix1);
    wpl1 = pixGetWpl(pix1);
    wpl2 = pixGetWpl(pix2);
    data1 = pixGetData(pix1);
    data2 = pixGetData(pix2);

    fullwords = w / 32;
    endbits = w & 31;
    endmask = 0xffffffff << (32 - endbits);

    for (i = 0; i < h; i++) {
	line1 = data1 + wpl1 * i;
	line2 = data2 + wpl2 * i;
	for (j = 0; j < fullwords; j++) {
	    if (*line1 ^ *line2)
		return 0;
	    line1++;
	    line2++;
	}
	if (endbits) {
	    if ((*line1 ^ *line2) & endmask)
		return 0;
	}
    }

    *pbool = 1;
    return 0;
}



/*-------------------------------------------------------------*
 *                         Pixel counting                      *
 *-------------------------------------------------------------*/
/*!
 *  pixZero()
 *
 *       Input: pix
 *              &bool  (<return> boolean: 1 if no ON pixels, 0 otherwise)
 *       Return: 0 if OK; 1 on error.
 */
INT32
pixZero(PIX    *pix,
        INT32  *pbool)
{
INT32        w, h, wpl, i, j, fullwords, endbits;
UINT32       endmask;
UINT32      *data, *line;

    PROCNAME("pixZero");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (!pbool)
	return ERROR_INT("pbool not defined", procName, 1);

    w = pixGetWidth(pix) * pixGetDepth(pix);
    h = pixGetHeight(pix);
    wpl = pixGetWpl(pix);
    data = pixGetData(pix);

    fullwords = w / 32;
    endbits = w & 31;
    endmask = 0xffffffff << (32 - endbits);

    *pbool = 1;
    for (i = 0; i < h; i++) {
	line = data + wpl * i;
	for (j = 0; j < fullwords; j++)
	    if (*line++) {
		*pbool = 0;
		return 0;
	    }
	if (endbits) {
	    if (*line & endmask) {
		*pbool = 0;
		return 0;
	    }
	}
    }

    return 0;
}


/*!
 *  pixCountPixels()
 *
 *       Input: binary pix
 *              &count (<return> count of ON pixels)
 *              tab8  (<optional> 8-bit pixel lookup table)
 *       Return: 0 if OK; 1 on error.
 */
INT32
pixCountPixels(PIX    *pix,
	       INT32  *pcount,
	       INT32  *tab8)
{
UINT8        endmask;
INT32        w, h, wpl, i, j;
INT32        fullbytes, endbits, sum;
INT32       *tab;
UINT32      *line, *data;

    PROCNAME("pixCountPixels");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (!pcount)
	return ERROR_INT("pcount not defined", procName, 1);
    if (pixGetDepth(pix) != 1)
	return ERROR_INT("pix not 1 bpp", procName, 1);

    *pcount = 0;

    if (!tab8) {
	if ((tab = makePixelSumTab8()) == NULL)
	    return ERROR_INT("tab not made", procName, 1);
    }
    else
	tab = tab8;

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);
    wpl = pixGetWpl(pix);
    data = pixGetData(pix);

    fullbytes = w / 8;
    endbits = w & 7;
    endmask = 0xff << (8 - endbits);

    sum = 0;
    for (i = 0; i < h; i++) {
	line = data + wpl * i;
	for (j = 0; j < fullbytes; j++)
	    sum += tab[GET_DATA_BYTE(line, j)];
	if (endbits)
	    sum += tab[GET_DATA_BYTE(line, fullbytes) & endmask];
    }
    *pcount = sum;

    if (!tab8)
	FREE((void *)tab);
    return 0;
}


/*!
 *  pixaCountPixels()
 *
 *       Input: binary pixa
 *       Return: na of ON pixels in each pix, or null on error
 */
NUMA *
pixaCountPixels(PIXA   *pixa)
{
INT32        d, i, n, count;
NUMA        *na;
INT32       *tab;
PIX         *pix;

    PROCNAME("pixaCountPixels");

    if (!pixa)
	return (NUMA *)ERROR_PTR("pix not defined", procName, NULL);

    if ((n = pixaGetCount(pixa)) == 0)
	return numaCreate(1);

    pix = pixaGetPix(pixa, 0, CLONE);
    d = pixGetDepth(pix);
    pixDestroy(&pix);
    if (d != 1)
	return (NUMA *)ERROR_PTR("pixa not 1 bpp", procName, NULL);

    if ((tab = makePixelSumTab8()) == NULL)
	return (NUMA *)ERROR_PTR("tab not made", procName, NULL);

    if ((na = numaCreate(n)) == NULL)
	return (NUMA *)ERROR_PTR("na not made", procName, NULL);
    for (i = 0; i < n; i++) {
	pix = pixaGetPix(pixa, i, CLONE);
	pixCountPixels(pix, &count, tab);
	numaAddNumber(na, count);
	pixDestroy(&pix);
    }
	
    FREE((void *)tab);
    return na;
}


/*!
 *  pixCountPixelsInRow()
 *
 *       Input: binary pix
 *              row number
 *              &count (<return> sum of ON pixels in raster line)
 *              tab8  (<optional> 8-bit pixel lookup table)
 *       Return: 0 if OK; 1 on error.
 */
INT32
pixCountPixelsInRow(PIX    *pix,
                    INT32   row,
	            INT32  *pcount,
	            INT32  *tab8)
{
UINT8        endmask;
INT32        j, w, h, wpl;
INT32        fullbytes, endbits, sum;
INT32       *tab;
UINT32      *line;

    PROCNAME("pixCountPixelsInRow");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (!pcount)
	return ERROR_INT("pcount not defined", procName, 1);
    if (pixGetDepth(pix) != 1)
	return ERROR_INT("pix not 1 bpp", procName, 1);

    *pcount = 0;

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);
    if (row < 0 || row >= h)
	return ERROR_INT("row out of bounds", procName, 1);
    wpl = pixGetWpl(pix);
    line = pixGetData(pix) + row * wpl;

    fullbytes = w / 8;
    endbits = w & 7;
    endmask = 0xff << (8 - endbits);

    if (!tab8) {
	if ((tab = makePixelSumTab8()) == NULL)
	    return ERROR_INT("tab not made", procName, 1);
    }
    else
	tab = tab8;

    sum = 0;
    for (j = 0; j < fullbytes; j++)
	sum += tab[GET_DATA_BYTE(line, j)];
    if (endbits)
	sum += tab[GET_DATA_BYTE(line, fullbytes) & endmask];
    *pcount = sum;

    if (!tab8)
	FREE((void *)tab);
    return 0;
}


/*!
 *  pixCountPixelsByRow()
 *
 *       Input: binary pix
 *              tab8  (<optional> 8-bit pixel lookup table)
 *       Return: na of counts, or null on error
 */
NUMA *
pixCountPixelsByRow(PIX    *pix,
	            INT32  *tab8)
{
INT32        w, h, i, count;
INT32       *tab;
NUMA        *na;

    PROCNAME("pixCountPixelsByRow");

    if (!pix)
	return (NUMA *)ERROR_PTR("pix not defined", procName, NULL);
    if (pixGetDepth(pix) != 1)
	return (NUMA *)ERROR_PTR("pix not 1 bpp", procName, NULL);

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);

    if (!tab8) {
	if ((tab = makePixelSumTab8()) == NULL)
	    return (NUMA *)ERROR_PTR("tab not made", procName, NULL);
    }
    else
	tab = tab8;

    if ((na = numaCreate(h)) == NULL)
	return (NUMA *)ERROR_PTR("na not made", procName, NULL);

    for (i = 0; i < h; i++) {
	pixCountPixelsInRow(pix, i, &count, tab);
	numaAddNumber(na, count);
    }

    if (!tab8)
	FREE((void *)tab);

    return na;
}


/*!
 *  pixThresholdPixels()
 *
 *       Input: binary pix
 *              threshold
 *              &boolval (<return> 1 if above threshold;
 *                                 0 if equal to or less than threshold)
 *              tab8  (8-bit pixel lookup table)
 *       Return: 0 if OK; 1 on error
 *
 *       This sums the ON pixels and returns immediately if the count
 *       goes above threshold.  It is therefore more efficient
 *       for matching images (by running this function on the xor of
 *       the 2 images)  than using pixCountPixels(), which counts all
 *       pixels before returning.
 */
INT32
pixThresholdPixels(PIX     *pix,
	           INT32    thresh,
	           INT32   *pboolval,
	           INT32   *tab8)
{
UINT8        endmask;
INT32        w, h, wpl, i, j;
INT32        fullbytes, endbits, sum;
UINT32      *line, *data;

    PROCNAME("pixThresholdPixels");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (!pboolval)
	return ERROR_INT("pboolval not defined", procName, 1);
    if (!tab8)
	return ERROR_INT("tab8 not defined", procName, 1);
    if (pixGetDepth(pix) != 1)
	return ERROR_INT("pix not 1 bpp", procName, 1);

    *pboolval = 0;  /* init */

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);
    wpl = pixGetWpl(pix);
    data = pixGetData(pix);

    fullbytes = w / 8;
    endbits = w & 7;
    endmask = 0xff << (8 - endbits);

    sum = 0;
    for (i = 0; i < h; i++) {
	line = data + wpl * i;
	for (j = 0; j < fullbytes; j++)
	    sum += tab8[GET_DATA_BYTE(line, j)];
	if (endbits)
	    sum += tab8[GET_DATA_BYTE(line, fullbytes) & endmask];
	if (sum > thresh) {
	    *pboolval = 1;
	    return 0;
	}
    }

    return 0;
}


/*!
 *  makePixelSumTab8()
 *
 *       Input: void
 *       Return: table of 256 INT32, or null on error
 *       Use: This table of integers gives the sum of 1 bits
 *            in the 8 bit index
 */
INT32 *
makePixelSumTab8(void)
{
UINT8        byte;
INT32        i;
INT32       *tab;

    PROCNAME("makePixelSumTab8");

    if ((tab = (INT32 *)CALLOC(256, sizeof(INT32))) == NULL)
	return (INT32 *)ERROR_PTR("tab not made", procName, NULL);

    for (i = 0; i < 256; i++) {
        byte = (UINT8)i;
	tab[i] = (byte & 0x1) +
	         ((byte >> 1) & 0x1) +
	         ((byte >> 2) & 0x1) +
	         ((byte >> 3) & 0x1) +
	         ((byte >> 4) & 0x1) +
	         ((byte >> 5) & 0x1) +
	         ((byte >> 6) & 0x1) +
	         ((byte >> 7) & 0x1);
    }

    return tab;
}


/*------------------------------------------------------------------*
 *                            Gray Histogram                        *
 *------------------------------------------------------------------*/
/*!
 *  pixGrayHistogram()
 *
 *     Input:  pix (1, 2, 4, 8, 16 bpp)
 *     Return: na (histogram), or null on error
 */
NUMA *
pixGrayHistogram(PIX   *pixs)
{
INT32    i, j, w, h, d, wpl, val, size;
UINT32  *data, *line;
FLOAT32 *array;
NUMA    *na;

    PROCNAME("pixGrayHistogram");

    if (!pixs)
	return (NUMA *)ERROR_PTR("pixs not defined", procName, NULL);
    d = pixGetDepth(pixs);
    if (d > 16)
	return (NUMA *)ERROR_PTR("depth not in {1,2,4,8,16}", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    wpl = pixGetWpl(pixs);
    data = pixGetData(pixs);

    size = 1 << d;

    na = numaCreate(size);
    na->n = size;  /* fake storage of zeroes */
    array = na->array;  /* don't do this at home */

    for (i = 0; i < h; i++) {
	line = data + i * wpl;
	switch (d) 
	{
	case 1:
	    for (j = 0; j < w; j++) {
		val = GET_DATA_BIT(line, j);
		array[val] += 1.0;
	    }
	    break;
	case 2:
	    for (j = 0; j < w; j++) {
		val = GET_DATA_DIBIT(line, j);
		array[val] += 1.0;
	    }
	    break;
	case 4:
	    for (j = 0; j < w; j++) {
		val = GET_DATA_QBIT(line, j);
		array[val] += 1.0;
	    }
	    break;
	case 8:
	    for (j = 0; j < w; j++) {
		val = GET_DATA_BYTE(line, j);
		array[val] += 1.0;
	    }
	    break;
	case 16:
	    for (j = 0; j < w; j++) {
		val = GET_DATA_TWO_BYTES(line, j);
		array[val] += 1.0;
	    }
	    break;
        default:
	    numaDestroy(&na);
	    return (NUMA *)ERROR_PTR("illegal depth", procName, NULL);
	}
    }

    return na;
}



/*-------------------------------------------------------------*
 *                    Pixel endian conversion                  *
 *-------------------------------------------------------------*/
/*!
 *  pixEndianByteSwap()
 *
 *     Input:  pix
 *     Return: 0 if OK, 1 on error
 *
 *     Action: flips the bytes within a 32-bit word for little-endian
 *             platforms only.  We don't do it on big-endians
 *             because the MSB-to-the-left rule applies to both,
 *             and serialization to left-to-right-byte-order only
 *             requires swapping for little-endians.
 */
INT32
pixEndianByteSwap(PIX   *pix)
{
UINT32     *data;
INT32       i, j, h, wpl;
UINT32      word;

    PROCNAME("pixEndianByteSwap");
        
#ifdef _BIG_ENDIAN_

    return 0;

#else   /* _LITTLE_ENDIAN_ */

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);

    data = pixGetData(pix);
    wpl = pixGetWpl(pix);
    h = pixGetHeight(pix);
    for (i = 0; i < h; i++) {
	for (j = 0; j < wpl; j++, data++) {
	    word = *data;
	    *data = (word >> 24) |
	            ((word >> 8) & 0x0000ff00) |
	            ((word << 8) & 0x00ff0000) |
	            (word << 24);
	}
    }

    return 0;

#endif   /* _BIG_ENDIAN_ */

}


/*-------------------------------------------------------------*
 *                Color sample setting and extraction          *
 *-------------------------------------------------------------*/
/*!
 *  pixCreateRGBImage()
 *
 *     Input: 8 bpp red pix
 *            8 bpp green pix
 *            8 bpp blue pix
 *     Return: 32 bpp pix, interleaved with 4 samples/pixel,
 *             or null on error
 *
 *     Note: the 4th byte, sometimes called the "alpha channel",
 *           and which is often used for blending between different
 *           images, is left with 0 value.
 *     Note: see Note (4) in pix.h for details on storage of
 *           8-bit samples within each 32-bit word.
 */
PIX *
pixCreateRGBImage(PIX   *pixr,
                  PIX   *pixg,
                  PIX   *pixb)
{
INT32    w, h;
PIX     *pixd;

    PROCNAME("pixCreateRGBImage");

    if (!pixr)
        return (PIX *)ERROR_PTR("pixr not defined", procName, NULL);
    if (!pixg)
        return (PIX *)ERROR_PTR("pixg not defined", procName, NULL);
    if (!pixb)
        return (PIX *)ERROR_PTR("pixb not defined", procName, NULL);
    if (pixGetDepth(pixr) != 8)
        return (PIX *)ERROR_PTR("pixr not 8 bpp", procName, NULL);
    if (pixGetDepth(pixg) != 8)
        return (PIX *)ERROR_PTR("pixg not 8 bpp", procName, NULL);
    if (pixGetDepth(pixb) != 8)
        return (PIX *)ERROR_PTR("pixb not 8 bpp", procName, NULL);
 
    w = pixGetWidth(pixr);
    h = pixGetHeight(pixr);
    if (w != pixGetWidth(pixg) || w != pixGetWidth(pixb))
        return (PIX *)ERROR_PTR("widths not the same", procName, NULL);
    if (h != pixGetHeight(pixg) || h != pixGetHeight(pixb))
        return (PIX *)ERROR_PTR("heights not the same", procName, NULL);

    if ((pixd = pixCreate(w, h, 32)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    pixSetRGBComponent(pixd, pixr, COLOR_RED);
    pixSetRGBComponent(pixd, pixg, COLOR_GREEN);
    pixSetRGBComponent(pixd, pixb, COLOR_BLUE);

    return pixd;
}


/*!
 *  pixGetRGBComponent()
 *
 *     Input:  pixs  (32 bpp)
 *             color  (one of {COLOR_RED, COLOR_GREEN, COLOR_BLUE})
 *     Return:  pixd, the selected 8 bpp component image of the
 *              input 32 bpp image, or null on error
 */
PIX *
pixGetRGBComponent(PIX    *pixs,
	           INT32   color)
{
UINT8       srcbyte;
UINT32     *lines, *lined;
UINT32     *datas, *datad;
INT32       i, j, w, h;
INT32       wpls, wpld;
PIX 	   *pixd;

    PROCNAME("pixGetRGBComponent");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
        return (PIX *)ERROR_PTR("pixs not 32 bpp", procName, NULL);
    if (color != COLOR_RED && color != COLOR_GREEN && color != COLOR_BLUE)
        return (PIX *)ERROR_PTR("invalid color", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    if ((pixd = pixCreate(w, h, 8)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    wpls = pixGetWpl(pixs);
    wpld = pixGetWpl(pixd);
    datas = pixGetData(pixs);
    datad = pixGetData(pixd);

    for (i = 0; i < h; i++) {
	lines = datas + i * wpls;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    srcbyte = GET_DATA_BYTE(lines + j, color);
	    SET_DATA_BYTE(lined, j, srcbyte);
	}
    }

    return pixd;
}


/*!
 *  pixSetRGBComponent()
 *
 *     Input:  pixd  (32 bpp)
 *             pixs  (8 bpp)
 *             color  (one of {COLOR_RED, COLOR_GREEN, COLOR_BLUE})
 *     Return:  0 if OK; 1 on error
 *     Action:  places specificed 8 bpp color component 
 *              interleaved appropriately in 32 bpp dest
 */
INT32
pixSetRGBComponent(PIX   *pixd,
	           PIX   *pixs,
	           INT32  color)
{
UINT8       srcbyte;
UINT32     *lines, *lined;
UINT32     *datas, *datad;
INT32       i, j, w, h;
INT32       wpls, wpld;

    PROCNAME("pixSetRGBComponent");

    if (!pixd)
        return ERROR_INT("pixd not defined", procName, 1);
    if (!pixs)
        return ERROR_INT("pixs not defined", procName, 1);

    if (pixGetDepth(pixd) != 32)
        return ERROR_INT("pixd not 32 bpp", procName, 1);
    if (pixGetDepth(pixs) != 8)
        return ERROR_INT("pixs not 8 bpp", procName, 1);
    if (color != COLOR_RED && color != COLOR_GREEN && color != COLOR_BLUE)
        return ERROR_INT("invalid color", procName, 1);
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    if (w != pixGetWidth(pixd) || h != pixGetHeight(pixd))
	return ERROR_INT("sizes not commensurate", procName, 1);

    datas = pixGetData(pixs);
    datad = pixGetData(pixd);
    wpls = pixGetWpl(pixs);
    wpld = pixGetWpl(pixd);

    for (i = 0; i < h; i++) {
	lines = datas + i * wpls;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    srcbyte = GET_DATA_BYTE(lines, j);
	    SET_DATA_BYTE(lined + j, color, srcbyte);
	}
    }

    return 0;
}


/*!
 *  composeRGBPixel()
 *
 *     Input:  rval, gval, bval
 *             &rgbpixel  (<return> 32-bit pixel)
 *     Return:  0 if OK; 1 on error
 */
INT32
composeRGBPixel(INT32    rval,
	        INT32    gval,
		INT32    bval,
		UINT32  *ppixel)
{
    PROCNAME("composeRGBPixel");

    if (!ppixel)
        return ERROR_INT("&pixel not defined", procName, 1);

    SET_DATA_BYTE(ppixel, COLOR_RED, rval);
    SET_DATA_BYTE(ppixel, COLOR_GREEN, gval);
    SET_DATA_BYTE(ppixel, COLOR_BLUE, bval);
    return 0;
}


/*!
 *  pixGetRGBLine()
 *
 *     Input:  pixs  (32 bpp)
 *             row
 *             bufr  (array of red samples; size w bytes)
 *             bufg  (array of green samples; size w bytes)
 *             bufb  (array of blue samples; size w bytes)
 *     Return:  0 if OK; 1 on error
 *
 *     Action: puts rgb components from the input line in pixs
 *             into the given buffers
 */
INT32
pixGetRGBLine(PIX    *pixs,
	      INT32   row,
	      UINT8  *bufr,
	      UINT8  *bufg,
	      UINT8  *bufb)
{
UINT32     *lines;
INT32       j, w, h;
INT32       wpls;

    PROCNAME("pixGetRGBLine");

    if (!pixs)
        return ERROR_INT("pixs not defined", procName, 1);
    if (pixGetDepth(pixs) != 32)
        return ERROR_INT("pixs not 32 bpp", procName, 1);
    if (!bufr || !bufg || !bufb)
        return ERROR_INT("buffer not defined", procName, 1);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    if (row < 0 || row >= h)
        return ERROR_INT("row out of bounds", procName, 1);
    wpls = pixGetWpl(pixs);
    lines = pixGetData(pixs) + row * wpls;

    for (j = 0; j < w; j++) {
	bufr[j] = GET_DATA_BYTE(lines + j, COLOR_RED);
	bufg[j] = GET_DATA_BYTE(lines + j, COLOR_GREEN);
	bufb[j] = GET_DATA_BYTE(lines + j, COLOR_BLUE);
    }

    return 0;
}


/*-------------------------------------------------------------*
 *               Conversion from colormapped pix               *
 *-------------------------------------------------------------*/
/*!
 *   pixRemoveColormap()
 *
 *        Input:  pix
 *                type (CONVERT_TO_GRAYSCALE, CONVERT_TO_FULL_COLOR)
 *        Return: new pix, or null on error
 *
 *        Note:  The input pixs is restricted to 2, 4 or 8 bpp.
 *               For grayscale conversion, we use a weighted average
 *               of the squared RGB values.
 */
PIX *
pixRemoveColormap(PIX    *pixs,
                  INT32   type)
{
UINT8         sval, gval;
UINT8        *rmap, *gmap, *bmap;
INT32         i, j, numcolors, w, h, d, wpls, wpld;
INT32         mval;
UINT32       *datas, *lines, *datad, *lined;
PIXCMAP      *cmap;
PIX          *pixd;

    PROCNAME("pixRemoveColormap");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (type != CONVERT_TO_GRAYSCALE && type != CONVERT_TO_FULL_COLOR)
	return (PIX *)ERROR_PTR("invalid conversion type", procName, NULL);

    d = pixGetDepth(pixs);
    if (d != 8 && d != 4 && d != 2)
	return (PIX *)ERROR_PTR("pixs must be {8, 4, 2} bpp", procName, NULL);
    if ((cmap = pixGetColormap(pixs)) == NULL)
	return pixClone(pixs);

    if (colormapToArrays(cmap, &rmap, &gmap, &bmap, &numcolors))
	return (PIX *)ERROR_PTR("colormap arrays not made", procName, NULL);

    if (numcolors > (1 << d)) {
	fprintf(stderr,
	    "colormap table with %d colors is incompatible with pix depth %d\n",
	     numcolors, d);
	return pixClone(pixs);
    }
/*    fprintf(stderr, " removing colormap table w/ %d colors\n", numcolors); */

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    switch (type)
    {
    case CONVERT_TO_GRAYSCALE:
	if ((pixd = pixCreate(w, h, 8)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
	datad = pixGetData(pixd);
	wpld = pixGetWpl(pixd);
	for (i = 0; i < h; i++) {
	    lines = datas + i * wpls;
	    lined = datad + i * wpld;
	    for (j = 0; j < w; j++) {
		if (d == 8)
		    sval = GET_DATA_BYTE(lines, j);
		else if (d == 4)
		    sval = GET_DATA_QBIT(lines, j);
		else   /* d == 2 */
		    sval = GET_DATA_DIBIT(lines, j);

		mval = (rmap[sval] * rmap[sval] +
			2 * gmap[sval] * gmap[sval] +
			bmap[sval] * bmap[sval]) / 4;
		gval = (UINT8)sqrt((FLOAT64)mval);
		SET_DATA_BYTE(lined, j, gval);
	    }
	}
	break;

    case CONVERT_TO_FULL_COLOR:
	if ((pixd = pixCreate(w, h, 32)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
	datad = pixGetData(pixd);
	wpld = pixGetWpl(pixd);
	for (i = 0; i < h; i++) {
	    lines = datas + i * wpls;
	    lined = datad + i * wpld;
	    for (j = 0; j < w; j++) {
		if (d == 8)
		    sval = GET_DATA_BYTE(lines, j);
		else if (d == 4)
		    sval = GET_DATA_QBIT(lines, j);
		else   /* d == 2 */
		    sval = GET_DATA_DIBIT(lines, j);

		SET_DATA_BYTE(lined + j, COLOR_RED, rmap[sval]);
		SET_DATA_BYTE(lined + j, COLOR_GREEN, gmap[sval]);
		SET_DATA_BYTE(lined + j, COLOR_BLUE, bmap[sval]);
	    }
	}
	break;

    default:
        WARNING("invalid conversion type", procName);
	return pixClone(pixs);
    }

    FREE((void *)rmap);
    FREE((void *)gmap);
    FREE((void *)bmap);
    return pixd;
}


/*!
 *   colormapToArrays()
 *
 *      Input:  colormap
 *              &rmap, &gmap, &bmap  (<return> colormap arrays)
 *              &numcolors  (<return> size of colormap arrays)
 *      Return: 0 if OK; 1 on error
 */
INT32
colormapToArrays(PIXCMAP   *cmap,
                 UINT8    **prmap,
                 UINT8    **pgmap,
                 UINT8    **pbmap,
		 INT32     *pnumcolors)
{
UINT8                *rmap, *gmap, *bmap;
INT32                 i, numcolors;
GIFColorTableEntry   *pgcte;
RGBQUAD              *prgb;

    PROCNAME("colormapToArrays");

    if (!cmap)
	return ERROR_INT("cmap not defined", procName, 1);

    *prmap = *pgmap = *pbmap = NULL;
    *pnumcolors = 0;

    if (cmap->type == CM_BMP)
	numcolors = cmap->size / sizeof(RGBQUAD);
    else if (cmap->type == CM_GIF)
	numcolors = cmap->size / sizeof(GIFColorTableEntry);
    else
	return ERROR_INT("invalid colormap type", procName, 1);
    *pnumcolors = numcolors;

    if (((rmap = (UINT8 *)CALLOC(numcolors, sizeof(UINT8))) == NULL) ||
	((gmap = (UINT8 *)CALLOC(numcolors, sizeof(UINT8))) == NULL) ||
	((bmap = (UINT8 *)CALLOC(numcolors, sizeof(UINT8))) == NULL))
	    return ERROR_INT("calloc fail for *map", procName, 1);
    *prmap = rmap;
    *pgmap = gmap;
    *pbmap = bmap;

    switch (cmap->type)
    {
    case CM_BMP:
	prgb = (RGBQUAD *)cmap->array;
	for (i = 0; i < numcolors; i++, prgb++) {
	    rmap[i] = prgb->rgbRed;
	    gmap[i] = prgb->rgbGreen;
	    bmap[i] = prgb->rgbBlue;
	}
	break;

    case CM_GIF:
	pgcte = (GIFColorTableEntry *)cmap->array;
	for (i = 0; i < numcolors; i++, pgcte++) {
	    rmap[i] = pgcte->cmRed;
	    gmap[i] = pgcte->cmGreen;
	    bmap[i] = pgcte->cmBlue;
	}
	break;

    default:
	return ERROR_INT("shouldn't get here: invalid colormap type",
	                 procName, 1);
    }

    return 0;
}


/*-------------------------------------------------------------*
 *            Conversion from RGB color to grayscale           *
 *-------------------------------------------------------------*/
/*!
 *   pixConvertRGBToGray()
 *
 *        Input:  pix (32 bpp RGB)
 *        Return: 8 bpp pix, or null on error
 *
 *        Note:  we use a weighted average of the RGB values
 */
PIX *
pixConvertRGBToGray(PIX    *pixs)
{
UINT8       redbyte, greenbyte, bluebyte, graybyte;
INT32       i, j, w, h, wpls, wpld;
FLOAT32     norm;
UINT32     *datas, *words, *datad, *lined;
PIX        *pixd;

    PROCNAME("pixConvertRGBToGray");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
	return (PIX *)ERROR_PTR("pixs not 32 bpp", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

    if ((pixd = pixCreate(w, h, 8)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);
    
    norm = 1. / (2.0 + GREEN_WEIGHT);
    for (i = 0; i < h; i++) {
	words = datas + i * wpls;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++, words++) {
	    redbyte = GET_DATA_BYTE(words, COLOR_RED);
	    greenbyte = GET_DATA_BYTE(words, COLOR_GREEN);
	    bluebyte = GET_DATA_BYTE(words, COLOR_BLUE);
	    graybyte = (UINT8)(RED_WEIGHT * redbyte +
		               GREEN_WEIGHT * greenbyte +
			       BLUE_WEIGHT * bluebyte + 0.5);
	    SET_DATA_BYTE(lined, j, graybyte);
	}
    }

    return pixd;
}


/*-------------------------------------------------------------*
 *                     Add and remove border                   *
 *-------------------------------------------------------------*/
/*!
 *   pixAddBorder()
 *
 *         Input: pix
 *                npix  (number of pixels to be added to each side)
 *                val   (value of added border pixels)
 *         Return: pix with the input pix centered, or null on error.
 *
 *         For binary images,
 *               white:  val = 0
 *               black:  val = 1
 *         For grayscale images,
 *               white:  val = 2 ** d - 1
 *               black:  val = 0
 */
PIX *
pixAddBorder(PIX     *pixs,
	     INT32    npix,
	     INT32    val)
{
INT32        ws, hs, wd, hd, d;
PIX         *pixd;

    PROCNAME("pixAddBorder");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    if (npix == 0)
	return pixClone(pixs);

    d = pixGetDepth(pixs); 
    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    wd = ws + 2 * npix;
    hd = hs + 2 * npix;
    if ((pixd = pixCreate(wd, hd, d)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixSetAllArbitrary(pixd, val);   /* a little extra writing ! */

    pixRasterop(pixd, npix, npix, ws, hs, PIX_SRC, pixs, 0, 0);

    return pixd;
}


/*!
 *  pixRemoveBorder()
 *
 *         Input: pixs
 *                npix  (number to be removed from each of the 4 sides)
 *         Return: pixd, or null on error
 */
PIX *
pixRemoveBorder(PIX    *pixs,
	        INT32   npix)
{
INT32        ws, hs, wd, hd, d;
PIX         *pixd;

    PROCNAME("pixRemoveBorder");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    if (npix == 0)
	return pixClone(pixs);

    d = pixGetDepth(pixs); 
    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    wd = ws - 2 * npix;
    hd = hs - 2 * npix;
    if (wd <= 0)
	return (PIX *)ERROR_PTR("width must be > 0", procName, NULL);
    if (hd <= 0)
	return (PIX *)ERROR_PTR("height must be > 0", procName, NULL);
    if ((pixd = pixCreate(wd, hd, d)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

	/* rasterop from the center */
    pixRasterop(pixd, 0, 0, wd, hd, PIX_SRC, pixs, npix, npix);
    
    return pixd;
}


/*!
 *   pixAddBorderGeneral()
 *
 *         Input: pix
 *                leftpix, rightpix, toppix, bottompix  (number of pixels
 *                     to be added to each side)
 *                val   (value of added border pixels)
 *         Return: pix with the input pix placed properly, or null on error
 *
 *         For binary images,
 *               white:  val = 0
 *               black:  val = 1
 *         For grayscale images,
 *               white:  val = 2 ** d - 1
 *               black:  val = 0
 */
PIX *
pixAddBorderGeneral(PIX     *pixs,
	            INT32    leftpix,
	            INT32    rightpix,
	            INT32    toppix,
	            INT32    bottompix,
	            INT32    val)
{
INT32        ws, hs, wd, hd, d;
PIX         *pixd;

    PROCNAME("pixAddBorderGeneral");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    d = pixGetDepth(pixs); 
    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    wd = ws + leftpix + rightpix;
    hd = hs + toppix + bottompix;
    if ((pixd = pixCreate(wd, hd, d)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixSetAllArbitrary(pixd, val);   /* a little extra writing ! */

    pixRasterop(pixd, leftpix, toppix, ws, hs, PIX_SRC, pixs, 0, 0);

    return pixd;
}


/*!
 *  pixRemoveBorderGeneral()
 *
 *         Input: pixs
 *                leftpix, rightpix, toppix, bottompix  (number of pixels
 *                     to be removed from each side)
 *         Return: pixd with pixels removed around border, or null on error
 *
 */
PIX *
pixRemoveBorderGeneral(PIX    *pixs,
	               INT32   leftpix,
	               INT32   rightpix,
	               INT32   toppix,
	               INT32   bottompix)
{
INT32        ws, hs, wd, hd, d;
PIX         *pixd;

    PROCNAME("pixRemoveBorderGeneral");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);

    d = pixGetDepth(pixs); 
    ws = pixGetWidth(pixs);
    hs = pixGetHeight(pixs);
    wd = ws - leftpix - rightpix;
    hd = hs - toppix - bottompix;
    if (wd <= 0)
	return (PIX *)ERROR_PTR("width must be > 0", procName, NULL);
    if (hd <= 0)
	return (PIX *)ERROR_PTR("height must be > 0", procName, NULL);
    if ((pixd = pixCreate(wd, hd, d)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixRasterop(pixd, 0, 0, wd, hd, PIX_SRC, pixs, leftpix, toppix);
    
    return pixd;
}



/*-------------------------------------------------------------*
 *                Extract rectangular region                   *
 *-------------------------------------------------------------*/
/*!
 *  pixClipRectangle()
 *
 *      Input: pixs
 *             clipping box
 *      Return: clipped pix, or null on error
 */
PIX *
pixClipRectangle(PIX    *pixs,
	         BOX    *box)
{
INT32        d;
PIX         *pixd;

    PROCNAME("pixClipRectangle");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (!box)
	return (PIX *)ERROR_PTR("box not defined", procName, NULL);

    d = pixGetDepth(pixs);
    if ((pixd = pixCreate(box->w, box->h, d)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixRasterop(pixd, 0, 0, box->w, box->h, PIX_SRC, pixs, box->x, box->y);

    return pixd;
}


/*-------------------------------------------------------------*
 *              Extract min rectangle with ON pixels           *
 *-------------------------------------------------------------*/
/*!
 *  pixClipToForeground()
 *
 *      Input: pixs
 *             &pixd  (<optional return> clipped pix returned)
 *             &box   (<optional return> bounding box)
 *      Return: 0 if OK; 1 on error
 *
 *      Note: at least one of {&pixd, &box} must be specified
 */
INT32
pixClipToForeground(PIX    *pixs,
                    PIX   **ppixd,
                    BOX   **pbox)
{
INT32       w, h, d, wpl, nfullwords, extra, i, j;
INT32       minx, miny, maxx, maxy;
UINT32      result, mask;
UINT32     *data, *line;
BOX        *box;

    PROCNAME("pixClipToForeground");

    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);
    if ((d = pixGetDepth(pixs)) != 1)
	return ERROR_INT("pixs not binary", procName, 1);
    if (!ppixd && !pbox)
	return ERROR_INT("neither &pixd nor &pbox defined", procName, 1);

    w = pixGetWidth(pixs);
    nfullwords = w / 32;
    extra = w & 31;
    mask = ~rmask32[32 - extra];
    h = pixGetHeight(pixs);
    wpl = pixGetWpl(pixs);
    data = pixGetData(pixs);

    result = 0;
    for (i = 0, miny = 0; i < h; i++, miny++) {
	line = data + i * wpl;
	for (j = 0; j < nfullwords; j++)
	    result |= line[j];
	if (extra)
	    result |= (line[j] & mask);
	if (result)
	    break;
    }
    if (miny == h) {  /* no ON pixels */
        if (ppixd)
	    *ppixd = NULL;
	if (pbox)
	    *pbox = NULL;
        return 1;
    }

    result = 0;
    for (i = h - 1, maxy = h - 1; i >= 0; i--, maxy--) {
	line = data + i * wpl;
	for (j = 0; j < nfullwords; j++)
	    result |= line[j];
	if (extra)
	    result |= (line[j] & mask);
	if (result)
	    break;
    }

    minx = 0;
    for (j = 0, minx = 0; j < w; j++, minx++) {
	for (i = 0; i < h; i++) {
	    line = data + i * wpl;
	    if (GET_DATA_BIT(line, j))
		goto minx_found;
	}
    }

minx_found:
    for (j = w - 1, maxx = w - 1; j >= 0; j--, maxx--) {
	for (i = 0; i < h; i++) {
	    line = data + i * wpl;
	    if (GET_DATA_BIT(line, j))
		goto maxx_found;
	}
    }

maxx_found:
    box = boxCreate(minx, miny, maxx - minx + 1, maxy - miny + 1);

    if (ppixd)
	*ppixd = pixClipRectangle(pixs, box);
    if (pbox)
	*pbox = box;
    else
	boxDestroy(&box);

    return 0;
}


