/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *   pts.c
 *
 *      Pta creation, destruction, copy, clone, empty
 *           PTA    *ptaCreate()
 *           void    ptaDestroy()
 *           PTA    *ptaCopy()
 *           PTA    *ptaClone()
 *           INT32   ptaEmpty()
 *
 *      Pta array extension
 *           INT32   ptaAddPt()
 *           INT32   ptaExtendArrays()
 *
 *      Pta rearrangements
 *           INT32   ptaConcatenate()
 *           PTA    *ptaReverse()
 *           PTA    *ptaCyclicPerm()
 *
 *      Pta Accessors
 *           INT32   ptaGetRefcount()
 *           INT32   ptaChangeRefcount()
 *           INT32   ptaGetCount()
 *           INT32   ptaGetPt()
 *           INT32   ptaGetIPt()
 *
 *      Ptaa creation, destruction
 *           PTAA   *ptaaCreate()
 *           void    ptaaDestroy()
 *
 *      Ptaa array extension
 *           INT32   ptaaAddPta()
 *           INT32   ptaaExtendArray()
 *
 *      Ptaa Accessors
 *           INT32   ptaaGetCount()
 *           INT32   ptaaGetPta()
 *  
 *      Pta serialized I/O
 *           INT32   ptaWriteStream()
 *           PTA    *ptaReadStream()
 *
 *      In use
 *           PTA    *ptaFindCornerPixels()
 *           INT32   ptaContainsPt()
 *           INT32   ptaTestIntersection()
 */

#include <stdio.h>
#include <stdlib.h>

#include "allheaders.h"


#define  INITIAL_PTR_ARRAYSIZE      20    /* n'import quoi */


/*---------------------------------------------------------------------*
 *                PTA creation, destruction, copy, clone               *
 *---------------------------------------------------------------------*/
/*!
 *  ptaCreate()
 *       Input:  n  (initial array sizes)
 *       Return: pta, or null on error.
 */
PTA *
ptaCreate(INT32  n)
{
PTA   *pta;

    PROCNAME("ptaCreate");

    if (n <= 0)
	n = INITIAL_PTR_ARRAYSIZE;

    if ((pta = (PTA *)CALLOC(1, sizeof(PTA))) == NULL)
	return (PTA *)ERROR_PTR("pta not made", procName, NULL);
    pta->n = 0;
    pta->nalloc = n;
    ptaChangeRefcount(pta, 1);  /* sets to 1 */
    
    if ((pta->x = (FLOAT32 *)CALLOC(n, sizeof(FLOAT32))) == NULL)
	return (PTA *)ERROR_PTR("x array not made", procName, NULL);
    if ((pta->y = (FLOAT32 *)CALLOC(n, sizeof(FLOAT32))) == NULL)
	return (PTA *)ERROR_PTR("y array not made", procName, NULL);

    return pta;
}


/*!
 *  ptaDestroy()
 *
 *       Input:  &pta (<to be nulled>)
 *       Return:  void
 *       Action:  decrements the refcount and, if 0, destroys
 *                the pta and nulls in te input ptr
 */
void
ptaDestroy(PTA  **ppta)
{
INT32   i;
PTA    *pta;

    PROCNAME("ptaDestroy");

    if (ppta == NULL) {
        WARNING("ptr address is NULL!", procName);
	return;
    }

    if ((pta = *ppta) == NULL)
	return;

    ptaChangeRefcount(pta, -1);
    if (ptaGetRefcount(pta) > 0)
	return;

    FREE((void *)pta->x);
    FREE((void *)pta->y);
    FREE((void *)pta);
    *ppta = NULL;
    return;
}


/*!
 *  ptaCopy()
 *
 *     Input: pta
 *     Return: copy of pta, or null on error
 */
PTA *
ptaCopy(PTA  *pta)
{
INT32    i;
FLOAT32  x, y;
PTA     *npta;

    PROCNAME("ptaCopy");

    if (!pta)
	return (PTA *)ERROR_PTR("pta not defined", procName, NULL);

    if ((npta = ptaCreate(pta->nalloc)) == NULL)
	return (PTA *)ERROR_PTR("npta not made", procName, NULL);

    for (i = 0; i < pta->n; i++) {
	ptaGetPt(pta, i, &x, &y);
	ptaAddPt(npta, x, y);
    }

    return npta;
}


/*!
 *  ptaClone()
 *
 *    Input:  pta
 *    Return: ptr to same pta, or null on error
 */
PTA *
ptaClone(PTA  *pta)
{
    PROCNAME("ptaClone");

    if (!pta)
	return (PTA *)ERROR_PTR("pta not defined", procName, NULL);

    ptaChangeRefcount(pta, 1);
    return pta;
}


/*!
 *   ptaEmpty()
 *
 *         Input:  pta
 *         Return: 0 if OK, 1 on error
 *
 *         Note: this only resets the "n" field, for reuse
 */
INT32
ptaEmpty(PTA   *pta)
{

    PROCNAME("ptaEmpty");

    if (!pta)
	return ERROR_INT("ptad not defined", procName, 1);
    pta->n = 0;
    return 0;
}


/*---------------------------------------------------------------------*
 *                          PTA array extension                        *
 *---------------------------------------------------------------------*/
/*!
 *  ptaAddPt()
 *
 *       Input:  pta 
 *               x, y
 */
INT32
ptaAddPt(PTA     *pta,
         FLOAT32  x,
	 FLOAT32  y)
{
INT32   n;

    PROCNAME("ptaAddPt");

    if (!pta)
	return ERROR_INT("pta not defined", procName, 1);

    n = pta->n;
    if (n >= pta->nalloc)
	ptaExtendArrays(pta);
    pta->x[n] = x;
    pta->y[n] = y;
    pta->n++;

    return 0;
}


/*!
 *  ptaExtendArrays()
 *
 *       Input:  pta
 *       Return: 0 if OK; 1 on error
 */
INT32
ptaExtendArrays(PTA   *pta)
{

    PROCNAME("ptaExtendArrays");

    if (!pta)
	return ERROR_INT("pta not defined", procName, 1);

    if ((pta->x = (FLOAT32 *)reallocNew((void **)&pta->x,
                                4 * pta->nalloc,
				8 * pta->nalloc)) == NULL)
	return ERROR_INT("new x array not returned", procName, 1);
    if ((pta->y = (FLOAT32 *)reallocNew((void **)&pta->y,
                                4 * pta->nalloc,
				8 * pta->nalloc)) == NULL)
	return ERROR_INT("new y array not returned", procName, 1);

    pta->nalloc = 2 * pta->nalloc;
    return 0;
}



/*---------------------------------------------------------------------*
 *                           Pta rearrangements                        *
 *---------------------------------------------------------------------*/
/*!
 *   ptaJoin()
 *
 *         Input:  ptad  (dest pta; add to this one)
 *                 ptas  (source pta; add from this one)
 *                 istart  (starting index in ptas)
 *                 iend  (ending index in ptas; use 0 to cat all)
 *         Return: 0 if OK, 1 on error
 *         Note: this assumes integer values.  Also, istart < 0 is
 *               taken to be istart = 0, and ditto for iend (except
 *               iend = 0 means read to the end).
 */
INT32
ptaJoin(PTA   *ptad,
        PTA   *ptas,
        INT32  istart,
        INT32  iend)
{
INT32    ns, i, x, y;

    PROCNAME("ptaJoin");

    if (!ptad)
	return ERROR_INT("ptad not defined", procName, 1);
    if (!ptas)
	return ERROR_INT("ptas not defined", procName, 1);
    ns = ptaGetCount(ptas);
    if (istart < 0)
	istart = 0;
    if (istart >= ns)
	return ERROR_INT("istart out of bounds", procName, 1);
    if (iend <= 0)
	iend = ns - 1;
    if (iend >= ns)
	return ERROR_INT("iend out of bounds", procName, 1);
    if (istart > iend)
	return ERROR_INT("istart > iend; no pts", procName, 1);

    for (i = istart; i <= iend; i++) {
	ptaGetIPt(ptas, i, &x, &y);
	ptaAddPt(ptad, x, y);
    }

    return 0;
}


/*!
 *   ptaReverse()
 *
 *         Input:  ptas
 *                 type  (0 for float values; 1 for integer values)
 *         Return: ptad (reversed pta), or null on error
 */
PTA  *
ptaReverse(PTA   *ptas,
           INT32  type)
{
INT32    n, i, ix, iy;
FLOAT32  x, y;
PTA    *ptad;

    PROCNAME("ptaReverse");

    if (!ptas)
	return (PTA *)ERROR_PTR("ptas not defined", procName, NULL);

    n = ptaGetCount(ptas);
    if ((ptad = ptaCreate(n)) == NULL)
	return (PTA *)ERROR_PTR("ptad not made", procName, NULL);
    for (i = n - 1; i >= 0; i--) {
	if (type == 0) {
	    ptaGetPt(ptas, i, &x, &y);
	    ptaAddPt(ptad, x, y);
	}
	else {  /* type == 1 */
	    ptaGetIPt(ptas, i, &ix, &iy);
	    ptaAddPt(ptad, ix, iy);
	}
    }

    return ptad;
}


/*!
 *   ptaCyclicPerm()
 *
 *         Input:  ptas
 *                 xs, ys  (start point; must be in ptas)
 *         Return: ptad (cyclic permutation, starting and ending at (xs, ys),
 *                 or null on error
 *
 *         Note: we check to insure that ptas is a closed path where
 *               the first and last points are identical, and the
 *               resulting pta also starts and ends on the same point
 *               (which in this case is (xs, ys).
 */
PTA  *
ptaCyclicPerm(PTA   *ptas,
              INT32  xs,
              INT32  ys)
{
INT32    n, i, x, y, j, index, state;
INT32    x1, y1, x2, y2;
PTA     *ptad;

    PROCNAME("ptaCyclicPerm");

    if (!ptas)
	return (PTA *)ERROR_PTR("ptas not defined", procName, NULL);

    n = ptaGetCount(ptas);

	/* verify input data */
    ptaGetIPt(ptas, 0, &x1, &y1);
    ptaGetIPt(ptas, n - 1, &x2, &y2);
    if (x1 != x2 || y1 != y2)
	return (PTA *)ERROR_PTR("start and end pts not same", procName, NULL);
    state = NOT_FOUND;
    for (i = 0; i < n; i++) {
	ptaGetIPt(ptas, i, &x, &y);
	if (x == xs && y == ys) {
	    state = FOUND;
	    break;
	}
    }
    if (state == NOT_FOUND)
	return (PTA *)ERROR_PTR("start pt not in ptas", procName, NULL);

    if ((ptad = ptaCreate(n)) == NULL)
	return (PTA *)ERROR_PTR("ptad not made", procName, NULL);
    for (j = 0; j < n - 1; j++) {
	if (i + j < n - 1)
	    index = i + j;
	else
	    index = (i + j + 1) % n;
	ptaGetIPt(ptas, index, &x, &y);
	ptaAddPt(ptad, x, y);
    }
    ptaAddPt(ptad, xs, ys);

    return ptad;
}



/*---------------------------------------------------------------------*
 *                           Pta accessors                             *
 *---------------------------------------------------------------------*/
INT32
ptaGetRefcount(PTA  *pta)
{
    PROCNAME("ptaGetRefcount");

    if (!pta)
	return ERROR_INT("pta not defined", procName, 1);
    return pta->refcount;
}


INT32
ptaChangeRefcount(PTA   *pta,
                  INT32  delta)
{
    PROCNAME("ptaChangeRefcount");

    if (!pta)
	return ERROR_INT("pta not defined", procName, 1);
    pta->refcount += delta;
    return 0;
}


/*!
 *  ptaGetCount()
 *
 *       Input:  pta
 *       Return: count, or 0 if no pta
 */
INT32
ptaGetCount(PTA   *pta)
{

    PROCNAME("ptaGetCount");

    if (!pta)
	return ERROR_INT("pta not defined", procName, 0);

    return pta->n;
}


/*!
 *  ptaGetPt()
 *
 *       Input:  pta
 *               index  (into arrays)
 *               &x, &y  (<return> float values)
 *       Return: 0 if OK; 1 on error
 */
INT32
ptaGetPt(PTA      *pta,
         INT32     index,
         FLOAT32  *px,
	 FLOAT32  *py)
{

    PROCNAME("ptaGetPt");

    if (!pta)
	return ERROR_INT("pta not defined", procName, 1);

    *px = pta->x[index];
    *py = pta->y[index];
    return 0;
}


/*!
 *  ptaGetIPt()
 *
 *       Input:  pta
 *               index  (into arrays)
 *               &x, &y  (<return> integer values)
 *       Return: 0 if OK; 1 on error
 */
INT32
ptaGetIPt(PTA    *pta,
          INT32   index,
          INT32  *px,
	  INT32  *py)
{

    PROCNAME("ptaGetIPt");

    if (!pta)
	return ERROR_INT("pta not defined", procName, 1);

    *px = (INT32)(pta->x[index] + 0.5);
    *py = (INT32)(pta->y[index] + 0.5);
    return 0;
}



/*---------------------------------------------------------------------*
 *                     PTAA creation, destruction                      *
 *---------------------------------------------------------------------*/
/*!
 *  ptaaCreate()
 *       Input:  n  (initial number of ptrs)
 *       Return: ptaa, or null on error
 */
PTAA *
ptaaCreate(INT32  n)
{
PTAA   *ptaa;

    PROCNAME("ptaaCreate");

    if (n <= 0)
	n = INITIAL_PTR_ARRAYSIZE;

    if ((ptaa = (PTAA *)CALLOC(1, sizeof(PTAA))) == NULL)
	return (PTAA *)ERROR_PTR("ptaa not made", procName, NULL);
    ptaa->n = 0;
    ptaa->nalloc = n;

    if ((ptaa->pta = (PTA **)CALLOC(n, sizeof(PTA *))) == NULL)
	return (PTAA *)ERROR_PTR("pta ptrs not made", procName, NULL);
    
    return ptaa;
}


/*!
 *  ptaaDestroy()
 *
 *       Input:  &ptaa <to be nulled>
 *       Return:  void
 */
void
ptaaDestroy(PTAA  **pptaa)
{
INT32    i;
PTAA    *ptaa;

    PROCNAME("ptaaDestroy");

    if (pptaa == NULL) {
        WARNING("ptr address is NULL!", procName);
	return;
    }

    if ((ptaa = *pptaa) == NULL)
	return;

    for (i = 0; i < ptaa->n; i++) 
	ptaDestroy(&ptaa->pta[i]);
    FREE((void *)ptaa->pta);

    FREE((void *)ptaa);
    *pptaa = NULL;
    return;
}


/*---------------------------------------------------------------------*
 *                          PTAA array extension                       *
 *---------------------------------------------------------------------*/
/*!
 *  ptaaAddPta()
 *
 *      Input: ptaa
 *             pta  (to be added)
 *             copyflag  (INSERT, COPY, CLONE)
 *      Return: 0 if OK, 1 on error
 *
 */
 INT32
 ptaaAddPta(PTAA   *ptaa,
            PTA    *pta,
	    INT32   copyflag)
{
INT32    n;
PTA     *ptac;

    PROCNAME("ptaaAddPta");

    if (!ptaa)
	return ERROR_INT("ptaa not defined", procName, 1);
    if (!pta)
	return ERROR_INT("pta not defined", procName, 1);
    if (copyflag != INSERT && copyflag != COPY && copyflag != CLONE)
	return ERROR_INT("copyflag must be in {0,1,2}", procName, 1);

    if (copyflag == INSERT)
	ptac = pta;
    else if (copyflag == COPY) {
	if ((ptac = ptaCopy(pta)) == NULL)
	    return ERROR_INT("ptac not made", procName, 1);
    }
    else {  /* copyflag == CLONE */
	if ((ptac = ptaClone(pta)) == NULL)
	    return ERROR_INT("pta clone not made", procName, 1);
    }

    n = ptaaGetCount(ptaa);
    if (n >= ptaa->nalloc)
	ptaaExtendArray(ptaa);
    ptaa->pta[n] = ptac;
    ptaa->n++;

    return 0;
}


/*!
 *  ptaaExtendArray()
 *
 *      Input: ptaa
 *      Return: 0 if OK, 1 on error
 */
INT32
ptaaExtendArray(PTAA   *ptaa)
{

    PROCNAME("ptaaExtendArray");

    if (!ptaa)
	return ERROR_INT("ptaa not defined", procName, 1);

    if ((ptaa->pta = (PTA **)reallocNew((void **)&ptaa->pta,
                               4 * ptaa->nalloc,
			       8 * ptaa->nalloc)) == NULL)
	return ERROR_INT("new ptr array not returned", procName, 1);

    ptaa->nalloc = 2 * ptaa->nalloc;
    return 0;
}



/*---------------------------------------------------------------------*
 *                          Ptaa accessors                             *
 *---------------------------------------------------------------------*/
/*!
 *  ptaaGetCount()
 *
 *       Input:  ptaa
 *       Return: count, or 0 if no ptaa
 */
INT32
ptaaGetCount(PTAA  *ptaa)
{

    PROCNAME("ptaaGetCount");

    if (!ptaa)
	return ERROR_INT("ptaa not defined", procName, 0);

    return ptaa->n;
}


/*!
 *  ptaaGetPta()
 *
 *      Input:  ptaa
 *              index  (to the i-th pta)
 *              accessflag  (COPY or CLONE)
 *      Return: pta, or null on error
 */
PTA *
ptaaGetPta(PTAA   *ptaa,
           INT32   index,
	   INT32   accessflag)
{
    PROCNAME("ptaaGetPta");

    if (!ptaa)
	return (PTA *)ERROR_PTR("ptaa not defined", procName, NULL);
    if (index < 0 || index >= ptaa->n)
	return (PTA *)ERROR_PTR("index not valid", procName, NULL);
    if (accessflag != COPY && accessflag != CLONE)
	return (PTA *)ERROR_PTR("accessflag must be in {1,2}", procName, NULL);

    if (accessflag == COPY)
	return ptaCopy(ptaa->pta[index]);
    else  /* CLONE */
	return ptaClone(ptaa->pta[index]);
}



/*---------------------------------------------------------------------*
 *                         Pta serialized I/O                          *
 *---------------------------------------------------------------------*/
/*!
 *  ptaWriteStream()
 *
 *     Input:  stream
 *             pta
 *             type  (0 for float values; 1 for integer values)
 *     Return: 0 if OK; 1 on error
 */
INT32
ptaWriteStream(FILE   *fp,
               PTA    *pta,
	       INT32   type)
{
INT32     i, n, ix, iy;
FLOAT32   x, y;

    PROCNAME("ptaWriteStream");

    if (!fp)
	return ERROR_INT("stream not defined", procName, 1);
    if (!pta)
	return ERROR_INT("pta not defined", procName, 1);

    n = ptaGetCount(pta);
    if (type == 0)
	fprintf(fp, "Point array with %d float points\n", n);
    else  /* type == 1 */
	fprintf(fp, "Point array with %d integer points\n", n);
    for (i = 0; i < n; i++) {
	if (type == 0) {  /* data is float */
	    ptaGetPt(pta, i, &x, &y);
	    fprintf(fp, "   (%f, %f)\n", x, y);
	}
	else {   /* data is integer */
	    ptaGetIPt(pta, i, &ix, &iy);
	    fprintf(fp, "   (%d, %d)\n", ix, iy);
	}
    }

    return 0;
}


/*!
 *  ptaReadStream()
 *
 *     Input:  stream
 *     Return: pta, or null on error
 */
PTA *
ptaReadStream(FILE   *fp)
{
char      typestr[128];
INT32     i, n, ix, iy, type;
FLOAT32   x, y;
PTA      *pta;

    PROCNAME("ptaReadStream");

    if (!fp)
	return (PTA *)ERROR_PTR("stream not defined", procName, NULL);

    if (fscanf(fp, "Point array with %d %s points\n", &n, typestr) != 2)
	return (PTA *)ERROR_PTR("not a pta file", procName, NULL);
    if (strcmp(typestr, "float") == 0)
	type = 0;
    else  /* typestr is "integer" */
	type = 1;

    if ((pta = ptaCreate(n)) == NULL)
	return (PTA *)ERROR_PTR("pta not made", procName, NULL);
    for (i = 0; i < n; i++) {
	if (type == 0) {  /* data is float */
	    fscanf(fp, "   (%f, %f)\n", &x, &y);
	    ptaAddPt(pta, x, y);
	}
	else {   /* data is integer */
	    fscanf(fp, "   (%d, %d)\n", &ix, &iy);
	    ptaAddPt(pta, ix, iy);
	}
    }

    return pta;
}



/*---------------------------------------------------------------------*
 *                                In use                               *
 *---------------------------------------------------------------------*/
/*!
 *  pixFindCornerPixels()
 *
 *     Input: pixs (1 bpp)
 *     Return: pta, or null on error
 *     Action: finds the 4 corner-most pixels, as defined by
 *             search inward from each corner
 */
PTA *
pixFindCornerPixels(PIX   *pixs)
{
INT32     i, j, x, y, w, h, wpl, mindim, found;
UINT32   *data, *line;
PTA      *pta;

    PROCNAME("pixFindCornerPixels");

    if (!pixs)
	return (PTA *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 1)
	return (PTA *)ERROR_PTR("pixs not binary", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    mindim = MIN(w, h);
    data = pixGetData(pixs);
    wpl = pixGetWpl(pixs);

    if ((pta = ptaCreate(4)) == NULL)
	return (PTA *)ERROR_PTR("pta not made", procName, NULL);

    for (found = FALSE, i = 0; i < mindim; i++) {
	for (j = 0; j <= i; j++) {
	    y = i - j;
	    line = data + y * wpl;
	    if (GET_DATA_BIT(line, j)) {
		ptaAddPt(pta, j, y);
		found = TRUE;
		break;
	    }
	}
	if (found == TRUE)
	    break;
    }

    for (found = FALSE, i = 0; i < mindim; i++) {
	for (j = 0; j <= i; j++) {
	    y = i - j;
	    line = data + y * wpl;
	    x = w - 1 - j;
	    if (GET_DATA_BIT(line, x)) {
		ptaAddPt(pta, x, y);
		found = TRUE;
		break;
	    }
	}
	if (found == TRUE)
	    break;
    }

    for (found = FALSE, i = 0; i < mindim; i++) {
	for (j = 0; j <= i; j++) {
	    y = h - 1 - i + j;
	    line = data + y * wpl;
	    if (GET_DATA_BIT(line, j)) {
		ptaAddPt(pta, j, y);
		found = TRUE;
		break;
	    }
	}
	if (found == TRUE)
	    break;
    }

    for (found = FALSE, i = 0; i < mindim; i++) {
	for (j = 0; j <= i; j++) {
	    y = h - 1 - i + j;
	    line = data + y * wpl;
	    x = w - 1 - j;
	    if (GET_DATA_BIT(line, x)) {
		ptaAddPt(pta, x, y);
		found = TRUE;
		break;
	    }
	}
	if (found == TRUE)
	    break;
    }

    return pta;
}

		
/*!
 *  ptaContainsPt()
 *
 *     Input: pta
 *            x, y  (point)
 *     Return: 1 if contained, 0 otherwise or on error
 */
INT32
ptaContainsPt(PTA   *pta,
              INT32  x,
              INT32  y)
{
INT32     i, n, ix, iy;

    PROCNAME("ptaContainsPt");

    if (!pta)
	return ERROR_INT("pta not defined", procName, 0);

    n = ptaGetCount(pta);
    for (i = 0; i < n; i++) {
	ptaGetIPt(pta, i, &ix, &iy);
	if (x == ix && y == iy)
	    return 1;
    }
    return 0;
}


/*!
 *  ptaTestIntersection()
 *
 *     Input: pta1, pta2
 *     Return: bval  (1 if they have any elements in common; 
 *                   0 otherwise or on error)
 */
INT32
ptaTestIntersection(PTA   *pta1,
                    PTA   *pta2)
{
INT32     i, j, n1, n2, x1, y1, x2, y2;

    PROCNAME("ptaTestIntersection");

    if (!pta1)
	return ERROR_INT("pta1 not defined", procName, 0);
    if (!pta2)
	return ERROR_INT("pta2 not defined", procName, 0);

    n1 = ptaGetCount(pta1);
    n2 = ptaGetCount(pta2);
    for (i = 0; i < n1; i++) {
	ptaGetIPt(pta1, i, &x1, &y1);
	for (j = 0; j < n2; j++) {
	    ptaGetIPt(pta1, i, &x2, &y2);
	    if (x1 == x2 && y1 == y2)
		return 1;
	}
    }

    return 0;
}

