/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  rotateamlow.c
 *
 *         Grayscale rotation (area mapped)
 *
 *             Rotation about image center:
 *                  void        rotateAMLow()
 *
 *             Rotation about UL corner of image:
 *                  void        rotateAMCornerLow()
 *
 */

#include <stdio.h>
#include <string.h>
#include <math.h>   /* required for sin and tan */

#include "allheaders.h"


/*------------------------------------------------------------------*
 *                     Rotation about the center                    *
 *------------------------------------------------------------------*/
void
rotateAMLow(UINT32  *datad,
            INT32    w,
	    INT32    h,
	    INT32    wpld,
	    UINT32  *datas,
	    INT32    wpls,
	    FLOAT32  angle)
{
INT32         i, j, xcen, ycen, wm2, hm2;
INT32         xdif, ydif, xpm, ypm, xp, yp, xf, yf;
INT32         v00, v01, v10, v11;
UINT8         val;
UINT32       *lines, *lined;
FLOAT32       sina, cosa;

    xcen = w / 2;
    wm2 = w - 2;
    ycen = h / 2;
    hm2 = h - 2;
    sina = 16. * sin(angle);
    cosa = 16. * cos(angle);

    for (i = 0; i < h; i++) {
	ydif = ycen - i;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    xdif = xcen - j;
	    xpm = (INT32)(-xdif * cosa - ydif * sina + 0.5);
	    ypm = (INT32)(-ydif * cosa + xdif * sina + 0.5);
	    xp = xcen + (xpm >> 4);
	    yp = ycen + (ypm >> 4);
	    xf = xpm & 0x0f;
	    yf = ypm & 0x0f;

		/* if off the edge, just write white */
	    if (xp < 0 || yp < 0 || xp > wm2 || yp > hm2) {
		SET_DATA_BYTE(lined, j, 255);
		continue;
	    }

	    lines = datas + yp * wpls;

		/* do area weighting.  Without this, we would
		 * simply do:
	         *   SET_DATA_BYTE(lined, j, GET_DATA_BYTE(lines, xp));
		 * which is faster but gives lousy results!
		 */
	    v00 = (16 - xf) * (16 - yf) * GET_DATA_BYTE(lines, xp);
	    v10 = xf * (16 - yf) * GET_DATA_BYTE(lines, xp + 1);
	    v01 = (16 - xf) * yf * GET_DATA_BYTE(lines + wpls, xp);
	    v11 = xf * yf * GET_DATA_BYTE(lines + wpls, xp + 1);
	    val = (UINT8)((v00 + v01 + v10 + v11 + 128) / 256);
	    SET_DATA_BYTE(lined, j, val);
	}
    }

    return;
}



/*------------------------------------------------------------------*
 *                     Rotation about the corner                    *
 *------------------------------------------------------------------*/
void
rotateAMCornerLow(UINT32  *datad,
                  INT32    w,
		  INT32    h,
		  INT32    wpld,
		  UINT32  *datas,
		  INT32    wpls,
		  FLOAT32  angle)
{
INT32         i, j, wm2, hm2;
INT32         xpm, ypm, xp, yp, xf, yf;
INT32         v00, v01, v10, v11;
UINT8         val;
UINT32       *lines, *lined;
FLOAT32       sina, cosa;

    wm2 = w - 2;
    hm2 = h - 2;
    sina = 16. * sin(angle);
    cosa = 16. * cos(angle);

    for (i = 0; i < h; i++) {
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    xpm = (INT32)(j * cosa + i * sina + 0.5);
	    ypm = (INT32)(i * cosa - j * sina + 0.5);
	    xp = xpm >> 4;
	    yp = ypm >> 4;
	    xf = xpm & 0x0f;
	    yf = ypm & 0x0f;

		/* if off the edge, just write white */
	    if (xp < 0 || yp < 0 || xp > wm2 || yp > hm2) {
		SET_DATA_BYTE(lined, j, 255);
		continue;
	    }

	    lines = datas + yp * wpls;

		/* do area weighting.  Without this, we would
		 * simply do:
	         *   SET_DATA_BYTE(lined, j, GET_DATA_BYTE(lines, xp));
		 * which is faster but gives lousy results!
		 */
	    v00 = (16 - xf) * (16 - yf) * GET_DATA_BYTE(lines, xp);
	    v10 = xf * (16 - yf) * GET_DATA_BYTE(lines, xp + 1);
	    v01 = (16 - xf) * yf * GET_DATA_BYTE(lines + wpls, xp);
	    v11 = xf * yf * GET_DATA_BYTE(lines + wpls, xp + 1);
	    val = (UINT8)((v00 + v01 + v10 + v11 + 128) / 256);
	    SET_DATA_BYTE(lined, j, val);
	}
    }

    return;
}


