/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  scalelow.c
 *
 *         Grayscale (interpolated) scaling: general case
 *                  void       scaleGrayLILow()
 *
 *         Grayscale (interpolated) scaling: 2x upscaling
 *                  void       scaleGray2xLILow()
 *                  void       scaleGray2xLILineLow()
 *
 *         Grayscale (interpolated) scaling: 4x upscaling
 *                  void       scaleGray4xLILow()
 *                  void       scaleGray4xLILineLow()
 *
 *         Grayscale and color scaling by closest pixel sampling
 *                  INT32      scaleBySamplingLow()
 *
 *         Binary scaling by closest pixel sampling
 *                  INT32      scaleBinaryLow()
 *
 *         Scale-to-gray 3x
 *                  void       scaleToGray3Low()
 *                  INT32     *makeSumTabSG3()
 *                  UINT8     *makeValTabSG3()
 *
 *         Scale-to-gray 4x
 *                  void       scaleToGray4Low()
 *                  INT32     *makeSumTabSG4()
 *                  UINT8     *makeValTabSG4()
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"



/*------------------------------------------------------------------*
 *              General linear interpolated scaling                 *
 *------------------------------------------------------------------*/
/*
 *  scaleGrayLILow()
 *
 *     We choose to divide each pixel into 16 x 16 sub-pixels.
 *     Linear interpolation is equivalent to finding the 
 *     fractional area (i.e., number of sub-pixels divided
 *     by 256) associated with each of the four nearest src pixels,
 *     and weighting each pixel value by this fractional area.
 */
void
scaleGrayLILow(UINT32   *datad,
               INT32     wd,
               INT32     hd,
               INT32     wpld,
               UINT32   *datas,
	       INT32     ws,
               INT32     hs,
               INT32     wpls)
{
INT32         i, j, wm2, hm2;
INT32         xpm, ypm;  /* location in src image, to 1/16 of a pixel */
INT32         xp, yp, xf, yf;  /* src pixel and pixel fraction coordinates */
INT32         v00, v01, v10, v11;
UINT8         val;
UINT32       *lines, *lined;
FLOAT32       scx, scy;

	/* (scx, scy) are scaling factors that are applied to the
	 * dest coords to get the corresponding src coords.
	 * We need them because we iterate over dest pixels
	 * and must find the corresponding set of src pixels. */
    scx = 16. * (FLOAT32)ws / (FLOAT32)wd;
    scy = 16. * (FLOAT32)hs / (FLOAT32)hd;

    wm2 = ws - 2;
    hm2 = hs - 2;

	/* iterate over the destination pixels */
    for (i = 0; i < hd; i++) {
	ypm = (INT32)(scy * (FLOAT32)i + 0.5);
	yp = ypm >> 4;
	yf = ypm & 0x0f;
	lined = datad + i * wpld;
	lines = datas + yp * wpls;
	for (j = 0; j < wd; j++) {
	    xpm = (INT32)(scx * (FLOAT32)j + 0.5);
	    xp = xpm >> 4;
	    xf = xpm & 0x0f;

		/* if near the edge, just use the src pixel value */
	    if (xp > wm2 || yp > hm2) {
		SET_DATA_BYTE(lined, j, GET_DATA_BYTE(lines, xp));
		continue;
	    }

		/* do bilinear interpolation.  Without this, we could
		 * simply subsample:
	         *   SET_DATA_BYTE(lined, j, GET_DATA_BYTE(lines, xp));
		 * which is faster but gives lousy results!
		 */
	    v00 = (16 - xf) * (16 - yf) * GET_DATA_BYTE(lines, xp);
	    v10 = xf * (16 - yf) * GET_DATA_BYTE(lines, xp + 1);
	    v01 = (16 - xf) * yf * GET_DATA_BYTE(lines + wpls, xp);
	    v11 = xf * yf * GET_DATA_BYTE(lines + wpls, xp + 1);
	    val = (UINT8)((v00 + v01 + v10 + v11 + 128) / 256);
	    SET_DATA_BYTE(lined, j, val);
	}
    }

    return;
}


/*------------------------------------------------------------------*
 *                 2x linear interpolated scaling                   *
 *------------------------------------------------------------------*/
/*
 *  scaleGray2xLILow()
 *
 *     This is a special case of 2x expansion by linear
 *     interpolation.  Each src pixel contains 4 dest pixels.
 *     The 4 dest pixels in src pixel 1 are numbered at
 *     their UL corners.  The 4 dest pixels in src pixel 1
 *     are related to that src pixel and its 3 neighboring
 *     src pixels as follows:
 *
 *             1-----2-----|-----|-----|
 *             |     |     |     |     |
 *             |     |     |     |     |
 *  src 1 -->  3-----4-----|     |     |  <-- src 2
 *             |     |     |     |     |
 *             |     |     |     |     |
 *             |-----|-----|-----|-----|
 *             |     |     |     |     |
 *             |     |     |     |     |
 *  src 3 -->  |     |     |     |     |  <-- src 4
 *             |     |     |     |     |
 *             |     |     |     |     |
 *             |-----|-----|-----|-----|
 *
 *           dest      src
 *           ----      ---
 *           dp1    =  sp1
 *           dp2    =  (sp1 + sp2) / 2
 *           dp3    =  (sp1 + sp3) / 2
 *           dp4    =  (sp1 + sp2 + sp3 + sp4) / 4
 *
 *     We iterate over the src pixels, and unroll the calculation
 *     for each set of 4 dest pixels corresponding to that src
 *     pixel, caching pixels for the next src pixel whenever possible.
 */
void
scaleGray2xLILow(UINT32   *datad,
                 INT32     wpld,
                 UINT32   *datas,
                 INT32     ws,
                 INT32     hs,
                 INT32     wpls)
{
INT32         i, hsm;
UINT32       *lines, *lined;

    hsm = hs - 1;

	/* We're taking 2 src and 2 dest lines at a time,
	 * and for each src line, we're computing 2 dest lines.
	 * Call these 2 dest lines:  destline1 and destline2.
	 * The first src line is used for destline 1.
	 * On all but the last src line, both src lines are 
	 * used in the linear interpolation for destline2.
	 * On the last src line, both destline1 and destline2
	 * are computed using only that src line (because there
	 * isn't a lower src line). */

	/* iterate over all but the last src line */
    for (i = 0; i < hsm; i++) {
	lines = datas + i * wpls;
	lined = datad + 2 * i * wpld;
	scaleGray2xLILineLow(lined, wpld, lines, ws, wpls, 0);
    }
    
        /* last src line */
    lines = datas + hsm * wpls;
    lined = datad + 2 * hsm * wpld;
    scaleGray2xLILineLow(lined, wpld, lines, ws, wpls, 1);

    return;
}


/* 
 *  scaleGray2xLILineLow()
 *
 *      Input:  lined   (ptr to top destline, to be made from current src line)
 *              wpld
 *              lines   (ptr to current src line)
 *              ws
 *              wpls
 *              lastlineflag  (1 if last src line; 0 otherwise)
 *      Return: void
 *
 */
void
scaleGray2xLILineLow(UINT32   *lined,
                     INT32     wpld,
		     UINT32   *lines,
		     INT32     ws,
		     INT32     wpls,
		     INT32     lastlineflag)
{
INT32         j, jd, wsm;
INT32         sval1, sval2, sval3, sval4;
UINT32       *linesp, *linedp;

    wsm = ws - 1;

    if (lastlineflag == 0) {
	linesp = lines + wpls;
	linedp = lined + wpld;
	sval2 = GET_DATA_BYTE(lines, 0);
	sval4 = GET_DATA_BYTE(linesp, 0);
	for (j = 0, jd = 0; j < wsm; j++, jd += 2) {
	    sval1 = sval2;
	    sval3 = sval4;
	    sval2 = GET_DATA_BYTE(lines, j + 1);
	    sval4 = GET_DATA_BYTE(linesp, j + 1);
	    SET_DATA_BYTE(lined, jd, sval1);                     /* pix 1 */
	    SET_DATA_BYTE(lined, jd + 1, (sval1 + sval2) / 2);   /* pix 2 */
	    SET_DATA_BYTE(linedp, jd, (sval1 + sval3) / 2);      /* pix 3 */
	    SET_DATA_BYTE(linedp, jd + 1,
	                  (sval1 + sval2 + sval3 + sval4) / 4);  /* pix 4 */
        }  
	sval1 = sval2;
	sval3 = sval4;
	SET_DATA_BYTE(lined, 2 * wsm, sval1);                     /* pix 1 */
	SET_DATA_BYTE(lined, 2 * wsm + 1, sval1);                 /* pix 2 */
	SET_DATA_BYTE(linedp, 2 * wsm, (sval1 + sval3) / 2);      /* pix 3 */
	SET_DATA_BYTE(linedp, 2 * wsm + 1, (sval1 + sval3) / 2);  /* pix 4 */
    }
    else {   /* last row of src pixels: lastlineflag == 1 */
	linedp = lined + wpld;
	sval2 = GET_DATA_BYTE(lines, 0);
	for (j = 0, jd = 0; j < wsm; j++, jd += 2) {
	    sval1 = sval2;
	    sval2 = GET_DATA_BYTE(lines, j + 1);
	    SET_DATA_BYTE(lined, jd, sval1);                       /* pix 1 */
	    SET_DATA_BYTE(linedp, jd, sval1);                      /* pix 3 */
	    SET_DATA_BYTE(lined, jd + 1, (sval1 + sval2) / 2);     /* pix 2 */
	    SET_DATA_BYTE(linedp, jd + 1, (sval1 + sval2) / 2);    /* pix 4 */
	}  
	sval1 = sval2;
	SET_DATA_BYTE(lined, 2 * wsm, sval1);                     /* pix 1 */
	SET_DATA_BYTE(lined, 2 * wsm + 1, sval1);                 /* pix 2 */
	SET_DATA_BYTE(linedp, 2 * wsm, sval1);                    /* pix 3 */
	SET_DATA_BYTE(linedp, 2 * wsm + 1, sval1);                /* pix 4 */
    }
	
    return;
}


/*------------------------------------------------------------------*
 *                 4x linear interpolated scaling                   *
 *------------------------------------------------------------------*/
/*
 *  scaleGray4xLILow()
 *
 *     This is a special case of 4x expansion by linear
 *     interpolation.  Each src pixel contains 16 dest pixels.
 *     The 16 dest pixels in src pixel 1 are numbered at
 *     their UL corners.  The 16 dest pixels in src pixel 1
 *     are related to that src pixel and its 3 neighboring
 *     src pixels as follows:
 *
 *             1---2---3---4---|---|---|---|---|
 *             |   |   |   |   |   |   |   |   |
 *             5---6---7---8---|---|---|---|---|
 *             |   |   |   |   |   |   |   |   |
 *  src 1 -->  9---a---b---c---|---|---|---|---|  <-- src 2
 *             |   |   |   |   |   |   |   |   |
 *             d---e---f---g---|---|---|---|---|
 *             |   |   |   |   |   |   |   |   |
 *             |===|===|===|===|===|===|===|===|
 *             |   |   |   |   |   |   |   |   |
 *             |---|---|---|---|---|---|---|---|
 *             |   |   |   |   |   |   |   |   |
 *  src 3 -->  |---|---|---|---|---|---|---|---|  <-- src 4
 *             |   |   |   |   |   |   |   |   |
 *             |---|---|---|---|---|---|---|---|
 *             |   |   |   |   |   |   |   |   |
 *             |---|---|---|---|---|---|---|---|
 *
 *           dest      src
 *           ----      ---
 *           dp1    =  sp1
 *           dp2    =  (3 * sp1 + sp2) / 4
 *           dp3    =  (sp1 + sp2) / 2
 *           dp4    =  (sp1 + 3 * sp2) / 4
 *           dp5    =  (3 * sp1 + sp3) / 4
 *           dp6    =  (9 * sp1 + 3 * sp2 + 3 * sp3 + sp4) / 16 
 *           dp7    =  (3 * sp1 + 3 * sp2 + sp3 + sp4) / 8
 *           dp8    =  (3 * sp1 + 9 * sp2 + 1 * sp3 + 3 * sp4) / 16 
 *           dp9    =  (sp1 + sp3) / 2
 *           dp10   =  (3 * sp1 + sp2 + 3 * sp3 + sp4) / 8
 *           dp11   =  (sp1 + sp2 + sp3 + sp4) / 4 
 *           dp12   =  (sp1 + 3 * sp2 + sp3 + 3 * sp4) / 8
 *           dp13   =  (sp1 + 3 * sp3) / 4
 *           dp14   =  (3 * sp1 + sp2 + 9 * sp3 + 3 * sp4) / 16 
 *           dp15   =  (sp1 + sp2 + 3 * sp3 + 3 * sp4) / 8
 *           dp16   =  (sp1 + 3 * sp2 + 3 * sp3 + 9 * sp4) / 16 
 *
 *     We iterate over the src pixels, and unroll the calculation
 *     for each set of 16 dest pixels corresponding to that src
 *     pixel, caching pixels for the next src pixel whenever possible.
 */
void
scaleGray4xLILow(UINT32   *datad,
                 INT32     wpld,
                 UINT32   *datas,
                 INT32     ws,
                 INT32     hs,
                 INT32     wpls)
{
INT32         i, hsm;
UINT32       *lines, *lined;

    hsm = hs - 1;

	/* We're taking 2 src and 4 dest lines at a time,
	 * and for each src line, we're computing 4 dest lines.
	 * Call these 4 dest lines:  destline1 - destline4.
	 * The first src line is used for destline 1.
	 * Two src lines are used for all other dest lines,
	 * except for the last 4 dest lines, which are computed
	 * using only the last src line.

	/* iterate over all but the last src line */
    for (i = 0; i < hsm; i++) {
	lines = datas + i * wpls;
	lined = datad + 4 * i * wpld;
	scaleGray4xLILineLow(lined, wpld, lines, ws, wpls, 0);
    }
    
        /* last src line */
    lines = datas + hsm * wpls;
    lined = datad + 4 * hsm * wpld;
    scaleGray4xLILineLow(lined, wpld, lines, ws, wpls, 1);

    return;
}


/* 
 *  scaleGray4xLILineLow()
 *
 *      Input:  lined   (ptr to top destline, to be made from current src line)
 *              wpld
 *              lines   (ptr to current src line)
 *              ws
 *              wpls
 *              lastlineflag  (1 if last src line; 0 otherwise)
 *      Return: void
 *
 */
void
scaleGray4xLILineLow(UINT32   *lined,
                     INT32     wpld,
		     UINT32   *lines,
		     INT32     ws,
		     INT32     wpls,
		     INT32     lastlineflag)
{
INT32         j, jd, wsm, wsm4;
INT32         s1, s2, s3, s4, s1t, s2t, s3t, s4t;
UINT32       *linesp, *linedp1, *linedp2, *linedp3;

    wsm = ws - 1;
    wsm4 = 4 * wsm;

    if (lastlineflag == 0) {
	linesp = lines + wpls;
	linedp1 = lined + wpld;
	linedp2 = lined + 2 * wpld;
	linedp3 = lined + 3 * wpld;
	s2 = GET_DATA_BYTE(lines, 0);
	s4 = GET_DATA_BYTE(linesp, 0);
	for (j = 0, jd = 0; j < wsm; j++, jd += 4) {
	    s1 = s2;
	    s3 = s4;
	    s2 = GET_DATA_BYTE(lines, j + 1);
	    s4 = GET_DATA_BYTE(linesp, j + 1);
	    s1t = 3 * s1;
	    s2t = 3 * s2;
	    s3t = 3 * s3;
	    s4t = 3 * s4;
	    SET_DATA_BYTE(lined, jd, s1);                             /* d1 */
	    SET_DATA_BYTE(lined, jd + 1, (s1t + s2) / 4);             /* d2 */
	    SET_DATA_BYTE(lined, jd + 2, (s1 + s2) / 2);              /* d3 */
	    SET_DATA_BYTE(lined, jd + 3, (s1 + s2t) / 4);             /* d4 */
	    SET_DATA_BYTE(linedp1, jd, (s1t + s3) / 4);                /* d5 */
	    SET_DATA_BYTE(linedp1, jd + 1, (9*s1 + s2t + s3t + s4) / 16); /*d6*/
	    SET_DATA_BYTE(linedp1, jd + 2, (s1t + s2t + s3 + s4) / 8); /* d7 */
	    SET_DATA_BYTE(linedp1, jd + 3, (s1t + 9*s2 + s3 + s4t) / 16);/*d8*/
	    SET_DATA_BYTE(linedp2, jd, (s1 + s3) / 2);                /* d9 */
	    SET_DATA_BYTE(linedp2, jd + 1, (s1t + s2 + s3t + s4) / 8);/* d10 */
	    SET_DATA_BYTE(linedp2, jd + 2, (s1 + s2 + s3 + s4) / 4);  /* d11 */
	    SET_DATA_BYTE(linedp2, jd + 3, (s1 + s2t + s3 + s4t) / 8);/* d12 */
	    SET_DATA_BYTE(linedp3, jd, (s1 + s3t) / 4);               /* d13 */
	    SET_DATA_BYTE(linedp3, jd + 1, (s1t + s2 + 9*s3 + s4t) / 16);/*d14*/
	    SET_DATA_BYTE(linedp3, jd + 2, (s1 + s2 + s3t + s4t) / 8); /* d15 */
	    SET_DATA_BYTE(linedp3, jd + 3, (s1 + s2t + s3t + 9*s4) / 16);/*d16*/
        }  
	s1 = s2;
	s3 = s4;
	s1t = 3 * s1;
	s3t = 3 * s3;
	SET_DATA_BYTE(lined, wsm4, s1);                               /* d1 */
	SET_DATA_BYTE(lined, wsm4 + 1, s1);                           /* d2 */
	SET_DATA_BYTE(lined, wsm4 + 2, s1);                           /* d3 */
	SET_DATA_BYTE(lined, wsm4 + 3, s1);                           /* d4 */
	SET_DATA_BYTE(linedp1, wsm4, (s1t + s3) / 4);                 /* d5 */
	SET_DATA_BYTE(linedp1, wsm4 + 1, (s1t + s3) / 4);             /* d6 */
	SET_DATA_BYTE(linedp1, wsm4 + 2, (s1t + s3) / 4);             /* d7 */
	SET_DATA_BYTE(linedp1, wsm4 + 3, (s1t + s3) / 4);             /* d8 */
	SET_DATA_BYTE(linedp2, wsm4, (s1 + s3) / 2);                  /* d9 */
	SET_DATA_BYTE(linedp2, wsm4 + 1, (s1 + s3) / 2);              /* d10 */
	SET_DATA_BYTE(linedp2, wsm4 + 2, (s1 + s3) / 2);              /* d11 */
	SET_DATA_BYTE(linedp2, wsm4 + 3, (s1 + s3) / 2);              /* d12 */
	SET_DATA_BYTE(linedp3, wsm4, (s1 + s3t) / 4);                 /* d13 */
	SET_DATA_BYTE(linedp3, wsm4 + 1, (s1 + s3t) / 4);             /* d14 */
	SET_DATA_BYTE(linedp3, wsm4 + 2, (s1 + s3t) / 4);             /* d15 */
	SET_DATA_BYTE(linedp3, wsm4 + 3, (s1 + s3t) / 4);             /* d16 */
    }
    else {   /* last row of src pixels: lastlineflag == 1 */
	linedp1 = lined + wpld;
	linedp2 = lined + 2 * wpld;
	linedp3 = lined + 3 * wpld;
	s2 = GET_DATA_BYTE(lines, 0);
	for (j = 0, jd = 0; j < wsm; j++, jd += 4) {
	    s1 = s2;
	    s2 = GET_DATA_BYTE(lines, j + 1);
	    s1t = 3 * s1;
	    s2t = 3 * s2;
	    SET_DATA_BYTE(lined, jd, s1);                            /* d1 */
	    SET_DATA_BYTE(lined, jd + 1, (s1t + s2) / 4 );           /* d2 */
	    SET_DATA_BYTE(lined, jd + 2, (s1 + s2) / 2 );            /* d3 */
	    SET_DATA_BYTE(lined, jd + 3, (s1 + s2t) / 4 );           /* d4 */
	    SET_DATA_BYTE(linedp1, jd, s1);                          /* d5 */
	    SET_DATA_BYTE(linedp1, jd + 1, (s1t + s2) / 4 );         /* d6 */
	    SET_DATA_BYTE(linedp1, jd + 2, (s1 + s2) / 2 );          /* d7 */
	    SET_DATA_BYTE(linedp1, jd + 3, (s1 + s2t) / 4 );         /* d8 */
	    SET_DATA_BYTE(linedp2, jd, s1);                          /* d9 */
	    SET_DATA_BYTE(linedp2, jd + 1, (s1t + s2) / 4 );         /* d10 */
	    SET_DATA_BYTE(linedp2, jd + 2, (s1 + s2) / 2 );          /* d11 */
	    SET_DATA_BYTE(linedp2, jd + 3, (s1 + s2t) / 4 );         /* d12 */
	    SET_DATA_BYTE(linedp3, jd, s1);                          /* d13 */
	    SET_DATA_BYTE(linedp3, jd + 1, (s1t + s2) / 4 );         /* d14 */
	    SET_DATA_BYTE(linedp3, jd + 2, (s1 + s2) / 2 );          /* d15 */
	    SET_DATA_BYTE(linedp3, jd + 3, (s1 + s2t) / 4 );         /* d16 */
	}  
	s1 = s2;
	SET_DATA_BYTE(lined, wsm4, s1);                              /* d1 */
	SET_DATA_BYTE(lined, wsm4 + 1, s1);                          /* d2 */
	SET_DATA_BYTE(lined, wsm4 + 2, s1);                          /* d3 */
	SET_DATA_BYTE(lined, wsm4 + 3, s1);                          /* d4 */
	SET_DATA_BYTE(linedp1, wsm4, s1);                            /* d5 */
	SET_DATA_BYTE(linedp1, wsm4 + 1, s1);                        /* d6 */
	SET_DATA_BYTE(linedp1, wsm4 + 2, s1);                        /* d7 */
	SET_DATA_BYTE(linedp1, wsm4 + 3, s1);                        /* d8 */
	SET_DATA_BYTE(linedp2, wsm4, s1);                            /* d9 */
	SET_DATA_BYTE(linedp2, wsm4 + 1, s1);                        /* d10 */
	SET_DATA_BYTE(linedp2, wsm4 + 2, s1);                        /* d11 */
	SET_DATA_BYTE(linedp2, wsm4 + 3, s1);                        /* d12 */
	SET_DATA_BYTE(linedp3, wsm4, s1);                            /* d13 */
	SET_DATA_BYTE(linedp3, wsm4 + 1, s1);                        /* d14 */
	SET_DATA_BYTE(linedp3, wsm4 + 2, s1);                        /* d15 */
	SET_DATA_BYTE(linedp3, wsm4 + 3, s1);                        /* d16 */
    }
	
    return;
}



/*------------------------------------------------------------------*
 *       Grayscale and color scaling by closest pixel sampling      *
 *------------------------------------------------------------------*/
/*
 *  scaleBySamplingLow()
 *
 *    Note: the dest must be cleared prior to this operation,
 *          and we clear it here in the low-level code.
 *
 *    Note: we reuse dest pixels and dest pixel rows whenever
 *          possible.  This speeds the upscaling; downscaling
 *          is done by strict subsampling and is unaffected.
 *
 *    Note: because we are sampling and not interpolating, this
 *          routine works directly, without conversion to full
 *          RGB color, for 2, 4 or 8 bpp palette color images.
 */
INT32
scaleBySamplingLow(UINT32   *datad,
                   INT32     wd,
                   INT32     hd,
                   INT32     wpld,
                   UINT32   *datas,
	           INT32     ws,
                   INT32     hs,
                   INT32     d,
                   INT32     wpls)
{
INT32         i, j, bpld;
INT32         xs, prevxs, sval;
INT32        *srow, *scol;
UINT32        csval;
UINT32       *lines, *prevlines, *lined, *prevlined;
FLOAT32       wratio, hratio;

    PROCNAME("scaleBySamplingLow");

	/* clear dest */
    bpld = 4 * wpld;
    memset((char *)datad, 0, hd * bpld);
    
	/* the source row corresponding to dest row i ==> srow[i]
	 * the source col corresponding to dest col j ==> scol[j]  */
    if ((srow = (INT32 *)CALLOC(hd, sizeof(INT32))) == NULL)
	return ERROR_INT("srow not made", procName, 1);
    if ((scol = (INT32 *)CALLOC(wd, sizeof(INT32))) == NULL)
	return ERROR_INT("scol not made", procName, 1);

    wratio = (FLOAT32)ws / (FLOAT32)wd;
    hratio = (FLOAT32)hs / (FLOAT32)hd;
    for (i = 0; i < hd; i++)
	srow[i] = MIN((INT32)(hratio * i + 0.5), hs - 1);
    for (j = 0; j < wd; j++)
	scol[j] = MIN((INT32)(wratio * j + 0.5), ws - 1);

    prevlines = NULL;
    for (i = 0; i < hd; i++) {
	lines = datas + srow[i] * wpls;
	lined = datad + i * wpld;
	if (lines != prevlines) {  /* make dest from new source row */
	    prevxs = -1;
	    sval = 0;
	    switch (d)
	    {
	    case 2:
		for (j = 0; j < wd; j++) {
		    xs = scol[j];
		    if (xs != prevxs) {  /* get dest pix from source col */
			sval = GET_DATA_DIBIT(lines, xs);
			SET_DATA_DIBIT(lined, j, sval);
			prevxs = xs;
		    }
		    else   /* copy prev dest pix */
			SET_DATA_DIBIT(lined, j, sval);
		}
		break;
	    case 4:
		for (j = 0; j < wd; j++) {
		    xs = scol[j];
		    if (xs != prevxs) {  /* get dest pix from source col */
			sval = GET_DATA_QBIT(lines, xs);
			SET_DATA_QBIT(lined, j, sval);
			prevxs = xs;
		    }
		    else   /* copy prev dest pix */
			SET_DATA_QBIT(lined, j, sval);
		}
		break;
	    case 8:
		for (j = 0; j < wd; j++) {
		    xs = scol[j];
		    if (xs != prevxs) {  /* get dest pix from source col */
			sval = GET_DATA_BYTE(lines, xs);
			SET_DATA_BYTE(lined, j, sval);
			prevxs = xs;
		    }
		    else   /* copy prev dest pix */
			SET_DATA_BYTE(lined, j, sval);
		}
		break;
	    case 16:
		for (j = 0; j < wd; j++) {
		    xs = scol[j];
		    if (xs != prevxs) {  /* get dest pix from source col */
			sval = GET_DATA_TWO_BYTES(lines, xs);
			SET_DATA_TWO_BYTES(lined, j, sval);
			prevxs = xs;
		    }
		    else   /* copy prev dest pix */
			SET_DATA_TWO_BYTES(lined, j, sval);
		}
		break;
	    case 32:
		for (j = 0; j < wd; j++) {
		    xs = scol[j];
		    if (xs != prevxs) {  /* get dest pix from source col */
			csval = lines[xs];
			lined[j] = csval;
			prevxs = xs;
		    }
		    else   /* copy prev dest pix */
			lined[j] = csval;
		}
		break;
	    default:
		return ERROR_INT("pixel depth not supported", procName, 1);
		break;
	    }
	}
	else {  /* lines == prevlines; copy prev dest row */
	    prevlined = lined - wpld;
	    memcpy((char *)lined, (char *)prevlined, bpld);
	}
	prevlines = lines;
    }

    FREE((void *)srow);
    FREE((void *)scol);

    return 0;
}



/*------------------------------------------------------------------*
 *              Binary scaling by closest pixel sampling            *
 *------------------------------------------------------------------*/
/*
 *  scaleBinaryLow()
 *
 *    Note: the dest must be cleared prior to this operation,
 *          and we clear it here in the low-level code.
 *
 *    Note: we reuse dest pixels and dest pixel rows whenever
 *          possible.  This speeds the upscaling; downscaling
 *          is done by strict subsampling and is unaffected.
 *
 */
INT32
scaleBinaryLow(UINT32   *datad,
               INT32     wd,
               INT32     hd,
               INT32     wpld,
               UINT32   *datas,
	       INT32     ws,
               INT32     hs,
               INT32     wpls)
{
INT32         i, j, bpld;
INT32         xs, prevxs, sval;
INT32        *srow, *scol;
UINT32       *lines, *prevlines, *lined, *prevlined;
FLOAT32       wratio, hratio;

    PROCNAME("scaleBinaryLow");

	/* clear dest */
    bpld = 4 * wpld;
    memset((char *)datad, 0, hd * bpld);
    
	/* the source row corresponding to dest row i ==> srow[i]
	 * the source col corresponding to dest col j ==> scol[j]  */
    if ((srow = (INT32 *)CALLOC(hd, sizeof(INT32))) == NULL)
	return ERROR_INT("srow not made", procName, 1);
    if ((scol = (INT32 *)CALLOC(wd, sizeof(INT32))) == NULL)
	return ERROR_INT("scol not made", procName, 1);

    wratio = (FLOAT32)ws / (FLOAT32)wd;
    hratio = (FLOAT32)hs / (FLOAT32)hd;
    for (i = 0; i < hd; i++)
	srow[i] = MIN((INT32)(hratio * i + 0.5), hs - 1);
    for (j = 0; j < wd; j++)
	scol[j] = MIN((INT32)(wratio * j + 0.5), ws - 1);

    prevlines = NULL;
    prevxs = -1;
    sval = 0;
    for (i = 0; i < hd; i++) {
	lines = datas + srow[i] * wpls;
	lined = datad + i * wpld;
	if (lines != prevlines) {  /* make dest from new source row */
	    for (j = 0; j < wd; j++) {
		xs = scol[j];
		if (xs != prevxs) {  /* get dest pix from source col */
		    if (sval = GET_DATA_BIT(lines, xs))
			SET_DATA_BIT(lined, j);
		    prevxs = xs;
		}
		else {  /* copy prev dest pix, if set */
		    if (sval)
			SET_DATA_BIT(lined, j);
		}
	    }
	}
	else {  /* lines == prevlines; copy prev dest row */
	    prevlined = lined - wpld;
	    memcpy((char *)lined, (char *)prevlined, bpld);
	}
	prevlines = lines;
    }

    FREE((void *)srow);
    FREE((void *)scol);

    return 0;
}



/*------------------------------------------------------------------*
 *                         Scale-to-gray 3x                         *
 *------------------------------------------------------------------*/
/*
 *  scaleToGray3Low()
 *
 *       Input:  usual image variables
 *               sumtab  (made from makeSumTabSG3())
 *               valtab  (made from makeValTabSG3())
 *       Return: 0 if OK; 1 on error
 *
 *       Each set of 8 3x3 bit-blocks in the source image, which
 *       consist of 72 pixels arranged 24 pixels wide by 3 scanlines,
 *       is converted to a row of 8 8-bit pixels in the dest image.
 *       These 72 pixels of the input image are runs of 24 pixels
 *       in three adjacent scanlines.  Each run of 24 pixels is
 *       stored in the 24 LSbits of a 32-bit word.  We use 2 LUTs.
 *       The first, sumtab, takes 6 of these bits and stores
 *       sum, taken 3 bits at a time, in two bytes.  (See
 *       makeSumTabSG3).  This is done for each of the 3 scanlines,
 *       and the results are added.  We now have the sum of ON pixels
 *       in the first two 3x3 blocks in two bytes.  The valtab LUT
 *       then converts these values (which go from 0 to 9) to
 *       grayscale values between between 255 and 0.  (See makeValTabSG3).
 *       This process is repeated for each of the other 3 sets of
 *       6x3 input pixels, giving 8 output pixels in total.
 *
 *       Note: because the input image is processed in groups of
 *       24 x 3 pixels, the process clips the input height to
 *       (h - h % 3) and the input width to (w - w % 24).
 *
 */
void
scaleToGray3Low(UINT32    *datad,
		INT32      wd,
		INT32      hd,
		INT32      wpld,
                UINT32    *datas,
		INT32      wpls,
		UINT32    *sumtab,
		UINT8     *valtab)
{
INT32         i, j, l, k;
UINT32        threebytes1, threebytes2, threebytes3, sum;
UINT32       *lines, *lined;

	/* i indexes the dest lines
	 * l indexes the source lines
	 * j indexes the dest bytes
	 * k indexes the source bytes
	 * We take 9 bytes from the source (72 binary pixels
	 * in three lines of 24 pixels each) and convert it
	 * into 8 bytes of the dest (8 8bpp pixels in one line)   */
    for (i = 0, l = 0; i < hd; i++, l += 3) {
	lines = datas + l * wpls;
	lined = datad + i * wpld; 
	for (j = 0, k = 0; j < wd; j += 8, k += 3) {
	    threebytes1 = (GET_DATA_BYTE(lines, k) << 16) |
			  (GET_DATA_BYTE(lines, k + 1) << 8) |
			  GET_DATA_BYTE(lines, k + 2);
	    threebytes2 = (GET_DATA_BYTE(lines + wpls, k) << 16) |
			  (GET_DATA_BYTE(lines + wpls, k + 1) << 8) |
			  GET_DATA_BYTE(lines + wpls, k + 2);
	    threebytes3 = (GET_DATA_BYTE(lines + 2 * wpls, k) << 16) |
			  (GET_DATA_BYTE(lines + 2 * wpls, k + 1) << 8) |
			  GET_DATA_BYTE(lines + 2 * wpls, k + 2);

	    sum = sumtab[(threebytes1 >> 18)] +
		  sumtab[(threebytes2 >> 18)] +
		  sumtab[(threebytes3 >> 18)];
	    SET_DATA_BYTE(lined, j, valtab[GET_DATA_BYTE(&sum, 2)]);
	    SET_DATA_BYTE(lined, j + 1, valtab[GET_DATA_BYTE(&sum, 3)]);

	    sum = sumtab[((threebytes1 >> 12) & 0x3f)] +
		  sumtab[((threebytes2 >> 12) & 0x3f)] +
		  sumtab[((threebytes3 >> 12) & 0x3f)];
	    SET_DATA_BYTE(lined, j + 2, valtab[GET_DATA_BYTE(&sum, 2)]);
	    SET_DATA_BYTE(lined, j + 3, valtab[GET_DATA_BYTE(&sum, 3)]);

	    sum = sumtab[((threebytes1 >> 6) & 0x3f)] +
		  sumtab[((threebytes2 >> 6) & 0x3f)] +
		  sumtab[((threebytes3 >> 6) & 0x3f)];
	    SET_DATA_BYTE(lined, j + 4, valtab[GET_DATA_BYTE(&sum, 2)]);
	    SET_DATA_BYTE(lined, j + 5, valtab[GET_DATA_BYTE(&sum, 3)]);

	    sum = sumtab[(threebytes1 & 0x3f)] +
		  sumtab[(threebytes2 & 0x3f)] +
		  sumtab[(threebytes3 & 0x3f)];
	    SET_DATA_BYTE(lined, j + 6, valtab[GET_DATA_BYTE(&sum, 2)]);
	    SET_DATA_BYTE(lined, j + 7, valtab[GET_DATA_BYTE(&sum, 3)]);
	}
    }

    return;
}



/*
 *   makeSumTabSG3()
 *         
 *       Returns a table of 64 UINT32s, giving the two output
 *       8-bit grayscale sums corresponding to 6 input bits of a binary
 *       image, for a 3x scale-to-gray op.  In practice, this would
 *       be used three times (on adjacent scanlines), and the sums would
 *       be added and then transformed to output 8 bpp pixel values,
 *       using makeValTabSG3().
 */
UINT32  *
makeSumTabSG3(void)
{
INT32    i;
INT32    sum[] = {0, 1, 1, 2, 1, 2, 2, 3};
UINT32  *tab;

    PROCNAME("makeSumTabSG3");

    if ((tab = (UINT32 *)CALLOC(64, sizeof(UINT32))) == NULL)
	return (UINT32 *)ERROR_PTR("calloc fail for tab", procName, NULL);

	/* pack the two sums separately in two bytes */
    for (i = 0; i < 64; i++) {
	tab[i] = (sum[i & 0x07]) | (sum[(i >> 3) & 0x07] << 8);
    }

    return tab;
}


/*
 *   makeValTabSG3()
 *         
 *       Returns an 8 bit value for the sum of ON pixels
 *       in a 3x3 square, according to
 *
 *              val = 255 - (255 * sum)/9
 *     
 *       where sum is in set {0, ... ,9}
 */
UINT8 *
makeValTabSG3(void)
{
INT32    i;
UINT8   *tab;

    PROCNAME("makeValTabSG3");

    if ((tab = (UINT8 *)CALLOC(10, sizeof(UINT8))) == NULL)
	return (UINT8 *)ERROR_PTR("calloc fail for tab", procName, NULL);

    for (i = 0; i < 10; i++)
	tab[i] = 0xff - (i * 255) / 9;

    return tab;
}



/*------------------------------------------------------------------*
 *                         Scale-to-gray 4x                         *
 *------------------------------------------------------------------*/
/*
 *  scaleToGray4Low()
 *
 *       Input:  usual image variables
 *               sumtab  (made from makeSumTabSG4())
 *               valtab  (made from makeValTabSG4())
 *       Return: 0 if OK; 1 on error.
 *
 *       The output is processed in sets of 2 output bytes on a row,
 *       corresponding to 2 4x4 bit-blocks in the input image.
 *       Two lookup tables are used.  The first, sumtab, gets the
 *       sum of ON pixels in two sets of four adjacent bits,
 *       storing the result in 2 adjacent bytes.  After sums from
 *       four rows have been added, the second table, valtab,
 *       converts from the sum of ON pixels in the 4x4 block to
 *       an 8 bpp grayscale value between 0 (for 16 bits ON)
 *       and 255 (for 0 bits ON).
 */
void
scaleToGray4Low(UINT32    *datad,
		INT32      wd,
		INT32      hd,
		INT32      wpld,
                UINT32    *datas,
		INT32      wpls,
		UINT32    *sumtab,
		UINT8     *valtab)
{
INT32         i, j, l, k;
UINT32        sbyte1, sbyte2, sbyte3, sbyte4, sum;
UINT32       *lines, *lined;

	/* i indexes the dest lines
	 * l indexes the source lines
	 * j indexes the dest bytes
	 * k indexes the source bytes
	 * We take four bytes from the source (in 4 lines of 8 pixels
	 * each) and convert it into two 8 bpp bytes of the dest. */
    for (i = 0, l = 0; i < hd; i++, l += 4) {
	lines = datas + l * wpls;
	lined = datad + i * wpld; 
	for (j = 0, k = 0; j < wd; j += 2, k++) {
	    sbyte1 = GET_DATA_BYTE(lines, k);
	    sbyte2 = GET_DATA_BYTE(lines + wpls, k);
	    sbyte3 = GET_DATA_BYTE(lines + 2 * wpls, k);
	    sbyte4 = GET_DATA_BYTE(lines + 3 * wpls, k);
	    sum = sumtab[sbyte1] + sumtab[sbyte2] +
	          sumtab[sbyte3] + sumtab[sbyte4];
	    SET_DATA_BYTE(lined, j, valtab[GET_DATA_BYTE(&sum, 2)]);
	    SET_DATA_BYTE(lined, j + 1, valtab[GET_DATA_BYTE(&sum, 3)]);
	}
    }

    return;
}


/*
 *   makeSumTabSG4()
 *         
 *       Returns a table of 256 UINT32s, giving the two output
 *       8-bit grayscale sums corresponding to 8 input bits of a binary
 *       image, for a 4x scale-to-gray op.  The sums from four
 *       adjacent scanlines are then added and transformed to
 *       output 8 bpp pixel values, using makeValTabSG4().
 */
UINT32  *
makeSumTabSG4(void)
{
INT32    i;
INT32    sum[] = {0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4};
UINT32  *tab;

    PROCNAME("makeSumTabSG4");

    if ((tab = (UINT32 *)CALLOC(256, sizeof(UINT32))) == NULL)
	return (UINT32 *)ERROR_PTR("calloc fail for tab", procName, NULL);

	/* pack the two sums separately in two bytes */
    for (i = 0; i < 256; i++) {
	tab[i] = (sum[i & 0xf]) | (sum[(i >> 3) & 0xf] << 8);
    }

    return tab;
}


/*
 *   makeValTabSG4()
 *         
 *       Returns an 8 bit value for the sum of ON pixels
 *       in a 4x4 square, according to
 *
 *              val = 255 - (255 * sum)/16
 *     
 *       where sum is in set {0, ... ,16}
 */
UINT8 *
makeValTabSG4(void)
{
INT32    i;
UINT8   *tab;

    PROCNAME("makeValTabSG4");

    if ((tab = (UINT8 *)CALLOC(17, sizeof(UINT8))) == NULL)
	return (UINT8 *)ERROR_PTR("calloc fail for tab", procName, NULL);

    for (i = 0; i < 17; i++)
	tab[i] = 0xff - (i * 255) / 16;

    return tab;
}



