#/*====================================================================*
# -  Copyright (C) 2001 Leptonica.  All rights reserved.
# -  This software is distributed in the hope that it will be
# -  useful, but with NO WARRANTY OF ANY KIND.
# -  No author or distributor accepts responsibility to anyone for the
# -  consequences of using this software, or for whether it serves any
# -  particular purpose or works at all, unless he or she says so in
# -  writing.  Everyone is granted permission to copy, modify and
# -  redistribute this source code, for commercial or non-commercial
# -  purposes, with the following restrictions: (1) the origin of this
# -  source code must not be misrepresented; (2) modified versions must
# -  be plainly marked as such; and (3) this notice may not be removed
# -  or altered from any source or modified source distribution.
# *====================================================================*/

/*
 *  skew.c
 *
 *      Simple top-level deskew interface
 *          PIX       *pixDeskew()
 *
 *      Simple top-level angle-finding interface
 *          INT32      pixFindSkew()
 *
 *      Basic angle-finding functions with all parameters
 *          INT32      pixFindSkewSweep()
 *          INT32      pixFindSkewSweepAndSearch()
 *
 *      Differential square sum function for scoring
 *          INT32      pixFindDifferentialSquareSum()
 *
 *
 *      Skew is determined by pixel profiles, which are computed
 *      as pixel sums along the raster line for each line in the
 *      image.  By vertically shearing the image by a given angle,
 *      the sums can be computed quickly along the raster lines
 *      rather than along lines at that angle.  The score is
 *      computed from these line sums by taking the square of
 *      the DIFFERENCE between adjacent line sums, summed over
 *      all lines.  The skew angle is then found as the angle
 *      that maximizes the score.  The actual computation for
 *      any sheared image is done in the function
 *      pixFindDifferentialSquareSum().
 *
 *      The search for the angle that maximizes this score is
 *      most efficiently performed by first sweeping coarsely
 *      over angles, using a significantly reduced image (say, 4x
 *      reduction), to find the approximate maximum within a half
 *      degree or so, and then doing an interval-halving binary
 *      search at higher resolution to get the skew angle to 
 *      within 1/20 degree or better.
 *      
 *      The differential signal is used (rather than just using
 *      that variance of line sums) because it rejects the
 *      background noise due to total number of black pixels,
 *      and has maximum contributions from the baselines and
 *      x-height lines of text when the textlines are aligned
 *      with the raster lines.  It also works well in multicolumn
 *      pages where the textlines do not line up across columns.
 *
 *      The method is fast, accurate to within an angle (in radians)
 *      of approximately the inverse width in pixels of the image, 
 *      and will work on a surprisingly small amount of text data
 *      (just a couple of text lines).  Consequently, it can
 *      also be used to find local skew if the skew were to vary
 *      significantly over the page.  Local skew determination
 *      is not very important except for locating lines of
 *      handwritten text that may be mixed with printed text.
 */

#include <stdio.h>
#include <stdlib.h>

#include "allheaders.h"

    /* Default sweep angle parameters for pixFindSkew() */
#define   DEFAULT_SWEEP_RANGE           5.     /* degrees */
#define   DEFAULT_SWEEP_DELTA           1.     /* degrees */

    /* Default final angle difference parameter for binary
     * search in pixFindSkew().  The expected accuracy is
     * not better than the inverse image width in pixels,
     * say, 1/2000 radians, or about 0.03 degrees. */
#define   DEFAULT_MINBS_DELTA           0.01   /* degrees */

    /* Default scale factors for pixFindSkew() */
#define   DEFAULT_SWEEP_REDUCTION       4   /* sweep part; don't change */
#define   DEFAULT_BS_REDUCTION          2   /* binary search part */

    /* Minimum angle for deskewing in pixDeskew() */
#define  MIN_DESKEW_ANGLE               0.1   /* degree */

    /* Minimum allowed confidence (ratio) for deskewing in pixDeskew() */
#define  MIN_ALLOWED_CONFIDENCE         3.0


#ifndef  NO_CONSOLE_IO
#define  DEBUG_PRINT_SCORES     0
#define  DEBUG_PRINT_ANGLE      1
#define  DEBUG_PRINT_CONFIDENCE 1
#define  DEBUG_PLOT_SCORES      0
#endif  /* ~NO_CONSOLE_IO */


/*----------------------------------------------------------------*
 *                       Top-level interfaces                     *
 *----------------------------------------------------------------*/
/*!
 *  pixDeskew()
 *
 *      Input:  pixs
 *              redsearch  (for binary search: reduction factor = 1, 2 or 4)
 *      Return: deskewed pix, or NULL on error
 */      
PIX *
pixDeskew(PIX    *pixs,
          INT32   redsearch)
{
FLOAT32    angle, conf, deg2rad;
PIX       *pixd;

    PROCNAME("pixDeskew");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (redsearch != 1 && redsearch != 2 && redsearch != 4)
	return (PIX *)ERROR_PTR("redsearch not in {1,2,4}", procName, NULL);

    deg2rad = 3.1415926535 / 180.;
    if (pixFindSkewSweepAndSearch(pixs, &angle, &conf,
	       	DEFAULT_SWEEP_REDUCTION, redsearch,
	       	DEFAULT_SWEEP_RANGE, DEFAULT_SWEEP_DELTA,
	       	DEFAULT_MINBS_DELTA))
	return pixClone(pixs);

    if (ABS(angle) < MIN_DESKEW_ANGLE || conf < MIN_ALLOWED_CONFIDENCE)
	return pixClone(pixs);

    if ((pixd = pixRotateShear(pixs, 0, 0, deg2rad * angle)) == NULL)
	return pixClone(pixs);
    else
	return pixd;
}
	


/*!
 *  pixFindSkew()
 *
 *      Input:  pixs
 *              &angle   (<returned> angle required to deskew, in degrees)
 *              &conf    (<returned> confidence value is ratio max/min scores)
 *      Return: 0 if OK, 1 on error or if angle measurment not valid
 *
 *      This is a simple higher-level interface, that uses default
 *      values of the parameters for reasonable speed and accuracy.
 *      Note that the angle returned is the negative of the skew
 *      angle of the image; it is the angle required for deskew.
 *      Clockwise rotations are positive angles.
 */      
INT32
pixFindSkew(PIX      *pixs,
            FLOAT32  *pangle,
            FLOAT32  *pconf)
{
    PROCNAME("pixFindSkew");

    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);
    if (!pangle)
	return ERROR_INT("&angle not defined", procName, 1);
    if (!pconf)
	return ERROR_INT("&conf not defined", procName, 1);

    return pixFindSkewSweepAndSearch(pixs, pangle, pconf, 
				     DEFAULT_SWEEP_REDUCTION,
				     DEFAULT_BS_REDUCTION,
				     DEFAULT_SWEEP_RANGE,
				     DEFAULT_SWEEP_DELTA,
				     DEFAULT_MINBS_DELTA);
}


/*----------------------------------------------------------------*
 *         Basic angle-finding functions with all parameters      *
 *----------------------------------------------------------------*/
/*!
 *   pixFindSkewSweep()
 *
 *      Input:  pixs
 *              &angle   (<returned> angle required to deskew, in degrees)
 *              reduction  (factor = 1, 2, 4 or 8)
 *              sweeprange   (half the full range; assumed about 0; in degrees)
 *              sweepdelta   (angle increment of sweep; in degrees)
 *      Return: 0 if OK, 1 on error or if angle measurment not valid
 */
INT32
pixFindSkewSweep(PIX      *pixs,
		 FLOAT32  *pangle,
		 INT32     reduction,
		 FLOAT32   sweeprange,
		 FLOAT32   sweepdelta)
{
INT32      bzero, i, nangles;
FLOAT32    deg2rad, rad2deg, theta;
FLOAT32    sum, maxscore, maxangle;
GPLOT     *gplot;
NUMA      *natheta, *nascore;
PIX       *pix, *pixt;

    PROCNAME("pixFindSkewSweep");

    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);
    if (!pangle)
	return ERROR_INT("&angle not defined", procName, 1);
    if (reduction != 1 && reduction != 2 && reduction != 4 && reduction != 8)
	return ERROR_INT("reduction must be in {1,2,4,8}", procName, 1);

    *pangle = 0.0;  /* init */
    deg2rad = 3.1415926535 / 180.;
    rad2deg = 1. / deg2rad;

	/* Generate reduced image, if requested */
    if (reduction == 1)
	pix = pixClone(pixs);
    else if (reduction == 2)
	pix = pixReduceRankBinaryCascade(pixs, 1, 0, 0, 0);
    else if (reduction == 4)
	pix = pixReduceRankBinaryCascade(pixs, 1, 1, 0, 0);
    else /* reduction == 8 */
	pix = pixReduceRankBinaryCascade(pixs, 1, 1, 2, 0);

    pixZero(pix, &bzero);
    if (bzero) {
	pixDestroy(&pix);
	return 1;
    }

    nangles = (INT32)((2. * sweeprange) / sweepdelta + 1);
    if ((natheta = numaCreate(nangles)) == NULL)
	return ERROR_INT("natheta not made", procName, 1);
    if ((nascore = numaCreate(nangles)) == NULL)
	return ERROR_INT("nascore not made", procName, 1);

    if ((pixt = pixCreateTemplate(pix)) == NULL)
	return ERROR_INT("pixt not made", procName, 1);

    for (i = 0; i < nangles; i++) {
	theta = -sweeprange + i * sweepdelta;   /* degrees */

	    /* Shear pix about the UL corner and put the result in pixt */
	pixVShearCorner(pixt, pix, deg2rad * theta);

	    /* Get score */
	pixFindDifferentialSquareSum(pixt, &sum);

#if  DEBUG_PRINT_SCORES
	fprintf(stderr, "sum(%7.2f) = %7.0f\n", theta, sum);
#endif  /* DEBUG_PRINT_SCORES */

	    /* Save the result in the output arrays */
	numaAddNumber(nascore, sum);
	numaAddNumber(natheta, theta);
    }

	/* Find the location of the maximum (i.e., the skew angle)
	 * by fitting the largest data point and its two neighbors
	 * to a quadratic, using lagrangian interpolation.  */
    numaFitMax(nascore, &maxscore, natheta, &maxangle);
    *pangle = maxangle;

#if  DEBUG_PRINT_ANGLE
    fprintf(stderr, " From sweep: angle = %7.3f, score = %7.3f\n",
            maxangle, maxscore);
#endif  /* DEBUG_PRINT_ANGLE */

#if  DEBUG_PLOT_SCORES
        /* Plot the result -- the scores versus rotation angle --
	 * using gnuplot with GPLOT_LINES (lines connecting data points).
	 * The GPLOT data structure is first created, with the
	 * appropriate data incorporated from the two input NUMAs,
	 * and then the function gplotMakeOutput() uses gnuplot to
	 * generate the output plot.  This can be either a .png file
	 * or a .ps file, depending on whether you use GPLOT_PNG
	 * or GPLOT_PS.  */
    gplot = gplotCreate(natheta, nascore, "sweep_output",
                        GPLOT_PNG, GPLOT_LINES,
			"Sweep. Variance of difference of ON pixels vs. angle",
			"plot1", "angle (deg)", "score");
    gplotAddPlot(gplot, natheta, nascore, GPLOT_POINTS, "plot2");
    gplotMakeOutput(gplot);
    gplotDestroy(&gplot);
#endif  /* DEBUG_PLOT_SCORES */

    numaDestroy(&nascore);
    numaDestroy(&natheta);
    pixDestroy(&pix);
    pixDestroy(&pixt);
    return 0;
}


/*!
 *   pixFindSkewSweepAndSearch()
 *
 *      Input:  pixs
 *              &angle   (<returned> angle required to deskew, in degrees)
 *              &conf    (<returned> confidence given by ratio of max/min score)
 *              redsweep  (sweep reduction factor = 1, 2, 4 or 8)
 *              redsearch  (binary search reduction factor = 1, 2, 4 or 8;
 *                          and must not exceed redsweep)
 *              sweeprange   (half the full range; assumed about 0; in degrees)
 *              sweepdelta   (angle increment of sweep; in degrees)
 *              minbsdelta   (min binary search increment angle; in degrees)
 *      Return: 0 if OK, 1 on error or if angle measurment not valid
 */
INT32
pixFindSkewSweepAndSearch(PIX      *pixs,
		          FLOAT32  *pangle,
			  FLOAT32  *pconf,
		          INT32     redsweep,
		          INT32     redsearch,
		          FLOAT32   sweeprange,
		          FLOAT32   sweepdelta,
		          FLOAT32   minbsdelta)
{
INT32      bzero, i, nangles, n, ratio, maxindex, minloc;
FLOAT32    deg2rad, rad2deg, theta, delta;
FLOAT32    sum, maxscore, maxangle;
FLOAT32    centerangle, leftcenterangle, rightcenterangle;
FLOAT32    lefttemp, righttemp;
FLOAT32    bsearchscore[5];
FLOAT32    minscore, scoreratio;
GPLOT     *gplot;
NUMA      *natheta, *nascore;
PIX       *pixsw, *pixsch, *pixt1, *pixt2;

    PROCNAME("pixFindSkewSweepAndSearch");

    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);
    if (!pangle)
	return ERROR_INT("&angle not defined", procName, 1);
    if (!pconf)
	return ERROR_INT("&conf not defined", procName, 1);
    if (redsweep != 1 && redsweep != 2 && redsweep != 4 && redsweep != 8)
	return ERROR_INT("redsweep must be in {1,2,4,8}", procName, 1);
    if (redsearch != 1 && redsearch != 2 && redsearch != 4 && redsearch != 8)
	return ERROR_INT("redsearch must be in {1,2,4,8}", procName, 1);
    if (redsearch > redsweep)
	return ERROR_INT("redsearch must not exceed redsweep", procName, 1);

    *pangle = 0.0;  /* init */
    deg2rad = 3.1415926535 / 180.;
    rad2deg = 1. / deg2rad;

	/* Generate reduced image for binary search, if requested */
    if (redsearch == 1)
	pixsch = pixClone(pixs);
    else if (redsearch == 2)
	pixsch = pixReduceRankBinaryCascade(pixs, 1, 0, 0, 0);
    else if (redsearch == 4)
	pixsch = pixReduceRankBinaryCascade(pixs, 1, 1, 0, 0);
    else  /* redsearch == 8 */
	pixsch = pixReduceRankBinaryCascade(pixs, 1, 1, 2, 0);

    pixZero(pixsch, &bzero);
    if (bzero) {
	pixDestroy(&pixsch);
	return 1;
    }

	/* Generate reduced image for sweep, if requested */
    ratio = redsweep / redsearch;
    if (ratio == 1)
	pixsw = pixClone(pixsch);
    else {  /* ratio > 1 */
	if (ratio == 2)
	    pixsw = pixReduceRankBinaryCascade(pixsch, 1, 0, 0, 0);
	else if (ratio == 4)
	    pixsw = pixReduceRankBinaryCascade(pixsch, 1, 2, 0, 0);
	else  /* ratio == 8 */
	    pixsw = pixReduceRankBinaryCascade(pixsch, 1, 2, 2, 0);
    } 

    if ((pixt1 = pixCreateTemplate(pixsw)) == NULL)
	return ERROR_INT("pixt1 not made", procName, 1);
    if (ratio == 1)
	pixt2 = pixClone(pixt1);
    else {
	if ((pixt2 = pixCreateTemplate(pixsch)) == NULL)
	    return ERROR_INT("pixt2 not made", procName, 1);
    }

    nangles = (INT32)((2. * sweeprange) / sweepdelta + 1);
    if ((natheta = numaCreate(nangles)) == NULL)
	return ERROR_INT("natheta not made", procName, 1);
    if ((nascore = numaCreate(nangles)) == NULL)
	return ERROR_INT("nascore not made", procName, 1);

	/* Do sweep */
    for (i = 0; i < nangles; i++) {
	theta = -sweeprange + i * sweepdelta;   /* degrees */

	    /* Shear pix about the UL corner and put the result in pixt1 */
	pixVShearCorner(pixt1, pixsw, deg2rad * theta);

	    /* Get score */
	pixFindDifferentialSquareSum(pixt1, &sum);

#if  DEBUG_PRINT_SCORES
	fprintf(stderr, "sum(%7.2f) = %7.0f\n", theta, sum);
#endif  /* DEBUG_PRINT_SCORES */

	    /* Save the result in the output arrays */
	numaAddNumber(nascore, sum);
	numaAddNumber(natheta, theta);
    }

	/* Find the largest of the set (maxscore at maxangle) */
    numaGetMax(nascore, &maxscore, &maxindex);
    numaGetFValue(natheta, maxindex, &maxangle);

#if  DEBUG_PRINT_ANGLE
    fprintf(stderr, " From sweep: angle = %7.3f, score = %7.3f\n",
            maxangle, maxscore);
#endif  /* DEBUG_PRINT_ANGLE */

	/* Empty the numas */
    numaEmpty(nascore);
    numaEmpty(natheta);

	/* Check if the max is at the end of the sweep. */
    if (maxindex == 0)
	fprintf(stderr, "Warning: max found at left edge of sweep\n");
    n = numaGetCount(natheta);
    if (maxindex == n - 1)
	fprintf(stderr, "Warning: max found at right edge of sweep\n");

	/* Do binary search to find skew angle.
	 * First, set up initial three points. */
    centerangle = maxangle;
    pixVShearCorner(pixt2, pixsch, deg2rad * centerangle);
    pixFindDifferentialSquareSum(pixt2, &bsearchscore[2]);
    pixVShearCorner(pixt2, pixsch, deg2rad * (centerangle - sweepdelta));
    pixFindDifferentialSquareSum(pixt2, &bsearchscore[0]);
    pixVShearCorner(pixt2, pixsch, deg2rad * (centerangle + sweepdelta));
    pixFindDifferentialSquareSum(pixt2, &bsearchscore[4]);

    numaAddNumber(nascore, bsearchscore[2]);
    numaAddNumber(natheta, centerangle);
    numaAddNumber(nascore, bsearchscore[0]);
    numaAddNumber(natheta, centerangle - sweepdelta);
    numaAddNumber(nascore, bsearchscore[4]);
    numaAddNumber(natheta, centerangle + sweepdelta);

        /* Start the search */
    delta = 0.5 * sweepdelta;
    while (delta >= minbsdelta)
    {
	    /* Get the left intermediate score */
        leftcenterangle = centerangle - delta;
	pixVShearCorner(pixt2, pixsch, deg2rad * leftcenterangle);
	pixFindDifferentialSquareSum(pixt2, &bsearchscore[1]);
	numaAddNumber(nascore, bsearchscore[1]);
	numaAddNumber(natheta, leftcenterangle);
	
	    /* Get the right intermediate score */
	rightcenterangle = centerangle + delta;
	pixVShearCorner(pixt2, pixsch, deg2rad * rightcenterangle);
	pixFindDifferentialSquareSum(pixt2, &bsearchscore[3]);
	numaAddNumber(nascore, bsearchscore[3]);
	numaAddNumber(natheta, rightcenterangle);
	
	    /* Find the maximum of the five scores and its location.
	     * Note that the maximum must be in the center
	     * three values, not in the end two. */
	maxscore = bsearchscore[1];
	maxindex = 1;
	for (i = 2; i < 4; i++) {
	    if (bsearchscore[i] > maxscore) {
	        maxscore = bsearchscore[i];
		maxindex = i;
	    }
	}

	    /* Set up score array to interpolate for the next iteration */
	lefttemp = bsearchscore[maxindex - 1];
	righttemp = bsearchscore[maxindex + 1];
	bsearchscore[2] = maxscore;
	bsearchscore[0] = lefttemp;
	bsearchscore[4] = righttemp;

	    /* Get new center angle and delta for next iteration */
	centerangle = centerangle + delta * (maxindex - 2);
        delta = 0.5 * delta;
    }
    *pangle = centerangle;

#if  DEBUG_PRINT_ANGLE
    fprintf(stderr, " From binary search: angle = %7.3f, score = %7.3f\n",
            centerangle, bsearchscore[2]);
#endif  /* DEBUG_PRINT_ANGLE */

        /* Return the ratio of Max score over Min score
	 * as a confidence value */
    numaGetMin(nascore, &minscore, &minloc);
    if (minscore != 0.0)
	*pconf = maxscore / minscore;
    else
	*pconf = 0.0;

#if  DEBUG_PRINT_CONFIDENCE
    fprintf(stderr, " score ratio = %8.2f\n", *pconf);
#endif  /* DEBUG_PRINT_CONFIDENCE */

#if  DEBUG_PLOT_SCORES
        /* Plot the result -- the scores versus rotation angle --
	 * using gnuplot with GPLOT_POINTS.  Because the data
	 * points are not in order (increasing or decreasing theta),
	 * use of GPLOT_LINES would not be informative! */
    gplot = gplotCreate(natheta, nascore, "search_output",
		GPLOT_PNG, GPLOT_POINTS,
		"Binary search.  Variance of difference of ON pixels vs. angle",
		"plot1", "angle (deg)", "score");
    gplotMakeOutput(gplot);
    gplotDestroy(&gplot);
#endif  /* DEBUG_PLOT_SCORES */

    numaDestroy(&nascore);
    numaDestroy(&natheta);
    pixDestroy(&pixsw);
    pixDestroy(&pixsch);
    pixDestroy(&pixt1);
    pixDestroy(&pixt2);
    return 0;
}



/*----------------------------------------------------------------*
 *                  Differential square sum function              *
 *----------------------------------------------------------------*/
/*!
 *   pixFindDifferentialSquareSum()
 *        Input:  pixs
 *                &sum  (<returned> result)
 *        Return: 0 if OK, 1 on error
 */
INT32
pixFindDifferentialSquareSum(PIX      *pixs,
                             FLOAT32  *psum)
{
INT32      i, n;
FLOAT32    val1, val2, diff, sum;
NUMA      *na;

    PROCNAME("pixFindDifferentialSquareSum");

    if (!pixs)
	return ERROR_INT("pixs not defined", procName, 1);

	/* Generate a number array consisting of the sum
	 * of pixels in each row of pixt */
    if ((na = pixCountPixelsByRow(pixs, NULL)) == NULL)
	return ERROR_INT("na not made", procName, 1);

	/* Sum the squares of differential row sums */
    n = numaGetCount(na);
    sum = 0.0;
    for (i = 1; i < n; i++) {
	numaGetFValue(na, i - 1, &val1);
	numaGetFValue(na, i, &val2);
	diff = val2 - val1;
	sum += diff * diff;
    }
    numaDestroy(&na);
    *psum = sum;
    return 0;
}

