/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *      stack.c
 *
 *          Generic stack
 *
 *          The stack is an array of void * ptrs, onto which
 *          objects can be stored.  At any time, the number of
 *          stored objects is stack->n.  The object at the bottom
 *          of the stack is at array[0]; the object at the top of
 *          the stack is at array[n-1].  New objects are added
 *          to the top of the stack; i.e., the first available 
 *          location, which is at array[n].  The stack is expanded
 *          by doubling, when needed.  Objects are removed
 *          from the top of the stack.  When an attempt is made
 *          to remove an object from an empty stack, the result
 *          is null.
 *        
 *
 *          Create/Destroy
 *               STACK     *stackCreate()
 *               void       stackDestroy()
 *
 *          Accessors
 *               INT32      stackAdd()
 *               void      *stackRemove()
 *               INT32      stackExtendArray()
 *
 *          Text description
 *               INT32      stackPrint()
 *
 */


#include <stdio.h>
#include <stdlib.h>

#include "allheaders.h"

#define  INITIAL_PTR_ARRAYSIZE     20


/*---------------------------------------------------------------------*
 *                          Create/Destroy                             *
/*---------------------------------------------------------------------*/
/*!
 *  stackCreate()
 *
 *      Input: nalloc (initial ptr array size; use 0 for default)
 *      Return: stack, or null on error
 */
STACK *
stackCreate(INT32  nalloc)
{
STACK    *stack;

    PROCNAME("stackCreate");

    if (nalloc <= 0)
	nalloc = INITIAL_PTR_ARRAYSIZE;

    if ((stack = (STACK *)CALLOC(1, sizeof(STACK))) == NULL)
	return (STACK *)ERROR_PTR("stack not made", procName, NULL);

    if ((stack->array = (void **)CALLOC(nalloc, sizeof(void *))) == NULL)
	return (STACK *)ERROR_PTR("stack ptr array not made", procName, NULL);

    stack->nalloc = nalloc;
    stack->n = 0;
    
    return stack;
}


/*!
 *  stackDestroy()
 *
 *      Input: &stack (<to be nulled>
 *      Return: void
 *
 *      N.B.: stackDestroy() does NOT destroy the objects on the
 *            stack.  These objects (i.e., the objects pointed
 *            to by the first stack->n elements in the ptr array)
 *            must be destroyed BEFORE the stack is destroyed.
 */
void
stackDestroy(STACK  **pstack)
{
STACK    *stack;

    if (!pstack) return;
    if ((stack = *pstack) == NULL)
	return;

    if (stack->auxstack)
	stackDestroy(&stack->auxstack);

    if (stack->array)
	FREE((void *)stack->array);
    FREE((void *)stack);
    *pstack = NULL;
}



/*---------------------------------------------------------------------*
 *                               Accessors                             *
/*---------------------------------------------------------------------*/
/*!
 *  stackAdd()
 *
 *       Input:  stack
 *               ptr to new thing (to be put on the stack)
 *       Return: 0 if OK; 1 on error.
 *
 */
INT32
stackAdd(STACK  *stack,
	 void   *pnew)
{

    PROCNAME("stackAdd");

    if (!stack)
	return ERROR_INT("stack not defined", procName, 1);
    if (!pnew)
	return ERROR_INT("pnew not defined", procName, 1);

	/* do we need to extend the array? */
    if (stack->n >= stack->nalloc)
	stackExtendArray(stack);

	/* store the new pointer */
    stack->array[stack->n] = (void *)pnew;
    stack->n++;

    return 0;
}


/*!
 *  stackRemove()
 *
 *       Input:  stack 
 *       Return: the ptr previously at the "top" of the stack,
 *               or null on error (e.g., if the queue is empty).
 */
void *
stackRemove(STACK  *stack)
{
void   *pcatch;

    PROCNAME("stackRemove");

    if (!stack)
	return ERROR_PTR("stack not defined", procName, NULL);

    if (stack->n == 0) {
        WARNING("stack empty", procName);
	return NULL;
    }

    stack->n--;
    pcatch = stack->array[stack->n];
	
    return pcatch;
}


/*!
 *  stackExtendArray()
 *
 *     Input:   stack
 *     Return:  0 if OK; 1 on error
 */
INT32
stackExtendArray(STACK  *stack)
{

    PROCNAME("stackExtendArray");

    if (!stack)
	return ERROR_INT("stack not defined", procName, 1);

    if ((stack->array = (void **)reallocNew((void **)&stack->array,
                                4 * stack->nalloc,
				8 * stack->nalloc)) == NULL)
	return ERROR_INT("new stack array not defined", procName, 1);

    stack->nalloc = 2 * stack->nalloc;
    return 0;
}



/*---------------------------------------------------------------------*
 *                            Debug output                             *
/*---------------------------------------------------------------------*/
/*!
 *  stackPrint()
 *
 *       Input:  stream
 *               stack
 *       Return: 0 if OK; 1 on error
 */
INT32
stackPrint(FILE   *fp,
	   STACK  *stack)
{
INT32     i;

    PROCNAME("stackPrint");

    if (!fp)
	return ERROR_INT("stream not defined", procName, 1);
    if (!stack)
	return ERROR_INT("stack not defined", procName, 1);

    fprintf(fp, "\n Stack: nalloc = %d, n = %d, array = %x\n", 
           stack->nalloc, stack->n, stack->array);
    for (i = 0; i < stack->n; i++)
	fprintf(fp,   "array[%d] = %x\n", i, stack->array[i]);
    
    return 0;
}



