/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/

/*
 *   viewfiles.c
 *
 *     Generate smaller images for viewing and write html
 *        INT32    pixHtmlViewer()
 *
 *     Note: this is specialized for unix, not M$
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>     /* unix only */

#include "allheaders.h"


#define   BUF_SIZE                  512
#define   DEFAULT_THUMB_WIDTH       120
#define   DEFAULT_VIEW_WIDTH        800
#define   MIN_THUMB_WIDTH            50
#define   MIN_VIEW_WIDTH            300


/*---------------------------------------------------------------------*
 *            Generate smaller images for viewing and write html       *
 *---------------------------------------------------------------------*/
/*!
 *  pixHtmlViewer()
 *
 *      Input:   dirin:  directory of input image files
 *               dirout: directory for output files
 *               rootname: root name for output files
 *               thumbwidth:  width of thumb images
 *                            (in pixels; use 0 for default)
 *               viewwidth:  maximum width of view images (no up-scaling)
 *                            (in pixels; use 0 for default)
 *               copyorig:  1 to copy originals to dirout; 0 otherwise
 *      Return: 0 if OK; 1 on error
 *
 *      Action: the thumb and view reduced images are generated,
 *              along with two html files:
 *                    <rootname>.html and <rootname>-links.html
 *              The thumb and view files are named
 *                    <rootname>_thumb_xxx.jpg
 *                    <rootname>_view_xxx.jpg
 *              With this naming scheme, any number of input directories
 *              of images can be processed into views and thumbs
 *              and placed in the same output directory.
 */
INT32
pixHtmlViewer(char     *dirin,
              char     *dirout,
              char     *rootname,
	      INT32     thumbwidth,
	      INT32     viewwidth,
	      INT32     copyorig)
{
char            *fname, *fullname, *outname;
char            *mainname, *linkname, *linknameshort;
char            *viewfile, *thumbfile;
char            *shtml, *slink;
char             charbuf[BUF_SIZE];
INT32            i, nfiles, index, w, nimages;
FLOAT32          factor;
PIX             *pix, *pixthumb, *pixview;
SARRAY          *safiles, *sathumbs, *saviews, *sahtml, *salink;
DIR             *pdir;
struct dirent   *pdirentry;

    PROCNAME("pixHtmlViewer");

    if (!dirin)
        return ERROR_INT("dirin not defined", procName, 1);
    if (!dirout)
        return ERROR_INT("dirout not defined", procName, 1);
    if (!rootname)
        return ERROR_INT("rootname not defined", procName, 1);

    if (thumbwidth == 0)
	thumbwidth = DEFAULT_THUMB_WIDTH;
    if (thumbwidth < MIN_THUMB_WIDTH) {
	WARNING("thumbwidth too small; using min value", procName);
	thumbwidth = MIN_THUMB_WIDTH;
    }
    if (viewwidth == 0)
	viewwidth = DEFAULT_VIEW_WIDTH;
    if (viewwidth < MIN_VIEW_WIDTH) {
	WARNING("viewwidth too small; using min value", procName);
	viewwidth = MIN_VIEW_WIDTH;
    }

        /* make the output directory if it doesn't already exist */
    sprintf(charbuf, "mkdir -p %s", dirout);
    system(charbuf);

	/* capture the filenames in the indir */
    if ((safiles = sarrayCreate(0)) == NULL)
        return ERROR_INT("safiles not made", procName, 1);
    if ((pdir = opendir(dirin)) == NULL)
        return ERROR_INT("pdir not opened", procName, 1);
    while (pdirentry = readdir(pdir))  {
	if (pdirentry->d_type == DT_DIR)  /* ignore directories */
	    continue;
	sarrayAddString(safiles, pdirentry->d_name, 1);
    }
    closedir(pdir);
/*    sarrayWriteStream(stderr, safiles); */

        /* generate output text file names */
    sprintf(charbuf, "%s/%s.html", dirout, rootname);
    mainname = stringNew(charbuf);
    sprintf(charbuf, "%s/%s-links.html", dirout, rootname);
    linkname = stringNew(charbuf);
    linknameshort = stringJoin(rootname, "-links.html");

    if ((sathumbs = sarrayCreate(0)) == NULL)
        return ERROR_INT("sathumbs not made", procName, 1);
    if ((saviews = sarrayCreate(0)) == NULL)
        return ERROR_INT("saviews not made", procName, 1);

	/* generate the thumbs and views */
    nfiles = sarrayGetCount(safiles);
    index = 0;
    for (i = 0; i < nfiles; i++) {
	fname = sarrayGetString(safiles, i, 0);
	fullname = genPathname(dirin, fname);
	fprintf(stderr, "name: %s\n", fullname);
	if ((pix = pixRead(fullname)) == NULL) {
	    fprintf(stderr, "file %s not a readable image\n", fullname);
	    FREE((void *)fullname);
	    continue;
	}
	FREE((void *)fullname);
	if (copyorig) {
	    outname = genPathname(dirout, fname);
	    pixWrite(outname, pix, IFF_JFIF_JPEG);
	    FREE((void *)outname);
	}

	    /* make and store the thumb */
	w = pixGetWidth(pix);
	factor = (FLOAT32)thumbwidth / (FLOAT32)w;
	if ((pixthumb = pixScale(pix, factor, factor)) == NULL)
	    return ERROR_INT("pixthumb not made", procName, 1);
	sprintf(charbuf, "%s_thumb_%03d.jpg", rootname, index);
	sarrayAddString(sathumbs, charbuf, 1);
	outname = genPathname(dirout, charbuf);
	pixWrite(outname, pixthumb, IFF_JFIF_JPEG);
	FREE((void *)outname);
	pixDestroy(&pixthumb);

	    /* make and store the view */
	factor = (FLOAT32)viewwidth / (FLOAT32)w;
	if (factor >= 1.0)
	    pixview = pixClone(pix);   /* no upscaling */
	else {
	    if ((pixview = pixScale(pix, factor, factor)) == NULL)
		return ERROR_INT("pixview not made", procName, 1);
	}
	sprintf(charbuf, "%s_view_%03d.jpg", rootname, index);
	sarrayAddString(saviews, charbuf, 1);
	outname = genPathname(dirout, charbuf);
	pixWrite(outname, pixview, IFF_JFIF_JPEG);
	FREE((void *)outname);
	pixDestroy(&pixview);

	pixDestroy(&pix);
	index++;
    }

	/* generate the main html file */
    if ((sahtml = sarrayCreate(0)) == NULL)
        return ERROR_INT("sahtml not made", procName, 1);
    sarrayAddString(sahtml, "<html>", 1);
    sprintf(charbuf, "<frameset cols=\"%d, *\">", thumbwidth + 30);
    sarrayAddString(sahtml, charbuf, 1);
    sprintf(charbuf, "<frame name=\"thumbs\" src=\"%s\">", linknameshort); 
    sarrayAddString(sahtml, charbuf, 1);
    sprintf(charbuf, "<frame name=\"views\" src=\"%s\">",
            sarrayGetString(saviews, 0, 0)); 
    sarrayAddString(sahtml, charbuf, 1);
    sarrayAddString(sahtml, "</frameset></html>", 1);
    shtml = sarrayToString(sahtml, 1);
    arrayToFile(mainname, "w", shtml, strlen(shtml));
    FREE((void *)shtml);
    FREE((void *)mainname);

	/* generate the link html file */
    nimages = sarrayGetCount(saviews);
    fprintf(stderr, "num. images = %d\n", nimages);
    if ((salink = sarrayCreate(0)) == NULL)
        return ERROR_INT("salink not made", procName, 1);
    for (i = 0; i < nimages; i++) {
        viewfile = sarrayGetString(saviews, i, 0);
        thumbfile = sarrayGetString(sathumbs, i, 0);
	sprintf(charbuf, "<a href=\"%s\" TARGET=views><img src=\"%s\"></a>",
	    viewfile, thumbfile);
	sarrayAddString(salink, charbuf, 1);
    }
    slink = sarrayToString(salink, 1);
    arrayToFile(linkname, "w", slink, strlen(slink));
    FREE((void *)slink);
    FREE((void *)linkname);
    FREE((void *)linknameshort);

    sarrayDestroy(&safiles);
    sarrayDestroy(&sathumbs);
    sarrayDestroy(&saviews);
    sarrayDestroy(&sahtml);
    sarrayDestroy(&salink);

    return 0;
}


