package ca.bcgsc.rwong.LRUI;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.image.RescaleOp;

import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.OverlayLayout;

/**
 * A panel that holds the gel image and adjustable node overlay 
 * @author rwong
 */
public class GelPane extends JPanel {

	//scale factor in the x direction
	private double xScaleFactor;
	//scale factor in the y direction
	private double yScaleFactor;
	/**
	 * handle to AdjustPane
	 * @see AdjustPane
	 */
	private AdjustPane ap;
	/**
	 * handle to AdjustPane
	 * @see ImagePane
	 */
	private ImagePane im;
	/**gel image in memory*/
	private BufferedImage gelImage;

	private static final long serialVersionUID = 1L;

	/**
	 * creates a gel panel 
	 * @param bigel image in memory
	 * @param xsf initial x scale factor
	 * @param ysf initial y scale factor
	 * @param statusText line of text to reflect what node is selected and other information
	 * @param origLanes initial data from source lanes file
	 * @param mainFrame handle to the LaneRulerFrame 
	 */
	public GelPane(BufferedImage bi, double xsf, double ysf, JLabel statusText, LanesReader origLanes, LaneRulerFrame mainFrame){
		xScaleFactor = xsf;
		yScaleFactor = ysf;
		
		Dimension size = new Dimension((int)((double)bi.getWidth()*xScaleFactor), (int)((double)bi.getHeight()*yScaleFactor));
		setLayout(new OverlayLayout(this));
		setPreferredSize(new Dimension(size));
		setOpaque(true);
		
		//getting gel parameters
		AdjustPane.markerOffset = mainFrame.gelSettings.markerOffset;
		AdjustPane.lastMarkerOffset = mainFrame.gelSettings.lastMarkerOffset;
		AdjustPane.markerPeriod = mainFrame.gelSettings.markerPeriod;
		AdjustPane.numLanes = mainFrame.gelSettings.numLanes;		
		
		
		if(bi!=null)	{
			
			gelImage=bi;
			//create imagePane to hold image
			im =new ImagePane(new Dimension(bi.getWidth(), bi.getHeight()), xsf, ysf, gelImage);
			add(im);
			//create adjustPane for node position and adjustment overlay
			ap =new AdjustPane(new Dimension(bi.getWidth(), bi.getHeight()), xsf, ysf, statusText, origLanes, mainFrame);
			add(ap);
	
			addMouseListener(ap);
			addMouseMotionListener(ap);
		}
	}
	
	/**
	 * change the scale of the image display
	 * @param xsf new x scale factor
	 * @param ysf new y scale factor
	 */
	public void changeScale(double xsf, double ysf) {
		xScaleFactor = xsf;
		yScaleFactor = ysf;
		Dimension size = new Dimension((int)((double)gelImage.getWidth()*xScaleFactor), (int)((double)gelImage.getHeight()*yScaleFactor));
		setLayout(new OverlayLayout(this));
		setPreferredSize(new Dimension(size));
		
		//change scale for two children frames
		im.changeScale(xsf, ysf);
		ap.changeScale(xsf, ysf);
		
	}
	
	/**
	 * 
	 * @return current x scale factor
	 */
	public double getXScaleFactor(){
		return xScaleFactor;
	}
	/**
	 * 
	 * @return current y scale factor
	 */
	public double getYScaleFactor(){
		return yScaleFactor;
	}
	/**
	 * 
	 * @return handle to adjustPane
	 */
	public AdjustPane getAdjustPane(){
		return ap;
	}
    
	/**
	 * overloaded paint method
	 */
	public void paint(Graphics g)
    {
		super.paint(g);
		im.paint(g);
		ap.paint(g);
    }
	
	/**
	 * turns off optimized drawing as there is overlap in our setup
	 */
	 public boolean isOptimizedDrawingEnabled(){
		 return false;
	 }
}

/**
 * panel to hold gel image, has scaling support
 * @author rwong
 *
 */
class ImagePane extends JPanel {
	private static final long serialVersionUID = 1L;
	private BufferedImage image=null;
	private Dimension size;
	private double xScaleFactor;
	private double yScaleFactor;
	
	/**
	 * create a new ImagePane to hold an image, maximize dynamic range of the image
	 * @param d overall true dimension of the original image
	 * @param xsf initial x scale factor
	 * @param ysf initial y scale factor
	 * @param bi gel image in memory
	 */
	public ImagePane(Dimension d, double xsf, double ysf, BufferedImage bi)
    {
		size = d;
		xScaleFactor=xsf;
		yScaleFactor=ysf;
				
		//finding max and min of image, sampling every 10th pixel
		float max = 0;
		float min = 255;
		int x;
		int y;
		float temp;
		Raster raster = bi.getData();
		for(y=0;y<bi.getHeight();y+=10){
			for(x=0;x<bi.getWidth();x+=10){		
			temp = raster.getSampleFloat(x, y, 0);
				if (temp>max){max=temp;}
				if (temp<min){min=temp;}
			}
		}
		//maximizing dynamic range
		RescaleOp ro = new RescaleOp(255f/(max-min),-255f/(max-min)*min, null);
		image = new BufferedImage(bi.getWidth(),  bi.getHeight(), BufferedImage.TYPE_BYTE_GRAY); 
		ro.filter(bi, image);

		setBackground(Color.black);
        setPreferredSize(new Dimension((int)(size.getWidth()*xScaleFactor),(int)(size.getHeight()*yScaleFactor)));
        
    }
	
	/**
	 * Change to a new scale
	 * @param xsf new x scale factor
	 * @param ysf new y scale factor
	 */
	public void changeScale(double xsf, double ysf){
		xScaleFactor=xsf;
		yScaleFactor=ysf;
	    setPreferredSize(new Dimension((int)(size.getWidth()*xScaleFactor),(int)(size.getHeight()*yScaleFactor)));
	    repaint();  
	}
	
	/**
	 * overloaded paint method
	 */
    public void paint(Graphics g)
    {
         Graphics2D g2 = (Graphics2D)g;
         g2.drawImage((Image)image, 0,0,(int)(size.getWidth()*xScaleFactor),(int)(size.getHeight()*yScaleFactor), null);
    }
}

/**
 * panel with adjustable node display
 * @author rwong
 *
 */
class AdjustPane extends JComponent implements MouseListener, MouseMotionListener, ActionListener{

	 /**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	/**
	 * 2D node array holding nodes in all zones and all lanes
	 */
	Node[][] nodeArray;
	/** handled to selected node*/
	 Node selectedNode;

	 /** set state for dragging vs selecting*/
	 private boolean lock;
	 /** set state to unhighlight the last selectedNode*/
	 private boolean unhighlight;
	 /** turns half-transparent overlay on or off*/
	 private boolean transparency;
	 /**track if changed since save*/
	 public boolean changed;
	 /**track if input data is on screen*/
	 public boolean inLoaded;  
	 	 
	 /**x scale factor*/
	 private double xScaleFactor;
	 /**y scale factor*/ 
	 private double yScaleFactor;
	 /**height of original image*/  
	 private double imageHeight;
	 /**width of original image*/  
	 private double imageWidth;
	 /**holds temporary value for mouse events*/
	 private double x, y;
	 /**clip region to be updated*/
	 private Rectangle2D activeRegion;
	 /**handle to the text output*/
	 private JLabel statusText;	 
	 /**string of the original text output, further information is added onto this*/
	 private String origStatusText;
 
	 /**default values for gel parameters*/
	public static int markerOffset=0;
	public static int lastMarkerOffset=0;
	public static int markerPeriod=5;
	public static int numLanes=121; 
	
	/**used for finding top and bottom limits*/
	private static int PASS_TOP = 999;
	private static int PASS_BOTTOM = -1;
	 
	/**
	 * creates a panel with adjustable node overlay, with scaling support
	 * @param d overall true dimension of the original image
	 * @param xsf initial x scale factor
	 * @param ysf initial y scale factor@param 
	 * @param text handle to label for text output
	 * @param origLanes initial node positions information
	 * @param mainFrame handle to the LaneRulerFrame
	 */
	 public AdjustPane(Dimension d, double xsf, double ysf, JLabel text, LanesReader origLanes, LaneRulerFrame mainFrame){

		 LaneRulerFrame temp = mainFrame;
		 
		 if (origLanes.yVals[0].length != numLanes){
			 temp.displayError("Error: invalid lanes file, number of lanes different from specified");
			 return;
		 }
		 		 
		setPreferredSize(d);
		imageHeight=d.getHeight();
		imageWidth=d.getWidth();
		setOpaque(false);	 	
	 	xScaleFactor=xsf;
	 	yScaleFactor=ysf; 	
	 	
	 	statusText = text;
	 	origStatusText = text.getText();
	 	
		 Node.setXScaleFactor(xScaleFactor);
		 Node.setYScaleFactor(yScaleFactor);
		 setNodes(origLanes.xVals, origLanes.yVals);
		 
		 lock = false;
		 unhighlight=false;
		 transparency=true;
		 changed = false;
		 inLoaded =true;
	 }
	 
	 /**
	  * create a 2D array of nodes and initialize them as markers or not, 
	  * with given data and gel parameters
	  * @param xVals x node positions for all zones
	  * @param yVals y node positions for all zones and all lanes 
	  */
	 public void setNodes(double[] xVals, double[][] yVals){
		 
		 int zone;
		 int row;
		 boolean IsMarker;
		 
		 nodeArray = new Node[xVals.length][yVals[0].length]; 
		 	
		 for(zone=0; zone<nodeArray.length;zone++)	 {
			 for(row=0; row<nodeArray[0].length;row++) {
				 IsMarker=((row-markerOffset)%markerPeriod == 0 && row <= yVals[0].length-1-lastMarkerOffset  && row-markerOffset >= 0);
				 if(zone==0)
				 {
					 nodeArray[zone][row]=new Node(zone, row, IsMarker, xVals[zone], yVals[zone][row]);
				 } else{
					 nodeArray[zone][row]=new Node(zone, row, IsMarker, xVals[zone], yVals[zone][row], nodeArray[zone-1][row].lineFrom);
				 }
			 }		
		 }
		 //paint entire image
		 repaint(0,0,(int)(imageWidth*xScaleFactor), (int)(imageHeight*yScaleFactor));
	 }
	 
	 /**
	  * turn off optimized drawing as we have overlap 
	  */
	 public boolean isOptimizedDrawingEnabled(){
		 return false;
	 }
	 
	 /**
	  * Change to a new scale
	  * @param xsf new x scale factor
	  * @param ysf new y scale factor
	  */	 
	 public void changeScale(double xsf, double ysf){
		 xScaleFactor=xsf;
		 yScaleFactor=ysf;
		 
		 Node.setXScaleFactor(xsf);
		 Node.setYScaleFactor(ysf);
		 
		 int zone;
		 int row;
		//reset node positions 
		 for(zone=0; zone<nodeArray.length;zone++)	 {
			 for(row=0; row<nodeArray[0].length;row++) {
					 nodeArray[zone][row].redraw();
			}
		 }
		 repaint(0,0,(int)(imageWidth*xScaleFactor), (int)(imageHeight*yScaleFactor));
	 }
	 
	public void paintComponent(Graphics g){
		update(g);
	}
	
/**
 * defines how things are order in which things are drawn on screen
 */
	public void update(Graphics g){
		
		Graphics2D g2 = (Graphics2D)g;
		int zone, row;

		//draw everything except the selected node
		for(row=0; row<nodeArray[0].length;row++){		
			for(zone=0; zone<nodeArray.length; zone++){
				if (nodeArray[zone][row].IsMarker){ //marker lane
					g2.setColor(new Color(1f, 0, 1f));
				}
				else {
					g2.setColor(new Color(1f,0,0));		
				}
							
				g2.draw(nodeArray[zone][row].rect);	
				if(zone<nodeArray.length-1){

					g2.draw(nodeArray[zone][row].lineFrom);
				}
			}
		}
		//draw selected node with fill and transparency
		if(selectedNode!=null){
			g2.setColor(new Color(0,1f,.5f, .3f));
			g2.fill(nodeArray[selectedNode.getZone()][selectedNode.getRow()].rect);
	
			if(lock && transparency){ //draw in transparency
				g2.setColor(new Color(0,1f,.5f, .3f));
				g2.setStroke(new BasicStroke((float)Node.getScaledStrokeSize(), BasicStroke.CAP_BUTT, BasicStroke.JOIN_ROUND));

				zone = selectedNode.getZone();
				row=selectedNode.getRow();

				if(selectedNode.getZone()!=0){
					g2.draw(nodeArray[zone][row].lineTo);
				}

				if(selectedNode.getZone()!=nodeArray.length-1){
					g2.draw(nodeArray[zone][row].lineFrom);
				}
			}	
		}
	}
	
	public void mouseClicked(MouseEvent e) {
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	/**
	 * selects a node if not part of a drag with selected node
	 */
	public void mousePressed(MouseEvent e) {
		int i,j;
		//if not in a drag with selected node
		if(!lock){
			for (i=0; i<nodeArray.length; i++){
				for(j=0; j<nodeArray[0].length;j++){					
					if(nodeArray[i][j].rect.contains(e.getX(),e.getY())){
						//un-highlight the last selected node
						if(selectedNode!=null){unhighlight=true;}
						
						selectedNode=nodeArray[i][j];

						x = selectedNode.rect.getCenterX();
						y = selectedNode.rect.getCenterY();
						lock = true;
						statusText.setText(origStatusText + ",    zone: " + selectedNode.getZone() + " row: " + selectedNode.getRow() + " selected.");
						changed = true;
						inLoaded = false;
						updateLocation(e);
					}
				}
			}

		}
	}

	/**
	 * release lock as mouse button is released
	 */
	public void mouseReleased(MouseEvent e) {
		if(lock){
		lock = false;	
		updateLocation(e);
		}

	} 
	
	/**
	 * updateLocation as long as we are dragging
	 */
	public void mouseDragged(MouseEvent e){
		if(lock){
			updateLocation(e);
		}
	}
	public void mouseMoved(MouseEvent e){}
	
	/**track check box and turns transparent overlay on or off*/
	public void actionPerformed(ActionEvent e){
		if (((Component)e.getSource()).getName().equals("transparency")){
			JCheckBox temp = (JCheckBox) e.getSource();
			transparency = temp.isSelected();
		}
	}
	
	/**
	 * Method reponsible for moving nodes
	 * the method stops moving the node if it is against the limit node
	 * @param e mouseevent that triggered this call
	 */
	public void updateLocation(MouseEvent e){
	
		int zone, row, selectedRow;	
		int rowToUpperLimit, rowToLowerLimit;
		int maxOffset=0;
		int offset;
		
		zone = selectedNode.getZone();	
		selectedRow=selectedNode.getRow();
		
		//getting upper and lower limit row to determine how many nodes to move
		//based on if the node is a marker or not
		int lowerLimitRow=getLowerLimitRow(selectedNode);
		int upperLimitRow=getUpperLimitRow(selectedNode);	
		
		//if the selected Node is on the edge, it is its own limit in that side 
		if(upperLimitRow == PASS_TOP){
			rowToUpperLimit = 0; 
		}else{
			rowToUpperLimit = upperLimitRow-selectedNode.getRow();
		}
		
		if(lowerLimitRow == PASS_BOTTOM){
			rowToLowerLimit = 0;
		}else{
			rowToLowerLimit = selectedNode.getRow()-lowerLimitRow;
		}
		
		//checkPoint checks rather the move is beyond neighbouring nodes
		if(checkPoint(e)==1){
			selectedNode.setLocation(x, (nodeArray[zone][selectedRow].rect.getCenterY()+((double)e.getY()-y)));	
			//move data lanes around the marker
			if(selectedNode.IsMarker){
				//data lanes above marker
				for(offset=1; offset < rowToUpperLimit; offset ++){
					row=selectedRow+offset;
					nodeArray[zone][row].setLocation(x, (double)offset/(double)rowToUpperLimit*(nodeArray[zone][upperLimitRow].rect.getCenterY()-selectedNode.rect.getCenterY())+selectedNode.rect.getCenterY());				
				}
				//data lanes below marker
				for(offset=1; offset < rowToLowerLimit; offset ++){
					row=selectedRow-offset;
					nodeArray[zone][row].setLocation(x, (double)offset/(double)rowToLowerLimit*(nodeArray[zone][lowerLimitRow].rect.getCenterY()-selectedNode.rect.getCenterY())+selectedNode.rect.getCenterY());				
				}
			}
		}else if(checkPoint(e)==-1){ //past limit node
			if(e.getY()-y>0){//moving down, so capped at lower node
				
		//		selectedNode.setLocation(x, getLowerLimit(selectedNode)+selectedNode.rect.getHeight()/2);
				selectedNode.setLocation(x, getLowerLimit(selectedNode));			
				//move to lower limit node only if node was not at lower limit
				if(y!=getLowerLimit(selectedNode)){
					
					//move data lanes around the marker
					if(selectedNode.IsMarker){
						//data lanes above marker
						for(offset=1; offset < rowToUpperLimit; offset ++){
							row=selectedRow+offset;
							nodeArray[zone][row].setLocation(x, (double)offset/(double)rowToUpperLimit*(nodeArray[zone][upperLimitRow].rect.getCenterY()-selectedNode.rect.getCenterY())+selectedNode.rect.getCenterY());				
						}
						//data lanes below marker
						for(offset=1; offset < rowToLowerLimit; offset ++){
							row=selectedRow-offset;
							nodeArray[zone][row].setLocation(x, (double)offset/(double)rowToLowerLimit*(nodeArray[zone][lowerLimitRow].rect.getCenterY()-selectedNode.rect.getCenterY())+selectedNode.rect.getCenterY());				
						}	
					}
			
				}
			} else if(e.getY()-y<0){//moving up, so capped at upper node

				selectedNode.setLocation(x, getUpperLimit(selectedNode));	
				//move only if node was not at upper limit
				if(y!=getUpperLimit(selectedNode)){
					//move data lanes around the marker
					if(selectedNode.IsMarker){
						//data lanes above marker
						for(offset=1; offset < rowToUpperLimit; offset ++){
							row=selectedRow+offset;
							nodeArray[zone][row].setLocation(x, (double)offset/(double)rowToUpperLimit*(nodeArray[zone][upperLimitRow].rect.getCenterY()-selectedNode.rect.getCenterY())+selectedNode.rect.getCenterY());				
						}
						//data lanes below marker
						for(offset=1; offset < rowToLowerLimit; offset ++){
							row=selectedRow-offset;
							nodeArray[zone][row].setLocation(x, (double)offset/(double)rowToLowerLimit*(nodeArray[zone][lowerLimitRow].rect.getCenterY()-selectedNode.rect.getCenterY())+selectedNode.rect.getCenterY());				
						}	
					}
				}
			}
		}
		y = selectedNode.rect.getCenterY();
		//unhighlight last selected node if needed
		if (unhighlight) {
			repaint(0,0,(int)(imageWidth*xScaleFactor), (int)(imageHeight*yScaleFactor));
			unhighlight=false;
		}
	

		//clip rendering area, union of the 8-neighbour nodes 
		if(selectedNode!=null){
			
			activeRegion=(Rectangle2D)selectedNode.rect.clone();
			//union with node above and below if they exist, set to image extent otherwise
			if(selectedRow-rowToLowerLimit<=0){
				activeRegion.setFrameFromDiagonal(activeRegion.getMinX(), activeRegion.getMinY(), activeRegion.getMaxX(), imageHeight*yScaleFactor);
			}else{
				Rectangle2D.union(activeRegion, nodeArray[zone][selectedRow-rowToLowerLimit].rect, activeRegion);
			}
			if(selectedRow+rowToUpperLimit>=nodeArray[0].length-1){
				activeRegion.setFrameFromDiagonal(activeRegion.getMinX(), 0, activeRegion.getMaxX(), activeRegion.getMaxY());

			}else{
				Rectangle2D.union(activeRegion, nodeArray[zone][selectedRow+rowToUpperLimit].rect, activeRegion);
			}
			//union corner nodes to left if it exist		
			if(zone!=0){
				//union diagonal neighbours to left if they exist
				if(selectedRow-rowToLowerLimit>=0){
					Rectangle2D.union(activeRegion, nodeArray[zone-1][selectedRow-rowToLowerLimit].rect, activeRegion);
				}
				if(selectedRow+rowToUpperLimit<=nodeArray[0].length-1){
					Rectangle2D.union(activeRegion, nodeArray[zone-1][selectedRow+rowToUpperLimit].rect, activeRegion);
				}			
			}
//			union nodes to right if it exist
			if(zone!=nodeArray.length-1){
	
//			union diagonal neighbours to right if they exist
				if(selectedRow-rowToLowerLimit>=0){
					Rectangle2D.union(activeRegion, nodeArray[zone+1][selectedRow-rowToLowerLimit].rect, activeRegion);
				}
				if(selectedRow+rowToUpperLimit<=nodeArray[0].length-1){
					Rectangle2D.union(activeRegion, nodeArray[zone+1][selectedRow+rowToUpperLimit].rect, activeRegion);
				}
			}
			
			//limit to image area only
			Rectangle2D.intersect(activeRegion, (Rectangle2D) new Rectangle(0, 0, (int)(imageWidth*xScaleFactor+0.5), (int)(imageHeight*yScaleFactor)), activeRegion);
			repaint((int) activeRegion.getMinX(),(int)activeRegion.getMinY(), (int)activeRegion.getWidth(),(int) activeRegion.getHeight());
		
		} else {
			//if no node selected, repaint entire image
			repaint(0,0,(int)(imageWidth*xScaleFactor), (int)(imageHeight*yScaleFactor));
		}
	}
	
	/**
	 * Checks the upper and lower limit of the selected node to see 
	 * if a move will go beyond these limits
	 * @param e mouse event with co-ordinate of the latest move
	 * @return -1 if this move result in moving beyond the limit, 0 if the move continues to be beyond limit, 1 otherwise
	 */
	private int checkPoint(MouseEvent e)
	{
			
		double upperLimit=getUpperLimit(selectedNode);
		double lowerLimit=getLowerLimit(selectedNode);


		//moving up
		if(e.getY()-y<0){
			if(y>selectedNode.rect.getCenterY()) {//last event beyond location of node, node already stopped against another node
				return 0;
			}
			if(selectedNode.rect.getCenterY()+e.getY()-y<upperLimit) {//pass neighbour node after move
				return -1;
			}
		}
		//moving down
		if(e.getY()-y>0){
			if(y<selectedNode.rect.getCenterY()) {//last event beyond location of node, node already stopped against another node
				return 0;
			}
			if(selectedNode.rect.getCenterY()+e.getY()-y>lowerLimit) {//pass neighbour node after move
				return -1;
			}
		}
		return 1;
	}
	
	/**
	 * finds the lower limit  of the moving range of the node, 
	 * which is the position of the lower neighbour of data lanes
	 * or the position of the lower neighbour marker of marker lanes
	 * or the bottommost lane for the bottommost marker lane that is not itself the bottommost lane
	 * or the bottom edge of the image for the bottommost lane 
	 * @param node node in question 
	 * @return position of the lower limit
	 */
	private double getLowerLimit(Node node){
		//get the lower limit row
		int lowerLimitRow=getLowerLimitRow(selectedNode);
		
		double lowerLimit;
		//if very bottom row, lower limit is edge of the image
		if(lowerLimitRow == PASS_BOTTOM){
			lowerLimit = imageHeight*yScaleFactor-selectedNode.rect.getHeight()/2;
		} else { //otherwise position of the node in the lower limit row
			lowerLimit = nodeArray[selectedNode.getZone()][lowerLimitRow].rect.getCenterY()-selectedNode.rect.getHeight();
		}
		
		return lowerLimit;
	}
	
	/**
	 * finds the upper limit of the moving range of the node, 
	 * which is the position of the upper neighbour of data lanes
	 * or the position of the upper neighbour marker of marker lanes
	 * or the topmost lane for the topmost marker lane that is not itself the topmost lane
	 * or the top edge of the image for the topmost lane 
	 * @param node node in question 
	 * @return position of the upper limit
	 */
	private double getUpperLimit(Node node){
		//get the upper limit row
		int upperLimitRow=getUpperLimitRow(selectedNode);
		
		double upperLimit;
		//if very bottom row, lower limit is edge of the image
		if(upperLimitRow == PASS_TOP){
			upperLimit = 0+selectedNode.rect.getHeight()/2;
		} else {//otherwise position of the node in the lower limit row
			upperLimit = nodeArray[selectedNode.getZone()][upperLimitRow].rect.getCenterY()+selectedNode.rect.getHeight();
		}
		
		return upperLimit;
	}
	
	/**
	 * finds the row that defines the upper movement range of a node
 	 * which is the upper neighbour of data lanes
	 * or the upper neighbour marker of marker lanes
	 * or the topmost lane for the topmost marker lane that is not itself the topmost lane
	 * or defined as PASS_TOP for the topmost lane 
	 * @param node
	 * @return the row number of the upper limit row of the node, PASS_TOP if the node is on the topmost edge already
	 */
	private int getUpperLimitRow(Node node){

		if(node.getRow()==nodeArray[0].length-1){
			return PASS_TOP;
		}
		
		if(node.IsMarker){
			//test if it is last marker
			if(node.getRow()+markerPeriod>=nodeArray[0].length-lastMarkerOffset){
				return nodeArray[0].length-1;
			} else {
				return node.getRow()+markerPeriod;
			}
		} else {
			return node.getRow()+1; //already testes for pass top
		}
	}
	
	/**
	 * finds the row that defines the lower movement range of a node
 	 * which is the lower neighbour of data lanes
	 * or the lower neighbour marker of marker lanes
	 * or the bottommost lane for the bottommost marker lane that is not itself the bottommost lane
	 * or defined as PASS_BOTTOM for the bottommost lane 
	 * @param node
	 * @return the row number of the lower limit row of the node, PASS_BOTTOM if the node is on the bottommost edge already
	 */
	private int getLowerLimitRow(Node node){
		if(node.getRow()==0){
			return PASS_BOTTOM;
		}
		if(node.IsMarker){
			//test if it is first marker
			if(node.getRow()-markerPeriod<markerOffset){
				return 0;
			} else {
				return node.getRow()-markerPeriod;
			}
		} else {
			return node.getRow()-1; //already testes for pass bottom
		}	

		
	}
	
}
