package ca.bcgsc.rwong.LRUI;

import javax.imageio.ImageIO;
import javax.swing.JFrame;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JViewport;
import javax.swing.border.EmptyBorder;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics2D;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.URL;

/**Heavy weight frame in which the lane adjustment takes place
 * Has buttons for saving, reverting and finishing
 * as well as for zooming 
 */
public class LaneRulerFrame extends JFrame{
	
	private static final long serialVersionUID = 1L;
	private URL imageURL;
	
	private String lanesPath;
	private String lanesoutPath;
	
  	private BufferedImage gelImage;
  	private int imageWidth;
  	private int imageHeight;
    int initScrollStep=15;
    
    private LanesManager lm;
    
    private GelPane g;
    private ZoomPanel zoomBar;
    private JScrollPane scrollPane; 
    private JLabel bottomText;
    private LanesWriter lw;
    private JButton saveButton;
    private JButton reinitButton;
    private JButton revertButton;
    private JButton doneButton;
    private JFrame startFrame;
    private JLabel outLine;
    public GelSettings gelSettings;

    /**Constructor, initialize required values and set up a frame for lane tracking adjustment
     * 
     * @param frameName the text that appears on top of the frame
     * @param frame handle to the frame that launched this LaneRulerFrame
     * @param param hold a set of setup parameters for this instance of the LaneRulerFrame
     * @param label handle to the error output label
     */	
	public LaneRulerFrame(String frameName, JFrame frame, GelSettings param, JLabel label){
		super(frameName);
		startFrame=frame;		
		outLine=label;
		
		setPreferredSize(new Dimension(1200,800));
      	
    	//during closing operation, do nothing until query user for save before quiting if changed
    	setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
    	addWindowListener(new WindowAdapter() {
    	    public void windowClosing(WindowEvent e) {
    	    	lm.checkSave(LanesManager.CLOSING);
    	    }
    	});
    	
    	gelSettings=param;
    	//checking params, output error otherwise
    	if (gelSettings.lanesSource.equals(gelSettings.lanesDest)){
    		displayError("Error: source and destination of lanes file cannot be the same");
    	}
    	if (gelSettings.markerOffset>gelSettings.numLanes){
    		displayError("Error: marker offest larger than number of lanes");
			return;
    	}
    	if (gelSettings.lastMarkerOffset>gelSettings.numLanes-gelSettings.markerOffset){
    		displayError("Error: last marker offest larger than number of lanes minus marker offset");
			return;
     	}
    	if (gelSettings.markerPeriod>gelSettings.numLanes){
    		displayError("Error: marker period larger than number of lanes");	
			return;
     	}
    	
    	//read in lanes file
		LanesReader lr = new LanesReader();
		try {
			lanesPath = gelSettings.lanesSource;
			lr = new LanesReader(lanesPath);
		}catch (IOException e) {
			displayError("Error when reading lanes file at \"" + lanesPath  + "\"");	
			return;
		} 
		
		//test if output path is valid and write in original values
		try {
			lanesoutPath = gelSettings.lanesDest;
			lw = new LanesWriter(lanesoutPath);
			lw.write(lr.xVals, lr.yVals);
			lw.close();

		}catch (IOException e) {
			displayError("Error when trying to write to output lanes file at \"" + lanesoutPath + "\"");
			return;
		} 
		
		//read in image and reduce image to 8bit gray scale, rotate if specified		
		try {
			imageURL =  new URL("file://" + gelSettings.imageSource);
			BufferedImage inImage;
   			inImage = ImageIO.read(imageURL);
   			
   			int width=inImage.getWidth();
   			int height=inImage.getHeight();
   			
   			if(gelSettings.rotateAngle == 90 | gelSettings.rotateAngle == 270){
   				width=inImage.getHeight();
  				height=inImage.getWidth();
   			}
        	gelImage = new BufferedImage(width,  height, BufferedImage.TYPE_BYTE_GRAY);
           	Graphics2D gelGraphics = gelImage.createGraphics();
           	AffineTransform at = new AffineTransform();
           	if(gelSettings.rotateAngle!=0){       	
           		at.translate(-inImage.getWidth()/2, -inImage.getHeight()/2);
           		at.rotate(Math.toRadians(gelSettings.rotateAngle),inImage.getWidth()/2, inImage.getHeight()/2 );
           		double transX = inImage.getWidth()/2;
           		double transY = inImage.getHeight()/2;
           	
           		if(gelSettings.rotateAngle == 90){
           			transX = inImage.getWidth()/2;
           			transY = -inImage.getHeight()/2;
           		}else if (gelSettings.rotateAngle == 180){
           			transX = -inImage.getWidth()/2;
           			transY = -inImage.getHeight()/2;
           		}else if(gelSettings.rotateAngle == 270){
           			transX = -inImage.getWidth()/2;
           			transY = inImage.getHeight()/2;
           		}
           		at.translate(transX,transY);	
           	}
           	gelGraphics.drawImage(inImage, at, null);
       } catch (IOException e){			
    	   displayError("Error when reading image file at \"" + imageURL.getFile() + "\"");		
			return;
       }
       
       //setting up gel panel
       imageWidth=gelImage.getWidth();
       imageHeight=gelImage.getHeight();
       bottomText = new JLabel(gelSettings.imageSource + "(" + imageWidth + " x " + imageHeight + ")");           
         g = new GelPane(gelImage, gelSettings.initXZoom, gelSettings.initYZoom, bottomText, lr, this);
  
       
       //adding two borders around GelPane so nodes on the end 
       //of image is not close to the scroll bar
       JPanel scrollBody = new JPanel(new BorderLayout());
       scrollBody.add(g,BorderLayout.CENTER);
       scrollBody.setBorder(new EmptyBorder(0, 25, 0, 25));
       
       //wrap scroll pane around GelPane
       scrollPane = new JScrollPane(scrollBody, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
       scrollPane.getHorizontalScrollBar().setUnitIncrement(initScrollStep);
       scrollPane.getVerticalScrollBar().setUnitIncrement(initScrollStep);
       
       scrollPane.getViewport().setOpaque(true);
       scrollPane.getViewport().setScrollMode(JViewport.SIMPLE_SCROLL_MODE);
       
       //setting up save buttons
       JPanel savePane = new JPanel(new FlowLayout(FlowLayout.TRAILING));

       //set lanes manager to track change state
       lm = new LanesManager(lanesPath, lanesoutPath, this);
       //setting up save, revert, reinit and done buttons
       saveButton = new JButton("Save");
       saveButton.setName("Save");
       saveButton.addActionListener(lm);
       saveButton.setVerticalTextPosition(JButton.CENTER);
       saveButton.setHorizontalTextPosition(JButton.CENTER);
       saveButton.setToolTipText("Click this to save current nodes positions.");        
       savePane.add(saveButton);
       
       revertButton = new JButton("Revert to Last Save");
       revertButton.setName("Revert");
       revertButton.addActionListener(lm);
       revertButton.setVerticalTextPosition(JButton.CENTER);
       revertButton.setHorizontalTextPosition(JButton.CENTER);
       revertButton.setToolTipText("Click this to revert nodes positions to last save.");        
       savePane.add(revertButton);
       
       reinitButton = new JButton("Revert to Initial Results");
       reinitButton.setName("Reinit");
       reinitButton.addActionListener(lm);
       reinitButton.setVerticalTextPosition(JButton.CENTER);
       reinitButton.setHorizontalTextPosition(JButton.CENTER);
       reinitButton.setToolTipText("Click this to revert nodes positions to initial results.");        
       savePane.add(reinitButton);        
       
       doneButton = new JButton("Done");
       doneButton.setName("Done");
       doneButton.addActionListener(lm);
       doneButton.setVerticalTextPosition(JButton.CENTER);
       doneButton.setHorizontalTextPosition(JButton.CENTER);
       doneButton.setToolTipText("Click this to declare lanetracking good for bandcalling and exit.");        
       savePane.add(doneButton);
       
       //set up zoom buttons
       getContentPane().setPreferredSize(new Dimension(100, 50));        
       zoomBar = new ZoomPanel(this);
       
       JPanel toolBar = new JPanel(new BorderLayout());
       toolBar.add(zoomBar, BorderLayout.SOUTH);
       toolBar.add(savePane, BorderLayout.NORTH);
       
       getContentPane().setLayout(new BorderLayout());
       getContentPane().add(scrollPane, BorderLayout.CENTER);	
       getContentPane().add(toolBar, BorderLayout.NORTH);
       getContentPane().add(bottomText, BorderLayout.SOUTH);
		
	}
	
	/**
	 * @return handle to gelPane
	 */
	public GelPane getGelPane(){
		return g;
	}		
	/**
	 * @return handle to scrollPane
	 */
	public JScrollPane getScrollPane(){
		return scrollPane;
	}	
	/**
	 * @return handle to saveButton
	 */
	public JButton getSaveButton(){
		return saveButton;
	}
	
	/**
	 * write error message to label in outLine, and close this LaneRulerFrame
	 * @param errorText error message to write to screen
	 */
	public void displayError(String errorText){
 	   	outLine.setText(errorText);
		startFrame.setVisible(true);
		startFrame.pack();
		startFrame.repaint();
		dispose();		
		return;
	}
	
	/**
	 * called after save checking is performed, 
	 * goes back to paramFrame or close program depending on user action
	 * @param status can be LanesManager.DONE, return to ParamPane, 
	 * or LanesManager.CLOSING, close program
	 */
	public void finish(int status){
	
		if(status == LanesManager.DONE){
			setVisible(false);
			startFrame.setVisible(true);
			this.dispose();
		} else if(status == LanesManager.CLOSING){
			System.exit(0);
		}
	}
		
}
