/**@file border.c
 *\brief contains function to find the outer most lanes by 
 *first finding the outer most marker lanes.  
 *
 *These are tested for similary and roughness before accepted.
 *Correlation and roughness code are also found here, which is also used for
 *result checking.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "border.h"
#include "peaks.h"

/**brief calculates the uncentered correlation between two rows in a PIX while allowing for a offset.
 *
 *@param pixGel pix representation of the gel image
 *@param row1 the y position of the first row to be compared
 *@param row2 the y position of the second row to be compared
 *@param start the leftmost column included in the comparison
 *@param stop the rightmost column included in the comparison
 *@param maxoffset the maximum offset allowed between the two rows
 *@returns the uncentered correlation between the two specified vectors
 */
float correlation(PIX *pixGel, int row1, int row2, int start, int stop, int maxoffset) {

UINT32  *line1, *line2;
UINT32  *datas;
INT32   wpls;
INT32	i, j;
UINT8	value1, value2;
int	offset, bestoff;
INT32	cross, s1, s2;
float	cor, bestcor;

  datas = pixGetData(pixGel);
  wpls = pixGetWpl(pixGel);
  line1 = datas + row1 * wpls;
  line2 = datas + row2 * wpls;
  bestcor = -1.0;
  for (offset = -maxoffset; offset <= maxoffset; offset++) {
    cross = 0;
    s1 = 0;
    s2 = 0;
    for (i = start; i < stop; i++) {
      value1 = GET_DATA_BYTE(line1, i);
      value2 = GET_DATA_BYTE(line2, i+offset);
      cross += (INT32)value1 * (INT32)value2;
      s1 += (INT32)value1 * (INT32)value1;
      s2 += (INT32)value2 * (INT32)value2;
    }
    cor = (float)cross / sqrt((float)s1*(float)s2);
    if (cor > bestcor) {
      bestcor = cor;
      bestoff = offset;
    }
  }
  return bestcor;
}

/**\brief calculates a roughness measure in a row in a PIX, defined as the sum of the absolute difference between neighbouring points normalized by the dynamic range of the pix
 *
 *@param pixGel pix representation of the gel image
 *@param row the row to be measured
 *@param start the leftmost column included in the measure
 *@param stop the rightmost column included in the measure
 *@returns the roughness measure
 */
float roughness (PIX *pixGel, int row, int start, int stop)
{
int i;
INT32 wpl, difference;
UINT32 *data, *line;
UINT8 value1, value2;
float roughness;
  /*assume 8-bit data*/

  data = pixGetData(pixGel);
  wpl = pixGetWpl(pixGel);
  line = data + row * wpl;
  difference =0;

  /*summing up the absolute differences*/
  value1 = GET_DATA_BYTE(line, start);
    for (i = start+1; i < stop; i++) {
      value2 = value1;
      value1 = GET_DATA_BYTE(line, i);
      difference += abs(value1-value2);
    }

    roughness = (float) difference / (powf(2, pixGetDepth(pixGel)));
    return roughness;

}

/**\brief finds the outer most lanes by extrapolating from the outermost marker lanes.
 *
 *Using the generic spacing, we estimate the distance between the 
 *outermost markers.
 *A number of the first zones after the first data zone is chosen.  
 *Tone adjusted zones are not included, 
 *unless all zones considered have been tone adjusted
 *From the sum of zone projection of these zones, 
 *we find two peaks close to the estimated distance apart. 
 *The data in these two rows are check for correlation with each other (with translational offset allowed) and with their respective neighbour marker lanes.
 *The data is also checked for roughness.
 *The results will be stored in GelInfo->LowYBorder and  GelInfo->HighYBorder
 *
 *@param FLAGOUT_FILE writable file handle to data dump 
 *@param pixGel pix representation of the gel image
 *@param ZoneProjection points to 1D y-projection in each of the zone
 *@param GelInfo holds various gel parameters
 *@param AnaCutoff holds various analysis cutoff
 *@param ToneAdjusted a vector stating if a zone has been tone adjusted
 *@param Verbose turn on text output to screen
 */
void findBorder (FILE *FLAGOUT_FILE, PIX *pixGel, INT32 **ZoneProjection, struct GelInfo *GelInfo, struct AnaCutoff *AnaCutoff, int* ToneAdjusted, int Verbose) {

float	*pcentroid, *psize;
float	sigma;
int	length, loch, hich, ifwhm, maxpk, npeaks;
int	i, j, k;
int	startzone, stopzone, nzones, startcor, stopcor, maxoffset;
INT32	*Projection;
int	BorderTolerance, MarkerTolerance, Nleftover;	
float	currentcor, bestcor, expectedDistance, expectedMarkerSpacing, curtopcor, topcor, curbottomcor, bottomcor;
float   toproughness, bottomroughness;
float   besttopcor, bestbottomcor, besttoproughness, bestbottomroughness;

 int zone_included=0;

 /*setting peak founding parameters*/
  maxpk = GelInfo->NumberOfLanes*2;
  pcentroid = calloc(maxpk, sizeof(float));
  psize = calloc(maxpk, sizeof(float));
  length = GelInfo->ImageHeight;
  loch = (int)(GelInfo->LaneSpacing/2.0 + 0.5);
  hich = length - loch - 1;
  ifwhm = (int)(GelInfo->LaneSpacing * FWHM_FACTOR + 0.5);
  sigma = SIGMA_CUTOFF;

  nzones = GelInfo->NumberOfZones / ZONE_PROJECTION_FACTOR;
  if (nzones == 0) nzones = 1;
  startzone = GelInfo->FirstDataZone+1 - (nzones) / 2; /*only use zones close tofront of gel*/

  stopzone = startzone+nzones;
  if (startzone < GelInfo->FirstDataZone) startzone = GelInfo->FirstDataZone;
  if (stopzone >= GelInfo-> NumberOfZones) stopzone = GelInfo->NumberOfZones-1;

  /*summing up ZoneProjection of zones near front of the gel*/
  Projection = calloc(GelInfo->ImageHeight, sizeof(INT32));
  for (i = startzone; i < startzone + nzones; i++) {
    if (ToneAdjusted[i] == 0) /*use only if not tone adjusted*/
      {	
	for (j = 0; j < GelInfo->ImageHeight; j++) {
	  Projection[j] += ZoneProjection[i][j];
	  zone_included++;
	}
      }
  }/*summing all values along a line within zones that we are using, only if the zone is not tone adjusted (i.e. zones with low data density)*/
  if (zone_included ==0) /*all zones in range are tone adjusted, then  use all zones in range*/
    {
      for (i = startzone; i < startzone + nzones; i++) 
	{	
	  for (j = 0; j < GelInfo->ImageHeight; j++) {
	    Projection[j] += ZoneProjection[i][j];
	    zone_included++;
	  }
	}
    }

  /*find peaks in summed projection*/
  npeaks = specFindPeak(Projection, length, pcentroid, psize, loch, hich, ifwhm, sigma, maxpk);

  /*estimating the distance between the two outermost markers and two neighbouring markers*/
  BorderTolerance = (int) (GelInfo->LaneSpacing * (float) GelInfo->MarkerPeriod * BORDERTOLERANCEFACTOR);
  MarkerTolerance = (int) (GelInfo->LaneSpacing *  BORDERTOLERANCEFACTOR);
  expectedDistance = (float)(GelInfo->NumberOfMarkers-1)* GelInfo->LaneSpacing * (float)GelInfo->MarkerPeriod; /*between first and last marker*/
  expectedMarkerSpacing = GelInfo->LaneSpacing * (float)GelInfo->MarkerPeriod;

  bestcor = AnaCutoff->border_cor_hardcutoff;
  startcor = GelInfo->StopPreData;
  stopcor = startcor + ((GelInfo->StopData - startcor)/ZONE_PROJECTION_FACTOR);
  maxoffset = GelInfo->ImageWidth / OFFSETFACTOR;
  for (i = 0; i < npeaks - 1; i++) {
    for (j = i+1; j < npeaks; j++) {
      if (fabs(pcentroid[j]-pcentroid[i]-expectedDistance) < BorderTolerance) /*for two peaks within tolerance to the expectedDistance*/ {
        currentcor = correlation(pixGel, (int)(pcentroid[i]+0.5), (int)(pcentroid[j]+0.5), startcor, stopcor, maxoffset);	
        


	/*take the two lines that have values along X most alike to be the marker lanes*/
	if (currentcor > bestcor) {
	  /*check roughness of candidates*/
	  toproughness = roughness(pixGel, (int)(pcentroid[i]+0.5), startcor, stopcor);
	  bottomroughness = roughness(pixGel, (int)(pcentroid[j]+0.5), startcor, stopcor);
	  
	  /*check neighbour correlation of candidates*/
	  topcor=0;
	  bottomcor=0;
	  for(k=i+1; k<j;k++){
	    if (fabs(pcentroid[k]-pcentroid[i]-expectedMarkerSpacing)<(MarkerTolerance)){
	      curtopcor = correlation(pixGel, (int)(pcentroid[i]+0.5), (int)(pcentroid[k]+0.5), startcor, stopcor, maxoffset);
	      if (curtopcor > topcor) topcor= curtopcor; 
	    }
	  }
	  for(k=j-1; k>i; k--){
	    if (fabs(pcentroid[j]-pcentroid[k]-expectedMarkerSpacing)<(MarkerTolerance)){
	      curbottomcor = correlation(pixGel, (int)(pcentroid[k]+0.5), (int)(pcentroid[j]+0.5), startcor, stopcor, maxoffset);
	      if (curbottomcor > bottomcor) bottomcor= curbottomcor;
	    }
	  }

	  /*consider only if candidates both pass roughness and neighbour correlation tests*/
	  if (topcor>AnaCutoff->border_neighbour_cor_cutoff && bottomcor>AnaCutoff->border_neighbour_cor_cutoff && toproughness > AnaCutoff->border_roughness_cutoff && bottomroughness > AnaCutoff->border_roughness_cutoff){

	    bestcor = currentcor;
	    besttopcor = topcor;
	    bestbottomcor = bottomcor;
	    besttoproughness = toproughness;
	    bestbottomroughness = bottomroughness;

	    GelInfo->LowYBorder = (int)(pcentroid[i] + 0.5);
	    GelInfo->HighYBorder = (int)(pcentroid[j] + 0.5);
	  }
	}
      }
    }
  }
  /*convert results to be for first and last lanes, only if cor is higher than cutoff*/
  if (bestcor > AnaCutoff->border_cor_hardcutoff) {
    GelInfo->LowYBorder -= (int)(GelInfo->LaneSpacing * (float)GelInfo->MarkerOffset + 0.5);
    Nleftover = GelInfo->NumberOfLanes - ((GelInfo->NumberOfMarkers-1)*GelInfo->MarkerPeriod) - 1 - GelInfo->MarkerOffset;
    GelInfo->HighYBorder += (int)(GelInfo->LaneSpacing * (float)Nleftover + 0.5);
    if (Verbose) {
      fprintf(stderr, "Generic position of first and last lanes: %d %d\n",GelInfo->LowYBorder, GelInfo->HighYBorder);
      if (bestcor < AnaCutoff->border_cor_softcutoff) fprintf(stderr, "Warning correlation between first and last markers is rather low: %1.4f\n",bestcor);
      else fprintf(stderr, "Correlation between first and last lanes: %1.4f\n",bestcor);
      fprintf(stderr, "Neighbour marker correlation of first and last lanes: %1.4f %1.4f\n", besttopcor, bestbottomcor);
      fprintf(stderr, "Roughness of first and last lanes: %1.4f %1.4f\n",besttoproughness, bestbottomroughness);
    }
  }
  
  fprintf(FLAGOUT_FILE, "Generic position of first and last lanes:\t%d\t%d\n",GelInfo->LowYBorder, GelInfo->HighYBorder);
  fprintf(FLAGOUT_FILE, "Correlation between first and last lanes:\t%1.4f\n",bestcor);
  fprintf(FLAGOUT_FILE, "Neighbour marker correlation of first and last lanes:\t%1.4f\t%1.4f\n", besttopcor, bestbottomcor);
  fprintf(FLAGOUT_FILE, "Roughness of first and last lanes:\t%1.4f\t%1.4f\n",besttoproughness, bestbottomroughness);

  free(Projection);
  free(pcentroid);
  free(psize);
}
