/**@file findlanes.c
 *\brief file holding the main for findlanes
 *
 *Lane tracking module of the LaneRuler software
 *Includes lane tracking and results checking, flagging gels 
 *for having areas anormalies that may interfere with lane tracking
 */

#include <stdio.h>
#include <stdlib.h> 

#ifdef X86_LINUX_ENV
#include <getopt.h>
#else
#include <unistd.h>
#endif

#include <string.h>
#include "findlanes.h"

/** \brief display function help*/
static void usage()
{
  fprintf(stderr, "FINDLANES program usage:\n\n");
  fprintf(stderr, "Gel format options:\n");
  fprintf(stderr, "  --gelspan [GELSPANFACTOR]     Minimum ratio of gel width (across lane dimension) to the image size [0,1] (default=%.2f)\n",DEFAULT_GELSPANFACTOR);
  fprintf(stderr, "  --markeroffset [MARKEROFFSET]      Location of first marker lane (default=%d)\n",DEFAULT_MARKEROFFSET);
  fprintf(stderr, "  --lastmarkeroffset [LASTMARKEROFFSET]  Offset of last marker lane from the last lane, the location used will be \n                                                        the last valid marker position defined by marker period that is not pass this point (default=%d)\n",DEFAULT_LASTMARKEROFFSET);
  fprintf(stderr, "  --markerperiod [MARKERPERIOD]      Period between marker lanes (default=%d lanes)\n",DEFAULT_MARKERPERIOD);
  fprintf(stderr, "  --numlanes [NUMLANES]              Number of lanes (default=%d lanes)\n",DEFAULT_NLANES);
  fprintf(stderr, "  --numzones [NUMZONES]              Number of zones (default=%d zones)\n",DEFAULT_NZONES);
  fprintf(stderr, "  --orientation [WELL_EDGE]          Edge closest to the wells of the image, defining orientation 0=LEFT, 1=TOP 2=RIGHT 3=BOTTOM (default=%i)\n",DEFAULT_WELL_EDGE);

  fprintf(stderr, "\nInput file structure options:\n");
  fprintf(stderr, "  --projectdir [PROJECTDIR], -p [PROJECTDIR]   Project directory (default=%s)\n",DEFAULT_PROJECTDIR);
  fprintf(stderr, "  --imagepath  [GELIMAGE]            Name of gel image (default=%s)\n",DEFAULT_GELNAME);
  fprintf(stderr, "  --readlanes [LANESFILE]            Use lanes file to generate traces16, no lane tracking is done if used. (default file name=%s)\n", DEFAULT_IN_LANES_NAME);

  fprintf(stderr, "\nOutput file structure options:\n");
  fprintf(stderr, "  --tracespath [TRACESNAME]          Name of the traces file (default=%s)\n",DEFAULT_TRACES_NAME);
  fprintf(stderr, "  --lanespath  [LANESNAME]           Name of the lanes file (default=%s)\n",DEFAULT_LANES_NAME);

  fprintf(stderr, "\nProgram control options\n");  
  fprintf(stderr, "  --noadjust, -a                     Turn off tone adjusting in low contrast zones (default=tone adjusting on)\n"); 
  fprintf(stderr, "  --nosmooth, -s                     Turn off lane smoothing (default=smoothing on)\n"); 
  fprintf(stderr, "  --matchall  -i                     Switch to match peaks to all lanes even in zone with marker frequency dominance (default=match only marker lanes in marker dominated zones)\n");   
  fprintf(stderr, "  --nogrowoff,                       Turn of no grow and contamination detection\n");
  fprintf(stderr, "  --diagnostic [DIAGNOSTIC], -d [DIAGNOSTIC] Name of the output diagnostic image (png format)\n");
  fprintf(stderr, "  --outimage [OUTIMAGE_PATH], -o [OUTIMAGE_PATH] Name of the output lane straightened image (png format)\n");
  fprintf(stderr, "  --Verbose, -v                      Select verbose mode, information output to stderr (default = nonverbose)\n");
  fprintf(stderr, "  --help                             Show this usage information\n");

  fprintf(stderr, "\nAnalysis parameters:\n");  
  fprintf(stderr, "  --morphelement [ELEMENT]           Size of the morphological structural element (default=image width/%d)\n",DEFAULT_SIZEFACTOR);
  fprintf(stderr, "  --peaksizefactor [PEAKCUTOFF]      Minimum fraction of maximum peak size in the zone for a peak to be used (range [0,1]) (default=%f)\n",DEFAULT_MINPSIZEFACTOR);
  fprintf(stderr, "  --borderroughness [ROUGHNESS]      Minimum roughness of border marker lanes (default=%f)\n",DEFAULT_BORDER_ROUGHNESS_CUTOFF);
  fprintf(stderr, "  --bordercorrelation [CORRELATION]      Minimum correlation of border marker lanes to each other (default=%f)\n",DEFAULT_BORDER_COR_SOFTCUTOFF);
  fprintf(stderr, "  --borderneighbourcorrelation [CORRELATION]      Minimum correlation of border marker lanes to its neighbour (default=%f)\n",DEFAULT_BORDER_NEIGHBOUR_COR_CUTOFF);

  fprintf(stderr, "\nResults checking parameters:\n"); 
  fprintf(stderr, "  --anasegments [NUMSEGMENTS]        Number of segments in which the result checking is done. (default=%i)\n",DEFAULT_MARKERCORSEGMENTS);
  fprintf(stderr, "  --markerroughness [ROUGHNESS]      Minimum roughness of marker lanes within any segments (default=%f)\n",DEFAULT_MARKER_ROUGHNESS_CUTOFF);
  fprintf(stderr, "  --markercorrelation[CORRELATION]   Minimum correlation between neighbouring marker lanes within any segments (default=%f)\n",DEFAULT_MARKER_CORRELATION_CUTOFF);   
  fprintf(stderr, "  --totalintdata [INTERPOLATED LANES] Maximum number of interpolated lanes allowed in a zone (default=%i)\n",DEFAULT_INTERPOLATED_DATA_CUTOFF);
  fprintf(stderr, "  --totalintmarker [INTERPOLATED MARKERS] Maximum number of interpolated markers allowed in a zone (default=%i)\n",DEFAULT_INTERPOLATED_MARKER_CUTOFF); 
  fprintf(stderr, "  --consecintdata [INTERPOLATED LANES] Maximum number of consecutive interpolated lanes allowed in a zone (default=%i)\n",DEFAULT_CON_INTERPOLATED_DATA_CUTOFF);  
  fprintf(stderr, "  --consecintmarker [INTERPOLATED MARKERS] Maximum number of consecutive interpolated markers allowed in a zone (default=%i)\n",DEFAULT_CON_INTERPOLATED_MARKER_CUTOFF);
  fprintf(stderr, "  --lanespacecheck [LANESPACEFACTOR] Fraction of the generic lane space used as tolerance for actual lane spacings(default=%f)\n",DEFAULT_LANESPACINGCHECK_FACTOR);
  fprintf(stderr, "  --numoutliers [OUTLIER NUMBER] Maximum number of outliers allowed in the gel (default=%i)\n",DEFAULT_NOUTLIER_CUTOFF);
  fprintf(stderr, "  --asymwhisker [WHISKER FACTOR]  Factor used to calculate cutoff for asymmetry within each zone (default=%f)\n",DEFAULT_ASYM_WHISKER_FACTOR);
  fprintf(stderr, "  --consecasymsoft [CONSECUTIVE ASYMMETRY CUTOFF] Number of consecutive asymmetric lanes to warrant adjustment (default=%i)\n",DEFAULT_CON_ASYM_HARD_CUTOFF);
  fprintf(stderr, "  --consecasymhard [CONSECUTIVE ASYMMETRY CUTOFF] Number of consecutive asymmetric lanes to fail gel (default=%i)\n",DEFAULT_CON_ASYM_HARD_CUTOFF);
  fprintf(stderr, "  --slopethreshold [SLOPE_THRESHOLD] Threshold used to determine the thresholded roughness.  (default=%i)\n",DEFAULT_SLOPE_THRESHOLD); 
  fprintf(stderr, "  --noisetoFGratio [RATIO] ratio of noise compared to signal for a point to be identified to have noise (default=%f)\n",DEFAULT_NOISE_TO_FG_RATIO);
  fprintf(stderr, "  --minfg [INTENSITY] minimum intensity for the noise to fg ratio to be considered (default=%i)\n",DEFAULT_MIN_FG);
  fprintf(stderr, "  --noisespanfactor [FLOAT] Amount of noise as multiple of lane width identified in a lane to be qualified as noisy (default=%f)\n",DEFAULT_NOISE_SPAN_FACTOR);

  fprintf(stderr, "\n\n");
  exit(0);
}

/**\brief main function for findlanes*/
main(int argc, char **argv) {

char	fileGelImage[FILE_NAME_LEN];
char	fileDiagnosticImage[FILE_NAME_LEN];
char	fileFlagout_Name[FILE_NAME_LEN];
char	fileErrorout_Name[FILE_NAME_LEN];
char    fileNogrow_Name[FILE_NAME_LEN];
char    fileLIMS_Name[FILE_NAME_LEN];
char    tracespath[FILE_NAME_LEN];
 char    outimgpath[FILE_NAME_LEN];
char    lanespath[FILE_NAME_LEN];
char    inlanespath[FILE_NAME_LEN];
char    tempstring[FILE_NAME_LEN];
char	c;
int	itemp, index;
int	Diagnostic_present = 0;
int	Verbose = 0;
int     well_edge;
int	Smoothing = 1;
int     Markerinc = 1;
float   minpsizefactor, ftemp;
PIX	*pixGelRaw, *pixGel8bits, *pixGelFilter, *pixGelFilter1, *pixGelClose, *pixAdj, *pixNotCloseAdj;
PIX	*pixOutRGB;
struct GelInfo *GelInfo = (struct GelInfo *)malloc (sizeof(struct GelInfo));
struct AnaCutoff *AnaCutoff = (struct AnaCutoff *)malloc (sizeof(struct AnaCutoff));
INT32	**ZoneProjection;
int	*ZoneID, *ToneAdjusted;
float	*Spacing, *NodePosition;
float	**PeakPosition, **MarkerCorrelation, **MarkerRoughness, **Asym, **Laneoffset;
int Adjust = 1;
int NoTrack = 0;
int CheckNoGrow = 1;
 int OutImage = 0;

int num_outliers, zoneID, num_corrected_asym;
int *num_interpolated_markers, *num_interpolated_data, *con_interpolated_markers, *con_interpolated_data, *con_asym_lanes;
int nummarker, markercorsegments;
FILE *FLAGOUT_FILE;
FILE *ERROROUT_FILE;
FILE *NOGROW_FILE;
FILE *LIMS_FILE;


/*see findlanes.h for default values*/
 GelInfo->NumberOfLanes = DEFAULT_NLANES;
 GelInfo->MarkerPeriod = DEFAULT_MARKERPERIOD;
 GelInfo->MarkerOffset = DEFAULT_MARKEROFFSET;
 GelInfo->LastMarkerOffset = DEFAULT_LASTMARKEROFFSET;
 GelInfo->NumberOfZones = DEFAULT_NZONES;
 GelInfo->GelSpanFactor = DEFAULT_GELSPANFACTOR;
 GelInfo->SizeOfStructuralElement = 0;
 GelInfo->AsymSpan=0;
 GelInfo->NumLightMarker=0;
 GelInfo->NumInterpMarkerSpan=0;
 GelInfo->maxConsecNoise=0;
 GelInfo->GelPassed = 1;
 strcpy(GelInfo->ProjectDir, DEFAULT_PROJECTDIR);
 strcpy(fileGelImage,DEFAULT_GELNAME);
 strcpy(tracespath,DEFAULT_TRACES_NAME);
 strcpy(outimgpath,DEFAULT_OUTIMG_NAME);
 strcpy(lanespath,DEFAULT_LANES_NAME);
 strcpy(inlanespath,DEFAULT_IN_LANES_NAME);


 strcpy(fileFlagout_Name, DEFAULT_FLAGOUT_NAME);
 strcpy(fileErrorout_Name, DEFAULT_ERROROUT_NAME);
 strcpy(fileNogrow_Name, DEFAULT_NOGROW_NAME);
 strcpy(fileLIMS_Name, DEFAULT_LIMS_NAME);

 minpsizefactor = DEFAULT_MINPSIZEFACTOR;
 well_edge = DEFAULT_WELL_EDGE;
 
 /*see anacutoff.h for default cutoff values*/
 AnaCutoff->border_cor_hardcutoff=DEFAULT_BORDER_COR_HARDCUTOFF;
 AnaCutoff->border_cor_softcutoff=DEFAULT_BORDER_COR_SOFTCUTOFF;
 AnaCutoff->border_neighbour_cor_cutoff=DEFAULT_BORDER_NEIGHBOUR_COR_CUTOFF;
 AnaCutoff->border_roughness_cutoff=DEFAULT_BORDER_ROUGHNESS_CUTOFF;

 AnaCutoff->lanespacingcheck_factor =  DEFAULT_LANESPACINGCHECK_FACTOR;
 AnaCutoff->marker_roughness_cutoff = DEFAULT_MARKER_ROUGHNESS_CUTOFF;
 AnaCutoff->marker_correlation_cutoff = DEFAULT_MARKER_CORRELATION_CUTOFF;
 AnaCutoff->interpolated_data_cutoff = DEFAULT_INTERPOLATED_DATA_CUTOFF;
 AnaCutoff->interpolated_marker_cutoff = DEFAULT_INTERPOLATED_MARKER_CUTOFF;
 AnaCutoff->consec_data_cutoff = DEFAULT_CON_INTERPOLATED_DATA_CUTOFF;
 AnaCutoff->consec_marker_cutoff = DEFAULT_CON_INTERPOLATED_MARKER_CUTOFF;
 AnaCutoff->markercorsegments = DEFAULT_MARKERCORSEGMENTS;
 AnaCutoff->noutlier_cutoff = DEFAULT_NOUTLIER_CUTOFF;

 AnaCutoff->asym_window_factor = DEFAULT_ASYM_WINDOW_FACTOR;
 AnaCutoff->asym_whisker_factor = DEFAULT_ASYM_WHISKER_FACTOR;
 AnaCutoff->consec_asym_soft_cutoff = DEFAULT_CON_ASYM_SOFT_CUTOFF;
 AnaCutoff->consec_asym_hard_cutoff = DEFAULT_CON_ASYM_HARD_CUTOFF;
 AnaCutoff->asym_threshold = DEFAULT_ASYM_THRESHOLD;
 AnaCutoff->asym_max_offset = DEFAULT_ASYM_MAX_OFFSET; 

 AnaCutoff->slope_threshold = DEFAULT_SLOPE_THRESHOLD;
 AnaCutoff-> match_maxoffset_factor = DEFAULT_MATCH_MAXOFFSET_FACTOR;
 AnaCutoff-> match_min_roughness_factor = DEFAULT_MATCH_MIN_ROUGHNESS_FACTOR;
 AnaCutoff-> match_min_valid_threshold = DEFAULT_MATCH_MIN_VALID_THRESHOLD;
 AnaCutoff-> match_contam_cutoff = DEFAULT_MATCH_CONTAM_CUTOFF;
 AnaCutoff-> non_recom_thresrough_factor = DEFAULT_NON_RECOM_THRESROUGH_FACTOR;
 AnaCutoff-> non_recom_maxdiff = DEFAULT_NON_RECOM_MAXDIFF;

 AnaCutoff->noise_to_FG_ratio = DEFAULT_NOISE_TO_FG_RATIO;
 AnaCutoff->min_FG = DEFAULT_MIN_FG;
 AnaCutoff->noise_span_factor = DEFAULT_NOISE_SPAN_FACTOR;
 
 /*setting long option list*/
 static struct option long_options[] =
   {
     /*gel format options*/
     {"gelspan", 1, 0, 0},  
     {"markeroffset", 1, 0, 0},
     {"lastmarkeroffset", 1, 0, 0},
     {"markerperiod", 1, 0, 0},
     {"numlanes", 1, 0, 0},
     {"numzones", 1, 0, 0},
     {"orientation", 1, 0, 0}, 
     
     /*input file structure options*/
     {"projectdir", 1, 0,'p'},
     {"imagepath", 1, 0, 0},
     {"readlanes", 1, 0, 0},
      
     /*Output file structure options*/
     {"tracespath", 1, 0, 0},
     {"lanespath", 1, 0, 0},

     /*Program control options*/
     {"noadjust", 0, 0, 'a'},
     {"nosmooth", 0, 0, 's'}, 
     {"matchall", 0, 0, 'i'}, 
     {"nogrowoff", 0, 0, 0},     
     {"diagnostic", 1, 0,'d'},
     {"outimage", 1, 0,'o'},
     {"Verbose", 0, 0, 'v'},
     {"help", 0, 0, 'h'},
     
     /*Analysis parameters:*/
     {"morphelement", 1, 0, 0},  
     {"peaksizefactor", 1, 0, 0},  
     {"borderroughness", 1, 0, 0},    
     {"bordercorrelation", 1, 0, 0},
     {"borderneighbourcorrelation", 1, 0, 0},
     
     /*Result checking parameters:*/
     {"anasegments", 1, 0, 0},  
     {"markerroughness",1, 0, 0},    
     {"markercorrelation",1, 0, 0},  
     {"totalintdata",1, 0, 0},     
     {"totalintmarker",1, 0, 0},   
     {"consecintdata",1, 0, 0},     
     {"consecintmarker",1, 0, 0},     
     {"numoutliers",1, 0, 0}, 
     {"lanespacecheck",1, 0, 0},     
 
     {"asymwhisker",1, 0, 0},
     {"consecasymsoft",1, 0, 0},
     {"consecasymhard",1, 0, 0},

     {"slopethreshold",1, 0, 0},

     {"noisetoFGratio",1, 0, 0},
     {"minfg",1, 0, 0},
     {"noisespanfactor",1, 0, 0},

     {0, 0, 0, 0}
   };

 /* getopt_long stores the option index here.   */
 int option_index = 0;
 
 
 /*parsing command line input*/
 
 while ((c = getopt_long(argc, argv, "ahivsp:d:o:", long_options, &option_index))!=-1)
   {
     
     switch (c)
       {
       case 0:
	 /* If this option set a flag, do nothing else now.   */
	 if (long_options[option_index].flag != 0)
	   break;
	 
	 if (!strcmp(long_options[option_index].name, "gelspan"))
	   {
	     ftemp = atof(optarg);
	     if (ftemp >= 0 && ftemp <= 1) 
	       {
		 GelInfo->GelSpanFactor = ftemp;
	       }
	     else
	       {
		 fprintf(stderr, "ERROR: invalid gel span: must be between 0 and 1.\n");
		 exit(1);
	       }
	    }	 
	 if (!strcmp(long_options[option_index].name, "markeroffset"))
	   {
	     itemp = atoi(optarg);
	     if (itemp >= 0) 
	       {
		 GelInfo->MarkerOffset = itemp;
	       }
	     else
	       {
		 fprintf(stderr, "ERROR: invalid marker offset: marker offset must be greater than 0.\n");
		 exit(1);
	       }
	    }
	  if (!strcmp(long_options[option_index].name, "lastmarkeroffset"))
	   {
	     itemp = atoi(optarg);
	     if (itemp >= 0) 
	       {
		 GelInfo->LastMarkerOffset = itemp;
	       }
	     else
	       {
		 fprintf(stderr, "ERROR: invalid last marker offset: marker offset must be greater than 0.\n");
		 exit(1);
	       }
	    }
	  if (!strcmp(long_options[option_index].name, "markerperiod"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0) 
		{
		  GelInfo->MarkerPeriod = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid marker spacing: spacing must be greater than 0.\n");
		  exit(1);
		}
	    }
	  if (!strcmp(long_options[option_index].name, "numlanes"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0) 
		{
		  GelInfo->NumberOfLanes = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid number of lanes:  number of lanes must be greater than 0.\n");
		  exit(1);
		}
	    }
	  if (!strcmp(long_options[option_index].name, "numzones"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0) 
		{
		  GelInfo->NumberOfZones = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid number of zones:  number of zones must be greater than 0.\n");
		  exit(1);
		}
	    }
	  
	  if (!strcmp(long_options[option_index].name, "orientation"))
	    {
	      itemp = atoi(optarg);
	      if (itemp >= 0 && itemp < 4)  well_edge = itemp;
	    }
	  
	  if (!strcmp(long_options[option_index].name, "imagepath"))
	    {      
	      strcpy(fileGelImage,optarg);
	    }

	  if (!strcmp(long_options[option_index].name, "readlanes"))
	    {      
	      strcpy(inlanespath,optarg);
	      NoTrack=1;
	    }
	  
	  if (!strcmp(long_options[option_index].name, "tracespath"))
	    {      
	      strcpy(tracespath,optarg);
	    }
	  if (!strcmp(long_options[option_index].name, "lanespath"))
	    {      
	      strcpy(lanespath,optarg);
	    }
	  if (!strcmp(long_options[option_index].name, "nogrowoff"))
	    { 
	      CheckNoGrow = 0;
	    }
	  if (!strcmp(long_options[option_index].name, "morphelement"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0) 
		{
		  GelInfo->SizeOfStructuralElement = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid structural element size: must be greater than 0.\n");
		  exit(1);
		}
	    }
	  if (!strcmp(long_options[option_index].name, "peaksizefactor"))
	    {
	      ftemp = atof(optarg);
	      if (ftemp >= 0 && ftemp <= 1) 
		{
		  minpsizefactor = ftemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid peak size factors: must be in [0,1].\n");
		  exit(1);
		}
	    }
	  if (!strcmp(long_options[option_index].name, "borderroughness"))
	    {
	      ftemp = atof(optarg);
	      if (ftemp >=0) 
		{
		  AnaCutoff->border_roughness_cutoff = ftemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid border roughness cutoff: must be greater or equal to 0.\n");
		  exit(1);
		}
	    }
	  if (!strcmp(long_options[option_index].name, "bordercorrelation"))
	    {
	      ftemp = atof(optarg);
	      if (ftemp >= 0 && ftemp <= 1) 
		{
		   AnaCutoff->border_cor_hardcutoff = ftemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid correlation: must be in [0,1].\n");
		  exit(1);
		}
	    }
 	  if (!strcmp(long_options[option_index].name, "borderneighbourcorrelation"))
	    {
	      ftemp = atof(optarg);
	      if (ftemp >= 0 && ftemp <= 1) 
		{
		   AnaCutoff->border_neighbour_cor_cutoff = ftemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid correlation: must be in [0,1].\n");
		  exit(1);
		}
	    }

	  if (!strcmp(long_options[option_index].name, "anasegments"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0) 
		{
		  AnaCutoff->markercorsegments = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid analysis segments: must be greater than 0.\n");
		  exit(1);
		}
	    }
	  if (!strcmp(long_options[option_index].name, "markerroughness"))
	    {
	      ftemp = atof(optarg);
	      if (ftemp >=0) 
		{
		  AnaCutoff->marker_roughness_cutoff = ftemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid marker roughness cutoff: must be greater or equal to 0.\n");
		  exit(1);
		}
	    }
	  if (!strcmp(long_options[option_index].name, "markercorrelation"))
	    {
	      ftemp = atof(optarg);
	      if (ftemp >= 0 && ftemp <= 1) 
		{
		  AnaCutoff->marker_correlation_cutoff = ftemp;
		}
	      else
		{	
		  fprintf(stderr, "ERROR: invalid peak size factor: must be in [0,1].\n");
		  exit(1);
		}
	    }
	  if (!strcmp(long_options[option_index].name, "totalintdata"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0) 
		{
		  AnaCutoff->interpolated_data_cutoff = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid lanes cutoff: must be greater than 0.\n");
		  exit(1);
		}
	    } 
  	  if (!strcmp(long_options[option_index].name, "totalintmarker"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0) 
		{
		  AnaCutoff->interpolated_marker_cutoff = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid marker cutoff: must be greater than 0.\n");
		  exit(1);
		}
	    } 
  	  if (!strcmp(long_options[option_index].name, "consecintdata"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0) 
		{
		  AnaCutoff->consec_data_cutoff = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid lanes cutoff: must be greater than 0.\n");
		  exit(1);
		}
	    } 
  	  if (!strcmp(long_options[option_index].name, "consecintmarker"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0) 
		{
		  AnaCutoff->consec_marker_cutoff = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid marker cutoff: must be greater than 0.\n");
		  exit(1);
		}
	    }
	  if (!strcmp(long_options[option_index].name, "consecasymsoft"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0) 
		{
		  AnaCutoff->consec_asym_soft_cutoff = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid consecutive asymmetric lanes cutoff: must be greater than 0.\n");
		  exit(1);
		}
	    }
	  	  if (!strcmp(long_options[option_index].name, "consecasymhard"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0) 
		{
		  AnaCutoff->consec_asym_hard_cutoff = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid consecutive asymmetric lanes cutoff: must be greater than 0.\n");
		  exit(1);
		}
	    }
	  if (!strcmp(long_options[option_index].name, "asymwhisker"))
	    {
	      ftemp = atof(optarg);
	      if (ftemp >= 0) 
		{
		  AnaCutoff->asym_whisker_factor = ftemp;
		}
	      else
		{	
		  fprintf(stderr, "ERROR: invalid whisker factor: must be greater than or equal to 0.\n");
		  exit(1);
		}  
	    } 
	  if (!strcmp(long_options[option_index].name, "numoutliers"))
	    {
	      itemp = atoi(optarg);
	      if (itemp >= 0) 
		{
		  AnaCutoff->noutlier_cutoff = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid number of outliers: must be greater than 0.\n");
		  exit(1);
		}
	    } 

	  if (!strcmp(long_options[option_index].name, "lanespacecheck"))
	    {
	      ftemp = atof(optarg);
	      if (ftemp > 0) 
		{
		  AnaCutoff->lanespacingcheck_factor = ftemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid lanespacing check factor: must be greater than 0.\n");
		  exit(1);
		}
	    } 
	  if (!strcmp(long_options[option_index].name, "slopethreshold"))
	    {
	      itemp = atoi(optarg);
	      if (itemp >= 0) 
		{
		  AnaCutoff->slope_threshold=itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid number of slope threshold: must be greater than or equal to 0.\n");
		  exit(1);
		}
	    } 
	  if (!strcmp(long_options[option_index].name, "noisetoFGratio"))
	    {
	      ftemp = atof(optarg);
	      if (ftemp >= 0 ) 
		{
		  AnaCutoff->noise_to_FG_ratio = ftemp;
		}
	      else
		{	
		  fprintf(stderr, "ERROR: invalid noise to fg ratio: must be greater than or equal to 0.\n");
		  exit(1);
		}  
	    } 
	  if (!strcmp(long_options[option_index].name, "minfg"))
	    {
	      itemp = atoi(optarg);
	      if (itemp > 0 && itemp <= 255) 
		{
		  AnaCutoff->min_FG = itemp;
		}
	      else
		{
		  fprintf(stderr, "ERROR: invalid min foreground intensity: must be greater within (0,255].\n");
		  exit(1);
		}
	    }
	  if (!strcmp(long_options[option_index].name, "noisespanfactor"))
	    {
	      ftemp = atof(optarg);
	      if (ftemp >= 0 ) 
		{
		  AnaCutoff-> noise_span_factor = ftemp;
		}
	      else
		{	
		  fprintf(stderr, "ERROR: invalid noise span factor: must be greater than or equal to 0.\n");
		  exit(1);
		}  
	    } 
	  break;
	  
	case 'v':
	  Verbose = 1;
	  break;
	case 'a':
	  Adjust = 0;
	  break;	
	case 'h':
	  usage();
	  break;  
	case 'i':
	  Markerinc = 0;
	  break;	
	case 's':
	  Smoothing = 0;
	  break;	 
	case 'p':
	  strcpy(GelInfo->ProjectDir,optarg);
	  break;
	case 'd':
	  strcpy(fileDiagnosticImage,optarg);
	  Diagnostic_present = 1;
	  break;
	case 'o':
	  strcpy(outimgpath,optarg);
	  OutImage = 1;
	  break;
	default:
	  fprintf (stderr, "getopt() returned character code %o - not handled in code. Usage instructions follow.\n", c);
	  usage();
	  exit(1);
       }	
   }  

  GelInfo->NumberOfMarkers = (GelInfo->NumberOfLanes - GelInfo->MarkerOffset - GelInfo->LastMarkerOffset - 1)/ GelInfo->MarkerPeriod + 1;
  /*if less than 4 markers, markerneighbour not used*/
  if(GelInfo->NumberOfMarkers < 4) {
    AnaCutoff->border_neighbour_cor_cutoff = -1;
    if(Verbose) fprintf(stderr, "Only 2 marker lanes present, borderneighbourcorrelation set to -1.\n");
  }

  if (Verbose) {
    fprintf(stderr, "Project directory is %s\n",GelInfo->ProjectDir);
    fprintf(stderr, "Using the gel image %s\n", fileGelImage);
    if (Diagnostic_present)
      fprintf(stderr, "Will produce diagnostic image %s\n", fileDiagnosticImage);
    fprintf(stderr, "Expecting %d lanes, with a marker every %d lanes starting at lane %d\n",
	    GelInfo->NumberOfLanes, GelInfo->MarkerPeriod, GelInfo->MarkerOffset);
    fprintf(stderr, "The image will be analyzed in %d zones\n", GelInfo->NumberOfZones);
    fprintf(stderr, "Traces information will be stored in %s\n", tracespath);
    if(NoTrack)
      {
	fprintf(stderr, "Lanes information will taken from %s\n", inlanespath);
      }
    else
      {
	fprintf(stderr, "Lanes information will be stored in %s\n", lanespath);
      }
    fprintf(stderr, "Analysis parameters will be stored in %s\n", fileFlagout_Name);

  }

  strcat(GelInfo->ProjectDir,"/");

  /*open files for parameters output*/
  sprintf(fileFlagout_Name,"%s%s",GelInfo->ProjectDir, DEFAULT_FLAGOUT_NAME);

  if ((FLAGOUT_FILE = fopen(fileFlagout_Name,"w")) == NULL)
    {
      fprintf(stderr,"Cannot open output file at %s\n", fileFlagout_Name);
      exit(1);
    }
  /*open file for error output*/
  sprintf(fileErrorout_Name,"%s%s",GelInfo->ProjectDir, DEFAULT_ERROROUT_NAME); 

  if ((ERROROUT_FILE = fopen(fileErrorout_Name,"w")) == NULL)
    {
      fprintf(stderr,"Cannot open error output file at %s\n", fileErrorout_Name);
      exit(1);
    }

/*   output given gel parameters */
  fprintf(FLAGOUT_FILE, "Zones:\t%i\n", GelInfo->NumberOfZones);
  fprintf(FLAGOUT_FILE, "Marker Period:\t%i\n", GelInfo->MarkerPeriod);
  fprintf(FLAGOUT_FILE, "Marker Offset:\t%i\n", GelInfo->MarkerOffset);

  
  /*read and invert image*/
  if ((pixGelRaw = pixReadTIFF16Gray(FLAGOUT_FILE, fileGelImage, GelInfo, well_edge, Verbose)) == NULL) {
    fprintf(stderr, "Cannot read gel image %s\n", fileGelImage);
    exit(1);
  }
  
  if (!GelInfo->SizeOfStructuralElement) /*for morphological closing*/
    GelInfo->SizeOfStructuralElement = 2*(GelInfo->ImageWidth/(2*DEFAULT_SIZEFACTOR))+1;

    if (Verbose)
      fprintf(stderr, "Size of structural element for morphological analysis %d\n",GelInfo->SizeOfStructuralElement);
    fprintf(FLAGOUT_FILE, "Morphological Element Size:\t%i\n", GelInfo->SizeOfStructuralElement);
 
    pixGel8bits = pixMapTo8bits(FLAGOUT_FILE, pixGelRaw, GelInfo, Verbose); /*linearly fit image to 8 bit*/
    
    pixGelFilter1 = pixMedianFilter(pixGel8bits, GelInfo); /*3x3 median filter and 9x1 median filter*/
    
    pixGelFilter = pixMedianFilter(pixGelFilter1, GelInfo); /*3x3 median filter and 9x1 median filter*/

    pixGelClose = pixCloseGray(pixGelFilter, GelInfo->SizeOfStructuralElement, HORIZ); /*morph close in horizontal direction*/

    /*initialize arrays*/
    ZoneProjection = (INT32 **)malloc(GelInfo->NumberOfZones * sizeof(INT32*));
    for (index = 0; index < GelInfo->NumberOfZones; index++) 
      ZoneProjection[index] = calloc(GelInfo->ImageHeight, sizeof(INT32)); /*ZoneProjection[zone][yval], the sum of all data in each zone on a line*/

    ZoneID = (int *)malloc(GelInfo->ImageWidth * sizeof(int));
    NodePosition = (float *)calloc(GelInfo->NumberOfZones+2, sizeof(float)); /*NodePosition is X center of mass of data values of each zone, note that NodePosition[0]=0, NodePosition[NumberOfZones+1]=imagewidth-1*/
    Spacing = (float *)calloc(GelInfo->NumberOfZones, sizeof(float)); 
    PeakPosition = (float **)malloc((GelInfo->NumberOfZones+2) * sizeof(float*));
    for (index = 0; index < GelInfo->NumberOfZones+2; index++) 
      PeakPosition[index] = calloc(GelInfo->NumberOfLanes, sizeof(float));
    Asym = (float **)malloc((GelInfo->NumberOfZones) * sizeof(float*));
    for (index = 0; index < GelInfo->NumberOfZones; index++) 
      Asym[index] = calloc(GelInfo->NumberOfLanes, sizeof(float));
    Laneoffset = (float **)malloc((GelInfo->NumberOfZones) * sizeof(float*));
    for (index = 0; index < GelInfo->NumberOfZones; index++) 
      Laneoffset[index] = calloc(GelInfo->NumberOfLanes, sizeof(float));
    con_asym_lanes =  (int *)calloc(GelInfo->NumberOfZones, sizeof(int)); 

    /*can read in lanes file*/
    if (NoTrack) /*if we need to read lanes file*/
      { 

	fprintf(FLAGOUT_FILE, "%s used for generating traces16\n", inlanespath);
	/*read given lanes file*/
	readLanesFile(PeakPosition, inlanespath,  NodePosition, GelInfo);

	GelInfo->LaneSpacing = averageLaneSpacing(PeakPosition, GelInfo);
 
	/*write curvature information to data dump*/
	writeCurvature(FLAGOUT_FILE, PeakPosition, NodePosition, GelInfo);
	fprintf(FLAGOUT_FILE, "No interpolation check performed.\n");
	
	/*write diagnostic image if requested*/
	if (Diagnostic_present) {
	  int zone=0;
	  /*estimate zoneID with nodeposition*/
	  for(itemp = 0; itemp<GelInfo->ImageWidth; itemp++){
	    ZoneID[itemp]=zone-1;
	    if(itemp > NodePosition[zone]){
	      zone++;
	      if(zone > GelInfo->NumberOfZones+1){
		zone=GelInfo->NumberOfZones+1;
	      }
	    }
	  }
	  outputDiagnostic(pixGelFilter, fileDiagnosticImage, PeakPosition, NodePosition, ZoneID, GelInfo); 
	}

	/*symmetry checks*/
	checkSymmetry(ERROROUT_FILE, FLAGOUT_FILE, pixGelFilter, ZoneID, PeakPosition, NodePosition, GelInfo, AnaCutoff, Asym, Laneoffset, con_asym_lanes, Verbose);
      }
    else /*track the lanes*/
      {
	/*work out zones to get tone adjustment*/
	initZones(FLAGOUT_FILE, pixGelClose, GelInfo, ZoneID, Verbose);
	projectZones(pixGelClose, ZoneProjection, GelInfo, ZoneID);  
	
	ToneAdjusted = calloc(GelInfo->NumberOfZones, sizeof(int));
	/*enhance zones with low contrast data*/  
	if(Adjust)   
	  {
	pixAdj = adjustTone(pixGelClose, ZoneProjection, GelInfo, ZoneID, ToneAdjusted, Verbose);
	
	fprintf(FLAGOUT_FILE, "Tone Adjusted Zone:");
	for (zoneID = 0; zoneID < GelInfo->NumberOfZones; zoneID++)
	  {
	    if (ToneAdjusted[zoneID] == 1) fprintf(FLAGOUT_FILE, "\t%i", zoneID);
	  }
	fprintf(FLAGOUT_FILE, "\n");
	  }
	
	GelInfo->LaneSpacing = findSpacing(FLAGOUT_FILE, ZoneProjection, GelInfo, Spacing, Verbose);/*find overall lane Spacing, the minimum of all zones.  also stores marker spacing in Spacing*/
	
	if (GelInfo->LaneSpacing < 0) {
	  fprintf (stderr, "Could not find lane spacing with confidence...\n");
	  if (Verbose) fprintf(stderr, "Gel FAILED, see %s for description\n",fileErrorout_Name); 
	  fprintf (ERROROUT_FILE,"Could not find lane spacing with confidence...\n");    
	  fprintf (FLAGOUT_FILE, "FAILED\n");
	  
	  sprintf(fileLIMS_Name,"%s%s",GelInfo->ProjectDir, DEFAULT_LIMS_NAME);
	  
	  if ((LIMS_FILE = fopen(fileLIMS_Name,"w")) == NULL)
	    {
	      fprintf(stderr,"Cannot open no grow output file at %s\n", fileLIMS_Name);
	      exit(1);
	}
	  fprintf(LIMS_FILE, "NumInterpMarkerSpan:-1\n");
	  fprintf(LIMS_FILE, "AsymSpan:-1\n");
	  fprintf(LIMS_FILE, "NumLightMarkers:-1\n");
	  fprintf(LIMS_FILE, "LRpassed:-1\n");
	  
	  fclose(ERROROUT_FILE);
	  fclose(FLAGOUT_FILE);
	  fclose(LIMS_FILE);
	  exit(1);
	}

	/*set node positions based on zone division*/
	if(Adjust)   
	  {
	    setNodePositions (pixAdj, GelInfo, ZoneID, NodePosition, Verbose);
	  }
	else {  
	  setNodePositions (pixGelFilter, GelInfo, ZoneID, NodePosition, Verbose);
	}
	
	/*find the outermost lanes*/

	findBorder(FLAGOUT_FILE, pixGelFilter, ZoneProjection, GelInfo, AnaCutoff, ToneAdjusted, Verbose); /*finds first and last lanes (not necessarily marker lanes*/
	if (GelInfo->LowYBorder * GelInfo->HighYBorder == 0) {
	  fprintf (stderr, "Could not find outside marker lanes with confidence...\n");
	  if (Verbose) fprintf(stderr, "Gel FAILED, see %s for description\n",fileErrorout_Name);
	  fprintf (ERROROUT_FILE,"Could not find outside marker lanes with confidence...\n");
	  fprintf (FLAGOUT_FILE, "FAILED\n");
	  fclose(ERROROUT_FILE);
	  fclose(FLAGOUT_FILE);
	  exit(1);
	} 
	

	/* match found peaks with lanes, PeakPosition indexed by nodeID = zoneID +1, all estimated peaks have negative values*/
	findPeakPosition(FLAGOUT_FILE, PeakPosition, ZoneProjection, GelInfo, Spacing, minpsizefactor, Markerinc, Verbose);  

	
	/*extrapolate nodes out to edge*/
	extrapolateNodes(PeakPosition, NodePosition, GelInfo); 
	
	/*check correlation and roughness of markerlanes, and correct for anything*/
	nummarker = (GelInfo->NumberOfLanes-1-GelInfo->MarkerOffset-GelInfo->LastMarkerOffset)/GelInfo->MarkerPeriod+1;

	MarkerCorrelation = (float **)malloc((GelInfo->NumberOfMarkers-1) * sizeof(float*));
	for (index = 0; index < GelInfo->NumberOfMarkers-1; index++) 
	  {
	    MarkerCorrelation[index] =(float*) calloc(AnaCutoff->markercorsegments, sizeof(float)); /*MarkerCorrelation[markerID][segment], correlation between neighbouring markers*/
	  }
	MarkerRoughness = (float **)malloc((GelInfo->NumberOfMarkers) * sizeof(float*));
	for (index = 0; index < GelInfo->NumberOfMarkers; index++) 
	  {
	    MarkerRoughness[index] =(float*) calloc(AnaCutoff->markercorsegments, sizeof(float)); /*MarkerRoughness[markerID][segment], roughness of each marker segments*/
	  }


	/*create non-closed adjusted image for marker checking*/
	if (Adjust)
	  {
	    pixNotCloseAdj = adjustTone(pixGelFilter, ZoneProjection, GelInfo, ZoneID, ToneAdjusted, 0);  
	    checkMarkers(ERROROUT_FILE, FLAGOUT_FILE, pixNotCloseAdj, PeakPosition, NodePosition, ZoneID, GelInfo, AnaCutoff, MarkerCorrelation, MarkerRoughness, Verbose);
	  }
	else 
	  {
	    checkMarkers(ERROROUT_FILE, FLAGOUT_FILE, pixGelFilter, PeakPosition, NodePosition, ZoneID, GelInfo, AnaCutoff, MarkerCorrelation, MarkerRoughness, Verbose);
	  }
	
	/*filter zone projection for correction*/
	filterZoneProjection (ZoneProjection, GelInfo, Spacing); 

	/*correct nodes involved in markers with poor correlation score (assuming affected by noise)*/
	correctMarkers(FLAGOUT_FILE, ZoneProjection, PeakPosition, GelInfo, AnaCutoff, MarkerCorrelation, Verbose);
	writeCurvature(FLAGOUT_FILE, PeakPosition, NodePosition, GelInfo);

  /*takes out and fill all outlier peaks and peaks estimated based on outlier peaks*/
	num_outliers = correctOutlierPeaks(PeakPosition, NodePosition, GelInfo, Verbose); 
	if (Verbose) fprintf(stderr, "Corrected %d outlier peak(s) producing extreme curvature\n", num_outliers);
	fprintf(FLAGOUT_FILE, "Number of Outliers corrected:\t%i\n", num_outliers);
 	if (num_outliers >= AnaCutoff->noutlier_cutoff)
	  {
	    sprintf(tempstring, "Corrected %i outliers. Gel FAILED.\n", num_outliers);
	    if (Verbose) fprintf(stderr, "%s", tempstring);
	    fprintf(ERROROUT_FILE, "%s", tempstring);
	  }

	if (Smoothing) {
	  smoothPeaks(PeakPosition, GelInfo, Verbose); /*using a 3 lanes x 3 zones median filter to get peaks closer to average lane position of entire lane*/
	  if (Verbose) fprintf (stderr, "Smoothing performed\n");
	} else {
	  if (Verbose) fprintf (stderr, "No smoothing performed\n");
	}

	checkSymmetry(ERROROUT_FILE, FLAGOUT_FILE, pixGelFilter, ZoneID, PeakPosition, NodePosition, GelInfo, AnaCutoff, Asym, Laneoffset, con_asym_lanes, Verbose);
	
	num_corrected_asym=correctAsymmetry(ZoneProjection, PeakPosition, Spacing, GelInfo, Asym, Markerinc); 

	if (Verbose) fprintf(stderr, "Corrected %d peaks for low symmetry.\n", num_corrected_asym);
	fprintf(FLAGOUT_FILE, "Number of peaks corrected for low symmetry:\t%i\n", num_corrected_asym);

	/*check peak positions*/
	num_interpolated_markers =  (int *)calloc(GelInfo->NumberOfZones, sizeof(int)); 
	num_interpolated_data =  (int *)calloc(GelInfo->NumberOfZones, sizeof(int)); 
	con_interpolated_markers =  (int *)calloc(GelInfo->NumberOfZones, sizeof(int)); 
	con_interpolated_data =  (int *)calloc(GelInfo->NumberOfZones, sizeof(int)); 
	checkPeakPositions(ERROROUT_FILE, FLAGOUT_FILE, PeakPosition, Spacing, Markerinc, GelInfo, AnaCutoff, num_interpolated_markers, num_interpolated_data, con_interpolated_markers, con_interpolated_data, Verbose);
	
	if (Adjust)  {  
	  if (Diagnostic_present) outputDiagnostic(pixNotCloseAdj, fileDiagnosticImage, PeakPosition, NodePosition, ZoneID, GelInfo); /*output .png file with green lines at found peak, red lines at estimated peaks*/
	}
	else {
	  if (Diagnostic_present) outputDiagnostic(pixGel8bits, fileDiagnosticImage, PeakPosition, NodePosition, ZoneID, GelInfo); /*output .png file with green lines at found peak, red lines at estimated peaks*/
	}
	
	adjustBoundaryNodes(PeakPosition, NodePosition, GelInfo); /*changes all PeakPosition to positive, moves 1st node to StopPreData, and extrapolates outwardss to get peak at  borders of the image*/
	
	/*write lanes file*/
	writeLanes(PeakPosition, lanespath, NodePosition, GelInfo); 
      }/*end !NoTrack*/  

    /*check noise galaxy*/
    int* noisePresent = (int*) calloc(GelInfo->NumberOfLanes, sizeof(int));
    detectNoise(pixGelFilter, PeakPosition, NodePosition, GelInfo, AnaCutoff, noisePresent, Verbose);
    GelInfo->maxConsecNoise = countMaxConsec(noisePresent, GelInfo->NumberOfLanes, GelInfo->LaneSpacing*AnaCutoff->noise_span_factor);

     /*write traces16*/ 
	writeTraces16(pixGelRaw, tracespath, PeakPosition, NodePosition, GelInfo, Verbose); 
     /*write straighted image*/
	if(OutImage) writeStraightedImage(pixGel8bits, outimgpath, PeakPosition, NodePosition, GelInfo, Verbose); 

    /*declare if gel passes or fails*/
    if(GelInfo->GelPassed ==1)
      {
	fprintf(FLAGOUT_FILE, "PASSED\n");
	if (Verbose) fprintf(stderr, "Gel PASSED\n",fileErrorout_Name);
      }
    else
      {
	fprintf(FLAGOUT_FILE, "FAILED\n");
	if (Verbose) fprintf(stderr, "Gel FAILED, see %s for description\n",fileErrorout_Name);
      }
    /*open file for no grow output*/
   sprintf(fileNogrow_Name,"%s%s",GelInfo->ProjectDir, DEFAULT_NOGROW_NAME);

    if ((NOGROW_FILE = fopen(fileNogrow_Name,"w")) == NULL)
      {
	fprintf(stderr,"Cannot open no grow output file at %s\n", fileNogrow_Name);
	exit(1);
      }
    if(CheckNoGrow) { checkNoGrows(NOGROW_FILE, pixGelFilter, PeakPosition, NodePosition, GelInfo, AnaCutoff, Verbose);}

  /*open file for LIMS output*/
  
  sprintf(fileLIMS_Name,"%s%s",GelInfo->ProjectDir, DEFAULT_LIMS_NAME);

  if ((LIMS_FILE = fopen(fileLIMS_Name,"w")) == NULL)
    {
      fprintf(stderr,"Cannot open no grow output file at %s\n", fileLIMS_Name);
      exit(1);
    }


  fprintf(LIMS_FILE, "Interp_Marker_Span: %i\n", GelInfo->NumInterpMarkerSpan);
  fprintf(LIMS_FILE, "Asym_Span: %i\n", GelInfo->AsymSpan);
  fprintf(LIMS_FILE, "Num_Light_Markers: %i\n", GelInfo->NumLightMarker);
  fprintf(LIMS_FILE, "Noise_Span: %i\n", GelInfo->maxConsecNoise);
  fprintf(LIMS_FILE, "LR_passed: %i\n", GelInfo->GelPassed);

  pixDestroy(&pixGelRaw);
  pixDestroy(&pixGel8bits);
  pixDestroy(&pixGelFilter);
  pixDestroy(&pixGelFilter1);
  pixDestroy(&pixGelClose);
  if(Adjust && !NoTrack)  pixDestroy(&pixNotCloseAdj);
    
  if(Adjust && !NoTrack)  pixDestroy(&pixAdj);

  for (index = 0; index < GelInfo->NumberOfZones; index++) {
    free(ZoneProjection[index]);
  }
  for (index = 0; index < GelInfo->NumberOfZones+2; index++) {
    free(PeakPosition[index]);
  }
  for (index = 0; index < GelInfo->NumberOfZones; index++) {
    free(Asym[index]);
  }  
  for (index = 0; index < GelInfo->NumberOfZones; index++) {
    free(Laneoffset[index]);
  }

  free(GelInfo);
  free(AnaCutoff);
  free(ZoneProjection);
  free(PeakPosition);
  free(Asym);
  free(Laneoffset);
  free(ZoneID);
  free(NodePosition);
  free(Spacing);
  free(con_asym_lanes);
  free(noisePresent);

 if(!NoTrack)
    {
      free(num_interpolated_markers);
      free(num_interpolated_data);
      free(con_interpolated_markers);
      free(con_interpolated_data);
      for (index = 0; index < GelInfo->NumberOfMarkers-1; index++) {
	free(MarkerCorrelation[index]);
      } 
      for (index = 0; index < GelInfo->NumberOfMarkers; index++) {
	free(MarkerRoughness[index]);
      }
      free(MarkerCorrelation);
      free(MarkerRoughness);
    }

 fclose(LIMS_FILE);
 fclose(NOGROW_FILE);
 fclose(ERROROUT_FILE);
 fclose(FLAGOUT_FILE);
 exit(0);
}
