/**@file noisedetect.c
 *\brief contains functions to find multi-lane noise*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "noisedetect.h"

/**\brief gray scale subtract two 8 bit pixs, negative values are set to 0
 *
 *@param pix1 pix to subtract from
 *@param pix2 pix to subtract with
 *@returns pix of the difference
 */
PIX* pixGraySubtract(PIX* pix1, PIX* pix2){
  int width1 = pixGetWidth(pix1);
  int height1 = pixGetHeight(pix1);
  int depth1 = pixGetDepth(pix1);
  int width2 = pixGetWidth(pix2);
  int height2 = pixGetHeight(pix2);
  int depth2 = pixGetDepth(pix2);

  /*test that the two pix have same size and depth*/
  if(width1!=width2){
    fprintf(stderr, "Error in pixGraySubtract: two images have different width\n");
      return pix1;
      }
  if(height1!=height2){
    fprintf(stderr, "Error in pixGraySubtract: two images have different height\n");
      return pix1;
      }
  if(depth1!=depth2){
    fprintf(stderr, "Error in pixGraySubtract: two images have different depth\n");
      return pix1;
      }

 if(depth1 != 8) 
   {
     fprintf(stderr, "pixGraySubtract function only support 8-bit pix\n");
     exit(1);
   }

 PIX* pixResult = pixCreate(width1, height1, depth1);
 int row, col;
 UINT32 val1, val2;

 UINT32  *line1, *line2, *lineout, *data1, *data2, *dataout;
 INT32   wpl1, wpl2, wplout;
 
 data1 = pixGetData(pix1);
 wpl1 = pixGetWpl(pix1);
 
 data2 = pixGetData(pix2);
 wpl2 = pixGetWpl(pix2);

 dataout = pixGetData(pixResult);
 wplout = pixGetWpl(pixResult);
 for (row = 0; row < height1; row++) {
     line1 = data1 + row * wpl1;
     line2 = data2 + row * wpl2;
     lineout = dataout + row * wplout;

   for(col=0; col<width1; col++){
     val1=GET_DATA_BYTE(line1, col);
     val2=GET_DATA_BYTE(line2, col);
     if(val1>val2){
       SET_DATA_BYTE(lineout, col, val1-val2);
     }else{
        SET_DATA_BYTE(lineout, col, 0);
     }
   }
 }

  return pixResult;
}

/**\brief subtract background extracted with open operation
 * 
 *@param pixIn input pix
 *@param morphElement size of the structural element used for opening for background extraction
 *@param Verbose turns on text output
 *@returns background subtracted pix
*/
PIX* subtractBG(PIX* pixIn, int morphElement, int Verbose){

  PIX* pixGelOpen;
  PIX* pixOut;

  pixGelOpen = pixOpenGray(pixIn, morphElement, HORIZ);
  pixOut = pixGraySubtract(pixIn, pixGelOpen);

  pixDestroy(&pixGelOpen);
  return pixOut;
}

/**\brief extract fore ground with a closing operation
 * 
 *@param pixIn input pix
 *@param morphElement size of the structural element used for closing for foreground extraction
 *@param Verbose turns on text output
 *@returns foreground pix
*/
PIX* extractFG(PIX* pixIn, int morphElement, int Verbose){
  
  PIX* pixGelClose;
  
  pixGelClose = pixCloseGray(pixIn, morphElement, HORIZ);
  
  return pixGelClose;
}

/**\brief calculate average lane spacing over gel given a set of peak positions
 *
 *@param PeakPosition holds y position for all nodes
 *@param GelInfo holds various gel parameters
*/
float averageLaneSpacing(float** PeakPosition, struct GelInfo* GelInfo){
  
  int lane, zone;
  float sum=0;
  for(zone=1; zone<GelInfo->NumberOfZones+1; zone++){
    for(lane=0; lane < GelInfo->NumberOfLanes-1; lane++){
      sum+=(PeakPosition[zone][lane+1]-PeakPosition[zone][lane]);
    }
  }
  return sum/(float) (GelInfo->NumberOfLanes-1) /(float) (GelInfo->NumberOfZones);

}

/**\brief returns 8 bit pix with 255's labelled at points with noise galaxy content
 *
 *@param pixNoise pix of image with bands removed, leaving only large noise patches
 *@param pixFG pix of foreground (closed image)
 *@param noise_to_FG_ratio minimum noise to FG ratio for a point to be labelled as having noise
 *@param min_FG minimum foreground intensity for a the noise to FG ratio to be valid
 *@param laneWidth the width of a lane
 *@param Verbose turn on text output to screen
 *@returns pix with 255 highlighting where the noise is located on a gel
*/
PIX* pickBadPoints(PIX* pixNoise, PIX* pixFG, float noise_to_FG_ratio, int min_FG, int laneWidth, int Verbose){

  int width = pixGetWidth(pixNoise);
  int height = pixGetHeight(pixNoise);
  int depth = pixGetDepth(pixNoise);
  PIX* pixRawVal = pixCreate(width, height, depth);

  int row, col;
  float curRatio;

 if(pixGetDepth(pixNoise) != 8) 
   {
     fprintf(stderr, "extractLanes function only support 8-bit pix\n");
     exit(1);
   }
 
  UINT32 noiseVal, FGVal;

  UINT32  *lineNoise, *lineFG, *lineout, *dataNoise, *dataFG, *dataout;
  INT32   wplNoise, wplFG, wplout;
  
  dataNoise = pixGetData(pixNoise);
  wplNoise = pixGetWpl(pixNoise);
  
  dataFG = pixGetData(pixFG);
  wplFG = pixGetWpl(pixFG);
  
  dataout = pixGetData(pixRawVal);
  wplout = pixGetWpl(pixRawVal);
  /*a point must have high noise to foreground ratio and enough foreground intensity to be considered as having noise galaxy content*/
  for (row = 0; row < height; row++) {
     lineNoise = dataNoise + row * wplNoise;
     lineFG = dataFG + row * wplFG;
     lineout = dataout + row * wplout;
     
     for(col=0; col<width; col++){
       noiseVal=GET_DATA_BYTE(lineNoise, col);
       FGVal=GET_DATA_BYTE(lineFG, col);
       if(FGVal>min_FG){
	 curRatio = (float) noiseVal/ (float) FGVal;
	 if(curRatio > noise_to_FG_ratio){
	   SET_DATA_BYTE(lineout, col, 255);
	}else{
	   SET_DATA_BYTE(lineout, col, 0);
	}
      }else{
	 SET_DATA_BYTE(lineout, col, 0);
      }
     }
  }
  /*only keep large patches of noise*/
 PIX*  pixOpen1 = pixOpenGray(pixRawVal, laneWidth*2+1, VERT);
 PIX*  pixOpen = pixOpenGray(pixOpen1, laneWidth, HORIZ);

  pixDestroy(&pixRawVal);
  pixDestroy(&pixOpen1);

  return pixOpen;
}

/**brief extract a center of lane for every lane from an image
 *
 *@param pixIn pix to extract from
 *@param PeakPosition holds y position of all nodes
 *@param NodePosition holds x position of nodes in each zone
 *@param GelInfo holds various gel parameters
 *@param values 2D array where results are returned
 */
void extractLanes(PIX* pixIn, float** PeakPosition, float* NodePosition, struct GelInfo* GelInfo, int** values){

UINT32  *lineinlow, *lineinhigh, *datain;
INT32   wplin;
 int laneID, nodeID;
float	y, slope;
int	halfwidth, output_width;
int	xindex, xstop, ylow, yhigh;
 int itemp;
 float ftemp;

 if(pixGetDepth(pixIn) != 8) 
   {
     fprintf(stderr, "extractLanes function only support 8-bit pix\n");
     exit(1);
   }

 datain = pixGetData(pixIn);
 wplin = pixGetWpl(pixIn);
 for(laneID = 0; laneID < GelInfo->NumberOfLanes; laneID++){
   for(nodeID=0; nodeID < GelInfo->NumberOfZones+1; nodeID++){
     slope = (fabs(PeakPosition[nodeID+1][laneID])-fabs(PeakPosition[nodeID][laneID]))/(NodePosition[nodeID+1]-NodePosition[nodeID]);
     xindex = (int)(NodePosition[nodeID]+0.5);
     xstop = (int)(NodePosition[nodeID+1]+0.5);
     if (nodeID == GelInfo->NumberOfZones) xstop++;
     while (xindex < xstop) {
       y = slope * ((float)xindex-NodePosition[nodeID])+fabs(PeakPosition[nodeID][laneID])-0.5;
       ylow=floor(y);
       yhigh=ceil(y);
       lineinlow = datain + (ylow) * wplin;
       if (ylow == yhigh) {
	 itemp = GET_DATA_BYTE(lineinlow, xindex);
	 values[laneID][xindex]=itemp;
       } else {
	 lineinhigh = datain + (yhigh) * wplin;     
	 ftemp = floor((float)GET_DATA_BYTE(lineinlow, xindex)*((float)yhigh-y) + (float)GET_DATA_BYTE(lineinhigh, xindex)*(y-(float)ylow));    
	 values[laneID][xindex]=(int)(ftemp+0.5);
       }
       xindex++;
     }
   }
 }
 return;
}

/**\brief count the number of lanes with noise galaxy present
 *
 *@param pixIn pix of the gel on which noise is to be detected
 *@param PeakPosition holds y position of all nodes
 *@param NodePosition holds x position of nodes in each zone
 *@param GelInfo holds various gel parameters
 *@param AnaCutoff holds various analysis cutoff
 *@param noisePresent array to store amount of noisy pixels is found in the center of each lane
 *@param Verbose turn on text output to screen
*/
void detectNoise(PIX* pixIn, float** PeakPosition, float* NodePosition, struct GelInfo* GelInfo, struct AnaCutoff* AnaCutoff, int *noisePresent, int Verbose){
  PIX* pixNoBG = subtractBG(pixIn, GelInfo->SizeOfStructuralElement, Verbose);
  PIX* pixFG = extractFG(pixNoBG, GelInfo->SizeOfStructuralElement, Verbose);

  /*open in vertical direction to suppress bands while conserving multilane noise features*/
  PIX* pixNoise = pixOpenGray(pixNoBG, 2*GelInfo->LaneSpacing, VERT);
  PIX* pixBadPoints = pickBadPoints(pixNoise, pixFG, AnaCutoff->noise_to_FG_ratio, AnaCutoff->min_FG, GelInfo->LaneSpacing, Verbose);

    int index;
    int** laneValues = (int**) calloc(GelInfo->NumberOfLanes, sizeof(int*));
    for(index = 0; index <GelInfo->NumberOfLanes; index++){
      laneValues[index] = (int*) calloc(GelInfo->ImageWidth, sizeof(int));
    }
    /*find the center of lanes of all lanes in pixBadPoints*/
    extractLanes(pixBadPoints, PeakPosition, NodePosition, GelInfo, laneValues);
    int lane, col;
    for(lane = 0; lane < GelInfo->NumberOfLanes; lane++){
      noisePresent[lane]=0;
      for(col = NodePosition[2]; col<NodePosition[GelInfo->NumberOfZones]; col++){
	if(laneValues[lane][col]>0){
	  noisePresent[lane]++;
	}
      }
    }

  for (index = 0; index < GelInfo->NumberOfLanes; index++) {
    free(laneValues[index]);
  }

    pixDestroy(&pixNoBG);
    pixDestroy(&pixFG);
    pixDestroy(&pixNoise);
    pixDestroy(&pixBadPoints);
}

/**\brief counts the maximum number of consecutive values over a cutoff
 *
 *@param noise array of values
 *@param nlanes size of array
 *@param cutoff cutoff to use
 *@returns the maximum number of consecutive values over the cutoff
*/
int countMaxConsec(int *noise, int nlanes, float cutoff){
  int lane;
  int lastNoise=0;
  int consecCount=0;
  int maxConsecCount=0;
    /*count number and max span of noisy lanes*/
    for(lane = 0; lane < nlanes; lane++){
      if(((float) noise[lane]) > cutoff){
	consecCount++;
	if(lane<nlanes-1 && noise[lane+1]<=cutoff){
	  if(consecCount>maxConsecCount){
	    maxConsecCount = consecCount;
	  }
	  consecCount=0;
	}
      }
    }
    if(consecCount>maxConsecCount){
      maxConsecCount = consecCount;
    }
    return maxConsecCount;
}
