/**@file position.c
 *\brief contains functions for peak matching and curvature calculations and output*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "position.h"
#include "peaks.h"

/**\brief interpolate for any unplaced nodes (peak position == 0)
 *
 *@param PeakPosition holds y position to nodes for a zone, 0 denotes unmatched nodes, those nodes will be filled by linearly interpolation or extrapolation
 *@param GelInfo holds various gel parameters
*/
void fillEmpties(float *PeakPosition, struct GelInfo *GelInfo){

int	i, j, nemptyleft, nemptyright, nempty;
float	found1, found2, currentSpacing;
/*fill empties using data from different lanes in zone*/
 found1=0; 
 nemptyleft = 0;
 for (i = 0; i < GelInfo->NumberOfLanes; i++) {
   if (PeakPosition[i] == 0) {
     nemptyleft++;
   } else {
     found1 = fabs(PeakPosition[i]);
     break;
   }
 }
 
 /*no peak if found within a zone, use the template*/
 if(found1 == 0)
   {	 
     currentSpacing = (float)(GelInfo->HighYBorder - GelInfo->LowYBorder) / (float)(GelInfo->NumberOfLanes - 1);
     for (i = 0; i < GelInfo->NumberOfLanes; i++)
       {
	 /*estimate using position of first and last lane*/
	 PeakPosition[i] = -1. * ((currentSpacing * i) + (float)GelInfo->LowYBorder);
       }
     return;
   }
 
 
 /*fill empties up to the first matched lane using LaneSpacing from first matched lane*/
 for (i = 0; i < nemptyleft; i++) {
   PeakPosition[i] = -1. * (found1 - ((float)(nemptyleft - i)*GelInfo->LaneSpacing));
 }
 
 nemptyright = 0;
 for (i = GelInfo->NumberOfLanes-1; i >= 0; i--) {
   if (PeakPosition[i] == 0) {
     nemptyright++;
   } else {
       found1 = fabs(PeakPosition[i]);
       break;
   }
 }
 /*fill empties up from the last matched lane using LaneSpacing from last matched lane*/
   for (i = 0; i < nemptyright; i++) {
     PeakPosition[GelInfo->NumberOfLanes-i-1] = -1. * (found1 + ((float)(nemptyright - i)*GelInfo->LaneSpacing));
   }
   
   i = nemptyleft+1;
   nempty = 0;
   /*empties in between matched lanes, interpolate using matched lanes directly before and after the set of empties*/
   while (i < GelInfo->NumberOfLanes-nemptyright) {
     while (PeakPosition[i] == 0) {
       nempty++;
       i++;
     }
     /*count size of empty spot*/
     if (nempty) {
       found2 = fabs(PeakPosition[i]);
       found1 = fabs(PeakPosition[i-nempty-1]);
       currentSpacing = (found2 - found1) / (float)(nempty+1);
       for (j = i-nempty; j < i; j++) {
         PeakPosition[j] = -1. * (found2 - (currentSpacing * (float)nempty));
	 nempty--;
       }
       nempty = 0;
     }
     i++;
   }
}


#define PIX_SORT(a,b) { if ((a)>(b)) PIX_SWAP((a),(b)); }
#define PIX_SWAP(a,b) { float temp=(a);(a)=(b);(b)=temp; }
static float med9(float *p) {
    PIX_SORT(p[1], p[2]) ; PIX_SORT(p[4], p[5]) ; PIX_SORT(p[7], p[8]) ;
    PIX_SORT(p[0], p[1]) ; PIX_SORT(p[3], p[4]) ; PIX_SORT(p[6], p[7]) ;
    PIX_SORT(p[1], p[2]) ; PIX_SORT(p[4], p[5]) ; PIX_SORT(p[7], p[8]) ;
    PIX_SORT(p[0], p[3]) ; PIX_SORT(p[5], p[8]) ; PIX_SORT(p[4], p[7]) ;
    PIX_SORT(p[3], p[6]) ; PIX_SORT(p[1], p[4]) ; PIX_SORT(p[2], p[5]) ;
    PIX_SORT(p[4], p[7]) ; PIX_SORT(p[4], p[2]) ; PIX_SORT(p[6], p[4]) ;
    PIX_SORT(p[4], p[2]) ; return(p[4]) ;
}

/**\brief compare each peak with the mean position of all other nodes in the lane
 *and smooth the overall results by applying a 3 lane x 3 zones  median filters to these differences
 *
 *@param PeakPosition holds y position to nodes for all zones, interpolated nodes are denoted by negative position
 *@param GelInfo holds various gel parameters
 *@param Verbose turn on text output to screen
*/
void smoothPeaks(float **PeakPosition, struct GelInfo *GelInfo, int Verbose) {

float	**delta, *LaneAverage, vector[9], median;
int	index, i, j, im, jm;

  LaneAverage = calloc(GelInfo->NumberOfLanes, sizeof(float));
  delta = (float **)malloc((GelInfo->NumberOfZones+2) * sizeof(float*));
  for (index = 0; index < GelInfo->NumberOfZones+2; index++) 
    delta[index] = calloc(GelInfo->NumberOfLanes+2, sizeof(float));

  for (i = 0; i < GelInfo->NumberOfLanes; i++) {
    for (j = 1; j < GelInfo->NumberOfZones + 1; j++) LaneAverage[i] += fabs(PeakPosition[j][i]);
    LaneAverage[i] /= (float) GelInfo->NumberOfZones; /*average peak value of all zones in a lane*/
  }
  for(i = 0; i < GelInfo->NumberOfLanes; i++) {
    for (j = 1; j < GelInfo->NumberOfZones + 1; j++) delta[j][i+1] = fabs(PeakPosition[j][i]) - LaneAverage[i];
    delta[0][i+1] = delta[1][i+1];
    delta[GelInfo->NumberOfZones+1][i+1] = delta[GelInfo->NumberOfZones][i+1];
  }/*delta includes results of "before first" and "after last" zones, padding for filtering*/
  for (j = 0; j < GelInfo->NumberOfZones + 2; j++) {
    delta[j][0] = delta[j][1+GelInfo->MarkerPeriod];
    delta[j][GelInfo->NumberOfLanes+1] = delta[j][GelInfo->NumberOfLanes-GelInfo->MarkerPeriod];
  }/*use results from other lanes for 0th lane and "after last" lane*/
  
  /*first and last valid lanes only, 3 lanes x 3 zones median filter with delta*/
  for (j = 1; j < GelInfo->NumberOfZones+1; j++) {
    for (i = 0; i < GelInfo->NumberOfLanes; i+=GelInfo->NumberOfLanes-1) {
      index = 0;
      for (jm = -1; jm < 2; jm++) {
        for (im = 0; im < 3; im++) {
	  vector[index] = delta[j+jm][i+im];
	  index++;
	}
      }
      median = med9(vector);
      delta[j][i+1] = median;
    }
  }
  /*set filtered values to edge of data set*/
  delta[0][1] = delta[1][1];
  delta[GelInfo->NumberOfZones+1][1] = delta[GelInfo->NumberOfZones][1];
  delta[0][GelInfo->NumberOfLanes] = delta[1][GelInfo->NumberOfLanes];
  delta[GelInfo->NumberOfZones+1][GelInfo->NumberOfLanes] = delta[GelInfo->NumberOfZones][GelInfo->NumberOfLanes];
      
  /*3 lanes x 3 zone median filter to take median offset from LaneAverage*/
  for (j = 1; j < GelInfo->NumberOfZones+1; j++) {
    for (i = 0; i < GelInfo->NumberOfLanes; i++) {
      index = 0;
      for (jm = -1; jm < 2; jm++) {
        for (im = 0; im < 3; im++) {
	  vector[index] = delta[j+jm][i+im];
	  index++;
	}
      }
      median = med9(vector);
      if (PeakPosition[j][i] > 0) {
        PeakPosition[j][i] = LaneAverage[i] + median;
      } else {
        PeakPosition[j][i] = -1 * (LaneAverage[i] + median);
      }
    }
  }
  free(LaneAverage);
  for (index = 0; index < GelInfo->NumberOfZones+2; index++) free(delta[index]);
  free(delta);
}


/**\brief calculates curvature of a set of peak position and node position for all lanes and all zones
 *and output in the FLAGOUT_FILE
 *@param FLAGOUT_FILE writable file handle to data log file
 *@param PeakPosition holds y position of all nodes
 *@param NodePosition holds x position of nodes in each zone
 *@param GelInfo holds various gel parameters
 */
void writeCurvature(FILE *FLAGOUT_FILE, float **PeakPosition, float *NodePosition, struct GelInfo *GelInfo)
{
float	curvature, expected;
int	index, laneID, i1, i2, i3;
float	slope, offset;  

 fprintf(FLAGOUT_FILE, "Curvature of results before outlier correction and smoothing:\n"); 
 for (laneID = 0; laneID < GelInfo->NumberOfLanes; laneID++) 
   {
     for (index = 0; index < GelInfo->NumberOfZones-2; index++) 
       {
	 i1 = index + 1;
	 i2 = index + 2;
	 i3 = index + 3;
	 slope = (fabs(PeakPosition[i3][laneID])-fabs(PeakPosition[i1][laneID]))/(NodePosition[i3]-NodePosition[i1]);
	 offset = fabs(PeakPosition[i3][laneID]) - (slope * NodePosition[i3]);
	 expected = slope * NodePosition[i2] + offset;
	 curvature = fabs(fabs(PeakPosition[i2][laneID]) - expected);  
	 fprintf(FLAGOUT_FILE, "\t%.2f", curvature);
       }
     fprintf(FLAGOUT_FILE, "\n");
   }
}

/**\brief finds nodes showing high curvature and correct for them by forcing them to be in a straight line with their neighbours in neighbouring zones
 *
 *@param PeakPosition holds y position of all nodes
 *@param NodePosition holds x position of nodes in each zone
 *@param GelInfo holds various gel parameters
 *@param Verbose turn on text output to screen
 *@returns the number of outlier nodes corrected
*/
int correctOutlierPeaks(float **PeakPosition, float *NodePosition, struct GelInfo *GelInfo, int Verbose){

float	curvature, maxcurve, tolerance, expected, corrected;
int	index, laneID, i1, i2, i3, maxID, noutliers, anotherpass, i, needfilling, niter;
float	slope, offset;
  
  noutliers = 0;
  tolerance = GelInfo->LaneSpacing * MAXCURVATUREFACTOR / (float)GelInfo->NumberOfZones;
  for (laneID = 0; laneID < GelInfo->NumberOfLanes; laneID++) {
    anotherpass = 1;
    while(anotherpass) {
      maxcurve = 0.;
      maxID = 0;
      anotherpass = 0;
      niter = 0; 
      /*peak location is defined as (NodePosition, PeakPosition)
       *we draw a straight line through the peaks of zone before and after this zone
       *then find the y difference between the line and peak of this zone to estimate curvature
       */
      for (index = 0; index < GelInfo->NumberOfZones-2; index++) {
        i1 = index + 1;
        i2 = index + 2;
        i3 = index + 3;
        slope = (fabs(PeakPosition[i3][laneID])-fabs(PeakPosition[i1][laneID]))/(NodePosition[i3]-NodePosition[i1]);
        offset = fabs(PeakPosition[i3][laneID]) - (slope * NodePosition[i3]);
	expected = slope * NodePosition[i2] + offset;
        curvature = fabs(fabs(PeakPosition[i2][laneID]) - expected);
        if (curvature > maxcurve) {
          maxcurve = curvature;
	  maxID = i2;
	  corrected = -expected;
        }
      }
      /*if a zone has peak value off of the expected more than tolerance, set PeakPosition of highest curvature zone to be expected*/
      if ((maxcurve > tolerance) && (niter <= GelInfo->NumberOfZones)) {
        anotherpass = 1;
	niter++;
	noutliers++;
	PeakPosition[maxID][laneID] = corrected;
	needfilling = 0;
        i = laneID - 1;
	/*all estimated peaks based on outlier has to be refilled*/
        while (i >= 0) {
	  if (PeakPosition[maxID][i] < 0) {
	    PeakPosition[maxID][i] = 0;
	    i--;
	    needfilling = 1;
	  } else {
	    break;
	  }
	}
	i = laneID + 1;
	while (i < GelInfo->NumberOfLanes) {
	  if (PeakPosition[maxID][i] < 0) {
	    PeakPosition[maxID][i] = 0;
	    i++;
	    needfilling = 1;
	  } else {
	    break;
	  }
	}
	if (needfilling) fillEmpties(PeakPosition[maxID], GelInfo);
      }

      /*if every node (??) is off, destroy PeakPosition of all zone in the lane,
       * as well as any estimated lane positions based on values in this lane
      *then fill again
      */
      if (niter >= GelInfo->NumberOfZones) {
        anotherpass = 0;
	for (index = 1; index < GelInfo->NumberOfZones+1; index++) {
	  PeakPosition[index][laneID] = 0;
	  i = laneID - 1;
          while (i >= 0) {
	    if (PeakPosition[index][i] < 0) {
	      PeakPosition[index][i] = 0;
	      i--;
	    } else {
	      break;
	    }
	  }
	  i = laneID + 1;
	  while (i < GelInfo->NumberOfLanes) {
	    if (PeakPosition[index][i] < 0) {
	      PeakPosition[index][i] = 0;
	      i++;
	    } else {
	      break;
	    }
	  }
	  fillEmpties(PeakPosition[index], GelInfo);
	}
      }
    }
  }

  return noutliers;
}

/**\brief Find peaks in ZoneProjections and match with nodes by closest distance to the template
 *
 *unmatch nodes are interpolated, which are marked using negative numbers
 *@param FLAGOUT_FILE writable file handle to data log file
 *@param PeakPosition holds y position to nodes for all zones, interpolated nodes are denoted by negative position
 *@param ZoneProjection points to 1D y-projection in each of the zone
 *@param GelInfo holds various gel parameters
 *@param Spacing the dominant spacing in all zones
 *@param minpsizefactor peaks are rejected if their psize is smaller than this factor times the psize of the nearest peak matched to a marker node
 *@param Markerinc if 1, matches only marker lanes in marker-dominated zones
 *@param Verbose turn on text output to screen
 */
findPeakPosition (FILE *FLAGOUT_FILE, float **PeakPosition, INT32 **ZoneProjection, struct GelInfo *GelInfo, 
		float *Spacing, float minpsizefactor, int Markerinc, int Verbose) {

float	*pcentroid, *psize;
float	sigma;
int	length, loch, hich, ifwhm, maxpk, npeaks;
int	i, j, k;
int	startzone, zoneID, nodeID, laneID, incrementzone, incrementlane;
float	*template, *pretemplate, currentSpacing, tolerance, current_tolerance, previouspeak;
int	itemp, jtemp;
int     numlanefound, markerfound, nummarkerfound;
float	ftemp1, ftemp2;
float   maxpsize, currentpsize;


/*setting up peak matching parameters*/
  maxpk = GelInfo->NumberOfLanes*2;
  pcentroid = calloc(maxpk, sizeof(float));
  psize = calloc(maxpk, sizeof(float));
  length = GelInfo->ImageHeight;

  loch = (int) (GelInfo->LowYBorder - fabs(GelInfo->LaneSpacing) + 0.5); /*start finding peaks a bit before first lane*/
  hich = (int) (GelInfo->HighYBorder + fabs(GelInfo->LaneSpacing) + 0.5); /*and after last lane*/
  ifwhm = (int)(fabs(GelInfo->LaneSpacing) * FWHM_FACTOR + 0.5);
  tolerance = fabs(GelInfo->LaneSpacing) / TOLERANCEFACTOR;
  sigma = SIGMA_CUTOFF;

  startzone = GelInfo->FirstDataZone+1;
  zoneID = startzone;

  template = calloc(GelInfo->NumberOfLanes, sizeof(float));
  pretemplate = calloc(GelInfo->NumberOfLanes, sizeof(float));
  currentSpacing = (float)(GelInfo->HighYBorder - GelInfo->LowYBorder) / (float)(GelInfo->NumberOfLanes - 1);

  /*create a template using position of first and last lane*/
  pretemplate[GelInfo->MarkerOffset] = (currentSpacing * GelInfo->MarkerOffset) + (float)GelInfo->LowYBorder;
  
  nodeID = zoneID + 1;
  /*match marker lanes in start zone first, taking the peak closest to the template, in order to estimate the mean psize for determining the cutoff*/
  incrementlane = GelInfo->MarkerPeriod;
  laneID = GelInfo->MarkerOffset;

  npeaks = specFindPeak(ZoneProjection[zoneID], length, pcentroid, psize, loch, hich, ifwhm, sigma, maxpk);

  for (j=laneID; j<GelInfo->NumberOfLanes; j+=incrementlane)
    {
      current_tolerance = tolerance;
      markerfound = 0;
      for(i = 0; i < npeaks; i++)
	{
	  if (fabs(pretemplate[j]-pcentroid[i]) < current_tolerance)
	    {
	      current_tolerance = fabs(pretemplate[j]-pcentroid[i]);
	      PeakPosition[nodeID][j] = pcentroid[i];
	      pcentroid[i] = -pcentroid[i]; /*if close enough to a lane, should not belong to any other lanes*/
	      markerfound = 1;
	    }
	  if (pcentroid[i]>pretemplate[j]+current_tolerance||(i == npeaks-1 && markerfound ==1))
	    {
	      if (markerfound==1) 
		{
		  template[j] = PeakPosition[nodeID][j];/*updating template*/
		  pretemplate[j] = PeakPosition[nodeID][j];
		  currentSpacing = (float) (GelInfo->HighYBorder - pretemplate[j])/ (float)(GelInfo->NumberOfLanes -1 - j);
		}
	      if (j+incrementlane<GelInfo->NumberOfLanes)
		{
		  pretemplate[j+incrementlane]=pretemplate[j]+currentSpacing*incrementlane; /*updating pretemplate for next marker*/
		}
	      break; /*go to next lane when we are looking pass the peak*/
	    }
	}
    }


  /*fill unmatched lanes in template*/
  fillEmpties(template, GelInfo);

  /*reset all template to positive*/
  for (i = 0; i < GelInfo->NumberOfLanes; i++) {
    template[i] = fabs(template[i]);
  }
  /*subsequently, the peakposition of the zones before is used as the template*/

  incrementzone = 1;
  while (zoneID >= 0) /*start near front of gel, then move forward to each zone until the end, and move backwards to zones before the startzone*/

    {
      
      nodeID = zoneID + 1;
      loch = (int) (template[0] - fabs(GelInfo->LaneSpacing) + 0.5); /*start finding peaks a bit  before first lane*/
      hich = (int) (template[GelInfo->NumberOfLanes-1] + fabs(GelInfo->LaneSpacing) + 0.5); /*and after last lane*/
      ifwhm = (int)(fabs(GelInfo->LaneSpacing) * FWHM_FACTOR + 0.5);
      tolerance = fabs(GelInfo->LaneSpacing) / TOLERANCEFACTOR;
      npeaks = specFindPeak(ZoneProjection[zoneID], length, pcentroid, psize, loch, hich, ifwhm, sigma, maxpk);
      
      maxpsize = 0;
      nummarkerfound = 0;
      currentpsize = 0;
      numlanefound = 0;
      
      /*match marker lanes first, taking the peak closest to the template, in order to estimate the mean psize for determining the cutoff*/
      incrementlane = GelInfo->MarkerPeriod;
      laneID = GelInfo->MarkerOffset;
      
      for (j=laneID; j<GelInfo->NumberOfLanes; j+=incrementlane)
	{
	  current_tolerance = tolerance;
	  markerfound = 0;
	  for(i = 0; i < npeaks; i++) 
	    {
	      if (fabs(template[j]-pcentroid[i]) < current_tolerance) 
		{
		  current_tolerance = fabs(template[j]-pcentroid[i]);
		  PeakPosition[nodeID][j] = pcentroid[i];
		  pcentroid[i] = -pcentroid[i]; /*if close enough to a lane, should not belong to any other lanes*/
		  markerfound = 1;
		  currentpsize = psize[i];
		}
	      if (pcentroid[i]>template[j]+current_tolerance||(i == npeaks-1 && markerfound ==1)) 
		{
		  maxpsize+=currentpsize;
		  nummarkerfound+=markerfound; /*add found marker psize to sum for mean calculation*/
		  break; /*go to next lane when we are looking pass the peak*/
		}
	    }
	}

      if (nummarkerfound != 0)
	{
	  maxpsize /= nummarkerfound; /* use mean of found marker psizes to use as reference for psize cutoff*/
	} 
      else 
	{
	  /*use  max psize in the zone*/
	  for (i = 0; i < npeaks; i++) 
	    {
	      if (maxpsize < psize[i]) maxpsize = psize[i];
	    }   
	}
      
   
      /*find other lanes if zone is lane spacing dominate or marker inc option is not turned on*/
      if (Markerinc == 0 || fabs(Spacing[zoneID]-GelInfo->LaneSpacing) < tolerance) 
	{ 
	  incrementlane = 1;
	  laneID = 0;
	  
	  /*try to match lane with unmatched peaks, take the peak closest to the template*/
	  for (j=laneID; j<GelInfo->NumberOfLanes; j+=incrementlane)
	    {
	      current_tolerance = tolerance;

	      if (PeakPosition[nodeID][j] > (float) loch) continue; /*marker peak already found, skip to next lane*/
	      for(i = 0; i < npeaks; i++) 
		{
		  if (fabs(template[j]-pcentroid[i]) < current_tolerance && psize[i] > maxpsize*minpsizefactor) 
		    {
		      current_tolerance = fabs(template[j]-pcentroid[i]);
		      PeakPosition[nodeID][j] = pcentroid[i];
		      pcentroid[i] = -pcentroid[i]; /*if close enough to a lane, should not belong to any other lanes*/
		   
		    }
		  if (pcentroid[i]>template[j]+current_tolerance) 
		    {
		      break; /*go to next lane when we are looking pass the peak*/
		    }
		}
	    }
	}
            
    
      /*first zone, based on rough template,
       *so we use data from higher lanes for matching unmatched peaks
       *after the first pass
       */
      if (zoneID == startzone)
	{
	  if (  fabs(Spacing[zoneID]-GelInfo->LaneSpacing) > tolerance&& Markerinc == 1) 
	    {
	      laneID = GelInfo->MarkerOffset + GelInfo->MarkerPeriod; /*start with second marker*/
	    } 
	  else 
	    {
	      laneID = 1; /*0+incrementlane*/
	    }
	        
	  /*try to match lane with unmatched peaks, take the peak closest to the template*/
	  for (j=laneID; j<GelInfo->NumberOfLanes; j+=incrementlane)
	    {
	      if (PeakPosition[nodeID][j] > (float) loch) continue; /*marker peak already found, skip to next lane*/ 
	      current_tolerance = tolerance;
	      previouspeak = 0;	
	      markerfound = 0;
	      for (k = j - incrementlane; k >= 0; k-=incrementlane)  /*find the last lanes found before it*/
		{
		  previouspeak = PeakPosition[nodeID][k];
		  if (previouspeak > 0) /*one lane matched with a peak found, extrapolate and try to match*/ 
		    {
		      for(i = 0; i < npeaks; i++) 
			{
			  if (fabs(pcentroid[i]-((float) (j-k))*(GelInfo->LaneSpacing)-previouspeak)< current_tolerance && psize[i] > maxpsize*minpsizefactor) 
			    {
			      current_tolerance = fabs(pcentroid[i]-((float) (j-k))*(GelInfo->LaneSpacing)-previouspeak);
			      PeakPosition[nodeID][j] = pcentroid[i];
			      pcentroid[i] = -pcentroid[i];
			    }
			  if (pcentroid[i]>template[j]+current_tolerance) 
			    {
			      k = -1;
			      break; /*go to next peak when we are looking pass the peak*/
			    }
			}
		    }
		}
	    }
	}


      fillEmpties(PeakPosition[nodeID], GelInfo);
      if (zoneID == (GelInfo->NumberOfZones-1)) {
	zoneID = startzone;
	nodeID = zoneID + 1;
	incrementzone = -1;
      }
      /*use matched lanes as the template for next zone*/
      for (i = 0; i < GelInfo->NumberOfLanes; i++) {
	template[i] = fabs(PeakPosition[nodeID][i]);
      }
      zoneID += incrementzone;
    }
  

  free(pcentroid);
  free(psize);
  free(template);
  free(pretemplate);
}
