/**@file toneadjust.c
 *\brief holds functions for tone mapping for contrast
 *enhancement using a bezier curve
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "toneadjust.h"

/**\brief calculate a cubic bezier curve based on P0, P1, P2 and P3, 
 *will return two vectors, x, y with values of the curve sampled at regular intervals defined by numpoints
 *
 *@param x array to hold x values of the curve
 *@param y array to hold y values of the curve
 *@param px0 x value of first point defining the curve
 *@param py0 y value of first point defining the curve
 *@param px1 x value of second point defining the curve
 *@param py1 y value of second point defining the curve
 *@param px2 x value of third point defining the curve
 *@param py2 y value of third point defining the curve
 *@param numpoints number of points of the curve to sample
*/
static void bezier4(float* x, float* y, float px0, float py0,  float px1, float py1, float px2,  float py2, float px3, float py3, int numpoints) 
{

  int i;
  float f1, f13, f3;
  float *t;

  t = calloc(numpoints, sizeof(float));
  /*applying Bezier formula*/
  for (i=0; i<numpoints; i++)
    {
      t[i]=(float)i*1.0/(float)numpoints;

      f1 = 1 - t[i];
      f13 = f1 * f1 * f1;
      f3 = t[i] * t[i] * t[i];

      x[i] = f13*px0 + 3.0*t[i]*f1*f1*px1 + 3.0*t[i]*t[i]*f1*px2 + f3*px3;
      y[i] = f13*py0 + 3.0*t[i]*f1*f1*py1 + 3.0*t[i]*t[i]*f1*py2 + f3*py3;
    }

  /*setting first and last point to be true extreme*/
  x[0]=px0;
  y[0]=py0;
  
  x[numpoints-1]=px3;
  y[numpoints-1]=py3;

  free(t);
}

/** \brief finds y value for a particular x value in a monotonic parametric 
 *curve defined by the two vectors of x, y 
 *
 *@param x array holding sorted x value to points on the curve
 *@param y array holding y values to points on the curve
 *@param xnow the x value of the point in question
 *@param numpoints the length of the x and y vectors
 *@returns the y value corresponding to the x value given
*/
static float parametric_point(float *x, float *y, float xnow, int numpoints)
{
  int i=0;
  float ynow;

  while (x[i]<xnow && i < numpoints) i++;
  /*set y to y[0] if xnow is less than minimum x[i]*/
  /*leave y to y[numpoints] if xnow is greater than maximum x[i]*/
  
  if (i==0)
    {
      return y[0];
    }
  else
    {
      ynow = (y[i]-y[i-1])/(x[i]-x[i-1])*(xnow-x[i-1])+y[i-1];
      return ynow;
    }
}

/**\brief tone adjust a PIX in large zones to bring up contrast of faint bands
 *
 *The bezier curve is adjusted using the mean intensity of the zone
 *
 *@param pixGel pix of gel image
 *@param ZoneProjection points to 1D y-projection in each of the zone, to be replaced in tone adjusted zones
 *@param GelInfo holds various gel parameters
 *@param ZoneID array noting the zone number for each column
 *@param ToneAdjusted array to note which zone has been tone adjusted
 *@param Verbose turn on text output to screen
 *@returns version of pixGel that has been toneadjusted in large zones
 */
PIX *adjustTone(PIX *pixGel, INT32 **ZoneProjection, struct GelInfo *GelInfo, int *ZoneID, int *ToneAdjusted,  int Verbose)
{

float mean;
INT32 total=0;
int *zonesize=calloc(GelInfo->NumberOfZones, sizeof(int));
INT32 minzonesize = 0;
int i, j, row, col;
float ftemp;
float **x, **y;
float meandiff;
char	fullName[FILE_NAME_LEN];

UINT32  *line, *lineadj;
UINT32  *data, *dataadj;
INT32   wpl, wpladj;

 x = calloc(GelInfo->NumberOfZones, sizeof(float*));
 y = calloc(GelInfo->NumberOfZones, sizeof(float*));


 /*work out zone size for idenifying zones of low contrast (large zones means low intensity bands)*/ 
 for (i=0; i<GelInfo->ImageWidth; i++)
   {
     if (ZoneID[i]>=0) /*include only predata and data*/
       {
	 zonesize[ZoneID[i]]++;
       }
   }

 minzonesize = zonesize[0];
/*find min zonesize*/
 for (i=0; i<GelInfo->NumberOfZones; i++)
   {
     minzonesize=minzonesize<zonesize[i]?minzonesize:zonesize[i];
   }


 /*if zonesize is large (data density small), we do a pixel mapping using overall image mean.
  *the zoneprojection is recalculated for those zones for better spacing and peaks calculation
  */
 PIX *pixadj =  pixCopy(NULL, pixGel);

 dataadj = pixGetData(pixadj);
 wpladj = pixGetWpl(pixadj);



/*get mean intensity for each zone*/
 for (i=0; i<GelInfo->NumberOfZones; i++)
   {
     total = 0;
     for(j=0; j<GelInfo->ImageHeight; j++)
       {
	 total+=ZoneProjection[i][j];
       }
     mean = (float) total / (float) (zonesize[i]*GelInfo->ImageHeight)/(float) (pow(2, pixGetDepth(pixGel))-1);
 
     /*getting bezier curve*/
     x[i]=calloc(BEZIER_NUMPOINTS, sizeof(float));
     y[i]=calloc(BEZIER_NUMPOINTS, sizeof(float));

     bezier4(x[i], y[i], (float) 0, (float) 0, (float)BEZIER_P1X_FACTOR*mean, (float)BEZIER_P1X_FACTOR*mean,  (float)0,  (float) 1,  (float) 1, (float) 1, BEZIER_NUMPOINTS);

     if(zonesize[i]>(int)(0.5+ LARGE_ZONESIZE_CUTOFF_FACTOR*(float)minzonesize)) ToneAdjusted[i]=1;
     if(Verbose && ToneAdjusted[i]==1) fprintf(stderr, "Tone in zone %i adjusted.\n", i);
   }
 
 for (row=0; row<GelInfo->ImageHeight; row++)
   {
     lineadj = dataadj + row * wpladj;
     for (col=0; col<GelInfo->ImageWidth; col++)
       {
	 if(ZoneID[col]>=0)
	   {
	     if(ToneAdjusted[ZoneID[col]]==1)
	       {
		 ZoneProjection[ZoneID[col]][row]-=(INT32) GET_DATA_BYTE(lineadj, col);
		 ftemp = (float) GET_DATA_BYTE(lineadj, col)/ (float) (pow(2, pixGetDepth(pixGel))-1);

		 /*getting transformed value through Bezier curve*/
		 ftemp = parametric_point(x[ZoneID[col]],y[ZoneID[col]], ftemp, BEZIER_NUMPOINTS);
 		  
		 SET_DATA_BYTE(lineadj, col, (int) (0.5+ftemp*(float) (pow(2, pixGetDepth(pixGel))-1)));
		 ZoneProjection[ZoneID[col]][row]+=(INT32) GET_DATA_BYTE(lineadj, col);
	       }
	   }
       }
   }

 if (Verbose) fprintf(stderr, "Low contrast zones adjusted\n");

 for(i=0; i<GelInfo->NumberOfZones; i++)
   {
     free(x[i]);
     free(y[i]);
   }

 free(x);
 free(y);

 return pixadj;
}
