/**@file zone.c
 *\brief contains functions related to the 
 *initialization operations of zone boundary, nodes, and zoneprojection 
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "zone.h"

/**\brief sum intensity in vertical direction in each zone, effectively
 * getting a 1D profile of the gel along all the lanes
 *
 *@param pixGel pix representation of the gel
 *@param XProjection array for outputing the summed intensity trace
 *@param GelInfo holds various gel parameters
 */
static void projectXDirection (PIX *pixGel, INT32 *XProjection, struct GelInfo *GelInfo) {

UINT32  *lines;
UINT32  *datas;
INT32   wpls;
INT32	i, j;
UINT8	value;
  datas = pixGetData(pixGel);
  wpls = pixGetWpl(pixGel);
  for (i = 0; i < GelInfo->ImageHeight; i++) {
    lines = datas + i * wpls;
    for (j = 0; j < GelInfo->ImageWidth; j++) {
      value = GET_DATA_BYTE(lines, j);
      XProjection[j] += value;
    }
  }
}


/**\brief Project to 1D (sum all pixel with same X) to establish StartPreData, StopPreData, StopData
 *
 *The zoneID of pixels before StartPreData and StopData is set to be -1
 *divide data from StartPreData to StopData evenly into zones
 *The StopPreData is forced to be a zone boundary
 *find "node" of each zone - center of mass of data values in each zone
 *
 *@param FLAGOUT_FILE writable file handle to data log file
 *@param pixGel pix of gel image
 *@param GelInfo holds various gel parameters
 *@param ZoneID array to note the zone number for each column
  *@param Verbose turn on text output to screen
*/  
void initZones (FILE *FLAGOUT_FILE, PIX *pixGel, struct GelInfo *GelInfo, int *ZoneID, int Verbose) {

INT32	*XProjection, firstmin, secondmin, max, downslope;
UINT32	*SumProjection, total;

int	i, posfirstmin, possecondmin, wellwindow, posmax, posdownslope;
 int     firstdatazonesize=0;
 int     lastpredatazonesize=0;
 
  XProjection = calloc(GelInfo->ImageWidth, sizeof(INT32));
  SumProjection = calloc(GelInfo->ImageWidth, sizeof(UINT32));

  projectXDirection(pixGel, XProjection, GelInfo);

  /* find x position such that sum of all entries in the column is max*/
  /*FINDMAXFACTOR determines start and end of search area*/
  max = -1;
  for (i = GelInfo->ImageWidth / FINDMAXFACTOR; i < (FINDMAXFACTOR-1)*GelInfo->ImageWidth/FINDMAXFACTOR; i++) {
    if (XProjection[i] > max) {
      max = XProjection[i];
      posmax = i;
    }
  }
  wellwindow = GelInfo->ImageWidth / WELLWINDOWFACTOR;  /*estimated size of well*/
  downslope = 0;
  posdownslope = 0;
  /*find location of well*/
  for (i = wellwindow; i < posmax; i++) {
    if (XProjection[i-wellwindow]-XProjection[i] > downslope) {
      downslope = XProjection[i-wellwindow]-XProjection[i];
      posdownslope = i;
    }
  }
  /*start looking for minimum pass the well*/
  firstmin = XProjection[posdownslope];
  posfirstmin = posdownslope;
  for (i = posdownslope+1; i < posmax; i++) {
    if (XProjection[i] < firstmin) {
      firstmin = XProjection[i];
      posfirstmin = i;
    }
  }
  secondmin = XProjection[GelInfo->ImageWidth-1];
  possecondmin = GelInfo->ImageWidth-1;
  /*find a minimum after the maximum*/
  for (i = GelInfo->ImageWidth-1; i > posmax; i--) {
    if (XProjection[i] < secondmin) {
      secondmin = XProjection[i];
      possecondmin = i;
    }
  }
  /*predata include firstmin and starts after the well, and it includes only points touching first minimum with xprojection value close to the minimum*/
  for (i = posfirstmin; i >= posdownslope; i--) {
    if (XProjection[i] > firstmin * PREDATAFACTOR) break;
    GelInfo->StartPreData = i;
  }
  for (i = posfirstmin; i < GelInfo->ImageWidth; i++) {
    if (XProjection[i] > firstmin * PREDATAFACTOR) break;
    GelInfo->StopPreData = posfirstmin + (float) (i-posfirstmin) * PREDATASTOPFACTOR;  /*we don't go to the exact start of data to give room for guassian tail for BL*/
  }
  /*stopdata is after secondmin and is the first point after the secondmin that the xprojection is increase pass a certain value*/
  for (i = possecondmin; i < GelInfo->ImageWidth; i++) {
    if (XProjection[i] > secondmin * POSTDATAFACTOR) break;
    GelInfo->StopData = i;
  }

  /*if stop data is placed at the end of the image, assume that the end of gel is too far from the end of image and is taken as secondmin
   *try to find the edge and step back a given amount*/
  if(GelInfo->StopData == GelInfo->ImageWidth-1)
    {
      if (Verbose) fprintf(stderr, "Data stop placed at end of image, recalculating...\n");
      fprintf(FLAGOUT_FILE, "Stop data initially placed at end of image.\n");
      for (i = possecondmin; i > posmax; i--) {
	if (XProjection[i] > secondmin * POSTDATAFACTOR) break;
	GelInfo->StopData = i - (float)GelInfo->SizeOfStructuralElement * STOPDATASTEPBACKFACTOR;
      }
    }

 
  if (Verbose) {
    fprintf(stderr, "Predata located between %d and %d\n", GelInfo->StartPreData, GelInfo->StopPreData);
    fprintf(stderr, "Data stop at %d\n", GelInfo->StopData);
  }
  fprintf(FLAGOUT_FILE, "Start PreData:\t%d\nStop PreData:\t%d\nStop Data:\t%d\n", GelInfo->StartPreData, GelInfo->StopPreData, GelInfo->StopData);
  /*ZoneID for points before StartPreData and after StopData is -1 (not valid zone)*/
  for (i = 0; i < GelInfo->StartPreData; i++) {
    XProjection[i] = 0;
    ZoneID[i] = -1;
  }

  for (i = GelInfo->StopData; i < GelInfo->ImageWidth; i++) {
    XProjection[i] = 0;
    ZoneID[i] = -1;
  }
  for (i = 0; i < GelInfo->ImageWidth; i++) SumProjection[i] = XProjection[i]>>SHIFTTOFIT32; /*type cast*/
  for (i = GelInfo->StartPreData; i < GelInfo->ImageWidth; i++) SumProjection[i] += SumProjection[i-1]; /*running sum*/
  total = SumProjection[GelInfo->ImageWidth-1];


  for (i = GelInfo->StartPreData; i < GelInfo->StopData; i++) {
    ZoneID[i] =  GelInfo->NumberOfZones * (SumProjection[i]-1) / total;
  } /*split into zones based on the running sum between StartPreData and StopData. largest zoneID is NumberOfZones-1
     *set all predata to have zoneID 1 less than first zone after predata (stoppredata must be a zone boundary)
     *to avoid artificially small zones, the resulting first data zone must be of a certain size, or it will be merged into the zone after*/
  if (ZoneID[GelInfo->StopPreData] == 0) 
    {
      i = GelInfo->StopPreData + 1;
      while ( ZoneID[i] == 0 ) {
	ZoneID[i] = 1;
	i++;
      }
    } 
  else 
    {
      i = GelInfo->StopPreData+1;
      while(ZoneID[i] == ZoneID[GelInfo->StopPreData])
	{
	  firstdatazonesize++;
	  i++;
	}
      i = GelInfo->StopPreData;
      while(ZoneID[i] == ZoneID[GelInfo->StopPreData])
	{
	  lastpredatazonesize++;
	  i--;
	}
      if(firstdatazonesize > lastpredatazonesize)
	{
	  i = GelInfo->StopPreData - 1;
	  while ( ZoneID[i] ==  ZoneID[GelInfo->StopPreData]) {
	    ZoneID[i] = ZoneID[GelInfo->StopPreData] - 1;
	    i--;
	  }
	}
      else
	{
	  i = GelInfo->StopPreData + 1;
	  while ( ZoneID[i] == ZoneID[GelInfo->StopPreData] ) {
	    ZoneID[i] =  ZoneID[GelInfo->StopPreData]+1;
	    i++;
	  }
	}
    }
 
  GelInfo->FirstDataZone = ZoneID[GelInfo->StopPreData+1];
  if (Verbose) fprintf(stderr, "Data start in zone %d\n", GelInfo->FirstDataZone);
  fprintf(FLAGOUT_FILE, "First Data zone:\t%d\n", GelInfo->FirstDataZone);

  free(XProjection);
  free(SumProjection);
}

/**\brief set node position to be the center of data mass location of each zone except for first and last (border) nodes
 *
 *@param pixGel 8bit pix representation of the gel image
 *@param GelInfo holds various gel parameters
 *@param ZoneID array to note the zone number for each column
 *@param NodePosition to hold x position of nodes in each zone
 *@param Verbose turns on text output to screen
*/
void setNodePositions (PIX *pixGel, struct GelInfo *GelInfo, int *ZoneID, float *NodePosition, int Verbose) {

INT32	*XProjection;
float	*SumZones;
int     i;

 XProjection = calloc(GelInfo->ImageWidth, sizeof(INT32));
 SumZones = calloc(GelInfo->NumberOfZones, sizeof(float));
 projectXDirection(pixGel, XProjection, GelInfo);

  for (i = 0; i < GelInfo->NumberOfZones; i++) SumZones[i] = 0.;
  for (i = 0; i < GelInfo->ImageWidth; i++) {
    if (ZoneID[i] >= 0) {
      NodePosition[ZoneID[i]+1] += (float) (XProjection[i] * i); 
      SumZones[ZoneID[i]] += (float)XProjection[i]; /*sum of all data values in a zone*/
    }
  }

  for (i = 0; i < GelInfo->NumberOfZones; i++) NodePosition[i+1] /= SumZones[i]; /*nodeposition is center of mass of XProjection in a zone*/
  NodePosition[0] = 0.;
  NodePosition[GelInfo->NumberOfZones+1] = (float)(GelInfo->ImageWidth-1); /*first node is at 0, last node is at imagewidth-1*/

  free(XProjection); 
  free(SumZones);
}


/**\brief integrate data horizontally in each zone, effectively getting a 1D profile for each zone
 *@param pixGel pix representation of gel image
 *@param ZoneProjection 2d array to hold 1D y-projections in each of the zone
 *@param GelInfo holds various gel parameters
 *@param ZoneID array noting the zone number for each column
*/
void projectZones (PIX *pixGel, INT32 **ZoneProjection, struct GelInfo *GelInfo, int *ZoneID) {

UINT32  *lines;
UINT32  *datas;
INT32   wpls;
INT32	i, j;
UINT8	value;

  datas = pixGetData(pixGel);
  wpls = pixGetWpl(pixGel);

  for (i = 0; i < GelInfo->ImageHeight; i++) {
    lines = datas + i * wpls;
    for (j = 0; j < GelInfo->ImageWidth; j++) {
      value = GET_DATA_BYTE(lines, j);
      if (ZoneID[j] >= 0) ZoneProjection[ZoneID[j]][i] += value;
    }
  }/*ZoneProjection holds sum of all data value of pixels in the zone on each line*/
}

#define PIX_SORT(a,b) { if ((a)>(b)) PIX_SWAP((a),(b)); }
#define PIX_SWAP(a,b) { INT32 temp=(a);(a)=(b);(b)=temp; }
static INT32 opt_med3(INT32 *p)
{
    PIX_SORT(p[0],p[1]) ; PIX_SORT(p[1],p[2]) ; PIX_SORT(p[0],p[1]) ;
    return(p[1]) ;
}

/**\brief aggressive filtering of the zone projection using the generic spacing.
 *
 *Each value is subtracted with values a half a lane spacing from it to enhance the periodic
 *peak pattern
 *then, each value is median filter with the neighbour lanes or neighour markers
 *
 *@param ZoneProjection 2D array holding 1D traces of each zone, will be replaced by filtered version
 *@param GelInfo holds various gel parameters
 *@param Spacing the dominant spacing in all zones, used to identify marker dominant zones
 */
void filterZoneProjection (INT32 **ZoneProjection, struct GelInfo *GelInfo, float *Spacing) {

INT32	*temp;
INT32	vector[3], median, min;
int	i, j, k, ispacing, start, stop;

  temp = calloc(GelInfo->ImageHeight, sizeof(INT32));

  ispacing = (int)( (GelInfo->LaneSpacing/2.) + 0.5);  /*half a lane*/
  start = GelInfo->LowYBorder-(int)(FIXFLATFACTOR * GelInfo->LaneSpacing);
  if (start < ispacing) start = ispacing;
  stop = GelInfo->HighYBorder+(int)(FIXFLATFACTOR * GelInfo->LaneSpacing);
  if (stop > GelInfo->ImageHeight - ispacing) stop = GelInfo->ImageHeight - ispacing;
  for (i = GelInfo->FirstDataZone; i < GelInfo->NumberOfZones; i++) /*exclude predata*/
  {
    min = ZoneProjection[i][start];
    for (j = 0; j < GelInfo->ImageHeight; j++) {
      if (j >= start && j < stop) {
        temp[j] = ZoneProjection[i][j] - ((ZoneProjection[i][j-ispacing]+ZoneProjection[i][j+ispacing])/2);
	if (temp[j] < min) min = temp[j];
      } else {
        temp[j] = 0;
      }
    }
    for (j = 0; j < GelInfo->ImageHeight; j++) ZoneProjection[i][j] = temp[j] - min; /*ZoneProjection changed to have narrower peaks surrounded by small values*/
  }
  
  for (i = 0; i < GelInfo->NumberOfZones; i++) {
    if (fabs(Spacing[i]-GelInfo->LaneSpacing) < 1.) /*for zones with dominant spacing close to lane spacing,(marker lanes not dominant)  use lane spacing*/ {
      ispacing = (int)(GelInfo->LaneSpacing + 0.5);
    } else {
      ispacing = (int)((float)GelInfo->MarkerPeriod * GelInfo->LaneSpacing + 0.5);
    }/*else,(in zones where marker is visible, or result invalid)  use marker spacing*/
    /*median filter between lanes or marker lanes*/
    for (k = 0; k < SMOOTHROUNDS; k++) {
      for (j = 0; j < GelInfo->ImageHeight; j++) {
        if (j >= ispacing && j < GelInfo->ImageHeight - ispacing) {

          vector[0] = ZoneProjection[i][j-ispacing];
          vector[1] = ZoneProjection[i][j];
          vector[2] = ZoneProjection[i][j+ispacing];

          median = opt_med3(vector);
        } else {
          median = ZoneProjection[i][j];
        }
        temp[j] = median;
      }
      for (j = 0; j < GelInfo->ImageHeight; j++) ZoneProjection[i][j] = temp[j];
    }
  }
  
  free(temp);
}
